;+=----------------------------------------------------------------------------
; Name   : example_03.pro
; Desc   : Example no.3, covering the use of mrdisplay and the five
;          plugin mrcontrast, mrzoom, mrscroll, mrdataflash and mrcursorinfo
;          Use left mouse button for scrolling the image
;          Use middle mouse button for zooming
;          Use right mouse button for greyscale changes
;          Double click of the mouse buttons resets the corresponding item
; Author : (c) Oliver Thilmann, Aug 2003
;=-----------------------------------------------------------------------------
pro example_03

	mrlib_init

		; need some data to display

	mrdicom_setUIDs, 'mrimg', '0.0.00.0.0000.0.0'
	file = dialog_pickfile( path='c:\' )

	dcm  = obj_new('mrImageSetDicom', 1)
	res  = dcm->load( file, 0, /rotate )

		; in case of errors, print error message and create image

	if( res eq 0 ) then begin

		errortxt = dcm->errMsg()
		res = dialog_message(['Error loading file', errortxt, 'Using greyscale bar instead'])
		img = intarr( 256, 256 )
		for i = 0, 255 do begin
			img[*, i] = i
		end
        img = ptr_new( img, /no_copy )   ; images are always referenced via pointers
        obj_destroy, dcm                ; destroy data object as it is not used

	end else begin

		; loading was ok -> getImgData returns a pointer to an array of
		; pointers. As only one image was loaded, we use the first one

		img = (*(dcm->imgData()))[0]
	end

		; CREATE UI - keep that short, doesn't need to look good for
		; demonstration purpose

	base    = widget_base( /column )

		; first: a display widget, canvas size = visible size = 512
		; see the documentation for argument explanation
		; Interpolation is switched off
		; Switch it on for a test: during scrolling and zooming, interpolation
		; will be switched off automatically to speed up execution - after releasing
		; the mouse button, the interpolated image will be calculated. Terribly
		; slow for high zooming factors!

	display = obj_new('mrdisplay', base, 512, 512, 512, 512, interpolate=0 )


		; now we instanciate the four plugin objects:

		; MRCONTRAST:
		; the contrast object displays none, one or two color bars, depending
		; on the mode - here we display both, horizontally

	contrast = obj_new('mrcontrast', base, xsize = 512, ysize = 100 , colbars=2 )

		; MRZOOM:
		; zooming - mouse up/down changes scale, double click resets to '1'

	zoom    = obj_new('mrzoom')

		; MRSCOLL:
		; scrolling - mouse movements move the image, double click centers

	pan  = obj_new('mrpan')

		; MRDATAFLASH
		; will overlay patient/examination data from the dicom object
		; only possible if dicom data is available
		; if so, set the dicom object for dataflash

	if( obj_valid( dcm ) ) then begin
		dataflash = obj_new('mrdataflash')
		dataflash->setDicomSource, (*(dcm->dicomData()))[0]
	end else begin
		dataflash = obj_new()
	end

		; MRCURSORINFO
		; displays the current image position and -value next to the cursor
		; simple to use

	cursorinfo = obj_new('mrcursorinfo', color=[255,0,0] )

		; realize the UI

	widget_control, base, /realize

		; Now register the plugins with the mrdisplay object
		; By registering with the mrdisplay object, the plugins request
		; a certain type of events (in this case, mouse click, double click
		; and motion events, the mrcontrast object also receives image change
		; events) upon which they react according to their use.
		; See the next example for more on events
		; (i.e. they will change contrast/zoom/scroll the image)

	contrast->selfregister,   display, button=3, /bordermode   ; right 15%: gamma, rest: center/width
	zoom->selfregister,       display, button=2
	pan->selfregister,        display, button=1
	cursorinfo->selfregister, display, toggleKey='c'

		; the dataflash object could also register with a mouse button,
		; which then would toggle dicom data display on and off.
		; alas, we have no mouse button left...

	if( obj_valid( dataflash ) ) then dataflash->selfregister, display

		; Next set the image. The mrdisplay object makes an internal copy
		; of the image data - any change to the original data will thus have
		; no effect (i.e. even deleting the original data :)
		; Use the update method to update image data.

		; Note that setting a new image will automatically send a
		; corresponding event to the contrast object (imgchange). The contrast
		; object will subsequently get the new image's properties and update
		; itself. Due to that mechanism, it doesn't matter whether the
		; contrast object registers before or after setting of the image.

	display->setImg, img

		; scale image, so it fits the screen

	display->fitToScreen

		; and finally display the image

	display->display

		; If we have a dicom object, display the data
		; This will also create a call to display->display

	if( obj_valid( dataflash ) ) then dataflash->display

		; The display object does not create mouse events by default -
		; it needs to be activated

	display->activate

		; and we need an xmanager running with a dummy handler or the
		; program exits immediately

	xmanager, 'mrdisplay_example', base, event_handler='dummy_handler', /no_block

end

; dummy handler does nothing - just intercepts events
pro dummy_handler, evt
end
