;+=----------------------------------------------------------------------------
; Name   : example_06.pro
; Desc   : mrRoiUI
;          See the evaluate_rois method for an example how to evaluate
;          the defined rois.
;          Most functionality of the roi UI is hopefully more or less
;          intuitive.
;          - to define a polygonial roi, keep the mouse button pressed while
;            outlining the roi area
;          - when editing a roi: press the mouse button and hold it - the roi
;            closest to the mouse button will become active (changes color)
;            after releasing the mouse button, it reverts to its original color.
;            Esp. when many rois are on screen, this might be a bit slow. Don't
;            hurry...
;          - Undo works only one level back and only for adding of a roi
;          - Known bugs:
;            Rois may be scaled so that parts of the roi lie outside the
;            image - in such cases it may happen, that the roi is not displayed
;            any more. It's there (number remains visible), just not displayed.
;            Rois maybe moved outside the image
; Author : (c) Oliver Thilmann, Aug 2003
;=-----------------------------------------------------------------------------
pro example_06

	mrlib_init

		; lazy again

	common globals6, roiobj, text, img

		; load image

	mrdicom_setUIDs, 'mrimg', '0.0.00.0.0000.0.0'
	file = dialog_pickfile( path='c:\' )

	dcm  = obj_new('mrImageSetDicom', 1)
	res  = dcm->load( file, 0, /rotate )

		; in case of errors, print error message and return

	if( res eq 0 ) then begin
		x = dialog_message(['Error loading file', dcm->errMsg()])
		return
	end

	img = (*(dcm->imgData()))[0]

		; CREATE UI, consisting of an mrdisplay object with contrast
		; functionality, the mrroiui user interface and a text object for
		; result display

	base    = widget_base( /row )

	display = obj_new('mrdisplay', base, 512, 512, 512, 512, interpolate=0 )

		; this time no display of greyscale bars for the contrast object

	contrast = obj_new('mrcontrast')

		; create roi object incl. load/save buttons and display style buttons

	roiobj  = obj_new('mrroiui', base, /column, spacing=10, /frame )
	base2   = widget_base( base, /column )
	text    = widget_text( base2, scr_ysize=250, /frame, /scroll, font='fixedsys' )
	but1	= widget_button( base2, value='eval Rois', event_pro='evaluate_rois' )

		; realize the UI

	widget_control, base, /realize

		; Now register the plugins with the mrdisplay object
		; By registering with the mrdisplay object, the plugins request
		; a certain type of events (in this case, mouse click, double click
		; and motion events, the mrcontrast object also receives image change
		; events) upon which they react according to their use.
		; See the next example for more on events
		; (i.e. they will change contrast/zoom/scroll the image)

	contrast->selfregister, display, button=3
	display->setImg, img

		; scale image, so it fits the screen, display the image and activate events
		; the display must be activated, or the roi object will not receive any
		; events and thus be disfunctional.

	display->fitToScreen
	display->display
	display->activate

		; register the roi event with the mrdisplay object, so it receives
		; mouse events from the display.
		; Important: the roi object *must not* be registered before a valid
		; image is set in the display object.
		; (imgchange events only came later - will be cured in the future)

	roiobj->selfregister, display

		; activate roi buttons

	roiobj->activate

		; and start the xmanager

	xmanager, 'example_06', base, event_handler='dummy_handler', /no_block

end

; dummy handler does nothing - just intercepts events
pro dummy_handler, evt
end


; ---------------------------------------------------------------------------
; Name: evaluate_rois
; Desc: An example on how to evaluate the defined rois
; ---------------------------------------------------------------------------
pro evaluate_rois, evt
	common globals6, roiobj, text, img

	; rois are stored in an IDLanRoiGroup object. Get that object
	; from our roi object and count.

	roigroup = roiobj->roiGroup()
	nrois    = roigroup->count()

	; start new text in textwidget

	widget_control, text, set_value='  ROI EVALUATION'
	widget_control, text, set_value='------------------', /append

	; if no rois defined we're already done

	if( nrois eq 0 ) then return

	; otherwise loop over all rois, calculate values and display

	for i = 0, nrois-1 do begin

			; get the roi from the goigroup container
		roi = roigroup->get( position=i )

			; the position does not necessarily correspond to the roinumber
			; get that from the roi itself (roi objects extend IDLanRoi)

		roino = roi->id()

			; get the mask of the roi. getFramedMask would provide only
			; the outline - here we're interested in the full roi
			; npixels will hold the number of pixels in the roi
			; In the process of roi creation, it may happen that npixels equals
			; '0' - in this case we continue with the next roi
		mask  = where( *(roi->filledMask()) gt 0, npixels)
		if( npixels eq 0 ) then continue


			; calculate the mean value of image points in the roi
		meanval = mean( (*img)[mask] )

			; and display the result in the text widget
		widget_control, text, set_value=' ', /append
		widget_control, text, set_value=string( roino,   format='("Roi      :",i5)'), /append
		widget_control, text, set_value=string( npixels, format='("# Pixels :",i5)'), /append
		widget_control, text, set_value=string( meanval, format='("Mean     :",f7.1)'), /append

	end

end
