;+=----------------------------------------------------------------------------
; Name   : mrcontrast
; RCS    : $Id: mrcontrast__define.pro,v 2.2 2004/11/24 08:49:06 olivert Exp olivert $
; Methods: <b>Constructor</b>
;          mrContrast::init                : constructor, part II
;          <br>
;          <b>General</b>
;          mrContrast::display             : (re-)display contrast bars on screen
;          mrContrast::setSensitivity      : sensitivity to mouse movements
;          mrContrast::mode                : returns 0 for center/width changes, 1 for gamma
;          mrContrast::setMode             : set to 0 for center/width, to 1 for gamma changes
;          <br>
;          <b>Event handling</b>
;          mrContrast::selfregister        : register with a display object
;          mrContrast::unregister          : unregister
;          <br>
;          <b>Private</b>
;          mrContrast::reset               : reset contrast to full width and center
;          mrContrast::update              : update internal status from image information
;          mrContrast::mrdisplayEvtHandler : the event handler and broker for mr*display
;          mrContrast::imgChangeEvt        : handling image change events
;          mrContrast::buttonDblClickEvt   : handling of double clicks
;          mrContrast::buttonPressEvt      : handling of button presses
;          mrContrast::buttonReleaseEvt    : handling of button release
;          mrContrast::buttonMotionEvt     : handling of motion while button pressed
;          mrContrast__define              : constructor, part I
;          mrContrast::cleanup             : destructor
; Desc   : Plugin to be used with mrdisplay and mrmultidisplay to change
;          center/width or gamma value of the displayed images upon mouse movements.
;          Option /bordermode of the selfregister method allows to have both on
;          the same mouse button. Double click will reset C/W or gamma, respectively.
;          /bordermode doesn't really work well with mrmultidisplay - see below.
; Code  :
;         display  = obj_new('mrdisplay')
;         contrast = obj_new('mrcontrast')
;         contrast->selfregister, display, button=3
;         display->activate
; Author: (c) Oliver Thilmann, oliver.thilmann@web.de, Oct 2002
; Licns : This program is published under the GPL. See gpl.txt for details
;=-----------------------------------------------------------------------------


;==============================================================================
;
;                          CONSTRUCTORS AND DESTRUCTORS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrContrast::init
; Syntax: obj = obj_new('mrcontrast', parent, /vertical, /gamma, colbars=2, xsize=100, ysize=300 )
; Args  : <b>parent</b>
;         id of the parent widget. If not provided, no colorbars are displayed
;         <b>xsize</b>
;         size in x of the colorbar drawing window (default: 100)
;         <b>ysize</b>
;         size in y of the colorbar drawing window (default: 300)
;         <b>vertical</b>
;         If set, the colorbars are displayed vertically, default is horizontal
;         The default values for xsize and ysize are exchanged in that case.
;         <b>colbars</b>
;         Colorbars to display. Possible values:
;         0 : display constant range colorbar only
;         1 : display adaptive range colorbar only
;         2 : display both colorbars
;         <b>gamma</b>
;         If set, mouse movements don't change center/width, but the gamma value
;         <b>sensitivity</b>
;         Float value, describting the sensitivity to mouse movements. Default is 0. (=auto)
;         <b>color</b>
;         [r,g,b] integer array for the foreground color
;         <b>bgcolor</b>
;         [r,g,b] integer array for the background color
; Desc  : Initializes an mrcontrast object for use with mrdisplay or mrmultidisplay.
;         Reasonable sensitivity values lie around 1./canvasSize in pixels. By default
;         mrcontrast tries to set a reasonable value when registering with the
;         display object - this is however only possible if the window is already
;         mapped to the screen when registering.
;=-----------------------------------------------------------------------------
function mrContrast::init, parent, xsize=xsize, ysize=ysize, $
						   color=color, bgcolor=bgcolor, title=title, $
						   colbars=colbars, vertical=vertical, sensitivity=sensitivity, $
						   gamma=gamma

		; set some defaults and argument checks

	colbars   = ( n_elements(colbars) eq 0 ) ?      0  : colbars < 2
	mode 	  = keyword_set( gamma )
	baseid    = ( n_elements(parent)  eq 0 ) ?      0L : parent
	color     = ( n_elements(color)   eq 3 ) ? color   : [255,255,255]
    bgcolor   = ( n_elements(bgcolor) eq 3 ) ? bgcolor : [0,0,0]
	vertical  = keyword_set( vertical )

	if( n_elements(xsize) eq 0 ) then if( vertical eq 1 ) then xsize = 100 else xsize = 300
	if( n_elements(ysize) eq 0 ) then if( vertical eq 1 ) then ysize = 300 else ysize = 100

	sensitivity = ( n_elements(sensitivity) eq 0 ) ? 0. : sensitivity

		; set parameters and defaults

	self.xsize       = xsize
	self.ysize       = ysize
	self.wid_base    = baseid
	self.wid_draw    = 0
	self.color       = color
	self.bgcolor     = bgcolor
	self.center		 = 0.5
	self.width       = 1.0
	self.gamma       = 1.
	self.mode        = mode
	self.sensitivity = sensitivity

		; There's two ways to display the current contrast setting:
		; 1) colorbar center and width are displayed in the mr*display window (always)
		; 2) one or two colorbars visualize the current settings in a separate widget.

	if( self.wid_base ne 0 ) then begin
		self.wid_draw = widget_draw( self.wid_base, xsize=self.xsize, ysize=self.ysize, $
						x_scroll_size=self.xsize, y_scroll_size=self.ysize, $
						graphics_level=2, renderer=1, retain=2 )
		widget_control, self.wid_draw, set_uvalue=self   ; so far unused, but what the heck...

			; default palette is greyscale. During selfregister and
			; update, the palette is taken from the display object itself

		ct = indgen( 256 )
		self.palette    = obj_new('IDLgrPalette', ct, ct, ct)
		self.grView     = obj_new('IDLgrView', viewplane_rect=[0,0,1,1], color=self.bgcolor )

			; colbar 1 displays full range
			; colbar 2 displays only the current range (i.e. axis values change)

			; position the colorbars in the widget - width and height of one colorbar are
			; always '1'. There's a bug in vcolorbar - so do not set the position via
			; the set_properties command (Might be David fixed that in the latest
			; version - did not check)

		if( vertical eq 1 ) then begin
			self.grColbar1  = obj_new('vcolorbar', color=self.color, $
								position=[0.7,0.05,0.90,0.95], range=[0,1], palette=self.palette )
			self.grColbar2  = obj_new('vcolorbar', color=self.color, $
								position=[1.7,0.05,1.90,0.95], range=[0,1], palette=self.palette )
		end else begin
			self.grColbar1  = obj_new('hcolorbar', color=self.color, $
								position=[0.15,0.35,1.85,0.65], range=[0,1], palette=self.palette )
			self.grColbar2  = obj_new('hcolorbar', color=self.color, $
								position=[0.15,1.35,1.85,1.65], range=[0,1], palette=self.palette )
		end

		self.grView->add, self.grColbar1
		self.grView->add, self.grColbar2

			; colbars = 0: absolute bar only, 1: relative bar only, 2: both bars

		if( vertical eq 1 ) then begin
			case colbars of
			0: self.grView->setProperty, viewplane_rect=[0,0,1,1]
			1: self.grView->setProperty, viewplane_rect=[1,0,1,1]
			2: self.grView->setProperty, viewplane_rect=[0,0,2,1]
			end
		end else begin
			case colbars of
			0: self.grView->setProperty, viewplane_rect=[0,0,2,1]
			1: self.grView->setProperty, viewplane_rect=[0,1,2,1]
			2: self.grView->setProperty, viewplane_rect=[0,0,2,2]
			end
		end
	end

		; the following is for displaying contrast values in the user widget
		; as of now, this will not be used if wid_base is set, but we create it anyway...

	self.grDModel = obj_new('IDLgrModel', hide=1)
	self.grDText  = obj_new('IDLgrText', locations=[0.5,0.01], color=[0,255,0],$
							char_dimensions=[.025,.025], alignment=0.5 )
	self.grDModel->add, self.grDText

	return, 1
end

;+=----------------------------------------------------------------------------
; Name : mrContrast::cleanup
; Type : Private Procedure
; Desc : the destructor
;=-----------------------------------------------------------------------------
pro mrContrast::cleanup
	if( self.display ne obj_new() ) then self->unregister
	obj_destroy, self.grView
	obj_destroy, self.grColbar1
	obj_destroy, self.grColbar2
	obj_destroy, self.palette
	obj_destroy, self.grDModel
	obj_destroy, self.grDText
end

;==============================================================================
;
;                               GET AND SET METHODS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name : mrContrast::setSensitivity
; Type : Procedure
; Args : float sensitivity
; Desc : How sensitive should the contrast react to mouse movements? Reasonable
;        values lie around 1./canvas size in pixels
;        A value of '0.' sets sensitivity to 1./min(canvas size in x and y)
;=-----------------------------------------------------------------------------
pro mrContrast::setSensitivity, sensitivity
	if( sensitivity eq 0. ) then begin
		win = self.display->grwindow()
		win->getProperty, dimensions=res
		self.sensitivity = 1./min(res)
	end else begin
		self.sensitivity = sensitivity
	end
end

;+=----------------------------------------------------------------------------
; Name  : mrContrast::mode
; Syntax: mode = obj->mode()
; Ret   : '0', if mouse movement changes center/width, '1' for gamma change
;=-----------------------------------------------------------------------------
function mrContrast::mode
	return, self.mode
end

;+=----------------------------------------------------------------------------
; Name  : mrContrast::setMode
; Syntax: mode = obj->mode()
; Ret   : '0', if mouse movement changes center/width, '1' for gamma change
;         '2', for C/W in center of screen and gamma when mouse is in right 15%
;=-----------------------------------------------------------------------------
pro mrContrast::setMode, mode
	on_error, 2
	if( n_elements(mode) ne 1 ) then message, 'Invalid or missing argument'
	self.mode = mode < 2
end



;+=----------------------------------------------------------------------------
; Name : mrContrast::reset
; Type : Procedure
; Desc : reset contrast to full width and center center. Will cause a call
;        to mr*display->display and redisplay mrcontrast colorbars.
;=-----------------------------------------------------------------------------
pro mrContrast::reset
	self.center = .5
	self.width  = 1.
	self.gamma  = 1.
	if( not obj_valid( self.display ) ) then return
	self.display->setContrast, 0.5, 1.0
	self.display->setGamma, 1.
	self.display->display
	self->update
	self->display
	return
end


;==============================================================================
;
;                             PUBLIC OBJECT FUNCTIONS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name : mrContrast::display
; Type : Procedure
; Desc : Redraw the colorbars
;=-----------------------------------------------------------------------------
pro mrContrast::display
	if( self.wid_draw eq 0L ) then return
	widget_control, self.wid_draw, get_value=grWindow
	if( grWindow eq obj_new() ) then return
	grWindow->draw, self.grView
end


;+=----------------------------------------------------------------------------
; Name : mrContrast::update
; Type : Procedure
; Desc : Update the color bar scale values according to stored center and
;        width values. Called from image and palette change events.
;        Does not redisplay!
;=-----------------------------------------------------------------------------
pro mrContrast::update
	if( self.display eq obj_new() ) then return

	contrast    = self.display->contrast()
	self.center = contrast[0]
	self.width  = contrast[1]
	contr_low   = self.center-self.width/2. > 0.
	contr_hig   = self.center+self.width/2. < 1.

	self.gamma  = self.display->gamma()

		; get palette scaling range - whether we display relative or absolute
		; values depends on the behaviour of the mr*display object

	imgmax  = self.display->imgMaxval()
	imgmin  = self.display->imgMinval()
	imglow  = imgmin + (imgmax-imgmin)*contr_low
	imghigh = imgmin + (imgmax-imgmin)*contr_hig
 	width   = self.width * (imgmax-imgmin)
	center  = self.center * (imgmax-imgmin)

		; update colorbars (if they exist) - get the palette each time: the
		; user might have changed that in the meantime!
		; (this was before changePalette events were introduced in mrdisplay)

	if( self.wid_draw ne 0L ) then begin
		palette = self.display->palette()
		palette->getProperty, red=red, green=green, blue=blue
		self.palette->setProperty, red=red, green=green, blue=blue, gamma=self.gamma
		self.grColbar1->setProperty, range=[imgmin, imgmax]
		self.grColbar1->clamp, [imglow, imghigh]
		self.grColbar2->setProperty, range=[imglow,imghigh]
	end
	case self.action of
		0: str = ''
		1: str = 'W '+ strtrim(fix(width),2) + '  C ' + strtrim(fix(center),2)
		2: str = 'Gamma '+ strtrim(string(self.gamma, format='(f4.2)'),2)
	end
	self.grDText->setProperty, strings = str
end

;==============================================================================
;
;                                  EVENT HANDLING
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrContrast::selfregister
; Syntax: obj->selfregister, display, button=int, /bordermode
; Args  : <b>display</b>
;         An object of class mrdisplay or mrmultidisplay with which to register
;         <b>button</b>
;         The button number to use: left=1, middle=2, right=3 (def)
;         <b>bordermode</b>
;         If set, beginning a mouse button in the right 15% of the display object
;         will cause a change in gamma, otherwise center and width are changed.
;         Unluckily the 15% refer to the <b>total</b> canvas size, which might be
;         considerably larger than the displayed canvas size. I.e. when using
;         scroll bars, /bordermode might not be such a good option.
; Desc :  The contrast objects signs in for Button-press, -release and -motion
;         events with the specified display object.
;         To switch off interactive contrast change, set button to '0'.
;=-----------------------------------------------------------------------------
pro mrContrast::selfregister, display, button=button, bordermode=bordermode
	if( n_elements(button) eq 0 ) then button = 3
	if( button lt 0 ) then begin
		print, 'Illegal button number in mrContrast::selfregister'
		return
	end
	if( keyword_set(bordermode) ) then self.mode=2
	display->registerEvt, self, $
		['Press', 'DblClick', 'Release','Motion','ImgChange','PaletteChange']
	self.display = display
	self.button  = 2^(button-1)        ; 2^-1 = 0 in integer arithmetics

		; register our textmodel with the display object
		; we remove it before, so the user may call self-registration
		; twice without hurting us

	(display->labelView())->remove, self.grDModel
	(display->labelView())->add, self.grDModel

		; set text size depending on canvas size

	win = self.display->grwindow()
	if( win eq obj_new() ) then begin
		res = 300
	end else begin
		win->getProperty, dimensions=res
	end
		charsize = 13./min(res)
	self.grDText->setProperty, char_dimensions=[charsize,charsize]

		; set sensitivity if not done by the user

	if( self.sensitivity eq 0. ) then self.sensitivity = 1./min(res)

		; reset object, this includes redisplay

	self->reset
	return
end

;+=----------------------------------------------------------------------------
; Name : mrContrast::unregister
; Desc : unregisters with a mr*display object. It is a good idea to unregister
;        before destroying the mr*display object.
;=-----------------------------------------------------------------------------
pro mrContrast::unregister
	if( not obj_valid(self.display) ) then return
	self.display->unregisterObj, self
	(self.display->labelView())->remove, self.grDModel
	self.display = obj_new()
end


;+=----------------------------------------------------------------------------
; Name : mrContrast::mrdisplayEvtHandler
; Args : *mrdisplayEvt event
; Desc : Event dispatcher
;=-----------------------------------------------------------------------------
pro mrContrast::mrdisplayEvtHandler, evt
		; image change and palette change events
	if( (*evt).type eq 5 or (*evt).type eq 6 ) then begin
		self->imgChangeEvt, evt
		return
	end
		; all other are button events
	if( self.button eq 0 ) then return                 ; no user interaction
	if( ((*evt).button and self.button) ne self.button ) then return  ; wrong button
	case (*evt).type of
		0:	self->buttonPressEvt,    evt
		1:	self->buttonDblClickEvt, evt
		2:	self->buttonReleaseEvt,  evt
		3:	self->buttonMotionEvt,   evt
		else: return
	end
end

;+=----------------------------------------------------------------------------
; Name : mrContrast::imgChangeEvt
; Args : mrdisplayEvt *evt : pointer to the mrdisplay event structure
; Calls: mrContrast::update
;        mrContrast::display
; Desc : Event handler for image change events.
;=-----------------------------------------------------------------------------
pro mrContrast::imgChangeEvt, evt
	self->update
	self->display
end

;+=----------------------------------------------------------------------------
; Name : mrContrast::buttonDblClickEvt
; Args : mrdisplayEvt *evt : pointer to the mrdisplay event structure
; Calls: mrContrast::reset
; Desc : Event handler for button double click events resets center and width
;        or gamma, depending on mode and (for mode = 2) mouse position
;=-----------------------------------------------------------------------------
pro mrContrast::buttonDblClickEvt, evt
	case self.mode of
		0: rmode = 0
		1: rmode = 1
		2: if( (*evt).x gt 0.85*self.display->screenXsize() ) then rmode=1 else rmode=0
	end
	if( rmode eq 0 ) then begin						; reset C/W
		self.center = .5
		self.width  = 1.
		self.display->setContrast, 0.5, 1.0
	end else begin									; reset gamma
		self.gamma = 1.
		self.display->setGamma, 1.
	end
	self.display->requestDisplay
	self->update
	self->display
end

;+=----------------------------------------------------------------------------
; Name : mrContrast::buttonPressEvt
; Args : mrdisplayEvt *evt : pointer to the mrdisplay event structure
; Calls: mrContrast::update
;        mrContrast::display
;        mrdisplay::display
; Desc : Event handler for button press events
;        Displays the current contrast scaling vals
;=-----------------------------------------------------------------------------
pro mrContrast::buttonPressEvt, evt
	case self.mode of
		0: self.action = 1
		1: self.action = 2
		2: if( (*evt).x gt 0.85*self.display->screenXsize() ) then self.action=2 else self.action=1
	end
	self.grDModel->setProperty, hide=0
	self->update
	self.display->requestDisplay
	self->display
end

;+=----------------------------------------------------------------------------
; Name : mrContrast::buttonReleaseEvt
; Args : mrdisplayEvt *evt : pointer to the mrdisplay event structure
; Desc : Event handler for button release events
;        Remove center/width values from display object
;=-----------------------------------------------------------------------------
pro mrContrast::buttonReleaseEvt, evt
	self.action=0
	self.grDModel->setProperty, hide=1
	self.display->requestDisplay
end

;+=----------------------------------------------------------------------------
; Name : mrContrast::buttonMotionEvt
; Args : mrdisplayEvt *evt : pointer to the mrdisplay event structure
; Calls: mrContrast::update
; Desc : Event handler for motion events while button is pressed
;        A bit unclean right now as we use low/high and center/width
;        simultaneously - this part needs some cleanup. Later.
;=-----------------------------------------------------------------------------
pro mrContrast::buttonMotionEvt, evt

		; recalculate and set color bar scalings and set grDText strings

	dx = (*evt).dx
	dy = (*evt).dy

	; if( abs(dx) le 0 ) then dx = 0
	; if( abs(dy) le 0 ) then dy = 0
	; we don't return here, even if both dx and dy are zero -
	; the routine would be "too fast" and above threshold would not work.
	; A better solution would probably mean creating some timestamp of the
	; last moves and do some integration or set the threshold not on relative
	; but absolute cursor displacements - or subtract a threshold from the
	; movement??? Needs some experimenting

	case self.action of
		0: ; do nothing
		1: begin								   ; change center and width
			contrast    = self.display->contrast()
			self.center = contrast[0]
			self.width  = contrast[1]

			self.center =               0. > (contrast[0] + dy * self.sensitivity) < 1.
			self.width  = self.sensitivity > (contrast[1] + dx * self.sensitivity) < 1.

				; get the display's image palette, load new table and redisplay

			(*evt).display->setContrast, self.center, self.width
		end
		2: begin									; change gamma
			gamma        = self.display->gamma()
			self.gamma   = 0.1 > (self.gamma + (dx+dy)*self.sensitivity ) < 10.
			self.display->setGamma, self.gamma
		end
	end

	self.display->requestDisplay

		; redraw contrast colorbars using the new palette (*if* we have one)

	self->update
	self->display

end


;==============================================================================
;
;                          PRIVATE OBJECT FUNCTIONS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name : mrContrast__define
; Type : Private procedure
; Desc : The constructor
;=-----------------------------------------------------------------------------
pro mrcontrast__define
	obj = { MRCONTRAST       		  ,$
			display     :	obj_new() ,$ ; the display object we're associated with
			button		:			0 ,$ ; the button on which to react
			center		:			0.,$ ; relative center of the contrast
			width		:			0.,$ ; relative width of the contrast
			gamma       :			0.,$ ; current gamma value
			mode		:           0 ,$ ; 0: C/W, 1: gamma, 2:border
			action		:			0 ,$ ; currently in action? 0: no 1:C/W, 2: gamma
			sensitivity :           0.,$ ; change in low/high for each screen pixel moved
			wid_base    :	0L        ,$ ; widget id of the base widget
			wid_draw    :	0L        ,$ ; widget id of the draw widget
			color       :	intarr(3) ,$ ; color for text and borders
			bgcolor     :	intarr(3) ,$ ; background border for the drawing widget
			grView      :	obj_new() ,$ ; the view - needed for cleanup and drawing
			grColbar1   :	obj_new() ,$ ; the colorbar displaying the full bar
			grColbar2   :	obj_new() ,$ ; the colorbar displaying from min to max
			palette     :	obj_new() ,$ ; palette for both the image and the two colorbars
			xsize       :	0         ,$ ; its size in x
			ysize       :	0         ,$ ; and its size in y
			grDModel    :	obj_new() ,$ ; model and
			grDText     :	obj_new()  $ ; text for display in mr*display object
		  }
end
