;+=----------------------------------------------------------------------------
; Name   : mrdicom
; RCS    : $Id: mrdicom__define.pro,v 2.8 2004/11/25 14:42:42 olivert Exp olivert $
; Methods:
;          <b>Constructors and destructor</b>
;          mrdicom::init               : the constructor part II
;          mrdicom_setUIDs             : must be called once before instanciating the first object
;          <br>
;          <b>Public variables (get)</b>
;          mrdicom::img                : return pointer to image data
;          mrdicom::errno              : returns error number of last write (0 if ok)
;          mrdicom::errMsg             : returns error message of last read or write ('' if ok)
;          mrdicom::filename           : returns the full filename associated with this object
;          <br>
;          <b>Public functions and procedures</b>
;          mrdicom::setImg             : set image data to be written
;          mrdicom::getTagValue        : get the value of a dicom header field by name
;          mrdicom::setTagValue        : set the value of a dicom header field
;          mrdicom::anonymize          : anonymize patient related data in the header
;          mrdicom::copy               : copy all dicom info (except image data) from some other object
;          mrdicom::write              : write the new dicom file to disc
;          mrdicom::generateUID        : generate unique ids for the new header
;          <br>
;          <b>Private</b>
;          mrdicom::fillCache          : fill the local structure from the dicom header
;          mrdicom::setBasicTags       : sets some std dicom tags - used by the constructor
;          mrdicom::getbytes           : private function for big/little endian handling
;          mrdicom::setVRtag           : sets data type for a certain header data element
;          mrdicom::generate_pixeltag  : private function, used when writing data
;          mrdicom::generate_anytag    : private function, used when writing data
;          mrdicom__define             : the constructor
;          mrdicom::cleanup            : the destructor
; Desc   : An object extending IDLffDicom to provide
;          - writing of dicom files
;          - direct access to user data
;          <br>
;          <b>Unique identifiers</b>
;          Please see the documentation of mrdicom_setUIDs for a list of all UIDs and how to
;          set and use them.
;          <br>
;          <b>Restrictions:</b>
;          mrdicom only supports dicom default transfer syntax which is
;          Dicom implicit VR little endian, UID = 1.2.840.1008.1.2
;          &nbsp;
;          There's no support for multi-valued tags (like e.g. ACQ aquisition matrix)
;          &nbsp;
;          Concerning getTagValue/setTagValue:
;          As of now mrdicom does not support sequences (and probably never will).
;          All data stored in sequences is silently ignored - if you need it, you
;          will have to access the data directly via getValue(group/element) and
;          select the ones with parent != -1.
;          &nbsp;
;          A word on Dicom data types:
;          Data elemens (DE) of type 1 are mandatory and must be present with a valid value,
;          DEs of type 2 must be present, but the value may be empty.
;          DEs of type 1C and 2C are mandatory in the above sense under certain conditions.
;          mrDicom does not know about such conditions and thus cannot check them. DEs of
;          types 1C and 2C are not written, if their value is empty - it is therefore the user's
;          responsibility to provide a value, if the conditions require the DE to be present.
;          &nbsp;
;          According to my printout of the DICOM3.0 standard, the VR type 'DS' is limited
;          to 16 bytes. However the latest Siemens software writes (at least for tag x20,x37)
;          more than 100 bytes. Since no other software seems to complain, we set no limit
;          on the size of elements of VR type 'DS'.
;          That's the nice thing about standards - there's so many ways of interpreting them.
;          &nbsp;
;          For lack of complete documentation (and time) only the MR-relevant DICOM
;          tags are covered. Here's the full list (also see mrdicom::setTagValue).
;          The two tags marked 'U' are not implemented and only listed for completeness.
;          With the standard at hand, it should be easy to add tags as needed.
;          <br>
;   GROUP  ELEMENT TYPE   DESCRIPTION                  FORMAT    INTERNAL VARIABLE
; ================================================================================
;   0002x  0000x  META Group Length                      UL  m_GroupLength
;   0002x  0001x  META File Meta Information Version     OB  m_FileMetaInfoVersion
;   0002x  0002x  META Media Stored SOP Class UID        UI  m_MediaStoredSOPClassUID
;   0002x  0003x  META Media Stored SOP Instance UID     UI  m_MediaStoredSOPInstanceUID
;   0002x  0010x  META Transfer Syntax UID               UI  m_TransferSyntaxUID
;   0002x  0012x  META Implementation Class UID          UI  m_ImplementationClassUID
;   0002x  0013x  META Implementation Version Name       SH  m_ImplementationVersionName
;   0002x  0016x  META Source application entity title   AE  m_SourceApplicationEntityTitle
;   0008x  0005x  ID Specific Character Set              CS  i_CharacterSet
;   0008x  0008x  ID Image Type                          CS  i_ImageType
;   0008x  0016x  ID SOP Class UID                       UI  i_SOPClassUID
;   0008x  0018x  ID SOP Instance UID                    UI  i_SOPInstanceUID
;   0008x  0020x  ID Study Date                          DA  i_StudyDate
;   0008x  0021x  ID Series Date                         DA  i_SeriesDate
;   0008x  0022x  ID Acquisition Date                    DA  i_AcquisitionDate
;   0008x  0023x  ID Image Date                          DA  i_ImageDate
;   0008x  0030x  ID Study Time                          TM  i_StudyTime
;   0008x  0031x  ID Series Time                         TM  i_SeriesTime
;   0008x  0032x  ID Acquisition Time                    TM  i_AcquisitionTime
;   0008x  0033x  ID Image Time                          TM  i_ImageTime
;   0008x  0050x  ID Accession Number                    SH  i_AccessionNumber
;   0008x  0060x  ID Modality                            CS  i_Modality
;   0008x  0070x  ID Manufacturer                        LO  i_Manufacturer
;   0008x  0080x  ID Institution Name                    LO  i_InstitutionName
;   0008x  0081x  ID Institution Address                 ST  i_InstitutionAddress
;   0008x  0090x  ID Referring Physician's Name          LO  i_PhysiciansName
;   0008x  1010x  ID Station Name                        SH  i_StationName
;   0008x  1030x  ID Study Description                   LO  i_StudyDescription
;   0008x  103Ex  ID Series Description                  LO  i_SeriesDescription
;   0008x  1070x  ID Operator's Name                     PN  i_OperatorsName
;   0008x  1090x  ID Manufacturer Model Name             LO  i_ManufacturerModelName
;U  0008x  1111x  ID Referenced Study Component Sequence SQ  i_RefStudyComponentSequence
;   0008x  1150x  ID Referenced SOP Class UID            UI  i_RefSOPClassUID
;   0008x  1155x  ID Referenced SOP Instance UID         UI  i_RefSOPInstanceUID
;U  0008x  1140x  ID Referenced Image Sequence           SQ  i_RefImageSequence
;   0010x  0010x  PAT Patient Name                       LO  p_PatientName
;   0010x  0020x  PAT Patient ID                         LO  p_PatientId
;   0010x  0030x  PAT Patient Birthdate                  DA  p_PatientBirthdate
;   0010x  0040x  PAT Patient Sex                        CS  p_PatientSex
;   0010x  1010x  PAT Patient Age                        AS  p_PatientAge
;   0010x  1030x  PAT Patient Weight                     DS  p_PatientWeight
;   0010x  2000x  PAT Medical Alerts                     LO  p_PatientAlerts
;   0010x  21C0x  PAT Pregnancy Status                   US  p_PregnancyStatus
;   0010x  4000x  PAT Patient Comments                   LO  p_PatientComments
;   0018x  0010x  ACQ Contrast/Bolus Agent               LO  a_ContrastBolusAgent
;   0018x  0015x  ACQ Body Part Examined                 CS  a_BodyPartExamined
;   0018x  0020x  ACQ Scanning Sequence                  CS  a_ScanningSequence
;   0018x  0021x  ACQ Sequence Variant                   CS  a_SequenceVariant
;   0018x  0022x  ACQ Scan Options                       CS  a_ScanOptions
;   0018x  0023x  ACQ MR Acquisition Type                CS  a_AcquisitionType
;   0018x  0024x  ACQ Sequence Name                      SH  a_SequenceName
;   0018x  0025x  ACQ Angio Flag                         CS  a_AngioFlag
;   0018x  0050x  ACQ Slice Thickness                    DS  a_SliceThickness
;   0018x  0080x  ACQ Repetition Time                    DS  a_RepetitionTime
;   0018x  0081x  ACQ Echo Time                          DS  a_EchoTime
;   0018x  0082x  ACQ Inversion Time                     DS  a_InversionTime
;   0018x  0083x  ACQ Number of Averages                 DS  a_NumberOfAverages
;   0018x  0084x  ACQ Imaging Frequency                  DS  a_ImagingFrequency
;   0018x  0085x  ACQ Imaged Nucleus                     SH  a_ImagedNucleus
;   0018x  0086x  ACQ Echo Number                        IS  a_EchoNumber
;   0018x  0087x  ACQ Magnetic Field Strength            DS  a_MagneticFieldStrength
;   0018x  0088x  ACQ Spacing Between Slices             DS  a_SpacingBetweenSlices
;   0018x  0089x  ACQ Number of Phase Encoding Steps     IS  a_NumberOfPhaseEncodingSteps
;   0018x  0091x  ACQ Echo Train Length                  IS  a_EchoTrainLength
;   0018x  0093x  ACQ Percent Sampling                   DS  a_PercentSampling
;   0018x  0094x  ACQ Percent Phase Field of View        DS  a_PercentPhaseFieldOfView
;   0018x  0095x  ACQ Pixel Bandwidth                    DS  a_PixelBandwidth
;   0018x  1000x  ACQ Device Serial Number               LO  a_DeviceSerialNumber
;   0018x  1020x  ACQ Software Version                   LO  a_SoftwareVersion
;   0018x  1030x  ACQ Protocol Name                      LO  a_ProtocolName
;   0018x  1041x  ACQ Contrast/Bolus Volume              DS  a_ContrastBolusVolume
;   0018x  1200x  ACQ Date of Last Calibration           DA  a_DateOfLastCalibration
;   0018x  1201x  ACQ Time of Last Calibration           TM  a_TimeOfLastCalibration
;   0018x  1250x  ACQ Receiving Coil                     SH  a_ReceivingCoil
;   0018x  1251x  ACQ Transmitting Coil                  SH  a_TransmittingCoil
;   0018x  1312x  ACQ Phase Encoding Direction           CS  a_PhaseEncodingDirection
;   0018x  1314x  ACQ Flip Angle                         DS  a_FlipAngle
;   0018x  1315x  ACQ Variable Flip Angle                CS  a_VariableFlipAngle
;   0018x  1316x  ACQ SAR                                DS  a_SAR
;   0018x  1318x  ACQ DB/DT                              DS  a_DBDT
;   0018x  5100x  ACQ Patient Position                   CS  a_PatientPosition
;   0020x  000Dx  REL Study Instance UID                 UI  r_StudyInstanceUID
;   0020x  000Ex  REL Series Instance UID                UI  r_SeriesInstanceUID
;   0020x  0010x  REL Study ID                           SH  r_StudyID
;   0020x  0011x  REL Series Number                      IS  r_SeriesNumber
;   0020x  0012x  REL Acquisition Number                 IS  r_AcquisitionNumber
;   0020x  0013x  REL Image Number                       IS  r_ImageNumber
;   0020x  0030x  REL Image Position                     RT  r_ImagePosition
;   0020x  0032x  REL Image Position Patient             DS  r_ImagePositionPatient
;   0020x  0035x  REL Image Orientation                  RT  r_ImageOrientation
;   0020x  0037x  REL Image Orientation (Patient)        DS  r_ImageOrientPatient
;   0020x  0050x  REL Location (RET)                     RT  r_Location
;   0020x  0052x  REL Frame of Reference UID             UI  r_FrameOfReference
;   0020x  0070x  REL Image Geometry Type (RET)          RT  r_ImageGeometryType
;   0020x  0080x  REL Masking Image (RET)                RT  r_MaskingImage
;   0020x  1001x  REL Acquisitions in Series             RT  r_AcqusitionInSeries
;   0020x  1040x  REL Position Reference Indicator       LO  r_PositionReferenceIndicator
;   0020x  1041x  REL Slice Location                     DS  r_SliceLocation
;   0020x  4000x  REL Image Comments                     LT  r_ImageComments
;   0028x  0002x  IMG Samples Per Pixel                  US  g_SamplesPerPixel
;   0028x  0004x  IMG Photometric Interpretation         CS  g_PhotometricInterpretation
;   0028x  0005x  IMG Image Dimensions (RET)             RT  g_ImageDimension
;   0029x  0006x  IMG Planar Configuration               US  g_PlanarConfiguration
;   0028x  0010x  IMG Rows                               US  g_Rows
;   0028x  0011x  IMG Columns                            US  g_Columns
;   0028x  0030x  IMG Pixel Spacing                      DS  g_PixelSpacing
;   0028x  0040x  IMG Image Format (RET)                 RT  g_ImageFormat
;   0028x  0100x  IMG Bits Allocated                     US  g_BitsAllocated
;   0028x  0101x  IMG Bits Stored                        US  g_BitsStored
;   0028x  0102x  IMG High Bit                           US  g_HighBit
;   0028x  0103x  IMG Pixel Representation               US  g_PixelRepresentation
;   0028x  0106x  IMG Smallest Image Pixel Value         US  g_SmallestImagePixelValue
;   0028x  0107x  IMG Largest Image Pixel Value          US  g_LargestImagePixelValue
;   0028x  1055x  IMG Window Center & Width Explanation  LO  g_WindowCenterWidthExplanation
;   0028x  0200x  IMG Image Location                     RT  g_ImageLocation
;   0028x  1050x  IMG Window Center                      DS  g_WindowCenter
;   0028x  1051x  IMG Window Width                       DS  g_WindowWidth
;   7FE0x  0010x  PXL Pixel Data                         OW  (none) ... pixel data
;
; Author: (c) Oliver Thilmann, Feb 03
;         Some of the writing code (generateAnyTag, setVRtag, getbytes) has
;         been taken from Bhautik Joshi's dicom_write - please see his copyright
;         and disclaimer at the end of the source file.
; Licns : This program is published under the GPL. See gpl.txt for details
;=-----------------------------------------------------------------------------


; Not as ugly a code any more as it used to be, but the frequent repetitions
; of the full list of tags is still not very elegant.
; But to speak with Bart Simpson: "I didn't do it", and besides that it works.


;==============================================================================
;
;                          CONSTRUCTORS AND DESTRUCTORS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrdicom::init
; Syntax: obj = obj_new('mrdicom', filename)
; Args  : <b>filename</b>
;         Optional. Specifies a DICOM3.0 file. If set, the file is read upon
;         instanciation via IDLffDicom->read. If reading fails, the object is not
;         created and obj_new will return obj_new()
; Desc  : Instanciates an mrdicom object.
;         <b>Important</b>
;         You have to call mrdicom_setUIDs before instanciating the first mrdicom
;         object, or instanciation will fail.
;=-----------------------------------------------------------------------------
function mrdicom::init, filename, _Extra=e

		; Our own pathetic little attempt to create a class variable in IDL
		; Initialized in mrdicom_setUIDs

	common mrdicomv16, uniqid, implementationClassUID, SopClassUID

	if( n_elements(implementationClassUID) eq 0 ) then message, 'implementation class UID not set - call mrdicom_setUIDs'
	if( n_elements(SopClassUID )           eq 0 ) then message, 'SOP class UID not set - call mrdicom_setUIDs'

	if( keyword_set(filename) ) then begin
	 	if (self->IDLffDicom::Init(_EXTRA=e) ne 1) then return, 0
	 	x = self->IdlffDicom::read( filename )       ; 1 if ok, 0 if error
	 	if( x eq 0 ) then return, 0
	 	self.filename = filename
	end else begin
	 	if (self->IDLffDicom::Init(_EXTRA=e) ne 1) then return, 0
		self->setBasicTags
	end
	self.validCache = 0                                        ; the internal 'cache' is empty

		; who wrote this file?

	self.implementationVersionName  = 'mrDicom-v2.0'       ; max 16 chars

	return, 1
end

;+=----------------------------------------------------------------------------
; Name  : mrdicom_setUIDs
; Syntax: mrdicom_setUIDs, imageType, implementationClassUID
; Args  : <b>imageType</b>
;         One of 'mrimg', 'mrspec', 'ctimg' or 'petimg'. Defines storedSOPClassUID
;         Default is 'mrimg'. All mrdicom objects within a program share the same
;         imageType. As of now, this shouldn't hurt much, since only MR image tags
;         are implemented anyway.
;         Ok, even setting the image type is a blatant overstatement :)<br>
;         <b>implementationClassUID</b>
;         A string - see below.
; Desc  : This 'class' method needs to be called exactly once in the program
;         before any instanciation of an mrDicom object as it sets a couple of UIDs
;         which are specific to your installation and the program. Unless these
;         variables have been set, all calls to mrdicom::init will fail.
;         <b>UIDs</b>
;         Dicom data contains a couple of unique identifiers (UIDs). Some of them
;         are predefined (e.g. the imageType argument is 'translated' into such a UID),
;         by a central organization. Three of these identifiers are set here and
;         used throughout the rest of the code. Some more are set or generated elsewhere.
;         <b>SOP class UID</b>
;         This identifier denotes the type of data written (MR images, SPEC images,
;         and so forth) and is derived from the imageType argument above. The
;         identifier for MR images e.g. is 1.2.840.10008.5.1.4.1.1.4
;         <b>Implementation class UID</b>
;         This is a unique identifier for the application which writes the file.
;         As this library has no own UID, it is suggested to use either the one
;         of the machine that generated the images (works for me and our PACS
;         system) or '0.0.000.00000.0.0.0.0.0.0' (though I have no whether the latter
;         is ok, and it's a bad idea anyway - see below).
;         To obtain this number, load a dicom file via IDLffDicom and dumpElements.
;         META Implementation Class UID (tag 0002,0012) contains the value to use.
;         <b>SOP instance UID</b>
;         A unique ID for each single file written. This is automatically created
;         in mrdicom::write and uses the implementation class UID as a substring.
;         That's why it is a bad idea to use the '0.0....' option - the UIDs would
;         not be unique anymore.
;         <b>Study SOP instance UID</b>
;         A unique ID for a study. It should be ok to use the same study UID for
;         derived images. If you want to create a new study UID, use mrdicom::generateUID('study')
;         <b>Series instance UID</b>
;         A unique identifier for the series. When writing data with a new series
;         number, create a new series UID via mrdicom::generateUID('series')
;         <b>Implementation version name</b>
;         A string holding information about the program that wrote the file.
;         This is internally set to something like 'mrDicom-v2.0' (or whatever version)
;         &nbsp;
;         Honestly, I'm not 100% confident of all these UIDs, though it works for
;         me. Nevertheless - if someone should know about a document that explains all
;         this service-object pair klimbim and has less pages than the 1200+ of the
;         dicom standard, I'd be thankful to hear about it.
;=-----------------------------------------------------------------------------
pro mrdicom_setUIDs, imageType, icUID, ssiUID

	common mrdicomv16, uniqid, implementationClassUID, SopClassUID

		; As of now, all mrdicom objects in a program share the same imageType,
		; (which is ok as only MR image tags are implemented anyway).
		; This could easily be changed, if needed.

		; set some defaults for now - these will go!

	implementationClassUID = (n_elements(icUID )    eq 1 ) ? icUID     : '0.0.00.0.0000.0.0'
	it                     = (n_elements(imageType) eq 1 ) ? imageType : 'mrimg'
	uniqid = 0

	case strlowcase(it) of
		'mrimg'  : SopClassUID  = '1.2.840.10008.5.1.4.1.1.4'
		'mrspec' : SopClassUID  = '1.2.840.10008.5.1.4.1.1.4.2'
		'ctimg'  : SopClassUID  = '1.2.840.10008.5.1.4.1.1.2'
		'petimg' : SopClassUID  = '1.2.840.10008.5.1.4.1.1.128'
		else     : begin
			message, 'Error: unsupported image type in mrdicom'
			return
		end
	end
end




;+=----------------------------------------------------------------------------
; Name : mrdicom::cleanup
; Desc : The destructor - free image data
;        <b>YAIF</b> (the sound I made when I learned that one or acronym
;        for 'yet another IDL feature'):
;        The cleanup method of any superclass is not called automatically
;        but needs to be called explicitly from the subclasses' destructor.
;=-----------------------------------------------------------------------------
pro mrdicom::cleanup
	self->IDLffDicom::cleanup
	if( ptr_valid(self.imgdata) ) then ptr_free, self.imgdata
end


;+=----------------------------------------------------------------------------
; Name  : mrdicom::setBasicTags
; Syntax: obj->setBasicTags
; Desc  : Fills the basic tags necessary for creating a DICOM3.0 file
;=-----------------------------------------------------------------------------
pro mrdicom::setBasicTags

	common mrdicomv16, uniqid, implementationClassUID, SopClassUID

		; first the essentials

	self.m_MediaStoredSOPClassUID    = SOPClassUID
	self.m_MediaStoredSOPInstanceUID = self->generateUID('image')
	self.m_TransferSyntaxUID         = '1.2.840.10008.1.2'
	self.m_ImplementationClassUID    = implementationClassUID
	self.m_ImplementationVersionName = self.implementationVersionName

		; second some basics

	time = string(BIN_DATE())

		; make each time field a two-digit number - starting with '0' if necessary
	if fix(time(0)) le 9 then  y = '0'+ strcompress(string(time(0)),/remove_all) else $
		y = strcompress(string(time(0)),/remove_all)
	if fix(time(1)) le 9 then  m = '0'+ strcompress(string(time(1)),/remove_all) else $
		m = strcompress(string(time(1)),/remove_all)
	if fix(time(2)) le 9 then  d = '0'+ strcompress(string(time(2)),/remove_all) else $
		d = strcompress(string(time(2)),/remove_all)
	date = y + '.' + m + '.' + d

	if fix(time(3)) le 9 then  hr = '0'+ strcompress(string(time(3)),/remove_all) else $
		hr = strcompress(string(time(3)),/remove_all)
	if fix(time(4)) le 9 then  minut = '0'+ strcompress(string(time(4)),/remove_all) else $
		minut = strcompress(string(time(4)),/remove_all)
	if fix(time(5)) le 9 then  sec = '0'+ strcompress(string(time(5)),/remove_all) else $
		sec = strcompress(string(time(5)),/remove_all)
	hour = hr + minut + sec

	self.i_ImageTime 				 = hour
	self.i_ImageDate 				 = date
	self.i_ImageType 				 = 'ORIGINAL\\PRIMARY\\OTHER'
	self.i_Modality	 				 = 'MR'
	self.i_Manufacturer 			 = 'NN'
	self.i_PhysiciansName			 = 'Internal'
	self.p_PatientName				 = 'Not Defined'
	self.p_PatientId				 = ' '
	self.p_PatientBirthdate		 	 = ' '
	self.p_PatientSex				 = 'O'
	self.a_AcquisitionType  		 = '2D'
	self.g_PhotometricInterpretation = 'MONOCHROME2'
	self.g_SamplesPerPixel			 = 1
	self.r_StudyID					 = 1000
	self.r_SeriesNumber				 = 1000
	self.r_ImageNumber	        	 = 1000
	self.r_AcquisitionNumber		 = 1000
	self.r_SliceLocation			 = 0.00
	self.g_PixelSpacing				 = '1\1'

	return
end


;+=----------------------------------------------------------------------------
; Name  : mrdicom::copy
; Syntax: obj->copy, source, studyUID=string, seriesUID=string
; Args  : <b>source</b>
;         An mrDicom object from which to copy all tag values
;         <b>studyUID</b>
;         A string holding a unique study id
;         <b>seriesUID</b>
;         A string holding a unique series id
; Desc  : Copy all tag values, except the image data, from another IDLffDicom or mrdicom
;         object.
;         <b>Unique ids</b>
;         studyUID and seriesUID are identifiers which have to be unique for
;         each study and series, respectively and must not be mixed up with the
;         study number or series number (s.b.)
;         By default, the unique ids are copied from the source dicom object.
;         If the user writes a new series or study, it is his responsibility to create
;         new unique ids and either set them via mrdicom::setTagValue (the
;         field names are rel_study_instance_uid and rel_series_instance_uid) or
;         provide them here as arguments. Suitable unique ids can be obtained
;         via mrdicom::generateUID
;         Do not confuse studyUID/seriesUID (i.e. REL_study/series_instance_UID)
;         with the fields study_id and series_number, which contain 'human readable'
;         ids - e.g. a typical value for series_number would be '6' while the
;         corresponding rel_study_series_UID might be 1.2.840.10008.5.1.4.1.1.4.1234.
;         However they go hand in hand, i.e. if you save to a new study_id, you should
;         also create a new studyUID (dito for series)
;         <b>Copying from a mrdicom object</b>
;         Please note that the copy function works on the underlying
;         dicom object directly and bypasses the mrdicom data cache used for writing.
;         This means that possible changes to data tags on an mrdicom object will not
;         be taken into account when copying tag values. !This should be changed!
;         Example:
;         1) User creates a mrdicom object and reads dicom data from file1
;         2) User then uses the setTagValue method to change some data
;         3) User creates a second mrdicom object
;         4) User copies header data from first mrdicom to second mrdicom object
;         Any changes done in 2) will not be reflected in 4) as setTagValue works
;         on an internal cache while header data is copied from the IDLffDicom object
;         directly. <b>Consequence:</b> This will be solved in a future version.
;         As for now you are discouraged to use IDLffDicom objects and mrdicom objects
;         in parallel. Always use mrdicom.
;=-----------------------------------------------------------------------------
pro mrdicom::copy, source, studyUID=studyUID, seriesUID=seriesUID

	refNr	 = source->GetReference()

	if( refNr[0] eq -1 ) then begin          ; invalid source
		self.validCache = 0                  ; with invalid source, cache is invalid, too
		return
	end

	group 	 = source->getgroup()
	element  = source->Getelement()
	VR       = source->GetVR()
	desc	 = source->GetDescription()
	length 	 = source->GetLength()
	valuePtr = source->Getvalue()

		; Still don't know what this string conversion is good for...

	for i = 0 , n_elements(refNr)-1 do begin
	if( self->getParent(i) ne -1 ) then continue    ; we only use non-sequence data!
	if( valuePtr[i] ne ptr_new() ) then begin
		if( n_elements(*valuePtr[i]) ne 1 ) then continue 	; *valuePtr[i] = string(*valuePtr[i])
		case group[i] of
		2: begin
			case element[i] of
			0:  self.m_GroupLength               = *ValuePtr[i]
			1:  self.m_FileMetaInfoVersion       = *ValuePtr[i]
			2:  self.m_MediaStoredSOPClassUID    = *ValuePtr[i]
			3:  self.m_MediaStoredSOPInstanceUID = *ValuePtr[i]
			16: self.m_TransferSyntaxUID         = *ValuePtr[i]
			18: self.m_ImplementationClassUID    = *ValuePtr[i]
			19: self.m_ImplementationVersionName = *ValuePtr[i]
			else:
			end
		end
		8: begin
			case element[i] of
			5:    self.i_CharacterSet          = *ValuePtr[i] ; x0005
			8:    self.i_ImageType             = *ValuePtr[i] ; x0008
			22:   self.i_SOPClassUID           = *ValuePtr[i] ; x0016
			24:   self.i_SOPInstanceUID        = *ValuePtr[i] ; x0018
			32:   self.i_StudyDate             = *ValuePtr[i] ; x0020
			33:   self.i_SeriesDate            = *ValuePtr[i] ; x0021
			34:   self.i_AcquisitionDate       = *ValuePtr[i] ; x0022
			35:   self.i_ImageDate             = *ValuePtr[i] ; x0023
			48:   self.i_StudyTime             = *ValuePtr[i] ; x0030
			49:   self.i_SeriesTime            = *ValuePtr[i] ; x0031
			50:   self.i_AcquisitionTime       = *ValuePtr[i] ; x0032
			51:   self.i_ImageTime             = *ValuePtr[i] ; x0033
			80:   self.i_AccessionNumber       = *ValuePtr[i] ; x0050
			96:   self.i_Modality              = *ValuePtr[i] ; x0060
			112:  self.i_Manufacturer          = *ValuePtr[i] ; x0070
			128:  self.i_InstitutionName       = *ValuePtr[i] ; x0080
			129:  self.i_InstitutionAddress    = *ValuePtr[i] ; x0081
			144:  self.i_PhysiciansName        = *ValuePtr[i] ; x0090
			4112: self.i_StationName           = *ValuePtr[i] ; x1010
			4144: self.i_StudyDescription      = *ValuePtr[i] ; x1030
			4158: self.i_SeriesDescription     = *ValuePtr[i] ; x103E
			4208: self.i_OperatorsName         = *ValuePtr[i] ; x1070
			4240: self.i_ManufacturerModelName = *ValuePtr[i] ; x1090
;			4369: self.i_RefStudyComponentSequence = *ValuePtr[i] ; x1111
			4432: self.i_RefSOPClassUID        = *ValuePtr[i] ; x1150
			4437: self.i_RefSOPInstanceUID     = *ValuePtr[i] ; x1155
;			4416: self.i_RefImageSequence      = *ValuePtr[i] ; x1140
			else:
			end
		end
		16: begin
			case element[i] of
			16:    self.p_PatientName      = *ValuePtr[i] ; x0010
			32:    self.p_PatientId        = *ValuePtr[i] ; x0020
			48:    self.p_PatientBirthdate = *ValuePtr[i] ; x0030
			64:    self.p_PatientSex       = *ValuePtr[i] ; x0040
			4112:  self.p_PatientAge       = *ValuePtr[i] ; x1010
			4144:  self.p_PatientWeight    = *ValuePtr[i] ; x1030
			8192:  self.p_PatientAlerts    = *ValuePtr[i] ; x2000
			8640:  self.p_PregnancyStatus  = *ValuePtr[i] ; x21C0
			16384: self.p_PatientComments  = *ValuePtr[i] ; x4000
			else:
			end
		end
		24: begin
			case element[i] of
			16:    self.a_ContrastBolusAgent         = *ValuePtr[i] ; x0010
			21:    self.a_BodyPartExamined           = *ValuePtr[i] ; x0015
			32:    self.a_ScanningSequence           = *ValuePtr[i] ; x0020
			33:    self.a_SequenceVariant            = *ValuePtr[i] ; x0021
			34:    self.a_ScanOptions                = *ValuePtr[i] ; x0022
			35:    self.a_AcquisitionType            = *ValuePtr[i] ; x0023
			36:    self.a_SequenceName               = *ValuePtr[i] ; x0024
			37:    self.a_AngioFlag                  = *ValuePtr[i] ; x0025
			80:    self.a_SliceThickness             = *ValuePtr[i] ; x0050
			128:   self.a_RepetitionTime             = *ValuePtr[i] ; x0080
			129:   self.a_EchoTime                   = *ValuePtr[i] ; x0081
			130:   self.a_InversionTime              = *ValuePtr[i] ; x0082
			131:   self.a_NumberOfAverages           = *ValuePtr[i] ; x0083
			132:   self.a_ImagingFrequency           = *ValuePtr[i] ; x0084
			133:   self.a_ImagedNucleus              = *ValuePtr[i] ; x0085
			134:   self.a_EchoNumber                 = *ValuePtr[i] ; x0086
			135:   self.a_MagneticFieldStrength      = *ValuePtr[i] ; x0087
			136:   self.a_SpacingBetweenSlices       = *ValuePtr[i] ; x0088
			137:   self.a_NumberOfPhaseEncodingSteps = *ValuePtr[i] ; x0089
			145:   self.a_EchoTrainLength            = *ValuePtr[i] ; x0091
			147:   self.a_PercentSampling            = *ValuePtr[i] ; x0093
			148:   self.a_PercentPhaseFieldOfView    = *ValuePtr[i] ; x0094
			149:   self.a_PixelBandwidth             = *ValuePtr[i] ; x0095
			4096:  self.a_DeviceSerialNumber         = *ValuePtr[i] ; x1000
			4128:  self.a_SoftwareVersion            = *ValuePtr[i] ; x1020
			4144:  self.a_ProtocolName               = *ValuePtr[i] ; x1030
			4161:  self.a_ContrastBolusVolume        = *ValuePtr[i] ; x1041
			4608:  self.a_DateOfLastCalibration      = *ValuePtr[i] ; x1200
			4609:  self.a_TimeOfLastCalibration      = *ValuePtr[i] ; x1201
			4688:  self.a_ReceivingCoil              = *ValuePtr[i] ; x1250
			4689:  self.a_TransmittingCoil           = *ValuePtr[i] ; x1251
			4882:  self.a_PhaseEncodingDirection     = *ValuePtr[i] ; x1312
			4884:  self.a_FlipAngle                  = *ValuePtr[i] ; x1314
			4885:  self.a_VariableFlipAngle          = *ValuePtr[i] ; x1315
			4886:  self.a_SAR                        = *ValuePtr[i] ; x1316
			4888:  self.a_DBDT                       = *ValuePtr[i] ; x1318
			20736: self.a_PatientPosition            = *ValuePtr[i] ; x5100
			else:
			end
		end
		32: begin
			case element[i] of
			13:    self.r_StudyInstanceUID           = *ValuePtr[i] ; x000D
			14:    self.r_SeriesInstanceUID          = *ValuePtr[i] ; x000E
			16:    self.r_StudyID                    = *ValuePtr[i] ; x0010
			17:    self.r_SeriesNumber               = *ValuePtr[i] ; x0011
			18:    self.r_AcquisitionNumber          = *ValuePtr[i] ; x0012
			19:    self.r_ImageNumber                = *ValuePtr[i] ; x0013
			48:    self.r_ImagePosition              = *ValuePtr[i] ; x0030
			50:    self.r_ImagePositionPatient       = *ValuePtr[i] ; x0032
			53:    self.r_ImageOrientation           = *ValuePtr[i] ; x0035
			55:    self.r_ImageOrientPatient         = *ValuePtr[i] ; x0037
			80:    self.r_Location                   = *ValuePtr[i] ; x0050
			82:    self.r_FrameOfReference           = *ValuePtr[i] ; x0052
			112:   self.r_ImageGeometryType          = *ValuePtr[i] ; x0070
			128:   self.r_MaskingImage               = *ValuePtr[i] ; x0080
			4097:  self.r_AcqusitionInSeries         = *ValuePtr[i] ; x1001
			4160:  self.r_PositionReferenceIndicator = *ValuePtr[i] ; x1040
			4161:  self.r_SliceLocation              = *ValuePtr[i] ; x1041
			16384: self.r_ImageComments              = *ValuePtr[i] ; x4000
			else:
			end
		end
		40: begin
			case element[i] of
			2:    self.g_SamplesPerPixel              = *ValuePtr[i] ; x0002
			4:    self.g_PhotometricInterpretation    = *ValuePtr[i] ; x0004
			5:    self.g_ImageDimension               = *ValuePtr[i] ; x0005
			6:    self.g_PlanarConfiguration          = *ValuePtr[i] ; x0006
			16:   self.g_Rows                         = *ValuePtr[i] ; x0010
			17:   self.g_Columns                      = *ValuePtr[i] ; x0011
			48:   self.g_PixelSpacing                 = *ValuePtr[i] ; x0030
			64:   self.g_ImageFormat                  = *ValuePtr[i] ; x0040
			256:  self.g_BitsAllocated                = *ValuePtr[i] ; x0100
			257:  self.g_BitsStored                   = *ValuePtr[i] ; x0101
			258:  self.g_HighBit                      = *ValuePtr[i] ; x0102
			259:  self.g_PixelRepresentation          = *ValuePtr[i] ; x0103
			262:  self.g_SmallestImagePixelValue      = *ValuePtr[i] ; x0106
			263:  self.g_LargestImagePixelValue       = *ValuePtr[i] ; x0107
			512:  self.g_ImageLocation                = *ValuePtr[i] ; x0200  RET!!!
			4176: self.g_WindowCenter                 = *ValuePtr[i] ; x1050
			4177: self.g_WindowWidth                  = *ValuePtr[i] ; x1051
			4181: self.g_WindowCenterWidthExplanation = *ValuePtr[i] ; x1055
			else:
			end
		end
		32736: begin                                                 ; the image data
			case element[i] of
			16:   self.imgdata = valuePtr[i]
			else:
			end
		end
		else:
		endcase

	endif
	endfor

		; now set series and study id if the user supplied these

	if( keyword_set(seriesUID) ) then self.r_SeriesInstanceUID = seriesUID
	if( keyword_set(studyUID)  ) then self.r_StudyInstanceUID  = studyUID

	self.validCache = 1

	ptr_free, valuePtr(*)
end


;+=----------------------------------------------------------------------------
; Name  : mrdicom::anonymize
; Syntax: obj->anonymize
; Desc  : Anonymizes/deletes patient related data from the object
;=-----------------------------------------------------------------------------
pro mrdicom::anonymize

		; only makes sense to anonymize patient data if tag values are valid
		; This might fail if no dicom data is attached - in which case we
		; insert the data anyway.

	if( self.validCache eq 0 ) then self->fillCache

	self.p_PatientName			=	'Anonymous'
	self.p_PatientId			=	' '
	self.p_PatientBirthDate		=	' '
	self.p_PatientSex			=	' '
	self.p_PatientAge			=	' '
	self.p_PatientWeight		=	0.0
	self.p_PatientAlerts  		=   ' '
	self.p_PatientComments		=	' '
end


;+=----------------------------------------------------------------------------
; Name  : mrdicom::setImg
; Syntax: obj->setImg, img, /no_copy
; Args  : <b>img</b>
;         A pointer to the image, either an [n,m] matrix or a [3,n,m] matrix
;         in case of true color images
;         <b>no_copy</b>
;         Do not create an internal copy of the image data
; Desc  : Set the image data
;         By default mrDicom makes an internal copy of the image data. As this
;         might be an overhead in many cases, this behaviour can be overridden
;         by setting no_copy to '1'.
;         However bear in mind that
;         a) when the mrdicom object is destroyed or
;         b) a different image is set
;         the stored image is destroyed.
; Code  :
;         > myimg1 = ptr_new( intarr(256,256), /no_copy )
;         > myimg2 = ptr_new( intarr(256,256), /no_copy )
;         > ; alternative 1:
;         > mrdicom->setImg, myimg1, /no_copy
;         > mrdicom->setImg, myimg2    ; this will free myimg1!
;         > ; alternative 2:
;         > mrdicom->setImg, myimg1
;         > mrdicom->setImg, myimg2    ; will leave myimg1 intact
;=-----------------------------------------------------------------------------
pro mrdicom::setImg, img, no_copy=no_copy

	no_copy = (n_elements(no_copy) eq 0) ? 0 : no_copy

		; when the user sets an image, we assume he will write it, so filling
		; the header will not produce any overhead

	if( self.validCache eq 0 ) then self->fillCache
	if( ptr_valid(self.imgdata) ) then ptr_free, self.imgdata
	if( no_copy eq 1 ) then begin
		self.imgdata = img
	end else begin
		self.imgdata = ptr_new( *img )
	end
end

;+=----------------------------------------------------------------------------
; Name : mrdicom::img
; Type : Function
; Desc : returns pointer to the image data or ptr_new() if no image data present
;=-----------------------------------------------------------------------------
function mrdicom::img
	if( ptr_valid( self.imgdata )) then return, self.imgdata

	vals    = self->getValue( 32736, 16, /no_copy )
	refs    = self->getReference( 32736, 16 )
	parents = self->getParent( refs )
	id = where( parents eq -1 )

	if( ptr_valid( vals[id] ) ) then begin
		self.imgdata = vals[id]
	end else begin
		self.imgdata = ptr_new()
	end

	return, self.imgdata
end

;+=----------------------------------------------------------------------------
; Name  : mrdicom::setTagValue
; Syntax: obj->setTagValue, tagname, value
; Args  : <b>tagname</b>
;         A string - one of the list below
;         <b>value</b>
;         The value to set of appropriate type
; Desc  : Fill a DICOM data element with the corresponding value.
;         There is absolutely no error checking whatsoever - esp. no checking
;         whether the filled data value matches the DICOM tag's value representation
;         (i.e. type). It is the user's responsibility alone that VR and
;         value match.
;         Here's a full list of available data tags - the tag names should
;         be pretty self explanatory:
;         <b>Tag list:</b>
;         <b>Meta tags (x0002)</b> (reading only!)
;         meta_group_length                    meta_file_meta_information_version
;         meta_media_stored_sop_class_uid      meta_media_stored_sop_instance_uid
;         meta_transfer_syntax_uid             meta_implementation_class_uid
;         meta_implementation_version_name     meta_source_application_entity_title
;         <b>General module attributes (x0008)</b>
;         id_specific_character_set            id_image_type
;         id_sop_class_uid                     id_sop_instance_uid
;         id_study_date                        id_series_date
;         id_acquisition_date                  id_image_date
;         id_study_time                        id_series_time
;         id_acquisition_time                  id_image_time
;         id_accession_number                  id_modality
;         id_manufacturer                      id_institution_name
;         id_institution_address               id_referring_physicians_name
;         id_station_name                      id_study_description
;         id_series_description                id_operators_name
;         id_manufacturer_model_name           id_referenced_study_component_sequence
;         id_referenced_sop_class_uid          id_referenced_sop_instance_uid
;         id_referenced_image_sequence
;         <b>Patient tags (x0010):</b>
;         pat_patient_name                     pat_patient_id
;         pat_patient_birthdate                pat_patient_sex
;         pat_patient_age                      pat_patient_weight
;         pat_medical_alerts                   pat_pregnancy_status
;         pat_patient_comments
;         <b>Acquisition tags (x0018)</b>
;         acq_contrast_bolus_agent             acq_body_part_examined
;         acq_scanning_sequence                acq_sequence_variant
;         acq_scan_options                     acq_mr_acquisition_type
;         acq_sequence_name                    acq_angio_flag
;         acq_slice_thickness                  acq_inversion_time
;         acq_repetition_time                  acq_echo_time
;         acq_number_of_averages               acq_imaging_frequency
;         acq_imaged_nucleus                   acq_echo_number
;         acq_magnetic_field_strength          acq_spacing_between_slices
;         acq_number_of_phase_encoding_steps   acq_echo_train_length
;         acq_percent_sampling                 acq_percent_phase_field_of_view
;         acq_pixel_bandwidth                  acq_device_serial_number
;         acq_software_version                 acq_protocol_name
;         acq_contrast_bolus_volume            acq_date_of_last_calibration
;         acq_time_of_last_calibration         acq_receiving_coil
;         acq_transmitting_coil                acq_phase_encoding_direction
;         acq_flip_angle                       acq_variable_flip_angle
;         acq_sar                              acq_db_dt
;         acq_patient_position
;         <b>Relating tags (x0020)</b>
;         rel_study_instance_uid               rel_series_instance_uid
;         rel_study_id                         rel_series_number
;         rel_acquisition_number               rel_image_number
;         rel_image_position                   rel_image_position_patient
;         rel_image_orientation                rel_image_orientation_patient
;         rel_location                         rel_frame_of_reference_uid
;         rel_image_geometry_type              rel_masking_image
;         rel_acquisitions_in_series           rel_position_reference_indicator
;         rel_slice_location                   rel_image_comments
;         <b>Image tags (x0028)</b>
;         img_samples_per_pixel                img_photometric_interpretation
;         img_image_dimensions                 img_planar_configuration
;         img_rows                             img_columns
;         img_pixel_spacing                    img_image_format
;         img_bits_allocated                   img_bits_stored
;         img_high_bit                         img_pixel_representation
;         img_smallest_image_pixel_value       img_largest_image_pixel_value
;         img_window_center_width_explanation  img_image_location
;         img_window_center                    img_window_width
;=-----------------------------------------------------------------------------
pro mrdicom::setTagValue, tagname, value

		; setting tag values requires a valid cache!

	if( self.validCache eq 0 ) then self->fillCache

		; we allow setting of meta tags as well - the user will know what
		; he does

	tagname = strlowcase( tagname )
	case tagname of
		'meta_group_length':                      self.m_GroupLength = value
		'meta_file_meta_information_version':     self.m_FileMetaInfoVersion = value
		'meta_media_stored_sop_class_uid':        self.m_MediaStoredSOPClassUID = value
		'meta_media_stored_sop_instance_uid':     self.m_MediaStoredSOPInstanceUID = value
		'meta_transfer_syntax_uid':               self.m_TransferSyntaxUID = value
		'meta_implementation_class_uid':          self.m_ImplementationClassUID = value
		'meta_implementation_version_name':       self.m_ImplementationVersionName = value
		'meta_source_application_entity_title':   self.m_SourceApplicationEntityTitle = value
		'id_specific_character_set':              self.i_CharacterSet = value
		'id_image_type':                          self.i_ImageType = value
		'id_sop_class_uid':                       self.i_SOPClassUID = value
		'id_sop_instance_uid':                    self.i_SOPInstanceUID = value
		'id_study_date':                          self.i_StudyDate = value
		'id_series_date':                         self.i_SeriesDate = value
		'id_acquisition_date':                    self.i_AcquisitionDate = value
		'id_image_date':                          self.i_ImageDate = value
		'id_study_time':                          self.i_StudyTime = value
		'id_series_time':                         self.i_SeriesTime = value
		'id_acquisition_time':                    self.i_AcquisitionTime = value
		'id_image_time':                          self.i_ImageTime = value
		'id_accession_number':                    self.i_AccessionNumber = value
		'id_modality':                            self.i_Modality = value
		'id_manufacturer':                        self.i_Manufacturer = value
		'id_institution_name':                    self.i_InstitutionName = value
		'id_institution_address':                 self.i_InstitutionAddress = value
		'id_referring_physicians_name':           self.i_PhysiciansName = value
		'id_station_name':                        self.i_StationName = value
		'id_study_description':                   self.i_StudyDescription = value
		'id_series_description':                  self.i_SeriesDescription = value
		'id_operators_name':                      self.i_OperatorsName = value
		'id_manufacturer_model_name':             self.i_ManufacturerModelName = value
;		'id_referenced_study_component_sequence': self.i_RefStudyComponentSequence = value
		'id_referenced_sop_class_uid':            self.i_RefSOPClassUID = value
		'id_referenced_sop_instance_uid':         self.i_RefSOPInstanceUID = value
;		'id_referenced_image_sequence':           self.i_RefImageSequence = value
		'pat_patient_name':                       self.p_PatientName = value
		'pat_patient_id':                         self.p_PatientId = value
		'pat_patient_birthdate':                  self.p_PatientBirthdate = value
		'pat_patient_sex':                        self.p_PatientSex = value
		'pat_patient_age':                        self.p_PatientAge = value
		'pat_patient_weight':                     self.p_PatientWeight = value
		'pat_medical_alerts':                     self.p_PatientAlerts = value
		'pat_pregnancy_status':                   self.p_PregnancyStatus = value
		'pat_patient_comments':                   self.p_PatientComments = value
		'acq_contrast_bolus_agent':               self.a_ContrastBolusAgent = value
		'acq_body_part_examined':                 self.a_BodyPartExamined = value
		'acq_scanning_sequence':                  self.a_ScanningSequence = value
		'acq_sequence_variant':                   self.a_SequenceVariant = value
		'acq_scan_options':                       self.a_ScanOptions = value
		'acq_mr_acquisition_type':                self.a_AcquisitionType = value
		'acq_sequence_name':                      self.a_SequenceName = value
		'acq_angio_flag':                         self.a_AngioFlag = value
		'acq_slice_thickness':                    self.a_SliceThickness = value
		'acq_repetition_time':                    self.a_RepetitionTime = value
		'acq_echo_time':                          self.a_EchoTime = value
		'acq_inversion_time':                     self.a_InversionTime = value
		'acq_number_of_averages':                 self.a_NumberOfAverages = value
		'acq_imaging_frequency':                  self.a_ImagingFrequency = value
		'acq_imaged_nucleus':                     self.a_ImagedNucleus = value
		'acq_echo_number':                        self.a_EchoNumber = value
		'acq_magnetic_field_strength':            self.a_MagneticFieldStrength = value
		'acq_spacing_between_slices':             self.a_SpacingBetweenSlices = value
		'acq_number_of_phase_encoding_steps':     self.a_NumberOfPhaseEncodingSteps = value
		'acq_echo_train_length':                  self.a_EchoTrainLength = value
		'acq_percent_sampling':                   self.a_PercentSampling = value
		'acq_percent_phase_field_of_view':        self.a_PercentPhaseFieldOfView = value
		'acq_pixel_bandwidth':                    self.a_PixelBandwidth = value
		'acq_device_serial_number':               self.a_DeviceSerialNumber = value
		'acq_software_version':                   self.a_SoftwareVersion = value
		'acq_protocol_name':                      self.a_ProtocolName = value
		'acq_contrast_bolus_volume':              self.a_ContrastBolusVolume = value
		'acq_date_of_last_calibration':           self.a_DateOfLastCalibration = value
		'acq_time_of_last_calibration':           self.a_TimeOfLastCalibration = value
		'acq_receiving_coil':                     self.a_ReceivingCoil = value
		'acq_transmitting_coil':                  self.a_TransmittingCoil = value
		'acq_phase_encoding_direction':           self.a_PhaseEncodingDirection = value
		'acq_flip_angle':                         self.a_FlipAngle = value
		'acq_variable_flip_angle':                self.a_VariableFlipAngle = value
		'acq_sar':                                self.a_SAR = value
		'acq_db_dt':                              self.a_DBDT = value
		'acq_patient_position':                   self.a_PatientPosition = value
		'rel_study_instance_uid':                 self.r_StudyInstanceUID = value
		'rel_series_instance_uid':                self.r_SeriesInstanceUID = value
		'rel_study_id':                           self.r_StudyID = value
		'rel_series_number':                      self.r_SeriesNumber = value
		'rel_acquisition_number':                 self.r_AcquisitionNumber = value
		'rel_image_number':                       self.r_ImageNumber = value
		'rel_image_position':                     self.r_ImagePosition = value
		'rel_image_position_patient':             self.r_ImagePositionPatient = value
		'rel_image_orientation':                  self.r_ImageOrientation = value
		'rel_image_orientation_patient':          self.r_ImageOrientPatient = value
		'rel_location':                           self.r_Location = value
		'rel_frame_of_reference_uid':             self.r_FrameOfReference = value
		'rel_image_geometry_type':                self.r_ImageGeometryType = value
		'rel_masking_image':                      self.r_MaskingImage = value
		'rel_acquisitions_in_series':             self.r_AcqusitionInSeries = value
		'rel_position_reference_indicator':       self.r_PositionReferenceIndicator = value
		'rel_slice_location':                     self.r_SliceLocation = value
		'rel_image_comments':                     self.r_ImageComments = value
		'img_samples_per_pixel':                  self.g_SamplesPerPixel = value
		'img_photometric_interpretation':         self.g_PhotometricInterpretation = value
		'img_image_dimensions':                   self.g_ImageDimension = value
		'img_planar_configuration':               self.g_PlanarConfiguration = value
		'img_rows':                               self.g_Rows = value
		'img_columns':                            self.g_Columns = value
		'img_pixel_spacing':                      self.g_PixelSpacing = value
		'img_image_format':                       self.g_ImageFormat = value
		'img_bits_allocated':                     self.g_BitsAllocated = value
		'img_bits_stored':                        self.g_BitsStored = value
		'img_high_bit':                           self.g_HighBit = value
		'img_pixel_representation':               self.g_PixelRepresentation = value
		'img_smallest_image_pixel_value':         self.g_SmallestImagePixelValue = value
		'img_largest_image_pixel_value':          self.g_LargestImagePixelValue = value
		'img_window_center_width_explanation':    self.g_WindowCenterWidthExplanation = value
		'img_image_location':                     self.g_ImageLocation = value
		'img_window_center':                      self.g_WindowCenter = value
		'img_window_width':                       self.g_WindowWidth = value
	else: x = dialog_message('Unknown tagname ' + tagname + ' in mrdicom::setTagValue')
	end
end

;+=----------------------------------------------------------------------------
; Name  : mrdicom::getTagValue
; Syntax: value = obj->getTagValue( tagname )
; Args  : <b>tagname</b>
;         String (case insensitive) - one of the list above.
; Ret   : The value of the corresponding field
; Desc  : For a list of available tagnames see mrdicom::setTagValue
;=-----------------------------------------------------------------------------
function mrdicom::getTagValue, tagname

		; reading from tag values requires a valid cache

	if( self.validCache eq 0 ) then self->fillCache

	tagname = strlowcase( tagname )
	case tagname of
		'meta_group_length':                      return, self.m_GroupLength
		'meta_file_meta_information_version':     return, self.m_FileMetaInfoVersion
		'meta_media_stored_sop_class_uid':        return, self.m_MediaStoredSOPClassUID
		'meta_media_stored_sop_instance_uid':     return, self.m_MediaStoredSOPInstanceUID
		'meta_transfer_syntax_uid':               return, self.m_TransferSyntaxUID
		'meta_implementation_class_uid':          return, self.m_ImplementationClassUID
		'meta_implementation_version_name':       return, self.m_ImplementationVersionName
		'meta_source_application_entity_title':   return, self.m_SourceApplicationEntityTitle
		'id_specific_character_set':              return, self.i_CharacterSet
		'id_image_type':                          return, self.i_ImageType
		'id_sop_class_uid':                       return, self.i_SOPClassUID
		'id_sop_instance_uid':                    return, self.i_SOPInstanceUID
		'id_study_date':                          return, self.i_StudyDate
		'id_series_date':                         return, self.i_SeriesDate
		'id_acquisition_date':                    return, self.i_AcquisitionDate
		'id_image_date':                          return, self.i_ImageDate
		'id_study_time':                          return, self.i_StudyTime
		'id_series_time':                         return, self.i_SeriesTime
		'id_acquisition_time':                    return, self.i_AcquisitionTime
		'id_image_time':                          return, self.i_ImageTime
		'id_accession_number':                    return, self.i_AccessionNumber
		'id_modality':                            return, self.i_Modality
		'id_manufacturer':                        return, self.i_Manufacturer
		'id_institution_name':                    return, self.i_InstitutionName
		'id_institution_address':                 return, self.i_InstitutionAddress
		'id_referring_physicians_name':           return, self.i_PhysiciansName
		'id_station_name':                        return, self.i_StationName
		'id_study_description':                   return, self.i_StudyDescription
		'id_series_description':                  return, self.i_SeriesDescription
		'id_operators_name':                      return, self.i_OperatorsName
		'id_manufacturer_model_name':             return, self.i_ManufacturerModelName
;		'id_referenced_study_component_sequence': return, self.i_RefStudyComponentSequence
		'id_referenced_sop_class_uid':            return, self.i_RefSOPClassUID
		'id_referenced_sop_instance_uid':         return, self.i_RefSOPInstanceUID
;		'id_referenced_image_sequence':           return, self.i_RefImageSequence
		'pat_patient_name':                       return, self.p_PatientName
		'pat_patient_id':                         return, self.p_PatientId
		'pat_patient_birthdate':                  return, self.p_PatientBirthdate
		'pat_patient_sex':                        return, self.p_PatientSex
		'pat_patient_age':                        return, self.p_PatientAge
		'pat_patient_weight':                     return, self.p_PatientWeight
		'pat_medical_alerts':                     return, self.p_PatientAlerts
		'pat_pregnancy_status':                   return, self.p_PregnancyStatus
		'pat_patient_comments':                   return, self.p_PatientComments
		'acq_contrast_bolus_agent':               return, self.a_ContrastBolusAgent
		'acq_body_part_examined':                 return, self.a_BodyPartExamined
		'acq_scanning_sequence':                  return, self.a_ScanningSequence
		'acq_sequence_variant':                   return, self.a_SequenceVariant
		'acq_scan_options':                       return, self.a_ScanOptions
		'acq_mr_acquisition_type':                return, self.a_AcquisitionType
		'acq_sequence_name':                      return, self.a_SequenceName
		'acq_angio_flag':                         return, self.a_AngioFlag
		'acq_slice_thickness':                    return, self.a_SliceThickness
		'acq_repetition_time':                    return, self.a_RepetitionTime
		'acq_echo_time':                          return, self.a_EchoTime
		'acq_inversion_time':                     return, self.a_InversionTime
		'acq_number_of_averages':                 return, self.a_NumberOfAverages
		'acq_imaging_frequency':                  return, self.a_ImagingFrequency
		'acq_imaged_nucleus':                     return, self.a_ImagedNucleus
		'acq_echo_number':                        return, self.a_EchoNumber
		'acq_magnetic_field_strength':            return, self.a_MagneticFieldStrength
		'acq_spacing_between_slices':             return, self.a_SpacingBetweenSlices
		'acq_number_of_phase_encoding_steps':     return, self.a_NumberOfPhaseEncodingSteps
		'acq_echo_train_length':                  return, self.a_EchoTrainLength
		'acq_percent_sampling':                   return, self.a_PercentSampling
		'acq_percent_phase_field_of_view':        return, self.a_PercentPhaseFieldOfView
		'acq_pixel_bandwidth':                    return, self.a_PixelBandwidth
		'acq_device_serial_number':               return, self.a_DeviceSerialNumber
		'acq_software_version':                   return, self.a_SoftwareVersion
		'acq_protocol_name':                      return, self.a_ProtocolName
		'acq_contrast_bolus_volume':              return, self.a_ContrastBolusVolume
		'acq_date_of_last_calibration':           return, self.a_DateOfLastCalibration
		'acq_time_of_last_calibration':           return, self.a_TimeOfLastCalibration
		'acq_receiving_coil':                     return, self.a_ReceivingCoil
		'acq_transmitting_coil':                  return, self.a_TransmittingCoil
		'acq_phase_encoding_direction':           return, self.a_PhaseEncodingDirection
		'acq_flip_angle':                         return, self.a_FlipAngle
		'acq_variable_flip_angle':                return, self.a_VariableFlipAngle
		'acq_sar':                                return, self.a_SAR
		'acq_db_dt':                              return, self.a_DBDT
		'acq_patient_position':                   return, self.a_PatientPosition
		'rel_study_instance_uid':                 return, self.r_StudyInstanceUID
		'rel_series_instance_uid':                return, self.r_SeriesInstanceUID
		'rel_study_id':                           return, self.r_StudyID
		'rel_series_number':                      return, self.r_SeriesNumber
		'rel_acquisition_number':                 return, self.r_AcquisitionNumber
		'rel_image_number':                       return, self.r_ImageNumber
		'rel_image_position':                     return, self.r_ImagePosition
		'rel_image_position_patient':             return, self.r_ImagePositionPatient
		'rel_image_orientation':                  return, self.r_ImageOrientation
		'rel_image_orientation_patient':          return, self.r_ImageOrientPatient
		'rel_location':                           return, self.r_Location
		'rel_frame_of_reference_uid':             return, self.r_FrameOfReference
		'rel_image_geometry_type':                return, self.r_ImageGeometryType
		'rel_masking_image':                      return, self.r_MaskingImage
		'rel_acquisitions_in_series':             return, self.r_AcqusitionInSeries
		'rel_position_reference_indicator':       return, self.r_PositionReferenceIndicator
		'rel_slice_location':                     return, self.r_SliceLocation
		'rel_image_comments':                     return, self.r_ImageComments
		'img_samples_per_pixel':                  return, self.g_SamplesPerPixel
		'img_photometric_interpretation':         return, self.g_PhotometricInterpretation
		'img_image_dimensions':                   return, self.g_ImageDimension
		'img_planar_configuration':               return, self.g_PlanarConfiguration
		'img_rows':                               return, self.g_Rows
		'img_columns':                            return, self.g_Columns
		'img_pixel_spacing':                      return, self.g_PixelSpacing
		'img_image_format':                       return, self.g_ImageFormat
		'img_bits_allocated':                     return, self.g_BitsAllocated
		'img_bits_stored':                        return, self.g_BitsStored
		'img_high_bit':                           return, self.g_HighBit
		'img_pixel_representation':               return, self.g_PixelRepresentation
		'img_smallest_image_pixel_value':         return, self.g_SmallestImagePixelValue
		'img_largest_image_pixel_value':          return, self.g_LargestImagePixelValue
		'img_window_center_width_explanation':    return, self.g_WindowCenterWidthExplanation
		'img_image_location':                     return, self.g_ImageLocation
		'img_window_center':                      return, self.g_WindowCenter
		'img_window_width':                       return, self.g_WindowWidth
	else: x = dialog_message('Unknown tagname ' + tagname + ' in mrdicom::getTagValue')
	end
end

;+=----------------------------------------------------------------------------
; Name  : mrdicom::filename
; Syntax: filename = obj->filename()
; Ret   : Returns the filename associated with this object. If the object has
;         been newly created and no data has been loaded yet (or written),
;         the function returns an empty string
;=-----------------------------------------------------------------------------
function mrdicom::filename
	return, self.filename
end


;+=----------------------------------------------------------------------------
; Name  : mrdicom::read
; Syntax: res = obj->read( filename, ... )
; Args  : same as IDLffDicom::read
; Desc  : overloaded IDLffDicom::read to store filename and for some
;         internal bookkeeping
; Ret   : 0 in the case of error, 1 if success
;=-----------------------------------------------------------------------------
function mrdicom::read, filename, _Extra=e
	if( keyword_set(filename) ) then begin
	 	result = self->IDLffDicom::read( filename, _EXTRA=e ) ; 1 if ok, 0 if error
		self.filename   = filename                            ; filename set even if error
	end else begin
		return, 0
	end
	self.validCache = 0           ; cache is invalid after reading a new file
	self.imgData    = ptr_new()   ; image data is also invalid
	return, result
end

;+=----------------------------------------------------------------------------
; Name  : mrdicom::write
; Syntax: res = obj->write( filename, /flip, studyUID=string, seriesUID=string )
; Args  : string filename    : the filename to write to
;         flip      = int    : flag whether to (rotate, 7) the image before writing
;         studyUID  = string : a user supplied studyID
;         seriesUID = string : a user supplied seriesID
; Ret   : 1 if ok, else 0 - check errno and errmsg in the latter case
; Desc  : Write image data in DICOM format to disk. For a explanation of studyUID
;         and seriesUID see mrdicom::copy. The presence of these arguments
;         here provides the possibility for fast writing of a sequence of images
;         via the same mrdicom object.
;         The method automatically generates a unique SOP_Inst_UID for each image.
;         Known Bug: if an integer or floating point value equals zero and is not
;         mandatory, it will not be written to the file! However the quantities
;         for which this holds true do not make sense with values of '0' (except
;         for pregnancy status), so this should not hurt much.
;=-----------------------------------------------------------------------------
function mrdicom::write, filename, flip=flip, studyUID=studyUID, seriesUID=seriesUID

	common mrdicomv16, uniqid, implementationClassUID, SopClassUID

		; need valid cache values to write

	if( self.validCache eq 0 ) then self->fillCache

	self.errno    = 0
	self.errmsg   = ''
	self.filename = filename

		; do some argument and error checking

	if( not ptr_valid( self.imgData ) ) then x = self->img()  ; will set self.imgData
	if( not ptr_valid( self.imgData ) ) then begin
		self.errmsg = 'no image data available'
		self.errno  = -1
		return, 0
	end

	if( n_elements(flip) eq 0 ) then flip = 0
	if( self.filename eq '' ) then begin
		 self.errmsg = 'no filename set'
		 self.errno  = -1
		 return, 0
	end

		; check and set the uniq ids for image, study and series

	self.i_SOPClassUID 	    = SOPClassUID                     ; MR image storage, explicit VR, little endian, uncompressed
	self.i_SOPInstanceUID   = self->generateUID('image')
	if( keyword_set(studyUID) ) then self.r_StudyInstanceUID  = studyUID
	if( keyword_set(seriesUID)) then self.r_SeriesInstanceUID = seriesUID

	if( self.r_StudyInstanceUID eq ''  or self.r_SeriesInstanceUID eq '' ) then begin
		self.errmsg = 'Missing unique study or series id'
		self.errno  = -3
		return, 0
	end

		; prepare image and set image related data

	image     = (*self.imgdata) < 4095 > 0     ; image may be float - clip before conversion
	image     = fix(image)                       ; make sure it's integer
	image     = image < 4095 > 0                 ; clip again (not needed, I guess, but this conversion business...)
 	imagedim  = size(image, /n_dimensions)       ; the number of dimensions (2 or 3)
 	imagedims = size(image, /dimensions  )       ;
	case imagedim of
		2:  begin                                ; monochrome image, low vals = black
				self.g_PhotometricInterpretation = 'MONOCHROME2'
				self.g_SamplesPerPixel           = 1
				self.g_BitsAllocated             = 16
				self.g_BitsStored                = 12
				self.g_HighBit                   = 11
				self.g_Columns	                 = imagedims[0]
				self.g_Rows		                 = imagedims[1]
				ImaSort = 1
			 	position = 3
				if( flip eq 1 ) then image = rotate( image, 7 )
			end
		3:  begin                                ; rgb image, 3 * size_x * size_y
				self.g_PhotometricInterpretation = 'RGB'
				self.g_SamplesPerPixel           = 3
				self.g_BitsAllocated             = 8
				self.g_BitsStored                = 8
				self.g_HighBit                   = 7
				self.g_Columns                   = imagedims[1]
				self.g_Rows                      = imagedims[2]
				ImaSort 	= 3
				position 	= 4
				image       = bytscl(image)
				if( flip eq 1 ) then begin
					tmp = bytarr(self.g_Columns, self.g_Rows)
					for j = 0, 2 do begin
						tmp[*,*] = image[j,*,*]
						image[j,*,*] = rotate( tmp, 7 )
					end
				end
			end
		else: begin
			self.errno = -2
			self.errmsg = 'Illegal image dimension or no image data available'
			return, errno
		end
	endcase

		; find out about image data type and size

	imagetype = size( image, /type )    ; might have been bytescaled
	case imagetype of 					; data type stored in the image
		1:  	bytesperpixel = 1		; byte
		2:  	bytesperpixel = 2		; int
		3:  	bytesperpixel = 4		; long int
		12: 	bytesperpixel = 2		; unsigned int
		13: 	bytesperpixel = 4		; unsigned long int
		else: 	bytesperpixel = 0       ; unknown
	endcase

	if( bytesperpixel ge 2 ) then begin     ; not in use since we assume Intel inside
		little_endian = (BYTE(1, 0, 1))[0]
		if (little_endian ne 1) then begin
			;byteorder, image, /SWAP_IF_BIG_ENDIAN
		endif else begin
			;byteorder, image, /SWAP_IF_LITTLE_ENDIAN
		endelse
	endif

	maxval              = max(image, min=minval)      ; and color images???

	center              = strcompress(string(0.5 * (maxval+minval)),/remove_all)
	width               = strcompress(string(1.0 * (maxval-minval)),/remove_all)
	self.g_SmallestImagePixelValue = minval
	self.g_LargestImagePixelValue  = maxval

	self.g_WindowCenter	= center + '\' + center
	self.g_WindowWidth 	= width  + '\' + width

	imageSize 			= ulong64(self.g_rows) * self.g_columns * bytesperpixel

		; open the file

	openw, u, self.filename, /get_lun

		; write standard conformant header

	writeu, u, bytarr(128)    	; file preamble: 128 bytes of value 00x
	writeu, u, 'DICM'           ; dicom prefix

	self.m_FileMetaInfoVersion       = bytarr(2)     ; file meta information
	self.m_FileMetaInfoVersion[0]    = 0B
	self.m_FileMetaInfoVersion[1]    = 1B

	self.m_MediaStoredSOPClassUID    = SOPClassUID
    self.m_MediaStoredSOPInstanceUID = self.i_SOPInstanceUID
	self.m_TransferSyntaxUID         = '1.2.840.10008.1.2'
	self.m_ImplementationClassUID    = implementationClassUID
	self.m_ImplementationVersionName = self.implementationVersionName

		; calculate size of meta header data

	len = strlen( self.m_MediaStoredSOPClassUID )
	groupLength = 2 + len + (len and 1)
	len = strlen( self.m_MediaStoredSOPInstanceUID )
	groupLength = groupLength + len + (len and 1)
	len = strlen( self.m_TransferSyntaxUID )
	groupLength = groupLength + len + (len and 1)
	len = strlen( self.m_ImplementationClassUID )
	groupLength = groupLength + len + (len and 1)
	len = strlen( self.m_ImplementationVersionName )
	groupLength = groupLength + len + (len and 1) + 52    ; tags for the meta elements

	self.m_GroupLength = groupLength

                                                       writeu, u, self->setvrtag('0002'x, '0000'x, 'UL', self.m_GroupLength, /explicit )
                                                       writeu, u, self->setvrtag('0002'x, '0001'x, 'OB', self.m_FileMetaInfoVersion, /explicit )
                                                       writeu, u, self->setvrtag('0002'x, '0002'x, 'UI', self.m_MediaStoredSOPClassUID, /explicit )
                                                       writeu, u, self->setvrtag('0002'x, '0003'x, 'UI', self.m_MediaStoredSOPInstanceUID, /explicit )
                                                       writeu, u, self->setvrtag('0002'x, '0010'x, 'UI', self.m_TransferSyntaxUID, /explicit )
                                                       writeu, u, self->setvrtag('0002'x, '0012'x, 'UI', self.m_ImplementationClassUID, /explicit )
                                                       writeu, u, self->setvrtag('0002'x, '0013'x, 'SH', self.m_ImplementationVersionName, /explicit )
 	if( self.i_CharacterSet               ne '' ) then writeu, u, self->setvrtag('0008'x, '0005'x, 'CS', self.i_CharacterSet )
                                                       writeu, u, self->setvrtag('0008'x, '0008'x, 'CS', self.i_ImageType )
                                                       writeu, u, self->setvrtag('0008'x, '0016'x, 'UI', self.i_SOPClassUID )
                                                       writeu, u, self->setvrtag('0008'x, '0018'x, 'UI', self.i_SOPInstanceUID )
													   writeu, u, self->setvrtag('0008'x, '0020'x, 'DA', self.i_StudyDate )
	if( self.i_SeriesDate                 ne '' ) then writeu, u, self->setvrtag('0008'x, '0021'x, 'DA', self.i_SeriesDate )
                                                       writeu, u, self->setvrtag('0008'x, '0022'x, 'DA', self.i_AcquisitionDate )
                                                       writeu, u, self->setvrtag('0008'x, '0023'x, 'DA', self.i_ImageDate )
													   writeu, u, self->setvrtag('0008'x, '0030'x, 'TM', self.i_StudyTime )
	if( self.i_SeriesTime                 ne '' ) then writeu, u, self->setvrtag('0008'x, '0031'x, 'TM', self.i_SeriesTime )
	if( self.i_AcquisitionTime            ne '' ) then writeu, u, self->setvrtag('0008'x, '0032'x, 'TM', self.i_AcquisitionTime )
                                                       writeu, u, self->setvrtag('0008'x, '0033'x, 'TM', self.i_ImageTime )
													   writeu, u, self->setvrtag('0008'x, '0050'x, 'SH', self.i_AccessionNumber )
													   ; !!! Jimmy mentions an unspecified problem with accessionNumber !!!
                                                       writeu, u, self->setvrtag('0008'x, '0060'x, 'CS', self.i_Modality )
                                                       writeu, u, self->setvrtag('0008'x, '0070'x, 'LO', self.i_Manufacturer )
                                                       writeu, u, self->setvrtag('0008'x, '0080'x, 'LO', self.i_InstitutionName )
    if( self.i_InstitutionAddress         ne '' ) then writeu, u, self->setvrtag('0008'x, '0081'x, 'ST', self.i_InstitutionAddress )
													   writeu, u, self->setvrtag('0008'x, '0090'x, 'LO', self.i_PhysiciansName )
	if( self.i_StationName                ne '' ) then writeu, u, self->setvrtag('0008'x, '1010'x, 'SH', self.i_StationName )
	if( self.i_StudyDescription           ne '' ) then writeu, u, self->setvrtag('0008'x, '1030'x, 'LO', self.i_StudyDescription )
	if( self.i_SeriesDescription          ne '' ) then writeu, u, self->setvrtag('0008'x, '103E'x, 'LO', self.i_SeriesDescription )
                                                       writeu, u, self->setvrtag('0008'x, '1070'x, 'PN', self.i_OperatorsName )
	if( self.i_ManufacturerModelName      ne '' ) then writeu, u, self->setvrtag('0008'x, '1090'x, 'LO', self.i_ManufacturerModelName )
;                                                      writeu, u, self->setvrtag('0008'x, '1111'x, 'SQ', self.i_RefStudyComponentSequence )
                                                       writeu, u, self->setvrtag('0008'x, '1150'x, 'UI', self.i_RefSOPClassUID )
                                                       writeu, u, self->setvrtag('0008'x, '1155'x, 'UI', self.i_RefSOPInstanceUID )
;                                                      writeu, u, self->setvrtag('0008'x, '1140'x, 'SQ', self.i_RefImageSequence )
                                                       writeu, u, self->setvrtag('0010'x, '0010'x, 'LO', self.p_PatientName )
                                                       writeu, u, self->setvrtag('0010'x, '0020'x, 'LO', self.p_PatientId )
                                                       writeu, u, self->setvrtag('0010'x, '0030'x, 'DA', self.p_PatientBirthdate )
                                                       writeu, u, self->setvrtag('0010'x, '0040'x, 'CS', self.p_PatientSex )
	if( self.p_PatientAge                 ne '' ) then writeu, u, self->setvrtag('0010'x, '1010'x, 'AS', self.p_PatientAge )
	if( self.p_PatientWeight              ne '' ) then writeu, u, self->setvrtag('0010'x, '1030'x, 'DS', self.p_PatientWeight )
	if( self.p_PatientAlerts              ne '' ) then writeu, u, self->setvrtag('0010'x, '2000'x, 'LO', self.p_PatientAlerts )
	if( self.p_PregnancyStatus            ne '' ) then writeu, u, self->setvrtag('0010'x, '21C0'x, 'US', self.p_PregnancyStatus )
	if( self.p_PatientComments            ne '' ) then writeu, u, self->setvrtag('0010'x, '4000'x, 'LO', self.p_PatientComments )
													   writeu, u, self->setvrtag('0018'x, '0010'x, 'LO', self.a_ContrastBolusAgent )
	if( self.a_BodyPartExamined           ne '' ) then writeu, u, self->setvrtag('0018'x, '0015'x, 'CS', self.a_BodyPartExamined )
													   writeu, u, self->setvrtag('0018'x, '0020'x, 'CS', self.a_ScanningSequence )
													   writeu, u, self->setvrtag('0018'x, '0021'x, 'CS', self.a_SequenceVariant )
													   writeu, u, self->setvrtag('0018'x, '0022'x, 'CS', self.a_ScanOptions )
                                                       writeu, u, self->setvrtag('0018'x, '0023'x, 'CS', self.a_AcquisitionType )
	if( self.a_SequenceName               ne '' ) then writeu, u, self->setvrtag('0018'x, '0024'x, 'SH', self.a_SequenceName )
	if( self.a_AngioFlag                  ne '' ) then writeu, u, self->setvrtag('0018'x, '0025'x, 'CS', self.a_AngioFlag )
	if( self.a_SliceThickness             ne '' ) then writeu, u, self->setvrtag('0018'x, '0050'x, 'DS', self.a_SliceThickness )
	if( self.a_RepetitionTime             ne '' ) then writeu, u, self->setvrtag('0018'x, '0080'x, 'DS', self.a_RepetitionTime )
													   writeu, u, self->setvrtag('0018'x, '0081'x, 'DS', self.a_EchoTime )
	if( self.a_InversionTime              ne '' ) then writeu, u, self->setvrtag('0018'x, '0082'x, 'DS', self.a_InversionTime )
	if( self.a_NumberOfAverages           ne '' ) then writeu, u, self->setvrtag('0018'x, '0083'x, 'DS', self.a_NumberOfAverages )
	if( self.a_ImagingFrequency           ne '' ) then writeu, u, self->setvrtag('0018'x, '0084'x, 'DS', self.a_ImagingFrequency )
	if( self.a_ImagedNucleus              ne '' ) then writeu, u, self->setvrtag('0018'x, '0085'x, 'SH', self.a_ImagedNucleus )
	if( self.a_EchoNumber                 ne '' ) then writeu, u, self->setvrtag('0018'x, '0086'x, 'IS', self.a_EchoNumber )
	if( self.a_MagneticFieldStrength      ne '' ) then writeu, u, self->setvrtag('0018'x, '0087'x, 'DS', self.a_MagneticFieldStrength )
	if( self.a_SpacingBetweenSlices       ne '' ) then writeu, u, self->setvrtag('0018'x, '0088'x, 'DS', self.a_SpacingBetweenSlices )
	if( self.a_NumberOfPhaseEncodingSteps ne '' ) then writeu, u, self->setvrtag('0018'x, '0089'x, 'IS', self.a_NumberOfPhaseEncodingSteps )
													   writeu, u, self->setvrtag('0018'x, '0091'x, 'IS', self.a_EchoTrainLength )
	if( self.a_PercentSampling            ne '' ) then writeu, u, self->setvrtag('0018'x, '0093'x, 'DS', self.a_PercentSampling )
	if( self.a_PercentPhaseFieldOfView    ne '' ) then writeu, u, self->setvrtag('0018'x, '0094'x, 'DS', self.a_PercentPhaseFieldOfView )
	if( self.a_PixelBandwidth             ne '' ) then writeu, u, self->setvrtag('0018'x, '0095'x, 'DS', self.a_PixelBandwidth )
	if( self.a_DeviceSerialNumber         ne '' ) then writeu, u, self->setvrtag('0018'x, '1000'x, 'LO', self.a_DeviceSerialNumber )
	if( self.a_SoftwareVersion            ne '' ) then writeu, u, self->setvrtag('0018'x, '1020'x, 'LO', self.a_SoftwareVersion )
	if( self.a_ProtocolName               ne '' ) then writeu, u, self->setvrtag('0018'x, '1030'x, 'LO', self.a_ProtocolName )
	if( self.a_ContrastBolusVolume        ne '' ) then writeu, u, self->setvrtag('0018'x, '1041'x, 'DS', self.a_ContrastBolusVolume )
	if( self.a_DateOfLastCalibration      ne '' ) then writeu, u, self->setvrtag('0018'x, '1200'x, 'DA', self.a_DateOfLastCalibration )
	if( self.a_TimeOfLastCalibration      ne '' ) then writeu, u, self->setvrtag('0018'x, '1201'x, 'TM', self.a_TimeOfLastCalibration )
	if( self.a_ReceivingCoil              ne '' ) then writeu, u, self->setvrtag('0018'x, '1250'x, 'SH', self.a_ReceivingCoil )
	if( self.a_TransmittingCoil           ne '' ) then writeu, u, self->setvrtag('0018'x, '1251'x, 'SH', self.a_TransmittingCoil )
	if( self.a_PhaseEncodingDirection     ne '' ) then writeu, u, self->setvrtag('0018'x, '1312'x, 'CS', self.a_PhaseEncodingDirection )
	if( self.a_FlipAngle                  ne '' ) then writeu, u, self->setvrtag('0018'x, '1314'x, 'DS', self.a_FlipAngle )
	if( self.a_VariableFlipAngle          ne '' ) then writeu, u, self->setvrtag('0018'x, '1315'x, 'CS', self.a_VariableFlipAngle )
	if( self.a_SAR                        ne '' ) then writeu, u, self->setvrtag('0018'x, '1316'x, 'DS', self.a_SAR )
	if( self.a_DBDT                       ne '' ) then writeu, u, self->setvrtag('0018'x, '1318'x, 'DS', self.a_DBDT )
	if( self.a_PatientPosition            ne '' ) then writeu, u, self->setvrtag('0018'x, '5100'x, 'CS', self.a_PatientPosition )
                                                       writeu, u, self->setvrtag('0020'x, '000D'x, 'UI', self.r_StudyInstanceUID )
                                                       writeu, u, self->setvrtag('0020'x, '000E'x, 'UI', self.r_SeriesInstanceUID )
                                                       writeu, u, self->setvrtag('0020'x, '0010'x, 'SH', self.r_StudyID )
                                                       writeu, u, self->setvrtag('0020'x, '0011'x, 'IS', self.r_SeriesNumber )
                                                       writeu, u, self->setvrtag('0020'x, '0012'x, 'IS', self.r_AcquisitionNumber )
                                                       writeu, u, self->setvrtag('0020'x, '0013'x, 'IS', self.r_ImageNumber )
	if( self.r_ImagePosition              ne '' ) then writeu, u, self->setvrtag('0020'x, '0030'x, 'RT', self.r_ImagePosition )
													   writeu, u, self->setvrtag('0020'x, '0032'x, 'DS', self.r_ImagePositionPatient )
	if( self.r_ImageOrientation           ne '' ) then writeu, u, self->setvrtag('0020'x, '0035'x, 'RT', self.r_ImageOrientation )
													   writeu, u, self->setvrtag('0020'x, '0037'x, 'DS', self.r_ImageOrientPatient )
	if( self.r_Location                   ne '' ) then writeu, u, self->setvrtag('0020'x, '0050'x, 'RT', self.r_Location )
													   writeu, u, self->setvrtag('0020'x, '0052'x, 'UI', self.r_FrameOfReference )
	if( self.r_ImageGeometryType          ne '' ) then writeu, u, self->setvrtag('0020'x, '0070'x, 'RT', self.r_ImageGeometryType )
	if( self.r_MaskingImage               ne '' ) then writeu, u, self->setvrtag('0020'x, '0080'x, 'RT', self.r_MaskingImage )
	if( self.r_AcqusitionInSeries         ne '' ) then writeu, u, self->setvrtag('0020'x, '1001'x, 'RT', self.r_AcqusitionInSeries )
													   writeu, u, self->setvrtag('0020'x, '1040'x, 'LO', self.r_PositionReferenceIndicator )
	if( self.r_SliceLocation              ne '' ) then writeu, u, self->setvrtag('0020'x, '1041'x, 'DS', self.r_SliceLocation )
	if( self.r_ImageComments              ne '' ) then writeu, u, self->setvrtag('0020'x, '4000'x, 'LT', self.r_ImageComments )
                                                       writeu, u, self->setvrtag('0028'x, '0002'x, 'US', self.g_SamplesPerPixel )
                                                       writeu, u, self->setvrtag('0028'x, '0004'x, 'CS', self.g_PhotometricInterpretation )
	if( self.g_ImageDimension             ne '' ) then writeu, u, self->setvrtag('0028'x, '0005'x, 'RT', self.g_ImageDimension )
;	if( self.g_PlanarConfiguration        ne '' ) then writeu, u, self->setvrtag('0029'x, '0006'x, 'US', self.g_PlanarConfiguration )
                                                       writeu, u, self->setvrtag('0028'x, '0006'x, 'US',0)
                                                       writeu, u, self->setvrtag('0028'x, '0008'x, 'IS',1)
                                                       writeu, u, self->setvrtag('0028'x, '0010'x, 'US', self.g_Rows )
                                                       writeu, u, self->setvrtag('0028'x, '0011'x, 'US', self.g_Columns )
													   writeu, u, self->setvrtag('0028'x, '0030'x, 'DS', self.g_PixelSpacing )
	if( self.g_ImageFormat                ne '' ) then writeu, u, self->setvrtag('0028'x, '0040'x, 'RT', self.g_ImageFormat )
                                                       writeu, u, self->setvrtag('0028'x, '0100'x, 'US', self.g_BitsAllocated )
                                                       writeu, u, self->setvrtag('0028'x, '0101'x, 'US', self.g_BitsStored )
                                                       writeu, u, self->setvrtag('0028'x, '0102'x, 'US', self.g_HighBit )
                                                       writeu, u, self->setvrtag('0028'x, '0103'x, 'US', '0000'x ) ;hardcoded
;    if( self.g_PixelRepresentation        ne '' ) then writeu, u, self->setvrtag('0028'x, '0103'x, 'US', self.g_PixelRepresentation )
	if( self.g_SmallestImagePixelValue    ne '' ) then writeu, u, self->setvrtag('0028'x, '0106'x, 'US', self.g_SmallestImagePixelValue )
	if( self.g_LargestImagePixelValue     ne '' ) then writeu, u, self->setvrtag('0028'x, '0107'x, 'US', self.g_LargestImagePixelValue )
;	if( self.g_WindowCenterWidthExplanation ne '' ) then writeu, u, self->setvrtag('0028'x, '1055'x, 'LO', self.g_WindowCenterWidthExplanation )
; uncommenting the preceeding line leads to dicom read errors of the created file - no idea why that happens
; the written value should be ok
	if( self.g_ImageLocation              ne '' ) then writeu, u, self->setvrtag('0028'x, '0200'x, 'RT', self.g_ImageLocation )
                                                       writeu, u, self->setvrtag('0028'x, '1050'x, 'DS', self.g_WindowCenter )
                                                       writeu, u, self->setvrtag('0028'x, '1051'x, 'DS', self.g_WindowWidth )
                                                       writeu, u, self->generate_pixeltag('7FE0'x,'0010'x, imagesize[0] * ImaSort )
	                                                   writeu, u, image

	close, u
	free_lun, u

		; and check the file

	If( query_dicom(self.filename) eq 0) then begin
		self.errno = -5
		self.errmsg = 'Error writing file: query_dicom returned error'
	end
	return, 1
end


;+=----------------------------------------------------------------------------
; Name  : mrdicom::generateUID
; Syntax: uid = obj->generateUID( type )
; Args  : <b>type</b>
;         String, one of 'image', 'study' or 'series'
; Desc  : Create unique ids to be used with mrdicom::copy and mrdicom::write
;         The 'image' option is essentially private: the writing routine will
;         automatically create a unique image id for each written image.
; Code  :
;         uniqueStudyID  = dicomobj->generateUID('study')
;         uniqueSeriesID = dicomobj->generateUID('series')
;=-----------------------------------------------------------------------------
function mrdicom::generateUID, type

	common mrdicomv16, uniqid, implementationClassUID, SopClassUID

	time   = string( bin_date(ascii_time) )
	id     = strcompress(time(0)+time(1)+time(2)+time(3)+time(4)+time(5)+string(uniqId), /remove_all)
	uniqId = (uniqId + 1) AND 65535           ; loop 0 - 65535

	case strlowcase( type ) of
		'image':  return, implementationClassUID + '.1' + id
		'study':  return, implementationClassUID + '.1' + id
		'series': return, implementationClassUID + '.11' + id
		else: x = dialog_message('Unknown type ' + type + ' in mrdicom::generateUID' )
	end
end


;+=----------------------------------------------------------------------------
; Name : mrdicom::getbytes
; Args : val, num
; Desc : byteorder the data due to different processors. Note: big endian ONLY.
;=-----------------------------------------------------------------------------
function mrdicom::getbytes, val, num
	ret=BYTARR(num)
	offset=0
	;work in big endian ONLY
		;val=swap_endian(val)
	;if (!version.arch eq 'x86') then begin REPLACED with little endian test
	;thanks to David Fanning :)
	little_endian = (BYTE(1, 0, 1))[0]
	if (little_endian) then begin
		byteorder,val,/SWAP_IF_BIG_ENDIAN
	endif else begin
		byteorder,val,/SWAP_IF_LITTLE_ENDIAN
	endelse
	for i=0,(num-1) do begin
		tmpres=BYTE(ISHFT(val, offset) AND 255)
		ret[i]=tmpres  ; ndring -> lagt till [i]
		offset=offset-8
	endfor
	return, ret
end

;+=----------------------------------------------------------------------------
; Name : mrdicom::generate_anytag
; Args : group, element, data, STR = str
; Desc : generate data tag to write
;=-----------------------------------------------------------------------------
function mrdicom::generate_anytag, group, element, data, STR = str

	pad=BYTE(0)

		;check to see if string type is set - if it is, change
		;padding byte to a space

		; OT: according to the standard definition only VR type
		; UI is to be padded with the space character (x32) to
		; achieve even length, all others are padded with (x00)

	IF KEYWORD_SET(STR) then begin
		pad=BYTE(STRING(' '))
	endif

	rs=[self->getbytes(group,2),self->getbytes(element,2)]

		;correct to even length if necessary

	bs=BYTE(data)
	nl=n_elements(bs)
	if ((nl mod 2) ne 0) then begin
		bs=[bs,pad]
		nl=nl+1
	end

	rs=[rs,self->getbytes(nl,2)]   ;size of field !!! 2? Why not 4?
	rs=[rs,[0,0]]				   ;padding
	rs=[rs,bs]					   ;string itself

	return, rs
end


;+=----------------------------------------------------------------------------
; Name : mrdicom::setVRtag
; Args : group, element, VR, data
; Desc : sets the data type for a given header data element
;        oth: added explanations and removed a couple of typos
;
;        Explicit layout:
;        tag (2*16bit), vr(2 chars), length(16bit uint), value (even byte length)
;        for OB,OW,OF,SQ,UT and UN, length is x00, followed by ulong (32bit) for length
;
;        Implicit layout:
;        tag (2*16bit), length (32 bit unsigned int), value (even byte length)
;
;        vr is a two byte character string as below
;        length is 16 or 32 bit, depending on VR and whether VR is
;        explicit or implicit, ffffffff means undefined
;        value: even length
;
;		 Little endian (the dicom default encoding):
;        binary numbers: least significant byte encoded first, than in increasing order of significance
;        strings       : encoded from left to right
;
;        Big endian:
;        binary numbers: most significant byte encoded first, then decreasing significance
;        strings       : encoded from left to right
;
;        Length of type 1 data types must not be zero, absence of type 1 data types
;        is a protocol violation
;
;        !!! we rely on little endian format for the /explicit stuff!
;        explicit is only supported for the VR types of the meta tags
;        for the rest we write implicit transfer syntax (i.e. no VRs)
;=-----------------------------------------------------------------------------
function mrdicom::setVRtag, group, element, VR, data, explicit=explicit

	explicit = n_elements(explicit) eq 0  ? 0 : 1

	CASE VR of

		; Application Entity
		; maximum 16 bytes, default character repertoire
		'AE': begin
			dval=STRMID(data,0,16)
			rs=self->generate_anytag(group,element,dval,/STR)
		end

		; Age String
		; 4 bytes fixed - format nnnX, where X={D,W,M,Y} (days, weeks, months, years),
		; n from [0,9], X as above
		'AS': begin
			dval=STRMID(data,0,4)
			rs=self->generate_anytag(group,element,dval,/STR)
		end

		; Attribute tag
		; ordered pair of 16 bit unsigned integers representing a data element tag
		; element tag eg. ['0018'x,'00FF'x]
		; little endian representation: 18 00 ff 00
		; big endian representation   : 00 18 00 ff
		'AT': begin
			dval=[self->getbytes(UINT(data[0]),2), self->getbytes(UINT(data[1]),2)]
			rs=self->generate_anytag(group,element,dval)
		end

		; Code string
		; maximum 16 bytes, character string, elements [0-9,_, ]
		'CS': begin
			dval=STRMID(data,0,16)
			rs=self->generate_anytag(group,element,dval,/STR)
		end

		; Date
		; string - 8 bytes fixed, format yyyymmdd, or 10 bytes fixed
		; yyyy.mm.dd, which is compatible with versions prior dicom v3.0 -
		; so that's what will be used
		'DA': begin
			dval=STRMID(data,0,10)
			rs=self->generate_anytag(group,element,dval,/STR)
		end

		; Decimal string
		; string of characters representing a floating point number
		; chars from [+,-,0-9,.,eE], 16 bytes maximum
		'DS': begin
			dval=STRTRIM(STRING(data),1)
			; see comment above. Siemens does not adhere to the standard, so we can't either.
			; dval=STRMID(dval,0,16)
			rs=self->generate_anytag(group,element,dval,/STR)
		end

		; Date/time
		; string, 26 byte maximum string of format
		; YYMMDDHHMMSS.FFFFFF&ZZZZ - the &ZZZZ part ist optional, FFFF ist fraction
		; of seconds
		'DT': begin
			dval=STRMID(data,0,26)
			rs=self->generate_anytag(group,element,dval,/STR)
		end

		; single precision binary floating point number, 4 bytes fixed, IEEE 754,
		; stored here in little endian format (check!!!)
		'FL': begin
			dval=FLOAT(data)  ;explicily cast to bytes
			dvaltmp=BYTE(dval,0,4)
		;fix byteorder (little-endian) if necessary; word length is 16 bits (?is this correct?)
			dval=[self->getbytes(UINT(dvaltmp[0:1],0,1),2), $
			      self->getbytes(UINT(dvaltmp[2:3],0,1),2)]
			rs=self->generate_anytag(group,element,dval)
		end

		; double precision binary floating point number, 8 bytes fixed, IEEE 754
		; stored here in little endian format (check!!!)
		'FD': begin
			dval=DOUBLE(data)	;explicily cast to bytes
			dvaltmp=BYTE(dval,0,8)
			;fix byteorder (little-endian) if necessary; word length is 16 bits (?is this correct?)
			dval=[self->getbytes(UINT(dvaltmp[0:1],0,1),2), $
			      self->getbytes(UINT(dvaltmp[2:3],0,1),2), $
			      self->getbytes(UINT(dvaltmp[4:5],0,1),2), $
			      self->getbytes(UINT(dvaltmp[6:7],0,1),2)]
			rs=self->generate_anytag(group,element,dval)
		end

		; Integer String
		; string of characters representing integer in decimal with optional
		; leading + or -, 12 bytes maximum
		'IS': begin
			dval=STRTRIM(STRING(data),1)
			dval=STRMID(dval,0,12)
			rs=self->generate_anytag(group,element,dval,/STR)
		end

		; Long String
		; maximum 64 bytes, encoding as in [x8,x5], must contain no
		; backslash! IDL doesn't care (??)
		'LO': begin
			dval=STRMID(data,0,64)
			rs=self->generate_anytag(group,element,dval,/STR)
		end

		; Long text
		; 10240 bytes max, IDL doesn't care too much (??), cannot be
		; multivalued, so backslash is allowed
		'LT': begin
			dval=STRMID(data,0,10240)
			rs=self->generate_anytag(group,element,dval,/STR)
		end

		; Other byte string
		; encoding specified by the negotiated transfer syntax
		; insensitive to little/big endian, if necessary padded by 00H to achieve
		; even length
		'OB': begin
			dval=data
			rs=self->generate_anytag(group,element,dval)
			if( explicit eq 1 ) then begin    ; great! See PS 3.5 7.1.2, p.33
				n  = n_elements(rs)-1
				rs = byte([rs[0:3],79,66,0,0,rs[4:n]])
			end
		end

		; Other float string
		; a string of 32bit IEEE 754, requires byte swapping within each 32bit
		; word when changing from little to big endian, max 2^32-4 ! :)
		; 'OF':
		; not implemented

		; Other word string
		; a string of 16bit words, where encoding is specified by TF.
		; swap within 16bit when changing little/big endian
		'OW': begin
			dval=data
			rs=self->generate_anytag(group,element,dval)
		end

		; Person name - a bit complicated
		; for now we handle it like a simple string
		'PN': begin
			dval=data
			rs=self->generate_anytag(group,element,dval)
		end

		; Short string, 16 bytes maximum
		; default character repertoire, no backslash
		'SH': begin
			dval=STRMID(data,0,16)
			rs=self->generate_anytag(group,element,dval,/STR)
			if( explicit eq 1 ) then begin
				n  = n_elements(rs)-1
				rs = byte([rs[0:3],83,72,rs[4:5], rs[8:n]])
			end
		end

		; Signed long
		; signed 32 bit integer, 4 bytes fixed length
		'SL': begin
			dval=self->getbytes(LONG(data),4)
			rs=self->generate_anytag(group,element,dval)
		end

		; Sequence of items - unsupported
		'SQ': begin
			print, 'SQ currently unsupported!'
			rs=BYTE(0)
		end

		; Signed short
		; signed 16 bit integer, 2 bytes fixed length
		'SS': begin
			dval=self->getbytes(FIX(data),2)
			rs=self->generate_anytag(group,element,dval)
		end

		; Short text - 1024 bytes max
		; x5C (\) maybe used
		'ST': begin
			dval=STRMID(data,0,1024)
			rs=self->generate_anytag(group,element,dval,/STR)
		end

		; Time, character string
		; format hhmmss.frac, 16 bytes maximum, frac is 1 millionth of a second (0-999999)
		'TM': begin
			dval=STRMID(data,0,16)
			rs=self->generate_anytag(group,element,dval,/STR)
		end

		; Unique identifier
		; character set [0-9.], 64 bytes maximum
		'UI': begin
			dval=STRMID(data,0,64)
			rs=self->generate_anytag(group,element,dval)
			if( explicit eq 1 ) then begin
				n  = n_elements(rs)-1
				rs = byte([rs[0:3],85,73,rs[4:5], rs[8:n]])
			end
		end

		; Unsigned long
		; unsigned 32 bit integer, 4 bytes fixed
		'UL': begin
			dval=self->getbytes(ULONG(data),4)
			rs=self->generate_anytag(group,element,dval)
			if( explicit eq 1 ) then begin
				n  = n_elements(rs)-1
				rs = byte([rs[0:3],85,76,rs[4:5],rs[8:n]])
			end
		end

		; Unknown
		; string of bytes where encoding is unknown
		'UN': begin
			dval=data
			rs=self->generate_anytag(group,element,dval)
		end

		; Unsigned integer
		; unsigned 16 bit integer, 2 bytes fixed
		'US': begin
			if fix(group) eq 24 and fix(element) eq  4880 then begin
			   dval = ' ';data(0)
			endif else begin
				dval=self->getbytes(UINT(data),2)
			endelse

			rs=self->generate_anytag(group,element,dval)
		end

		; Unlimited text
		; well, limit is 2^32-2, may contain x5C
		'UT': begin
			dval=data
			rs=self->generate_anytag(group,element,dval)
		end

		; removed RT - not defined in the standard
		; unlimited text; could be huge!
		'RT': begin          ;     Rich text??
			dval=data
			rs=self->generate_anytag(group,element,dval)
		end

		else: Begin
			x = dialog_message(['Fatal: unsupported value representation in mrdicom',VR])
			Mdval=data
			rs=self->generate_anytag(group,element,dval)
		endelse
	ENDCASE
	return, BYTE(rs)
end


;+=----------------------------------------------------------------------------
; Name : mrdicom::generate_pixeltag
; Args : group, element, val
; Desc : generate pixel tag
;        we don't handle any overlay data elements here [60xx,3000], but
;        only the pixel tag [7FE0,0010].
;        VR is OW or OB, depending on the negotiated transfer syntax
;        OB shall be unaffected by byte ordering
;        Important other data elements are
;        [0028,0100] bits allocated - defines size of pixel in bits, must be 8 or 16
;        [0028,0101] bits stored    - 8 for '8' allocated, 12 or 16 for '16' allocated
;        [0028,0102] high bit       - shall be one less than bits stored
;
;        Overlay images:
;        bits stored is always '1', overlay images always heva sn image depth of '1'
;        bits allocated [60xx,0100]; bit position [60cc,0102]
;=-----------------------------------------------------------------------------
function mrdicom::generate_pixeltag, group, element, val
	return, BYTE([self->getbytes(group,2),self->getbytes(element,2), self->getbytes(val,4)])
end

;+=----------------------------------------------------------------------------
; Name  : mrdicom::fillCache
; Syntax: obj->fillCache
; Desc  : Private! Fills the tag values from the DICOM object into cache.
;         It might be the case that the user starts a new dicom file 'from scratch'
;         in which case no dicom data is present from which we might fill the cache.
;         Also in this case, the internal data cache is rated as 'valid' after
;         trying to fill the cache.
;=-----------------------------------------------------------------------------
pro mrdicom::fillCache
	self->copy, self
	self.validCache = 1                ; after fill header, tags are always valid!
end

; +=----------------------------------------------------------------------------
; Name : mrdicom::validate
; Type : Function
; Args : verbosity = int  : 0 (default) silent, 1: print error messages to stdout
; Desc : not implemented, just an idea
; Ret  : returns the number of found errors in the dicom file to be written
; =-----------------------------------------------------------------------------
;function mrdicom::validate, verbosity=v

	; [0028,0100] bits allocated must be either 8 or 16
	; [0028,0101] bits stored must be 8 for ba 8 or 12/16 for ba 16
	; [0028,0102] high bit should be bits allocated minus 1

;	return, 0
;end

;+=----------------------------------------------------------------------------
; Name : mrdicom::errno
; Type : Function
; Ret  : Returns the error number of last write command (0 if ok)
;=-----------------------------------------------------------------------------
function mrdicom::errno
	return, self.errno
end

;+=----------------------------------------------------------------------------
; Name : mrdicom::errMsg
; Type : Function
; Ret  : Returns the error description of last write command (empty string if ok)
;=-----------------------------------------------------------------------------
function mrdicom::errMsg
	return, self.errmsg
end


;+=----------------------------------------------------------------------------
; Name  : mrdicom__define
; Type  : Private Procedure (constructor)
;=-----------------------------------------------------------------------------
pro MRdicom__define

		; a couple of words on validCache:
		; the object contains a 'cache' for all dicom tags (well, almost all) that
		; might be present in a file. This cache is not filled immediately when
		; reading the file, as it takes quite a long time to fill it and in many
		; cases the user is only interested in the image data.
		; validCache serves as a flag whether the cache values are valid. Once
		; the cache has been filled, all tag handling is done via the cache.
		; Also see comments in copy and fillCache

	obj = { MRDICOM                                    , $
			inherits                        IDLffDicom , $
			validCache                      :    0     , $ ; flag whether fields below are valid
			filename                        :    ''	   , $ ; either file read from or file to write to
			imgdata                         : ptr_new(), $ ; image data - 2dim (greyscale) or 3dim RGB
			errno                           :    0	   , $ ; Error number of last command
			errmsg                          :    ''	   , $ ; the corresponding errortext
			implementationVersionName  		:    ''    , $ ; set in init
			m_GroupLength                   :    0     , $ ; UL
			m_FileMetaInfoVersion           : bytarr(2), $ ; OB
			m_MediaStoredSOPClassUID        :    ''    , $ ; UI
			m_MediaStoredSOPInstanceUID     :    ''    , $ ; UI
			m_TransferSyntaxUID             :    ''    , $ ; UI
			m_ImplementationClassUID        :    ''    , $ ; UI
			m_ImplementationVersionName     :    ''    , $ ; SH
			m_SourceApplicationEntityTitle  :    ''    , $ ; AE
			i_CharacterSet                  :    ''    , $ ; CS
			i_ImageType                     :    ''    , $ ; CS
			i_SOPClassUID                   :    ''    , $ ; UI
			i_SOPInstanceUID                :    ''    , $ ; UI
			i_StudyDate                     :    ''    , $ ; DA
			i_SeriesDate                    :    ''    , $ ; DA
			i_AcquisitionDate               :    ''    , $ ; DA
			i_ImageDate                     :    ''    , $ ; DA
			i_StudyTime                     :    ''    , $ ; TM
			i_SeriesTime                    :    ''    , $ ; TM
			i_AcquisitionTime               :    ''    , $ ; TM
			i_ImageTime                     :    ''    , $ ; TM
			i_AccessionNumber               :    ''    , $ ; SH
			i_Modality                      :    ''    , $ ; CS
			i_Manufacturer                  :    ''    , $ ; LO
			i_InstitutionName               :    ''    , $ ; LO
			i_InstitutionAddress            :    ''    , $ ; ST
			i_PhysiciansName                :    ''    , $ ; LO
			i_StationName                   :    ''    , $ ; SH
			i_StudyDescription              :    ''    , $ ; LO
			i_SeriesDescription             :    ''    , $ ; LO
			i_OperatorsName                 :    ''    , $ ; PN
			i_ManufacturerModelName         :    ''    , $ ; LO
;			i_RefStudyComponentSequence     :    ''    , $ ; SQ
			i_RefSOPClassUID                :    ''    , $ ; UI
			i_RefSOPInstanceUID             :    ''    , $ ; UI
;			i_RefImageSequence              :    ''    , $ ; SQ
			p_PatientName                   :    ''    , $ ; LO
			p_PatientId                     :    ''    , $ ; LO
			p_PatientBirthdate              :    ''    , $ ; DA
			p_PatientSex                    :    ''    , $ ; CS
			p_PatientAge                    :    ''    , $ ; AS
			p_PatientWeight                 :    ''    , $ ; DS
			p_PatientAlerts                 :    ''    , $ ; LO
			p_PregnancyStatus               :     0    , $ ; US
			p_PatientComments               :    ''    , $ ; LO
			a_ContrastBolusAgent            :    ''    , $ ; LO
			a_BodyPartExamined              :    ''    , $ ; CS
			a_ScanningSequence              :    ''    , $ ; CS
			a_SequenceVariant               :    ''    , $ ; CS
			a_ScanOptions                   :    ''    , $ ; CS
			a_AcquisitionType               :    ''    , $ ; CS
			a_SequenceName                  :    ''    , $ ; SH
			a_AngioFlag                     :    ''    , $ ; CS
			a_SliceThickness                :    ''    , $ ; DS
			a_RepetitionTime                :    ''    , $ ; DS
			a_EchoTime                      :    ''    , $ ; DS
			a_InversionTime                 :    ''    , $ ; DS
			a_NumberOfAverages              :    ''    , $ ; DS
			a_ImagingFrequency              :    ''    , $ ; DS
			a_ImagedNucleus                 :    ''    , $ ; SH
			a_EchoNumber                    :    0     , $ ; IS
			a_MagneticFieldStrength         :    ''    , $ ; DS
			a_SpacingBetweenSlices          :    ''    , $ ; DS
			a_NumberOfPhaseEncodingSteps    :    0     , $ ; IS
			a_EchoTrainLength               :    0     , $ ; IS
			a_PercentSampling               :    ''    , $ ; DS
			a_PercentPhaseFieldOfView       :    ''    , $ ; DS
			a_PixelBandwidth                :    ''    , $ ; DS
			a_DeviceSerialNumber            :    ''    , $ ; LO
			a_SoftwareVersion               :    ''    , $ ; LO
			a_ProtocolName                  :    ''    , $ ; LO
			a_ContrastBolusVolume           :    ''    , $ ; DS
			a_DateOfLastCalibration         :    ''    , $ ; DA
			a_TimeOfLastCalibration         :    ''    , $ ; TM
			a_ReceivingCoil                 :    ''    , $ ; SH
			a_TransmittingCoil              :    ''    , $ ; SH
			a_PhaseEncodingDirection        :    ''    , $ ; CS
			a_FlipAngle                     :    ''    , $ ; DS
			a_VariableFlipAngle             :    ''    , $ ; CS
			a_SAR                           :    ''    , $ ; DS
			a_DBDT                          :    ''    , $ ; DS
			a_PatientPosition               :    ''    , $ ; CS
			r_StudyInstanceUID              :    ''    , $ ; UI
			r_SeriesInstanceUID             :    ''    , $ ; UI
			r_StudyID                       :    ''    , $ ; SH
			r_SeriesNumber                  :    0     , $ ; IS - always written
			r_AcquisitionNumber             :    0     , $ ; IS - always written
			r_ImageNumber                   :    0     , $ ; IS - always written
			r_ImagePosition                 :    ''    , $ ; RT
			r_ImagePositionPatient          :    ''    , $ ; DS
			r_ImageOrientation              :    ''    , $ ; RT
			r_ImageOrientPatient            :    ''    , $ ; DS
			r_Location                      :    ''    , $ ; RT
			r_FrameOfReference              :    ''    , $ ; UI
			r_ImageGeometryType             :    ''    , $ ; RT
			r_MaskingImage                  :    ''    , $ ; RT
			r_AcqusitionInSeries            :    ''    , $ ; RT
			r_PositionReferenceIndicator    :    ''    , $ ; LO
			r_SliceLocation                 :    ''    , $ ; DS
			r_ImageComments                 :    ''    , $ ; LT
			g_SamplesPerPixel               :    0     , $ ; US - always written
			g_PhotometricInterpretation     :    ''    , $ ; CS
			g_ImageDimension                :    ''    , $ ; RT
			g_PlanarConfiguration           :    0     , $ ; US - not written
			g_Rows                          :    0     , $ ; US - always written
			g_Columns                       :    0     , $ ; US - always written
			g_PixelSpacing                  :    ''    , $ ; DS
			g_ImageFormat                   :    ''    , $ ; RT
			g_BitsAllocated                 :    0     , $ ; US - always written
			g_BitsStored                    :    0     , $ ; US - always written
			g_HighBit                       :    0     , $ ; US - always written
			g_PixelRepresentation           :    0     , $ ; US
			g_SmallestImagePixelValue       :    0     , $ ; US
			g_LargestImagePixelValue        :    0     , $ ; US
			g_WindowCenterWidthExplanation  :    ''    , $ ; LO
			g_ImageLocation                 :    ''    , $ ; RT
			g_WindowCenter                  :    ''    , $ ; DS
			g_WindowWidth                   :    ''      $ ; DS
	}
end



;
; Here's Bhautik Joshi's Copyright and disclaimer:
;
; NAME:
; 	DICOM_WRITER
;
; VERSION:
;	0.21
;
; PURPOSE:
;	Generate a dicom file from within RSI IDL
;
; AUTHOR:
;	Bhautik Joshi
;
; EMAIL:
;	bjoshi@geocities.com
;
; HOMEPAGE:
;	http://cow.mooh.org
;
; USE:
;	DICOM_WRITER, filename, image, VOXELSIZE = voxelsize, SSAI = ssai, $
;		PATIENT = patient, PHYSICIAN = physician, PATID = patid
;
; INPUT:
;	filename - string containing name of dicom file to be written to
;	image - Integer (BYTE, FIX, UINT, LONG or ULONG) image - type and bpp
;	is now automagically set
;
; OPTIONAL PARAMETERS
;	voxelsize - Array of 3 floating point values representing voxel size
;	with the format [x,y,z], otherwise set to default of [1.0,1.0,1.0]
;	ssai - Array of 4 integer values representing studyID, seriesnum,
;       acqnum,imagenum, with the format [studyID,seriesnum,acqnum,imagenum],
;	otherwise set to default of [0,0,0,0]
;	patient - patient name, if not defined, set to dummy name
;	physician - physician name, if not defined, set to dummy name
;	patid - patiend ID, if not defined, set to dummy name
;
; NOTES ON USAGE (READ! IMPORTANT!):
;	* At the moment the program only writes to a single slice
;	* Extra dicom tags can be easily added (see body of program, especially
;	  generate_VRtag function)
;	* There is little to no error-checking at the moment, so be
;	  careful!
;	* Analyse seems to need a minimum image size of somewhere around
;	  100x100
;	* IMPORTANT: The DICOM writer tries to write 'Implicit VR' type
;	  DICOM files - see DICOM standard PS 3.5-1999, part 7
;	* Can write most VR (Value Represenation) tags via new function,
;	  generate_VRtag. Currently supported:
;	  AE, AS, AT, CS, DA, DS, DT, FL, FD, FD, IS, LO, LT, OB, OW,
;	  SH, SL, SS, ST, TM, UI, UL, UN, US, UT
;	  and SQ, PN unsupported (I got away with using UI in place of PN)
;	* See comments near generate_VRtag function for notes on usage
;	  of the function for adding your own additional tags
;
; EXAMPLE:
;	Create a horrendously boring byte image and store it in a
;	dicom file, test.dcm, with voxel dimensions of [2.0,3.0,4.0],
;	and studyid=1,series number=2,acquisiton number=3 and image
;	number=4:
;
;	> rows = 200
;	> cols = 200
;	> image = indgen(rows,cols)
;	> dicom_writer, 'test.dcm', image, voxelsize=[2.0,3.0,4.0], ssai=[1,2,3,4]
;
; HISTORY:
;	Based on Marc O'Briens (m.obrien@sghms.ac.uk) TIFF_to_DICOM.c
;	version 0.1 	08-01-2002 - first working version produced
;	version 0.11	09-01-2002 - fixed endian-ness issue & added get_lun
;				     functionality
;	version 0.2	14-01-2002 - many fixes and additions, including:
;		* replaced most generate_* functions with generate_VRtag
;		* support for many VR (Value representation) types
;		* Autodetection of little/big endian architecture and
;		  automagic byte ordering as necessary (for tags/image)
;		* automagically detect image type and set bpp as necessary
;		* more data in the header can be set manually
;	version 0.21	15-01-2002 - uploaded all over the place & fixed bug
;				     that didn't update patient, patid etc.
;
; TODO:
;	* Allow for more robust dicom writing
;	* Part 10 compliance (!!!!!!!!!!!)
;	* Decent error checking
;
; DISCLAIMER:
;
; Permission to use, copy, modify, and distribute this software and its
; documentation for any purpose and without fee is hereby granted,
; provided that the above copyright notice appear in all copies and that
; both that copyright notice and this permission notice appear in
; supporting documentation.
;
; This file is provided AS IS with no warranties of any kind.  The author
; shall have no liability with respect to the infringement of copyrights,
; trade secrets or any patents by this file or any part thereof.  In no
; event will the author be liable for any lost revenue or profits or
; other special, indirect and consequential damages.
;
; The author accepts no responsibility for any action arising from use of
; this package. The software is not guaranteed to write compliant DICOM
; files. If it causes damage to you or your system, you have been warned -
; this is a work in progress. If it bites your dog, its not my fault. If
; it causes you to curl up on the floor in the foetal position muttering
; about pixies and mushrooms, its not my fault. If it causes you or someone
; else to spontaneously burst into song and dance, its not my fault but
; I'd like to hear about it. You have been warned.
;
