;+=----------------------------------------------------------------------------
; Name   : mrdisplay
; RCS    : $Id: mrdisplay__define.pro,v 2.3 2004/11/24 10:43:18 olivert Exp olivert $
; Methods:
;          <b>Constructor</b>
;          mrdisplay::init             : constructor, 2nd part
;          <br>
;          <b>Public variables (get)</b>
;          mrdisplay::img              : returns a pointer to currently used image data
;          mrdisplay::imgXsize         : return image size in x in pixels
;          mrdisplay::imgYsize         : return image size in y in pixels
;          mrdisplay::imgMinval        : return minimum absolute image value
;          mrdisplay::imgMaxval        : return maximum absolute image value
;          mrdisplay::interpolate      : return current interpolation mode
;          mrdisplay::scaling          : return the scaling factor
;          mrdisplay::location         : return current image location within window
;          mrdisplay::contrast         ; return current contrast center and width
;          mrdisplay::gamma            ; return the currently used gamma value
;          mrdisplay::grWindow         : return the main window object
;          mrdisplay::screenXsize      : return visible canvas size in x in pixels
;          mrdisplay::screenYsize      : return visible canvas size in y in pixels
;          mrdisplay::labelView        : return the view meant for labels
;          mrdisplay::imgView          : return the view which holds the image
;          mrdisplay::palette          : return the currently used palette
;          <br>
;          <b>Public variables (set)</b>
;          mrdisplay::setInterpolate   : switch interpolation on or off
;          mrdisplay::setScaling       : set the scaling factor
;          mrdisplay::setLocation      : set image location in window
;          mrdisplay::setContrast      : same, but via center and width
;          mrdisplay::setGamma         ; set the gamma value of the used palette
;          mrdisplay::setPalette       : set the palette colors by passing RGB vectors
;          mrdisplay::setPaletteRGB    : set RGB colorset of the palette or a predefined color table
;          <br>
;          <b>General</b>
;          mrdisplay::setImg           : set the image to be displayed
;          mrdisplay::updateImg        : as setImg, but keep all settings
;          mrdisplay::clear            : clear the display - automatically displays
;          mrdisplay::display          : display object content on screen
;          mrdisplay::scroll           : scroll image by the specified amount
;          mrdisplay::centerImg        : center image in the middle of the screen
;          mrdisplay::fitToScreen      : resize image so it fits screen size
;          <br>
;          <b>Event handling</b>
;          mrdisplay::activate         : generate events
;          mrdisplay::deactivate       : no events no more
;          mrdisplay::registerEvt      : register an event
;          mrdisplay::requestDisplay   : call display after callbacks are done
;          mrdisplay::unregisterEvt    : unregister an event
;          mrdisplay::unregisterObj    : unregister a whole object
;          mrdisplay::unregisterAll    : remove all entries from callback tables
;          <br>
;          <b>Private</b>
;          mrdisplay_evthandler        : non-object event handler calls the object handler
;          mrdisplay::evthandler       : the object handler
;          mrdisplay::resolveTypeStr   : translate an event string into an event type
;          mrdisplay__define           : constructor, 1st part
;          mrdisplay::cleanup          : the destructor
; Desc   : An object to display a single indexed image. Full support for true color
;          images will be available in the next version.
;          <br>
;          !!! events need cleanup to include only the fields which are defined !!!
;          <b>Event handling:</b>
;          mrDisplay issues events on different conditions. The following event types are
;          presently defined - objects may register for one or more event types:
;          <br>
;          <b>Event types:</b>
;          Type (number)     string             occurence
;          0                'press'             any button press event
;          1                'dblclick'          any button double click event
;          2                'release'           any button release event
;          3                'motion'            mouse has been moved inside window
;          4                'keypress'          a key has been pressed inside window
;          5                'imgchange'         an image has been set or deleted
;          6                'palettechange'     the used palette has been changed (setpalette)
;          <br>
;          <b>The event structure</b>
;          > x = { mrdisplayEvt          ,$
;          >     type      :   0           ,$ ; (all) an integer denoting the event type, see below
;          >     top       :   0L          ,$ ; (0-4) the top widget id
;          >     display   :   obj_new()   ,$ ; (0-4) the display object itself
;          >     grWindow  :   obj_new()   ,$ ; (0-4) the IDLgrWindow object within the draw widget
;          >     imgView   :   obj_new()   ,$ ; (0-4) this objects image view window
;          >     lblView   :   obj_new()   ,$ ; (0-4) this objects label view window
;          >     x         :   0           ,$ ; x position of the event, device coordinates
;          >     y         :   0           ,$ ; y position of the event, device coordinates
;          >     dx        :   0           ,$ ; x difference since last event
;          >     dy        :   0           ,$ ; y difference since last event
;          >     imgx      :   0           ,$ ; x position of the event, image coordinates
;          >     imgy      :   0           ,$ ; y position of the event, image coordinates
;          >     key       :   0L          ,$ ; key off an ASCII key press
;          >     modifiers :   0L          ,$ ; bitmask: 1 = shift, 2 = control
;          >     button    :   0           ,$ ; bitmap, left, middle right mouse button, pressed/released
;          >                                $ ; button for press/release events, current status in motion evts
;          >     buttonstatus: 0            $ ; current status of pressed/released mouse buttons, bitmap
;          > }
;         <br>
; Author : (c) Oliver Thilmann, oliver.thilmann@web.de
; Licns  : This program is published under the GPL. See gpl.txt for details
; TODO   : add a background view so that the background is black, not white
;=-----------------------------------------------------------------------------

;==============================================================================
;
;                          CONSTRUCTORS AND DESTRUCTORS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::init
; Syntax: obj = obj_new('mrdisplay', baseid, x, y, xscroll, yscroll, /interpolate)
; Args  : <b>baseid</b>
;         The base widget in which to create the drawing widget
;         <b>x, y</b>
;         Widget screen size in x and y in pixels
;         <b>xscroll, yscroll</b>
;         Canvas size in x and y in pixels. If canvas size is greater than
;         screen size, scroll bars are automatically added.
;         <b>interpolate</b>
;         Set to '1' for interpolation. Default is '0' (off)
; Desc  : initialization of the mrdisplay widget
; Code  : disp = obj_new('MRdisplay', baseid, 512, 512, 512, 512, interpolate=0)
;         img = indgen( 512, 512 )
;         imgptr = ptr_new( img, /no_copy )
;         disp->setImg, imgptr
;         disp->display
; Code  : For code examples on how to handle call back events, see e.g. the
;         source of the mrcontrast object, specfically mrcontrast::selfregister
;         and mrcontrast::mrdisplayEvtHandler
;         See mrdisplay::registerEvt for the available event types.
;=-----------------------------------------------------------------------------
function mrdisplay::init, baseid, x, y, xscroll, yscroll, interpolate=ip

       ; first three args are mandatory, xscroll,yscroll = x,y if unset

	if( n_elements(x) eq 0 or n_elements(y) eq 0 or n_elements(baseid) eq 0 ) then return, 0
	if( n_elements(xscroll) eq 0 ) then xscroll = x
	if( n_elements(yscroll) eq 0 ) then yscroll = y
	if( n_elements(ip) eq 0 )      then ip = 0       ; default is no interpolation

		; create widget base (so we might add more later, if we want to) and
 		; the draw widget. The draw widget's uvalue holds a pointer to this object
 		; this is later needed for event handling, color_model=0 means RGB colors

	self.wid_base = widget_base( baseid, xsize=x, ysize=y )
	self.wid_draw = widget_draw( self.wid_base, xsize=x, ysize=y, $
                   scroll=1, x_scroll_size=xscroll, y_scroll_size=yscroll, $
                   graphics_level=2, renderer=1, retain=2, color_model=0 )
	self.wid_xsize = x
	self.wid_ysize = y
	widget_control, self.wid_draw, set_uvalue=self           ; for evt. handling

       ; set some last object values and defaults

 	evtnames = ['press','dblclick','release','motion','keypress','imgchange','palettechange']
	cbo                     = objarr( 7 )                  ; create 7 lists of call back objects
	for i = 0, 6 do cbo[i]  = obj_new('IDL_Container')
	self.callBackObjects    = ptr_new( cbo )
	self.interpolate        = ip
	self.scaling            = 1.
	self.palette            = obj_new( 'IDLgrPalette' )      ; load and set default greyscale
	self.palette->loadct, 0                                  ; palette
	self.imgMinVal          = 0
	self.imgMaxVal          = 0
	self.center             = .5
	self.width              = 1.
	self.imgXsize           = 0
	self.imgYsize           = 0
	self.displayRequestFlag = 0

	self.lastx				= 0
	self.lasty				= 0
	self.buttonstatus		= 0

       ;. one view for the image, which scales and moves
       ;. one view for labels, fixed to the image frame,
       ;. one Scene to display it all,
       ;. in the land of IDL, where the shadows lie

	self.grScene   = obj_new('IDLgrScene')
	self.grImgView = obj_new('IDLgrView', viewplane_rect=[0.,0.,1.,1.] )
	self.grLblView = obj_new('IDLgrView', viewplane_rect=[0.,0.,1.,1.], location=[0,0], $
						dimensions=[self.wid_xsize, self.wid_ysize], /transparent)
	self.grScene->add, self.grImgView
	self.grScene->add, self.grLblView

	return, 1
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::cleanup
; Syntax: obj_destroy, obj
; Desc  : The destructor
;=-----------------------------------------------------------------------------
pro mrdisplay::cleanup
	obj_destroy, self.grImage
	obj_destroy, self.grModel
	obj_destroy, self.grimgView
	obj_destroy, self.grlblView
	obj_destroy, self.grScene
	obj_destroy, self.palette
		; empty container before destruction, so we don't kill the plugins!
	for i = 0, n_elements(*self.callBackObjects)-1 do begin
		(*self.callBackObjects)[i]->remove, /all
		obj_destroy, (*self.callBackObjects)[i]
	end
	ptr_free, self.callBackObjects
end

;==============================================================================
;
;                               GET AND SET METHODS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::setImg
; Syntax: obj->setImg, image, top=integer, /no_copy, /update
; Args  : <b>image</b>
;         Pointer to a two-dimensional matrix with the image data.
;         <b>top</b>
;         Integer: top byte value to which to scale the image in the range
;         [0,255], default is 255
;         <b>no_copy</b>
;         Don't make a copy of the image data, but use the reference directly.
;         Mind that the image will be free'd, when a new image is set or the
;         object destroyed!
;         <b>update</b>
;         Equivalent to calling mrdisplay::updateImg - which will go
;         eventually.
; Desc  : Hand the image over to mrdisplay. Unless the /no_copy keyword is set,
;         mrDisplay will make a copy of the image data. Note that the image
;         will not be displayed until the user calls mrdisplay::display
;         Setting an image causes a 'ImgChange' event.
;=-----------------------------------------------------------------------------
pro mrdisplay::setImg, imgptr, top=top, no_copy=no_copy, update=update

	if( not ptr_valid(imgptr[0]) ) then begin
		message, 'Invalid image pointer'
	end

	no_copy = keyword_set( no_copy )
	if( n_elements(top) ne 0 ) then begin
		top = top > 0
		top = top < 255
	end else begin
		top=255
	end

	if( keyword_set(update) ) then begin
		self->updateImg, imgptr, top=top, no_copy=no_copy
		return
	end

		; if self.grModel exists, remove it from the self.grView container

	if( self.grModel ne obj_new() ) then begin
		self.grImgView->remove, self.grModel
		obj_destroy, self.grImage
		obj_destroy, self.grModel
	end
	ptr_free, self.imgData

		; is that an indexed or a true color image?

	imgsize            = size( *imgptr, /dimensions )
	case size( *imgptr, /n_dimensions ) of
	2: begin
		self.imgData       = ptr_new( *imgptr, no_copy=no_copy )   ; create a copy or don't
		self.topbytscl     = top
	    img                = bytscl( *self.imgData, top=self.topbytscl )
		self.imgMinval     = min( *self.imgData, max=maxval )
	    self.imgMaxval     = maxval
		self.imgXsize      = imgsize[0]
		self.imgYsize      = imgsize[1]
	end
	3: begin
		self.imgData       = ptr_new( *imgptr, no_copy=no_copy )   ; create a copy or don't
		self.topbytscl     = 255
		img                = *self.imgData
		self.imgMinval     = min( *self.imgData, max=maxval )
	    self.imgMaxval     = maxval
		self.imgXsize      = imgsize[1]
		self.imgYsize      = imgsize[2]
	end
	else: begin
		message, 'mrDisplay only supports (n,m) and (3,n,m) (and permutations) image layout'
 	end
 	end

	self.center        = 0.5                        ; reset contrast settings
	self.width         = 1.0

       ; now create the object hierarchy and link the objects
       ; the image model must be @ position 0 in the grView container!

    self.grModel = obj_new('IDLgrModel')
    self.grImage = obj_new('IDLgrImage', img, $
                   sub_rect    = [0,0,self.imgXsize, self.imgYsize], $
                   xcoord_conv = [0.,1.], $
                   ycoord_conv = [0.,1.], $
                   dimensions  = [1.,1.], $
                   palette     = self.palette, $
                   interpolate = self.interpolate )

       ; As the dimensions of the view should cover the whole image (incl. scaling)
       ; but cover also the whole display area, so the user is free to move the
       ; image wherever he wants, panning is done by changing the 'location' property
       ; of the View

    self.scaling = 1
    self.grImgView->setProperty, $
         dimensions = [self.imgXsize, self.imgYsize], transparent=1, $
         location   = [(self.wid_xsize-self.imgXsize)/2,(self.wid_ysize-self.imgYsize)/2], $
         viewplane_rect=[0,0,1,1], color=[255,255,255]
    self.grImgView->add,  self.grModel, position=0
    self.grModel->add, self.grImage

       ; create ImgChange events (type 5)

	evt = ptr_new( {type:5, display:self}, /no_copy )
    objs = (*self.callbackobjects)[5]->get( /all, count=nobj )
    if( nobj gt 0 ) then begin
       for i = 0, nobj-1 do objs[i]->mrdisplayEvtHandler, evt
    end
    ptr_free, evt
	if( self.displayRequestFlag ) then self->display
end


;+=----------------------------------------------------------------------------
; Name  : mrdisplay::updateImg
; Syntax: <b>Deprecated!</b>
;         Use
;         obj->setImg, image, top=integer, /no_copy, /update
;         instead
; Desc  : As mrdisplay::setImg, but all image settings (scaling, location,
;         contrast) are kept - only the image data is replaced. The new image
;         must have the same size and the same palette as the old one, or the
;         results may be unexpected.
;         Calling this routine causes 'ImgChange' events.
;=-----------------------------------------------------------------------------
pro mrdisplay::updateImg, imgptr, top=top, no_copy=no_copy

       ; argument checking

	no_copy = keyword_set( no_copy )
	if( self.grImage eq obj_new() ) then message, 'No image present for update'
	if( imgptr eq ptr_new()       ) then message, 'Void image pointer argument'
	if( n_elements(top) ne 0 ) then begin
		top = top > 0
		top = top < 255
	end else begin
		top=255
	end

		; destroy current image and save the new one (possibly making a copy )

	ptr_free, self.imgData
    self.topbytscl = top
    self.imgMinval = min( *imgptr, max=maxval )
    self.imgMaxval = maxval
    self.imgData   = ptr_new( *imgptr, no_copy=no_copy )
    imgsize        = size( *self.imgData, /dimensions )
    self.imgXsize  = imgsize[0]
    self.imgYsize  = imgsize[1]

		; setting the contrast with the current values will copy the new
		; image data to the grImage object while respecting the new min/max values
		; i.e. calling this method, sets the image!

	self->setContrast, self.center, self.width

		; finally handle imageChange events (type 5)

	evt = ptr_new( {type:5, display:self}, /no_copy )
    objs = (*self.callbackobjects)[5]->get( /all, count=nobj )
	if( nobj gt 0 ) then begin
		for i = 0, nobj-1 do objs[i]->mrdisplayEvtHandler, evt
	end
	ptr_free, evt
	if( self.displayRequestFlag ) then self->display

end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::clear
; Syntax: obj->clear
; Desc  : Clear the screen. The currently displayed image is free'd, that is
;         if mrdisplay::setImg was called with keyword no_copy set, the
;         image data is gone.
;         This method creates imgchange events.
;=-----------------------------------------------------------------------------
pro mrdisplay::clear
	self.grImgView->remove, self.grModel
	obj_destroy, self.grImage
	obj_destroy, self.grModel
	ptr_free,    self.imgData
	self.imgXsize = 0
	self.imgYsize = 0
	self.grModel   = obj_new()
	self.grImage   = obj_new()
	self.imgData   = ptr_new()
	self.grImgView->setProperty, color=[0,0,0]
    widget_control, self.wid_draw, get_value=grWindow
    if( grWindow eq obj_new() ) then return          ; widget_draw not realized

       ; create ImgChange events (type 5)

	evt = ptr_new( {type:5, display:self}, /no_copy )
    objs = (*self.callbackobjects)[5]->get( /all, count=nobj )
    if( nobj gt 0 ) then begin
       for i = 0, nobj-1 do objs[i]->mrdisplayEvtHandler, evt
    end
    ptr_free, evt
	self->display  ; clear always redisplays
end


;+=----------------------------------------------------------------------------
; Name  : mrdisplay::img
; Syntax: img = obj->img()
; Desc  : Return pointer to currently used image data
;=-----------------------------------------------------------------------------
function mrdisplay::img
	return, self.imgData
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::interpolate
; Syntax: ip = obj->interpolate()
; Desc  : Return the current interpolation status
;=-----------------------------------------------------------------------------
function mrdisplay::interpolate
    return, self.interpolate
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::setInterpolate
; Syntax: obj->setInterpolate, /onoff
; Args  : <b>onoff</b>
;         It's either 1 (on) or 0 (off)
; Desc  : Set ot unset interpolation. Does not automatically redisplay.
;=-----------------------------------------------------------------------------
pro mrdisplay::setInterpolate, onoff
	if( onoff gt 0 ) then self.interpolate = 1 else self.interpolate = 0
	if( self.grImage ne obj_new() ) then $
		self.grImage->setProperty, interpolate=self.interpolate
	return
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::scaling
; Syntax: s = obj->scaling()
; Desc  : Returns the scaling factor
;=-----------------------------------------------------------------------------
function mrdisplay::scaling
    return, self.scaling
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::setScaling
; Syntax: obj->setScaling, scaling_factor
; Args  : <b>scaling_factor</b>
;         The new scaling factor
; Desc  : Set the scaling factor - this will not redisplay the image
;=-----------------------------------------------------------------------------
pro mrdisplay::setScaling, scale
    self.grImgView->getProperty, dimensions=dim, location=loc
    dim[0] = self.imgXsize * scale
    dim[1] = self.imgYsize * scale
    loc[0] = loc[0] - self.imgXsize/2. * (scale - self.scaling)
    loc[1] = loc[1] - self.imgYsize/2. * (scale - self.scaling)
    self.grImgView->setProperty, dimensions=dim, location=loc
    self.scaling = scale
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::imgXsize
; Syntax: xSize = obj->imgXsize()
; Ret   : Returns the image's original (i.e. unscaled) size in x
;=-----------------------------------------------------------------------------
function mrdisplay::imgXSize
    return, self.imgXsize
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::imgYsize
; Syntax: ySize = obj->imgYsize()
; Ret   : returns the image's original (i.e. unscaled) size in y
;=-----------------------------------------------------------------------------
function mrdisplay::imgYSize
    return, self.imgYsize
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::imgMinval
; Syntax: min = obj->imgMinval()
; Ret   : returns the image's real (i.e. not scaled to [0:top]) minimum pixel value
;=-----------------------------------------------------------------------------
function mrdisplay::imgMinval
    return, self.imgMinval
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::imgMaxval
; Syntax: max = obj->imgMaxval()
; Ret  : returns the image's (unscaled) maximum pixel value
;=-----------------------------------------------------------------------------
function mrdisplay::imgMaxval
    return, self.imgMaxval
end


;+=----------------------------------------------------------------------------
; Name  : mrdisplay::contrast
; Syntax: cw = obj->contrast()
; Ret   : Returns double array with two elements, both in range [0.,1.]
;         cw[0] = current contrast center
;         cw[1] = current contrast width
;=-----------------------------------------------------------------------------
function mrdisplay::contrast
    return, [self.center, self.width]
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::setContrast
; Syntax: obj->setContrast, center, width
; Args  : <b>center</b>
;         Relative contrast center, 0. <= center <= 1.
;         <b>width</b>
;         Relative contrast width, 0. <= width <= 1.
; Desc  : Set center and width of the image contrast. Both values are relative
;         (i.e. [0.,1.]) and internally scaled to absolute image values
; Code  : mrdisplay->setContrast, .5, 1.       ; resets to full width
;=-----------------------------------------------------------------------------
pro mrdisplay::setContrast, center, width
	on_error, 2
	if( n_elements(center) ne 1 or n_elements(width) ne 1 ) then message, 'Missing or wrong argument'
    if( self.imgData eq ptr_new() ) then return
    self.center = center < 1. > 0.
    self.width  = width  < 1. > 0.
    low = center-width > 0.
    hig = center+width < 1.
    dispmin            = self.imgMinval + (self.imgMaxval-self.imgMinval)*low
    dispmax            = self.imgMinval + (self.imgMaxval-self.imgMinval)*hig
    self.grImage->setProperty, data=bytscl( *(self.imgData), min=dispmin, max=dispmax, top=self.topbytscl )
end


;+=----------------------------------------------------------------------------
; Name  : mrdisplay::setGamma
; Syntax: obj->setGamma, gamma
; Args  : <b>gamma</b>
;         Gamma value for the image's palette 0.1 < gamma < 10.
; Desc  : Set the gamma value for the image's palette
;=-----------------------------------------------------------------------------
pro mrdisplay::setGamma, gamma
	self.palette->setProperty, gamma=gamma
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::gamma
; Syntax: gamma = obj->gamma()
; Desc  : Return the current gamma value of the image palette
;=-----------------------------------------------------------------------------
function mrdisplay::gamma
	self.palette->getProperty, gamma=gamma
	return, gamma
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::setLocation
; Syntax: obj->setLocation, x, y
; Args  : <b>x,y</b>
;         New x,y location of the image
; Desc  : Scroll the image position to the given values
;=-----------------------------------------------------------------------------
pro mrdisplay::setLocation, x, y
    self.grImgView->setProperty, location=[x,y]
end


;+=----------------------------------------------------------------------------
; Name  : mrdisplay::location
; Syntax: loc = obj->location()
; Desc  : Return the image position within the window in a 2-element array.
;         (It's the view's position, to be precise)
;=-----------------------------------------------------------------------------
function mrdisplay::location
     self.grImgView->getProperty, location=loc
     return, loc
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::screenXsize
; Syntax: win = obj->screenXsize()
; Ret   : Returns the visible canvas size in x in pixels ('x' argument in init)
;=-----------------------------------------------------------------------------
function mrdisplay::screenXsize
	return, self.wid_xsize
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::screenYsize
; Syntax: win = obj->screenYsize()
; Ret   : Returns the visible canvas size in y in pixels ('y' argument in init)
;=-----------------------------------------------------------------------------
function mrdisplay::screenYsize
	return, self.wid_ysize
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::grWindow
; Syntax: win = obj->grWindow()
; Ret   : The IDLgrWindow object associated with the drawing widget
;         The return value is obj_new(), if the draw widget has not been
;         realized yet.
;=-----------------------------------------------------------------------------
function mrdisplay::grWindow
    widget_control, self.wid_draw, get_value=grWindow
    return, grWindow
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::labelView
; Syntax: view = obj->labelView()
; Ret   : The used IDLgrView object for display of labels. The user may
;         add models to this object which then are displayed at a fixed position
;         on the screen. Changing any of the view's properties means asking for
;         trouble, so don't do it. You may only add models to this view.
;         The view's coordinates are normalized!
;=-----------------------------------------------------------------------------
function mrdisplay::labelView
    return, self.grLblView
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::imgView
; Syntax: view = obj->imgView()
; Ret   : The used IDLgrView object for display of the image. Again the
;         user may add image related objects to this view (e.g. rois) which will
;         be scaled and moved along with the image.
;=-----------------------------------------------------------------------------
function mrdisplay::imgView
    return, self.grimgView
end


;+=----------------------------------------------------------------------------
; Name  : mrdisplay::setPalette
; Syntax: obj->setPalette, palette
; Args  : <b>palette</b>
;         An object of class IDLgrPalette
; Desc :  Copies the RGB values of the argument palette object to mrdisplay's
;         internal palette object. Please note: copies! Otherwise there's the
;         risk that the user destroys the palette object.
;=-----------------------------------------------------------------------------
pro mrdisplay::setPalette, palette
	if( n_elements(palette) eq 0 ) then message, 'missing palette argument'
	if( not obj_isa( palette, 'IDLgrPalette' ) ) then message, 'wrong palette argument'
	palette->getProperty, red=r, green=g, blue=b
	self.palette->setProperty, red=r, green=g, blue=b

		; create palettchange events

	evt = ptr_new( {type:6, display:self}, /no_copy )
    objs = (*self.callbackobjects)[6]->get( /all, count=nobj )
    if( nobj gt 0 ) then begin
       for i = 0, nobj-1 do objs[i]->mrdisplayEvtHandler, evt
    end
    ptr_free, evt
	if( self.displayRequestFlag ) then self->display
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::setPaletteRGB
; Syntax: obj->setPaletteRGB, red=intarr, green=intarr, blue=intarr, colorTable=integer
; Args  : <b>red, green, blue</b>
;         Integer arrays with 256 elements holding the red/green/blue values of
;         the palette
;         <b>colorTable</b>
;         An integer value between 0 and 41 indexing one of IDL's predefined
;         color tables. If set to 42, a logarithmically divided colortable is
;         loaded. If colorTable is set, keywords red/green/blue are ignored.
; Desc :  Sets the red/green/blue values of the image palette used to display
;         the image.
;=-----------------------------------------------------------------------------
pro mrdisplay::setPaletteRGB, red=red, green=gre, blue=blu, colorTable=ct
	if( n_elements(ct) ne 0 ) then begin
		if( ct lt 0 or ct gt 42 ) then message, 'Color table index out of bounds'
		if( ct lt 42 ) then begin
			self.palette->loadct, ct
		end else begin
			red = intarr( 256 )
			gre = intarr( 256 )
			blu = intarr( 256 )
			rs = [0,0,0,0,0,0,0,0,0,255,255,255,255,255,230,210]
			gs = [0,32,32,64,0,160,160,208,255,255,204,128,102,0,26,0]
			bs = [0,112,144,192,255,255,160,64,0,0,102,64,0,0,0,0]
				; maximum ratio is '8'
			scalefactor = 16./alog( 255./32.+1.)
			for i = 0, 255 do begin
				id = alog (1.*i/32.+1. )*scalefactor < 15.
				red[i] = rs[id]
				gre[i] = gs[id]
				blu[i] = bs[id]
			end
			self.palette->setProperty, red=red, green=gre, blue=blu
		end
	end else begin
		self.palette->setProperty, red=red, green=gre, blue=blu
	end

         ; create PaletteChange events (type 6)

	evt = ptr_new( {type:6, display:self}, /no_copy )
    objs = (*self.callbackobjects)[6]->get( /all, count=nobj )
    if( nobj gt 0 ) then begin
       for i = 0, nobj-1 do objs[i]->mrdisplayEvtHandler, evt
    end
    ptr_free, evt
    if( self.displayRequestFlag ) then self->display

end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::palette
; Syntax: palette = obj->palette()
; Desc  : Returns an object reference to mrDicoms internal IDLgrPalette object.
;         Destroying this object means asking for trouble.
;=-----------------------------------------------------------------------------
function mrdisplay::palette
    return, self.palette
end

;==============================================================================
;
;                             PUBLIC OBJECT FUNCTIONS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrdisplay::display
; Syntax: obj->display
; Desc  : (re-)draw the current image
;=-----------------------------------------------------------------------------
pro mrdisplay::display
	if( self.grScene eq obj_new() ) then return      ; no image set
    widget_control, self.wid_draw, get_value=grWindow
    if( grWindow eq obj_new() ) then return          ; widget_draw not realized
    if( self.interpolate ) then widget_control, hourglass=1
    grWindow->draw, self.grScene
    if( self.interpolate ) then widget_control, hourglass=0
   	self.displayRequestFlag = 0                      ; request fulfilles -> reset flag
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::scroll
; Syntax: obj->scroll, dx, dy
; Args  : <b>dx, dy</b>
;         Integer. Scroll the image by the given offset
; Desc  : Scroll the image position by the given values in pixels
;=-----------------------------------------------------------------------------
pro mrdisplay::scroll, dx, dy
	self.grImage->getProperty, xcoord_conv=a, ycoord_conv=b
	self.grImgView->getProperty, location=loc
	loc[0] = loc[0]+dx
	loc[1] = loc[1]+dy
	self.grImgView->setProperty, location=loc
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::centerImg
; Syntax: obj->centerImg
; Desc  : Center the image on the canvas
;=-----------------------------------------------------------------------------
pro mrdisplay::centerImg
    self.grImgView->setProperty, location=[(self.wid_xsize-self.imgXsize*self.scaling)/2, $
                                        (self.wid_ysize-self.imgYsize*self.scaling)/2]
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::fitToScreen
; Syntax: obj->fitToScreen
; Desc  : Rescale the image, so it fits the canvas. This will also center the
;         image on the canvas.
;=-----------------------------------------------------------------------------
pro mrdisplay::fitToScreen
    self->setScaling, min([1.*self.wid_xsize/self.imgXsize, 1.*self.wid_ysize/self.imgysize])
    self->centerImg
end

;==============================================================================
;
;                                  EVENT HANDLING
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::requestDisplay
; Syntax: obj->requestDisplay
; Desc  : Request a redisplay, but don't call mrdisplay::display immediately.
;         Instead, a redisplay of the screen is only performed after all external
;         event handlers have been called - thus avoiding possible overhead by
;         multiple calls to mrdisplay::display from each event handler.
;=-----------------------------------------------------------------------------
pro mrdisplay::requestDisplay
	self.displayRequestFlag = 1
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::activate
; Syntax: obj->activate
; Desc  : Switch on event handling
;=-----------------------------------------------------------------------------
pro mrdisplay::activate
    widget_control, self.wid_draw, event_pro='mrdisplay_evthandler'
    widget_control, self.wid_draw, /draw_button_events, /draw_motion_events
    if( !VERSION.release gt 5.5 ) then $
       widget_control, self.wid_draw, draw_keyboard_events=2
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::deactivate
; Syntax: obj->deactivate
; Desc  : Switch off event handling
;=-----------------------------------------------------------------------------
pro mrdisplay::deactivate
    widget_control, self.wid_draw, draw_button_events=0,  $
                       draw_motion_events=0
    if( !VERSION.release gt 5.5 ) then $
       widget_control, self.wid_draw, draw_keyboard_events=0
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::unregisterAll
; Syntax: obj->unregisterAll
; Desc  : Clear all entries in the callback table
;         (done by deleting the old one and creating a new)
;=-----------------------------------------------------------------------------
pro mrdisplay::unregisterAll
    for i = 0, n_elements(*self.callBackObjects)-1 do $
       (*self.callBackObjects)[i]->remove, /all
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::registerEvt
; Syntax: obj->registerEvt, object, event_type
; Args  : <b>object</b>
;         The object which provides the call back function
;         <b>event_type</b>
;         String or string array denoting the event type (s.b.)
; Desc  : Register an event with mrdisplay.
;         Possible event types are:
;            'Press'         : any button press
;            'Release'       : any button release
;            'DblClick'      : any button double click
;            'Motion'        : any mouse motion
;            'KeyPress'      : any key press
;            'ImgChange'     : called after setImg or updateImg
;            'PaletteChange' : called after call to setImgPaletteRGB
;         Event type strings are case insensitive.
;         The object must provide a method named 'mrdisplayEvtHandler' which
;         dispatches the incoming events. This method will be called
;         with a pointer to mrdisplay's event structure as an argument (see
;         above for the event structure definition)
;         Registering the same object with the same event type a second
;         type will <b>not</b> cause the handler function to be called twice,
;         but implicitly only move the function call to the end of the list.
;
; Code:
; disp = new_object('mrdisplay', base, 256, 256, 256, 256)
; zoom = new_obj('mrzoom')
; disp->registerEvt, zoom, ['Press','Release']
;=-----------------------------------------------------------------------------
pro mrdisplay::registerEvt, object, typestr
    if( n_elements( typestr ) eq 0 ) then return
    for i = 0, n_elements(typestr)-1 do begin
       type = self->resolveTypeStr(typestr[i])
       if( type eq -1 ) then continue
       (*self.callBackObjects)[type]->remove, object
       (*self.callBackObjects)[type]->add, object
    end
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::unregisterEvt
; Syntax: obj->unregisterEvt, object, event_type
; Args  : <b>object</b>
;         The call back object to be removed
;         <b>event_type</b>
;         String (or string array) denoting the event type(s) to remove
; Desc  : unregister an event/events from mrdisplay callback
;=-----------------------------------------------------------------------------
pro mrdisplay::unregisterEvt, cbobj, typestr
    if( n_elements(typestr) eq 0 ) then return
    for i = 0, n_elements(typestr)-1 do begin
       type = self->resolveTypeStr(typestr)
       if( type eq -1 ) then continue
       (*self.callBackObjects)[type]->remove, cbobj
    end
    return
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::unregisterObj
; Syntax: obj->unregisterObj, object
; Args  : <b>object</b>
;         The call-back object to remove
; Desc  : Unregister a whole object from mrdisplay callback
;=-----------------------------------------------------------------------------
pro mrdisplay::unregisterObj, cbobj
    for i = 0, n_elements(*self.callBackObjects)-1 do begin
       ((*self.callBackObjects)[i])->remove, cbobj
    end
end


;==============================================================================
;
;                          PRIVATE OBJECT FUNCTIONS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::resolveTypeStr
; Syntax: type_id = self->resolveTypeStr, event_type
; Desc  : Convert the type string into an index, case insensitive
;=-----------------------------------------------------------------------------
function mrdisplay::resolveTypeStr, typestr
    case strlowcase(typestr) of
       'press'             : return, 0
       'dblclick'          : return, 1
       'release'           : return, 2
       'motion'            : return, 3
       'keypress'          : return, 4
       'imgchange'         : return, 5
       'palettechange'     : return, 6
       else                : return, -1
    end
end


;+=----------------------------------------------------------------------------
; Name  : mrdisplay_evthandler
; Syntax: mrdisplay_evthandler, evt
; Calls : mrdisplay::evtHandler
; Desc  : Get object from event and call the object's event handler
;=-----------------------------------------------------------------------------
pro mrdisplay_evthandler, evt
    widget_control, evt.id, get_uvalue=display
    if( not obj_valid(display) ) then message, 'Invalid object pointer in event handler'
    display->evtHandler, evt
end

;+=----------------------------------------------------------------------------
; Name  : mrdisplay::evthandler
; Syntax: self->evthandler, evt
; Desc  : Handles events from the window and passes them on to registered
;         objects, i.e. creates events of type '0' to '4'.
;         imageChange and paletteChange events are created by their respective
;         methods.
;=-----------------------------------------------------------------------------
pro mrdisplay::evthandler, evt

       ; draw events evt structures contain: type, x, y, press, release, clicks
       ; type=0: button press   1: button release  2: motion
       ; evt.press: bitmaps of currently held down mouse buttons (incl. the newly pressed)
       ; evt.release:
       ; press/release: bitmaps of the mousebuttons
       ; clicks: 1 for single, 2 for double clicks

       ; fill some the fields

    widget_control, self.wid_draw, get_value=grWindow

		; calculate image coordinates

    self.grImgView->getProperty, location=loc
    imgx      = -1 > (evt.x - loc[0] - 1)/self.scaling
    imgy      = -1 > (evt.y - loc[1] )/self.scaling
    if( imgx ge self.imgXsize ) then imgx = -1
    if( imgy ge self.imgYsize ) then imgy = -1

	mrevt = {type:-1, display:self, top:evt.top, grWindow:grWindow, imgView:self.grImgView, lblView:self.grLblView, $
		   dx:evt.x-self.lastx, dy:evt.y-self.lasty, imgx:imgx, imgy:imgy, x:evt.x, y:evt.y, modifiers:evt.modifiers, $
		   key:0L, button:0, buttonStatus:0}

       ; now dispatch the events

	case evt.type of
	0: begin                                                  ; button press
		button = (255-self.buttonStatus) and evt.press        ; wrong IDL docu!
		self.buttonStatus = self.buttonStatus or evt.press
		mrevt.type = fix(evt.clicks-1)
	end
	1: begin                                                  ; button release
		button = evt.release
		self.buttonStatus = self.buttonStatus and (255 - evt.release )
		mrevt.type = 2
	end
	2: begin                                                  ; motion
		button = self.buttonStatus
		mrevt.type = 3
	end
	5: begin                                                  ; ASCII press
		button = self.buttonStatus
		if( evt.press eq 1 ) then begin
			mrevt.type = 4
			mrevt.key  = long( evt.ch )
		end
	end
	6: begin
		button = self.buttonStatus
		if( evt.press eq 1 ) then begin
			mrevt.type = 4
			mrevt.key  = evt.key
		end
	end
	end

    if( mrevt.type eq -1 ) then return

 		; fill mouse button status into event structure

    mrevt.button       = button
    mrevt.buttonStatus = self.buttonStatus

		; remember position. Since mouse events will not occur while
		; an old event is processed, it doesn't matter whether we do that
		; before or after event processing.

	self.lastx = evt.x
	self.lasty = evt.y

		; finally call all objects in our list

    objs = (*self.callBackObjects)[mrevt.type]->get( /all, count=nobj )
    mrevt = ptr_new( mrevt, /no_copy )
    if( nobj gt 0 ) then begin
       for i = 0, nobj-1 do objs[i]->mrdisplayEvtHandler, mrevt
    end
    ptr_free, mrevt

		; and call display, if requested

	if( self.displayRequestFlag ) then self->display
end




;+=----------------------------------------------------------------------------
; Name : mrdisplay__define
; Type : Constructor
;=-----------------------------------------------------------------------------
pro mrdisplay__define
    obj = { MRDISPLAY, $
         wid_base:           0          ,$  ; the base widget's id
         wid_draw:           0          ,$  ; the drawing widget's id
         wid_xsize:          0          ,$  ; size in x of both base and drawing widget
         wid_ysize:          0          ,$  ; dito in y
         grScene:            obj_new()  ,$  ; IDLgrScene
         grImgView:          obj_new()  ,$  ; IDLgrView for the image - scaled and moved
         grLblView:          obj_new()  ,$  ; IDLgrView for labels - fixed to the frame
         grModel:            obj_new()  ,$  ; IDLgrModel
         grImage:            obj_new()  ,$  ; IDLgrImage
         palette:         	 obj_new()  ,$  ; the image's palette
         interpolate:        0          ,$  ; interpolate the image? 1=yes
         width:				 0.         ,$  ; width for contrast
         center:             0.         ,$  ; center for contrast
         scaling:            0.         ,$  ; scaling factor for the image
         topbytscl:          255        ,$  ; top for bytescale
         imgData:            ptr_new()  ,$  ; local storage of image data
         imgXsize:           0          ,$  ; size in pixels in x of the image
         imgYsize:           0          ,$  ; size in pixels in y of the image
         imgMinval:          0.         ,$  ; minimum data value of the image
         imgMaxval:          0.         ,$  ; dito, but maximum
         displayRequestFlag: 0			,$	; set to '1', if redisplay is requested
         callBackObjects: 	 ptr_new()  ,$  ; a container holding all call back objects
         lastx:				 0			,$	; x position of last event
         lasty:				 0			,$ 	; y position of last event
         buttonStatus:       0           $  ; mouse buttons pressed or released - bitmap
    }
end
