;+=----------------------------------------------------------------------------
; Name   : mrHash
; RCS    : $Id: mrhash__define.pro,v 2.1 2004/11/26 16:24:04 olivert Exp olivert $
; Methods:
;          <b>Constructor</b>
;          mrHash::init              : constructor, 2nd part
;          <br>
;          <b>Public methods</b>
;          mrHash::set               : key, value
;          mrHash::get               : key, returns value
;          mrHash::exists            : key, returns 0 or 1
;          mrHash::delete            : key, remove key from array
;          mrHash::count             : returns number of entries in the hash
;          mrHash::keys              : returns string array containing all keys
;          mrHash::load              : load a hash array from a file
;          mrHash::save              : write a hash array to file
;          <br>
;          <b>Private functions</b>
;          mrHash__define            : constructor, 1st part
;          mrHash::cleanup           : the destructor
; Desc   : Implementation of an associative array capable of storing int, real
;          (=double), string or object references.
;          Although this is only a quick hack (linear search via where), it is
;          reasonable fast for a couple of thousand entries and in that range
;          compares well with the more sophisticated implementation available
;          on the RSINC contribution pages.
;          If hashes get considerably bigger than that, IDL might anyway not
;          be the language of choice.
;          The written file format is key=value, i.e. keys may not contain the
;          '=' character.
; Code   : <tt>phone = obj_new('mrhash', 'string')
;          phone->set, 'Kay',   '+49 123 123456'
;          phone->set, 'Meike', '+46 987 9823'
;          print, "Kay's phone number: ", phone->get('Kay')
;          </tt>
; Author: (c) Oliver Thilmann, oliver.thilmann@web.de
; Licns : This program is published under the LGPL. See gpl.txt for details
;=-----------------------------------------------------------------------------


;==============================================================================
;
;                          CONSTRUCTORS AND DESTRUCTORS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrHash::init
; Syntax: obj = obj_new('mrhash', type, blocksize=blocksize)
; Args  : <b>type</b>
;         String denoting the type of the parameter. One of 'int', 'real',
;         'string' or 'object'
;         <b>blocksize</b>
;         Data blocks are allocated in chunks of size blocksize. Default is 100.
; Desc  : Instanciate a mrHash object with initally blocksize data elements.
;=-----------------------------------------------------------------------------
function mrhash::init, type, blocksize=blocksize
	if( not keyword_set(type) ) then return, 0
	blocksize = (n_elements(blocksize) eq 0) ? 100 : blocksize

	case type of
		'int'	: begin
			value = intarr(blocksize)
			self.type = 0
		end
		'real'	: begin
			value = dblarr(blocksize)
			self.type = 1
		end
		'string': begin
			value = strarr(blocksize)
			self.type = 2
		end
		'object': begin
			value = objarr(blocksize)
			self.type = 3
		end
		else: return, 0
	end
	self.key       = ptr_new( strarr(blocksize), /no_copy)
	self.value     = ptr_new( value, /no_copy )
	self.last      = -1
	self.size      = blocksize
	self.blocksize = blocksize
	return, 1
end

;+=----------------------------------------------------------------------------
; Name  : mrHash::cleanup
; Syntax: obj->cleanup, /destroy
; Args  : <b>destroy</b>
;         If the array stored objects and destroy is set to '1', the
;         objects are destroyed. Default is '0'.
; Desc  : The destructor
;=-----------------------------------------------------------------------------
pro mrhash::cleanup, destroy=destroy
	destroy = n_elements(destroy) eq 1 ? destroy : 0
	ptr_free, self.key
	if( self.type eq 3 and destroy eq 1 ) then begin
		for i = 0, self.last-1 do begin
			if( obj_valid( (*self.value)[i] ) ) then obj_destroy, (*self.value)[i]
		end
	end
	ptr_free, self.value
end


;==============================================================================
;
;                              PUBLIC CLASS METHODS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrHash::set
; Syntax: obj->set, key, value
; Args  : <b>key</b>
;         String with the key - if empty, the method returns immediately
;         <b>value</b>
;         A value of the correct type
; Desc  : Set a key-value pair
;=-----------------------------------------------------------------------------
pro mrhash::set, key, value
	if( key eq '' ) then return
	id = where( *self.key eq key )
	if( id ge 0 ) then begin
		(*self.value)[id] = value
		return
	end
	self.last = self.last+1
	if( self.last eq self.size ) then begin
		newsize = self.size + self.blocksize
		newkey = strarr( newsize )
		newkey[0:self.last-1] = *self.key
		ptr_free, self.key
		self.key = ptr_new( newkey, /no_copy )
		case self.type of
			0: newvalue = intarr( newsize )
			1: newvalue = dblarr( newsize )
			2: newvalue = strarr( newsize )
			3: newvalue = objarr( newsize )
		end
		newvalue[0:self.last-1] = *self.value
		ptr_free, self.value
		self.value = ptr_new( newvalue, /no_copy )
		self.size =  newsize
	end
	(*self.key)[self.last] = key
	(*self.value)[self.last] = value
end

;+=----------------------------------------------------------------------------
; Name  : mrHash::get
; Syntax: value = obj->get(key, defined)
; Args  : <b>key</b>
;         key string
;         <b>defined</b>
;         If present, set to '1' if the key existed or to '0' otherwise.
; Desc  : Retrieve a value from the array. If the key does not exist,
;         0, 0., '', or obj_new() are returned, depending on the array type.
;=-----------------------------------------------------------------------------
function mrhash::get, key, defined
	if( n_elements(key) ne 0 ) then begin
		id = where( *self.key eq key )
		if( id[0] ge 0 ) then begin
			defined = 1
			return, (*self.value)[id[0]]
		end
	end
	defined = 0
	case self.type of
		0: return, 0
		1: return, 0.
		2: return, ''
		3: return, obj_new()
	end
end

;+=----------------------------------------------------------------------------
; Name  : mrHash::delete
; Syntax: obj->delete, key
; Args  : <b>key</b>
;         key string
; Desc  : Remove the entry from the array.
;=-----------------------------------------------------------------------------
pro mrhash::delete, key
	id = where( *self.key eq key )
	if( id eq -1 ) then return
	if( id eq self.last ) then begin
		self.last = self.last-1
		return
	end
	(*self.key)[id]   = (*self.key)[self.last]
	(*self.value)[id] = (*self.value)[self.last]
	self.last = self.last-1
	return
end

;+=----------------------------------------------------------------------------
; Name  : mrHash::keys
; Syntax: keys = obj->keys( count=count )
; Ret   : A string array containing all currently known keys - the array
;         has no specific order. If the optional parameter count is set, the
;         number of elements in the array is returned.
;=-----------------------------------------------------------------------------
function mrhash::keys, count=count
	if( self.last eq -1 ) then begin
		count = 0
		return, ''
	end
	count = self.last+1
	keys = strarr( self.last+1 )
	keys[0:self.last] = (*self.key)[0:self.last]
	return, keys
end

;+=----------------------------------------------------------------------------
; Name  : mrHash::exists
; Syntax: res = obj->exists(key)
; Args  : <b>key</b>
;         String holding the key
; Desc  : Returns '1', if the array element exists, '0' otherwise
;=-----------------------------------------------------------------------------
function mrhash::exists, key
	id = where( *self.key eq key )
	if( id[0] eq -1 ) then return, 0 else return, 1
end

;+=----------------------------------------------------------------------------
; Name  : mrHash::count
; Syntax: nentries = obj->count()
; Desc  : Returns the number of entries in the hash
;=-----------------------------------------------------------------------------
function mrhash::count
	return, self.last+1
end

;+=----------------------------------------------------------------------------
; Name  : mrHash::load
; Syntax: res = obj->load( filename, count=count, /compress )
; Args  : <b>filename</b>
;         The file to load
;         <b>count</b>
;         If set, contains the number of loaded key-value pairs.
;         <b>compress</b>
;         Needs to be set if loading a compressed file.
; Ret   : 0 in case of failure, 1 otherwise
; Desc  : Loads a hashtable from disk. The file format is
;         key=value
;         keys thus may not contain the '=' character
;         Lines containing no '=' sign or starting with an '=' sign are invalid.
;         They are ignored and can thus be used as comment lines
;         If a value should not be of the specified type, the routine returns
;         with an error.
;=-----------------------------------------------------------------------------
function mrhash::load, filename, count=count, compress=compress

		; error handling

	error = 0
	catch, error
	if( error ne 0 ) then begin
		catch, /cancel
		if( n_elements(unit) ne 0 ) then free_lun, unit
		print, !ERROR_STATE.msg
		return, 0
	end

		; load

	count = 0
	if( n_elements(filename) eq 0 ) then message, 'missing filename argument'
	compress = (n_elements(compress) eq 0 ) ? 0 : compress
	openr, unit, filename, /get_lun, compress=compress, error=err
	if( err ne 0 ) then message, 'error opening file ' + filename

	line = ''
	while( not EOF(unit) ) do begin
		readf, unit, line
		sep = strpos( line, '=' )
		if( sep le 0 ) then continue   ; no '=' or comment line
		key = strmid( line, 0, sep )
		key = strtrim( key, 2 )        ; remove leading and trailing whitespace
		value = strmid( line, sep+1 )
		case self.type of
			0: self->set, key, fix(value)
			1: self->set, key, double(value)
			2: self->set, key, value
		end
	end
 	free_lun, unit
 	count = self->count()
	return, 1

end

;+=----------------------------------------------------------------------------
; Name  : mrHash::save
; Syntax: res = obj->save( filename, /compress )
; Args  : <b>filename</b>
;         The file to save to
;         <b>compress</b>
;         If set, compress the file via IDL's buildin compress option
; Desc  : Saves a hash to file - this obviously will not work
;         for hashes storing object references.
; Ret   : Returns '1' for success, '0' in case of failure
;=-----------------------------------------------------------------------------
function mrhash::save, filename, compress=compress

		; error handling

	error = 0
	catch, error
	if( error ne 0 ) then begin
		catch, /cancel
		print, !ERROR_STATE.msg
		return, 0
	end

		; save

	if( self.type eq 3 ) then message, 'object references cannot be saved'
	compress = ( n_elements(compress) eq 0 ) ? 0 : comrpess
	openw, unit, filename, /get_lun, compress=compress, error=err
	if( err ne 0 ) then message, 'error opening file ' + filename
	for i = 0, self.last do printf, unit, (*self.key)[i]+'='+strtrim( string((*self.value)[i]), 2)
	free_lun, unit
	return, 1
end


;+=----------------------------------------------------------------------------
; Name  : mrHash__define
; Desc  : Private Procedure (constructor)
;=-----------------------------------------------------------------------------
pro mrhash__define
	obj = { MRHASH, $
			key			:	ptr_new()	,$ ; keys
			value       :   ptr_new()   ,$ ; values
			type        :           0   ,$ ; int, double, string or object
			last        :           0   ,$ ; last current element
			size        :           0   ,$ ; array size
			blocksize	:			0	 $ ; array increment
		}
end