;+=----------------------------------------------------------------------------
; Name   : mrImageSet
; RCS    : $Id: mrimageset__define.pro,v 2.3 2004/11/24 10:43:39 olivert Exp olivert $
; Methods:
;         <b>Constructor</b>
;         mrImageSet::init             :
;         <br>
;         <b>Public variables (get)</b>
;         mrImageSet::name             : an image set may be given a name for identification
;         mrImageSet::class            : a 'class' name to group image sets
;         mrImageSet::xsize            : number of pixels in 'x'
;         mrImageSet::ysize            : number of pixels in 'y'
;         mrImageSet::d0               : matrix dimension 0
;         mrImageSet::nz               : alias for mrImageSet::d0
;         mrImageSet::nSlices          : alias for mrImageSet::d0
;         mrImageSet::d1               : matrix dimension 1
;         mrImageSet::nt               : alias for mrImageSet::d1
;         mrImageSet::nShots           : number of 'shots' (= images in time)
;         mrImageSet::d2               : matrix dimension 2
;         mrImageSet::size             : combination of the above, similar to idl's 'size'
;         mrImageSet::nImgs            : total number of images
;         mrImageSet::palette          : the palette object
;         mrImageSet::imgData          : pointer to matrix holding image pointers
;         mrImageSet::path             : last used path in load/save
;         mrImageSet::uvalue           : return the user definable pointer
;         mrImageSet::parent           : a reference to an image set from which this set is derived
;         <br>
;         <b>Public variables (set)</b>
;         mrImageSet::setName          : set the name
;         mrImageSet::setClass         : set the class name
;         mrImageSet::setUvalue        : set the user definable value (a pointer, sigh)
;         mrImageSet::setParent        : store reference to a parent mrImageSet(Dicom) object
;         mrImageSet::setPalette       : copies RGB values from a passed palette
;         mrImageSet::setPaletteRGB    : sets the RGB values of the internally stored palette object
;         <br>
;         <b>Image handling</b>
;         mrImageSet::set              : set an image
;         mrImageSet::get              : get a specific image
;         mrImageSet::delete           : delete one or all images
;		  mrImageSet::condense         : condense a matrix dimension into the image data
;         mrImageSet::load             : load an image from harddisc
;         mrImageSet::save             : save one or all images to harddisc
;         mrImageSet::errMsg           : returns an error description of the last I/O error
;         <br>
;         <b>Notifications</b>
;         mrImageSet::notify           : see mrNotify::notify and below
;         mrImageSet::enableNotify     : see mrNotify::enableNotify
;         mrImageSet::disableNotify    : see mrNotify::disableNotify
;         <br>
;         <b>Private functions</b>
;         mrImageSet__define           : the constructor
;         mrImageSet::cleanup          : the destructor
; Desc  : Load and save functionality are not fully implemented, yet.
;         <br>
;         <b>General</b>
;         mrImageSet serves as a container for a set of images that have the same
;         size and share a common palette. The size may be set upon instanciation
;         of the object - otherwise the first loaded/set image determines the image
;         size for this instance. The same mechanism is applied to the RGB palette.
;         An image set may have a user definable 'name' and 'class' for identification
;         and grouping.
;         <br>
;         <b>Image sets</b>
;         Image sets are stored as matrices of image pointers. These matrices may have
;         up to three dimensions, e.g. for perfusion sequences one dimension for the
;         slice number, a second dimension for point in time.
;         nx and ny denote the size of the images itself in pixels.
;         d0, d1 and d2 denote the extensions of above mentioned matrix. Since this
;         doesn't mean anything, there's a number of aliases for better readability
;         of the source code that uses this object:
;         d0 can also be accessed as nz or nSlices (number of images in z)
;         d1 can also be accessed as nt or nShots  (number of images in time)
;         d2 got no aliases yet
;         <br>
;         <b>Data organization</b>
;         There's no reason why the images should have only two dimensions (x,y) -
;         in many cases calculation routines can make better use of IDL syntax if
;         some of above listed matrix dimensions could be condensed into the image
;         data.
;         The mrImageSet::condense method serves this purpose by folding one dimension
;         of the pointer matrix into the images.
;         <br>
;         <b>Load and Save</b><i>(not implemented yet)</i>
;         The object supports some standard image formats for which IDL provides
;         loading and saving routines. Image sets based on the DICOM format are
;         special and are handled via mrImageSetDicom, which is a subclass to
;         mrImageSet and provides special load and save functionality for DICOM.
;         <br>
;		  <b>Notifications</b>
;         External objects can register with an mrImageSet object for notification
;         upon data change via mrImageSet::notify
;         Available events are 'set', 'delete' and 'palettechange'
; Author: (c) Oliver Thilmann, oliver.thilmann@radfys.lu.se
; Licns : This program is published under the GPL. See gpl.txt for details
;=-----------------------------------------------------------------------------

;==============================================================================
;
;                          CONSTRUCTORS AND DESTRUCTORS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::init
; Syntax: obj = obj_new('mrImageSet', d0, d1, d2, name=name, class=class, xsize=xsize,
;         ysize=ysize, red=red, green=green, blue=blue
; Args  : <b>d0, d1, d2</b>
;         Dimensions of the image pointer matrix
;         <b>name, class</b>
;         Two optional strings to set the class attributes 'name' and 'class'.
;         <b>xsize, ysize</b>
;         Size in pixels of the images in this set. If not set here, the first
;         image loaded/set will determine the values. All images in a set must be
;         of the same size.
;         <b>red, green, blue</b>
;         Image palette. If not set, the palette is determined by the first image
;         added to the set. See also mrImageSet::set and mrImageSet::load
;         All images in a set share the same palette ('cause it rhymes)
; Desc  : Instanciates an object of class mrImageSet. The optional arguments
;         <i>red, green and blue</i> must be set all at once or not at all.
; Code  : obj=obj_new('mrImageSet', 10, 60, name='signal', class='perfusion')
;=-----------------------------------------------------------------------------
function mrImageSet::init, d0, d1, d2, name=name, class=class, xsize=xsize, $
                     ysize=ysize, red=red, green=green, blue=blue

	x = self->mrNotify::init(['set','delete','palettechange'])

		; also allow for 0 dimensions - used by mrImageSetDicom
		; maybe we should check, as it doesn't make sense for mrImageSet objects

	self.ndims = 0
	self.dims  = [-1,-1,-1]
	if( n_elements(d0) ne 0 ) then if( d0 gt 0 ) then begin
			self.dims[self.ndims] = d0
			self.ndims = self.ndims + 1
	end
	if( n_elements(d1) ne 0 ) then if( d1 gt 0 ) then begin
			self.dims[self.ndims] = d1
			self.ndims = self.ndims + 1
	end
	if( n_elements(d2) ne 0 ) then if( d2 gt 0 ) then begin
			self.dims[self.ndims] = d2
			self.ndims = self.ndims + 1
	end

	self.name  = ( n_elements( name  ) eq 0 ) ?  '' : name
	self.class = ( n_elements( class ) eq 0 ) ?  '' : class
	self.xsize = ( n_elements( xsize ) eq 0 ) ?   0 : xsize
	self.ysize = ( n_elements( ysize ) eq 0 ) ?   0 : ysize

		; create palette either from user settings or use greyscale bar

	colsum = n_elements(red) + n_elements(green) + n_elements(blue)
	if( colsum eq 3*256 ) then begin
		self.palette = obj_new('IDLgrPalette', red, green, blue)
	end else begin
		if( colsum ne 0 ) then message, 'Illegal color settings'
		self.palette = obj_new('IDLgrPalette')
		self.palette->loadCT, 0
	end

	self.path         = ''
	self.errmsg       = ''

	if( self.ndims gt 0 ) then begin
		data         = ptrarr( self.dims[0:self.ndims-1] )
		self.imgdata = ptr_new( data, /no_copy )
	end else begin
		self.imgdata = ptr_new()
	end
	self.uvalue      = ptr_new()
	return, 1
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::cleanup
; Syntax: obj_destroy, obj
; Desc  : The destructor - deletes all data by calling mrImageSet::unload
;=-----------------------------------------------------------------------------
pro mrImageSet::cleanup
	self->delete, /all
	self->mrNotify::cleanup
	if( ptr_valid(self.uvalue) ) then ptr_free, self.uvalue
end

;==============================================================================
;
;                               GET AND SET METHODS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::name
; Ret   : This mrImageSet's user defined <i>name</i>, a string.
;=-----------------------------------------------------------------------------
function mrImageSet::name
	return, self.name
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::class
; Ret   : This mrImageSet's user defined <i>class name</i>, a string.
;=-----------------------------------------------------------------------------
function mrImageSet::class
	return, self.class
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::xsize
; Ret   : number of image pixels in x
;=-----------------------------------------------------------------------------
function mrImageSet::xsize
	return, self.xsize
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::ysize
; Ret   : number of image pixels in y
;=-----------------------------------------------------------------------------
function mrImageSet::ysize
	return, self.ysize
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::d0
; Ret   :
;=-----------------------------------------------------------------------------
function mrImageSet::d0
	return, self.dims[0]
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::nz
; Ret   :
;=-----------------------------------------------------------------------------
function mrImageSet::nz
	return, self.dims[0]
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::nSlices
; Ret   :
;=-----------------------------------------------------------------------------
function mrImageSet::nSlices
	return, self.dims[0]
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::d1
; Ret   :
;=-----------------------------------------------------------------------------
function mrImageSet::d1
	return, self.dims[1]
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::nt
; Ret   :
;=-----------------------------------------------------------------------------
function mrImageSet::nt
	return, self.dims[1]
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::nShots
; Ret   :
;=-----------------------------------------------------------------------------
function mrImageSet::nShots
	return, self.dims[1]
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::d2
; Ret   :
;=-----------------------------------------------------------------------------
function mrImageSet::d2
	return, self.dims[2]
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::size
; Syntax: res = obj->size( /dimensions, /n_dimensions, /n_elements )
; Ret   : An array denoting the dimensions of the image matrix.
;         The array format resembles the format returned by IDL's size function.
;         ret[0] = the number of dimensions
;         ret[1:#dims] = size of each dimension
;         second last element: -1
;         last array element : total number of images in this image set
;         If the number of dimensions is zero (i.e. the matrix consists of just
;         a single element/pointer), the method returns [0,-1,1], if the image
;         pointer is valid or [0,-1,0], if not.
;=-----------------------------------------------------------------------------
function mrImageSet::size, dimensions=dimensions, n_dimensions=n_dimensions, n_elements=n_elements

		; only returning the number of dimensions is straight forward

	if( n_elements(n_dimensions) ne 0 ) then return, self.ndims

		; special treatment in case we're a singularity

	if( self.ndims eq 0 ) then begin
		if( n_elements(dimensions) ne 0 ) then return, 0
		if( n_elements(n_elements) ne 0 ) then begin
			if( ptr_valid(self.imgdata) ) then return, 1 else return, 0
		end
		if( ptr_valid(self.imgdata) ) then return, [0,-1,1] else return, [0,-1,0]
	end

		; the rest is straight forward

	if( n_elements(dimensions)   ne 0 ) then begin
		return, self.dims[0:self.ndims-1]
	end

	nel = 1
	for i = 0, self.ndims-1 do nel = nel * self.dims[i]
	if( n_elements(n_elements)  ne 0 ) then begin
		return, nel
	end

	a = intarr( self.ndims + 3 )
	a[0] = self.ndims
	a[1:self.ndims] = self.dims[0:self.ndims-1]
	a[self.ndims+1] = -1             ; dummy for now to match IDL's size format
	a[self.ndims+2] = nel
	return, a

end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::nImgs
; Ret   : Returns the number of images.
;         Please be careful in case of mrimagsetdicom objects - after condensing
;         data, the number of images and the number of dicom objects does not
;         match any more! If you're looking for the number of DICOM objects,
;         use mrImageSetDicom::nDicom instead
;=-----------------------------------------------------------------------------
function mrImageSet::nImgs
	return, self->size(/n_elements)
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::palette
; Ret   : Returns a reference to the palette object. The user may modify the
;         object properties, but destroying the object means asking for
;         trouble.
;=-----------------------------------------------------------------------------
function mrImageSet::palette
	return, self.palette
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::setPalette
; Syntax: obj->setPalette, palette
; Args  : <b>palette</b>
;         An IDLgrPalette object
; Desc  : Copies (!) the rgb values from the palette object to the internal
;         palette.
;=-----------------------------------------------------------------------------
pro mrImageSet::setPalette
	if( n_elements(palette) eq 0 ) then message, 'missing palette argument'
	if( not obj_isa( palette, 'IDLgrPalette' ) ) then message, 'wrong palette argument'
	palette->getProperty, red=r, green=g, blue=b
	self.palette->setProperty, red=r, green=g, blue=b

		; palette change notifications

	self->sendNotification, 'palettechange'
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::setPaletteRGB
; Syntax: obj->setPaletteRGB, red=intarr, green=intarr, blue=intarr, colorTable=integer
; Args  : <b>red, green, blue</b>
;         Integer arrays with 256 elements holding the red/green/blue values of
;         the palette
;         <b>colorTable</b>
;         An integer value between 0 and 41 indexing one of IDL's predefined
;         color tables. If set to 42, a logarithmically divided colortable is
;         loaded.
; Desc :  Sets the red/green/blue values of the internal image palette.
;=-----------------------------------------------------------------------------
pro mrImageSet::setPaletteRGB, red=red, green=gre, blue=blu, colorTable=ct
	if( n_elements(ct) ne 0 ) then begin
		if( ct lt 0 or ct gt 42 ) then message, 'Color table index out of bounds'
		if( ct lt 42 ) then begin
			self.palette->loadct, ct
		end else begin
			red = intarr( 256 )
			gre = intarr( 256 )
			blu = intarr( 256 )
			rs = [0,0,0,0,0,0,0,0,0,255,255,255,255,255,230,210]
			gs = [0,32,32,64,0,160,160,208,255,255,204,128,102,0,26,0]
			bs = [0,112,144,192,255,255,160,64,0,0,102,64,0,0,0,0]
				; maximum ratio is '8'
			scalefactor = 16./alog( 255./32.+1.)
			for i = 0, 255 do begin
				id = alog (1.*i/32.+1. )*scalefactor < 15.
				red[i] = rs[id]
				gre[i] = gs[id]
				blu[i] = bs[id]
			end
			self.imgPalette->setProperty, red=red, green=gre, blue=blu
		end
	end else begin
		self.palette->setProperty, red=red, green=gre, blue=blu
	end

		; palette change notifications

	self->sendNotification, 'palettechange'
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::imgData
; Ret   : Returns the pointer to the image pointer matrix.
;=-----------------------------------------------------------------------------
function mrImageSet::imgdata
	return, self.imgdata
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::path
; Ret   :
;=-----------------------------------------------------------------------------
function mrImageSet::path
	return, self.path
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::uvalue
; Ret   :
;=-----------------------------------------------------------------------------
function mrImageSet::uvalue
	return, *self.uvalue
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::parent
; Ret   : The parent imageSet (of class mrImageSet or mrImageSetDicom) of this
;         image set - as set in mrImageSet::setParentSet.
;=-----------------------------------------------------------------------------
function mrImageSet::parent
	return, self.parent
end

;------------------------------------------------------------------------------
; Name  : mrImageSet::errNo
; Syntax: print, obj->errNo()
; Ret   : Returns the error number corresponding to the last I/O error.
;         The error number is not reset in case of a successful command - the
;         method is thus not suited to check for error conditions. Check the
;         method return values instead.
;         For a list of error numbers and the corresponding texts see the
;         introductory description.
;
;
;         NOT IMPLEMENTED
;
;------------------------------------------------------------------------------
function mrImageSet::errNo
	return, self.errNo
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::errMsg
; Syntax: print, obj->errMsg()
; Ret   : Returns a message text corresponding to the last I/O error.
;=-----------------------------------------------------------------------------
function mrImageSet::errMsg
	return, self.errMsg
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::setName
; Syntax: obj->setName, name
; Desc  : Sets the name of this object
;=-----------------------------------------------------------------------------
pro mrImageSet::setName, name
	if( n_elements(name) ne 1 ) then message, 'Illegal or missing argument name'
	self.name = name
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::setClass
; Syntax: obj->setClass, classname
; Desc  : Sets the classname of this object (which has nothing to do with
;         'classes' as used in OO programming, of course).
;=-----------------------------------------------------------------------------
pro mrImageSet::setClass, class
	if( n_elements(class) ne 1 ) then message, 'Illegal or missing argument class name'
	self.class = class
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::setUvalue
; Syntax: obj->setUvalue, value, /no_copy
; Args  : <b>value</b>
;         Value may be anything, a single value, an array, a structure.
;         The function will internally create a pointer to this value and store
;         the pointer (although this is transparent to the user when retrieving
;         the uvalue.
;         <b>no_copy</b>
;         Well, of course it is not transparent to the user. Normally the method
;         will create a copy of the specified value. if the no_copy keyword is
;         used, no copy will be created, but the value variable is invalid after
;         setting of the uvalue.
; Desc  : Sets a uvalue
;=-----------------------------------------------------------------------------
pro mrImageSet::setUvalue, uvalue, no_copy=no_copy
	nocopy = keyword_set( no_copy )
	if( self.uvalue ne ptr_new() ) then ptr_free, self.uvalue
	self.uvalue = ptr_new( uvalue, no_copy=nocopy )
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::setParent
; Syntax: obj->setParent, mrImageSetObject
; Desc  : mrImageSets might be created from other mrImageSet or mrImageSetDicom
;         objects. To store these dependencies it is possible to save the
;         parent set in an object variable.
;         Might be useful, e.g. to store a parent mrImageSetDicom object to
;         keep the dicom data without the need to create a full mrImageSetDicom
;         which would only mean an unnecessary overhead.
;=-----------------------------------------------------------------------------
pro mrImageSet::setParent, value
	self.parent = value
end


;==============================================================================
;
;                             SET/GET/DELETE, LOAD/SAVE
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrImageSet::set
; Syntax: obj->set, index, image, /no_copy
; Args  : <b>index</b>
;         integer array, containing the coordinates of the image, i.e.
;         the matrix indices of the image pointer matrix. Obviously the number
;         of elements of the array must match the number of dimensions of the
;         pointer matrix.
;         <b>image</b>
;         Pointer to the image
;         <b>no_copy</b>
;         flag. By default, the method makes a copy of the passed image data.
;         To conserve space, this may be suppressed. Note, however, that
;         destroying the mrImageSet object will free all stored image pointers!
;         Also changing the image size later means asking for trouble.
; Desc  : Adds an image to the dataset
;=-----------------------------------------------------------------------------
pro mrImageSet::set, index, image, no_copy=no_copy

		; check the command line arguments and whether enough positional
		; parameters are present

	if( n_elements(image) eq 0 ) then return
	if( not ptr_valid(image) ) then return

	if( n_elements(index) ne self.ndims ) then message, 'Mismatch in number of indexing parameters'
	nocopy = ( n_elements(no_copy) eq 0 ) ? 0 : no_copy
	for i = 0, self.ndims-1 do begin
		if( index[i] ge self.dims[i] ) then message, 'Index too large'
	end

		; for a start we only deal with 2-dim images (i.e. indexed color)

	dims = size( *image, /n_dimensions )
	if( dims ne 2 ) then begin
		x = dialog_message('Time to continue coding in mrImageSet')
		message, 'Stops here'
		return
	end

	dims = size( *image, /dimensions )
	if( self.xsize eq 0 ) then begin
		self.xsize = dims[0]
		self.ysize = dims[1]
	end else begin
		if( self.xsize ne dims[0] or self.ysize ne dims[1] ) then begin
			message, 'Mismatch in image size'
			return
		end
	end

		; calculate image position as 1-dim index

	id = index[self.ndims-1]
	for i = self.ndims-2, 0, -1 do id = index[i] + self.dims[i]*id
	if( ptr_valid( (*self.imgdata)[id] ) ) then ptr_free, (*self.imgdata)[id]
	if( nocopy ) then (*self.imgdata)[id] = image else (*self.imgdata)[id] = ptr_new( *image )

		; notify others

	self->sendNotification, 'set'
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::get
; Syntax: image = obj->get( index, /all )
; Args  : <b>index</b>
;         integer array, containing the coordinates of the image to set, i.e.
;         the matrix indices of the image pointer matrix. Obviously the number
;         of elements of the array must match the number of dimensions of the
;         pointer matrix.
;         <b>all</b>
;         If set, the method returns a pointer to the whole image matrix
;         i.e. the same as mrImageSet::imgdata() would return
; Desc  : Returns the image pointer
;=-----------------------------------------------------------------------------
function mrImageSet::get, index, all=all

	if( n_elements(all) ne 0 ) then return, self.imgdata
	if( n_elements(index) ne self.ndims ) then message, 'Mismatch in number of indexing parameters'

		; calculate image position as 1-dim index

	id = index[self.ndims-1]
	for i = self.ndims-2, 0, -1 do id = index[i] + self.dims[i]*id
	return, (*self.imgdata)[id]
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::delete
; Syntax: obj->delete, index, /progressbar, /all
; Args  : <b>index</b>
;         See mrImageSet::set
;         <b>progressbar</b>
;         If set, a progressbar will be instanciated and update to show
;         deletion progress. Only appropriate for very large data sets or very
;         slow machines.
;         <b>all</b>
;         If set to '1', all images will be 'free'd. Arguments <i>slice</i>
;         and <i>shot</i> are ignored in this case (and need not be set).
; Desc  : Deletes one or all image(s) from the data set.
;=-----------------------------------------------------------------------------
pro mrImageSet::delete, index, progressbar=progressbar, all=all

	all         = ( n_elements(all)         eq 0 ) ? 0 : all
	progressbar = ( n_elements(progressbar) eq 0 ) ? 0 : progressbar

	if( all eq 1 ) then begin
		if( progressbar ) then begin
			pb = obj_new('mrprogressbar', title=(mrI18n())->get('is_deleting') )
		end
		nimgs = self->nImgs()
		for i = 0, nimgs-1 do begin
			if( ptr_valid((*self.imgdata)[i]) ) then ptr_free, (*self.imgdata)[i]
			(*self.imgdata)[i] = ptr_new()
			if( progressbar ) then pb->update, (1.+i)/nimgs
		end
		if( obj_valid(pb) ) then obj_destroy, pb
		self.xsize     = 0
		self.ysize     = 0
	end else begin
		if( n_elements(pos) ne self.ndims ) then return
		id = pos[self.ndims-1]
		for i = self.ndims-2, 0, -1 do id = pos[i] + self.dims[i]*id
		if( ptr_valid((*self.imgdata)[id]) ) then ptr_free, (*self.imgdata)[id]
		(*self.imgdata)[id] = ptr_new()
	end
	self.errmsg    = ''

		; notify others

	self->sendNotification, 'delete'
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::condense
; Syntax: res = obj->condense( index, /progressbar, /no_copy )
; Args  : <b>index</b>
;         The matrix dimension to add to the image data
;         The image matrix itself will subsequently be reduced by that
;         dimension, e.g.
;         obj->condense, 0
;         will  turn former d1 into d0 (i.e. 'shots' into 'slices')
;         and d2 into d1 while the original d0 dimension is added to the
;         image data itself, which will thus become a 3-dim array with
;         dimensions [nx,ny,d0].
;         <b>progressbar</b>
;         Displays a progressbar while resorting the images. Only
;         appropriate for very large datasets or slow machines.
;         <b>no_copy</b>
;         By default, the new image data is stored in a newly created
;         mrImageSet object which is returned to the user. A reference to
;         the parent imageset is stored in the new set in this case.
;         If the no_copy keyword is set, the condensed data will replace
;         the existing.
; Ret   : either obj_new() or, if the /copy keyword is set, a reference to
;         the newly created mrImageSet object.
; Desc  : Condenses one matrix dimension into the image data itself.
;         Example: 128x128 images, 20 slices (d1), 60 shots(d2)
;         newset = obj->condense(1)
;         will leave you with 60 images of dimension [128,128,20]
;         The image matrix will afterwards only be 1-dimensional with 60
;         image pointers. If the no_copy keyword is set, a new mrImageSet object
;         is created with the new data. Otherwise, the original images are
;         lost (though they can of course be easily recreated)
;         <b>Limit</b>
;         Before calling the 'condense' method, the images must be two-
;         dimensional, i.e. at most one matrix dimension can be condensed
;         into an image as of now.
;=-----------------------------------------------------------------------------
function mrImageSet::condense, index, progressbar=pb, no_copy=no_copy

		; argument checks

	if( self.ndims        eq 0 )            then return, ptr_new()
	if( n_elements(index) ne 1 )            then message, 'Invalid index argument'
	if( index lt 0 or index ge self.ndims ) then message, 'Invalid index argument'
	no_copy = ( n_elements(no_copy)  eq 0 ) ? 0 : no_copy
	pb      = ( n_elements(pb)       eq 0 ) ? 0 : pb
	if( obj_class(self) eq 'MRIMAGESETDICOM' and no_copy eq 1 ) then $
		message, 'Keyword no_copy must not be used with mrImageSetDicom objects'

	if( pb ) then pbar = obj_new('mrProgressbar')

		; first create a new matrix to store the images - reduced by the index dimension

	ndims = 0
	dims  = [-1,-1,-1]
	for i = 0, self.ndims-1 do begin
		if( i ne index ) then begin
			dims[ndims] = self.dims[i]
			ndims = ndims+1
		end
	end
	newimgs = ptrarr( dims[0:ndims-1] )
	imgs    = *(self.imgdata)

		; ok, now we have a pointer array for the new set of images
		; now loop over the non-condensed dimensions

		; help, I knotted my brain - isn't there a better way
		; to do this? The permutations drive me nuts.

	case index of
	0: begin
		case self.ndims of
		1: begin
			img = intarr( self.xsize, self.ysize, self.dims[0] )
			newimgs[0] = ptr_new( img, /no_copy )
			for k = 0, self.dims[0]-1 do begin
				(*newimgs)[*,*,k] = *imgs[k]
			end
		end
		2: begin
			for i = 0, self.dims[1]-1 do begin
				img = intarr( self.xsize, self.ysize, self.dims[0] )
				for k = 0, self.dims[0]-1 do begin
					img[*,*,k] = *imgs[k,i]
				end
				newimgs[i] = ptr_new( img, /no_copy )
				if( pb ) then bar->update, 1.*i/self.dims[1]-1
			end
		end
		3: begin
			for i = 0, self.dims[1]-1 do begin
				for j = 0, self.dims[2]-1 do begin
					img = intarr( self.xsize, self.ysize, self.dims[0] )
					for k = 0, self.dims[0]-1 do begin
						img[*,*,k] = *imgs[k,i,j]
					end
					newimgs[i,j] = ptr_new( img, /no_copy )
				end
				if( pb ) then bar->update, 1.*i/self.dims[1]-1
			end
		end
		end
	end
	1: begin
		case self.ndims of
		2: begin
			for i = 0, self.dims[0]-1 do begin
				img = intarr( self.xsize, self.ysize, self.dims[1] )
				for k = 0, self.dims[1]-1 do begin
					img[*,*,k] = *imgs[i,k]
				end
				newimgs[i] = ptr_new( img, /no_copy )
			end
		end
		3: begin
			for i = 0, self.dims[0]-1 do begin
				for j = 0, self.dims[2]-1 do begin
					img = intarr( self.xsize, self.ysize, self.dims[1] )
					for k = 0, self.dims[1]-1 do begin
						img[*,*,k] = *imgs[i,k,j]
					end
					newimgs[i,j] = ptr_new( img, /no_copy )
				end
				if( pb ) then bar->update, 1.*i/self.dims[0]-1
			end
		end
		end
	end
	2: begin
		for i = 0, self.dims[0]-1 do begin
			for j = 0, self.dims[1]-1 do begin
				img = intarr( self.xsize, self.ysize, self.dims[2] )
				for k = 0, self.dims[2]-1 do begin
					img[*,*,k] = *imgs[i,j,k]
				end
				newimgs[i,j] = ptr_new( img, /no_copy )
			end
			if( pb ) then bar->update, 1.*i/self.dims[0]-1
		end
	end
	end

	if( pb eq 1 ) then bar->update, 1.

		; if no_copy was set, replace the data in the local imageSet object
		; otherwise create a new mrImageSet object and return it

	if( no_copy eq 0 ) then begin                                   ; create new mrImageSet object
		self.palette->getProperty, blue=b, green=g, red=r
		case ndims of
			0: is = obj_new('mrImageSet',       0,          blue=b, green=g, red=r, xsize=self.xsize, ysize=self.ysize )
			1: is = obj_new('mrImageSet', dims[0],          blue=b, green=g, red=r, xsize=self.xsize, ysize=self.ysize )
			2: is = obj_new('mrImageSet', dims[0], dims[1], blue=b, green=g, red=r, xsize=self.xsize, ysize=self.ysize )
		end
		is->setName,  self.name+' condensed'
		is->setClass, self.class+' condensed'
		is->setMatrix, ptr_new( newimgs, /no_copy )
		is->setParent, self
		if( pb eq 1 ) then obj_destroy, bar
		return, is
	end else begin                                                   ; else replace data
		self->delete, /all
		self.dims=dims
		self.ndims=ndims
		ptr_free, self.imgdata
		self.imgdata = ptr_new( newimgs, /no_copy )
		if( pb eq 1 ) then obj_destroy, bar
		return, obj_new()
	end

end

; secret method, very private
pro mrImageSet::setMatrix, matrix
	self.imgData = matrix
end

;+=----------------------------------------------------------------------------
; Name  : mrImageSet::load
; Syntax: res = obj->load( filename, position, type=type)
; Args  : !!! WIP !!!
;         <b>filename</b>
;         The file to load
;         <b>type</b>
;         One of the image file formats supported by IDL, i.e.
;         'BMP', 'PNG', 'TIFF' or 'DICOM'. Note that access to dicom data is
;         not possible - mrImageSetDicom is to be used for that.
; Ret   : 0 in case of error, 1 in case of success.
;         mrImageSet::lastErrMsg can be used to obtain the error message.
; Desc  : Loads a single image and stores a pointer to the data at the
;         specified matrix location.
;         The first image loaded defines the image size and RGB colors for
;         this imageset object. Subsequently loaded images are checked to
;         be of same size and are rejected if they are not. RGB colors are
;         not checked, though.
;         Not all IDL supported file types are supported here because of
;         their color handling. Can be extended when needed.
;         The method is called 'load', not 'read'. IMO, load/save methods
;         include opening and closing of a file, while read/write methods
;         read from/write to an already opened file.
;=-----------------------------------------------------------------------------
function mrImageSet::load, filename, pos, type=type, slice=slice, shot=shot

	error = 0
	catch, error
	if( error ne 0 ) then begin
		catch, /cancel
		print, !ERROR_STATE.msg
		self.errmsg = !ERROR_STATE.msg
		return, 0
	end

	if( n_elements(filename) eq          0 ) then message, 'Missing filename argument'
	if( n_elements(pos)      ne self.ndims ) then message, 'Missing or wrong position argument'
	type  = (n_elements(type)  eq 0) ? '' : type

	if( type eq '' ) then begin
		res = query_image( filename, type=type )
		if( res eq 0 ) then message, 'Format not supporter from IDL'
	end

	type = strlowcase(type)

	img = -1
	case type of
		'bmp': 	 img = read_bmp( filename, r, g, b )
		'png': 	 img = read_png( filename, r, g, b )
		'dicom': img = read_dicom( filename, r, g, b )
		'tiff':  img = read_tiff( filename, r, g, b )
		else: begin
			message, '%Format not supported by mrImageSet: ' + type
		end
	end

	dims = size( img )
	case dims[0] of
		2: begin
			xsize = dims[1]
			ysize = dims[2]
		end
		3: begin
			xsize = dims[2]
			ysize = dims[3]
		end
		else: message, 'Error loading image'
	end
	if( ( (self.xsize ne 0) and (xsize ne self.xsize)) or $
	    ( (self.ysize ne 0) and (ysize ne self.ysize)) ) then message, 'Image size does not match set'

	if( self.xsize eq 0 ) then begin
		self.xsize = xsize
		self.ysize = ysize
	end

	if( type eq 'bmp' and dims eq 3) then begin   ; bitmap true color are bgr!
		tmp = img[0,*,*]
		img[0,*,*] = img[2,*,*]
		img[2,*,*] = tmp
	end

		; first image determines rgb, if not set before

	if( self.red[0] eq -1 ) then begin
		self.red     = r
		self.green   = g
		self.blue    = b
	end
	self.type    = size( res, /n_dimensions) - 2  ; 0: rgb, 1: true color

	id = pos[self.ndims-1]
	for i = self.ndims-2, 0, -1 do id = pos[i] + self.dims[i]*id
	self.imgdata[id]  = ptr_new(res, /no_copy)

		; notify others

	self->sendNotification, 'set'

	return, 1
end


;+=----------------------------------------------------------------------------
; Name  : mrImageSet::save
; Syntax: not implemented, yet. Needed?
; Args  :
; Ret   :
; Desc  :
;=-----------------------------------------------------------------------------
function mrImageSet::save
	message, 'Not implemented'
end


;+=----------------------------------------------------------------------------
; Name : mrImageSet__define
; Desc : The constructor.
;=-----------------------------------------------------------------------------
pro mrImageSet__define
	obj = { MRIMAGESET,              $
			inherits mrNotify		,$
			name:				''	,$ ;
			class:				''	,$ ;
			type:				 0	,$ ;
			typestr:			''	,$ ;
			dims:        intarr(3)  ,$ ;
			ndims:               0  ,$ ;
			xsize:				 0	,$ ;
			ysize:				 0	,$ ;
			path:				''	,$ ;
			errno:				 0  ,$ ; error number of last error
			errmsg:				''	,$ ; error message of last error
			uvalue:		 ptr_new()	,$ ; user definable pointer
			imgdata:	 ptr_new()	,$ ; pointer to the image data matrix
			palette:	 obj_new()	,$ ; an IDLgrPalette object
			parent:		 obj_new()	 $ ; if derived from another set, store reference here
	}
end
