;+=----------------------------------------------------------------------------
; Name   : mrmultidisplay
; RCS    : $Id: mrmultidisplay__define.pro,v 2.2 2004/11/24 10:44:12 olivert Exp olivert $
; Methods:
;          <b>Constructor</b>
;          mrmultidisplay::init             : constructor, 2nd part
;          <br>
;          <b>Public variables (get)</b>
;          mrmultidisplay::palette          : returns the currently used palette object
;          mrmultidisplay::imgXsize         : returns the image size in x in pixels
;          mrmultidisplay::imgYsize         : returns the image size in y in pixels
;          mrmultidisplay::nx               : returns number of images in x direction
;          mrmultidisplay::ny               : returns number of images in y direction
;          mrmultidisplay::contrast         : as the mrdisplay function
;          mrmultidisplay::gamma            ; return the currently used gamma value
;          mrmultidisplay::screenXsize      : return visible canvas size in x in pixels
;          mrmultidisplay::screenYsize      : return visible canvas size in y in pixels
;          mrmultidisplay::imgMinval        : return the minimum and maximum absolute (i.e.
;          mrmultidisplay::imgMaxval        : unscaled) values of all images
;          <br>
;          <b>Public variables (set)</b>
;          mrmultidisplay::setContrast      : set contrast for all displayed images
;          mrmultidisplay::setGamma         ; set the gamma value of the used palette
;          mrmultidisplay::setPalette       : set the palette to use
;          mrmultidisplay::setPaletteRGB    ; set palette RGB values or load from predefined colortables
;          <br>
;          <b>General</b>
;          mrmultidisplay::display          : display object content on screen
;          <br>
;          <b>Image handling</b>
;          mrmultidisplay::setImg           : x, y, ptr  : set image at position x,y
;          mrmultidisplay::deleteImg        : x, y       : delete image at position x,y
;          mrmultidisplay::label            : x,y,text,/set,/delete,color=c
;          mrmultidisplay::frame            : x,y,/set,/delete,color=c
;          <br>
;          <b>Event handling</b>
;          mrmultidisplay::requestDisplay   ; display after all event handlers are done
;          mrmultidisplay::activate         : generate events
;          mrmultidisplay::deactivate       : no events no more
;          mrmultidisplay::registerEvt      : register an event
;          mrmultidisplay::unregisterEvt    : unregister an event
;          mrmultidisplay::unregisterObj    : unregister a whole object
;          mrmultidisplay::unregisterAll    : remove all entries from callback tables
;          <br>
;          <b>Private</b>
;          mrmultidisplay::labelView        : obsolete?
;          mrmultidisplay_evthandler        : non-object event handler calls the object handler
;          mrmultidisplay::evthandler       : the object handler
;          mrmultidisplay::resolveTypeStr   : translate an event string into an event type
;          mrmultidisplay__define           : constructor, 1st part
;          mrmultidisplay::cleanup          : the destructor
; Desc   : An object to display a set of images in an array like fashion, while
;          automatically scaling the images to the correct size
;          Many properties are similar to the ones of 'mrdisplay' as we want
;          to be able to use the same callback objects where possible. Actually
;          mrdisplay and mrmultidisplay call the same event handler on the object side.
;          <b>Event handling:</b>
;          mrDisplay issues events on different conditions. The following event types are
;          presently defined - objects may register for one or more event types:
;          <br>
;          <b>Event types:</b>
;          Type (number)     string             occurence
;          0                'press'             any button press event
;          1                'dblclick'          any button double click event
;          2                'release'           any button release event
;          3                'motion'            mouse has been moved inside window
;          4                'keypress'          a key has been pressed inside window
;          5                'imgchange'         an image has been set or deleted
;          6                'palettechange'     the used palette has been changed (setpalette)
;          <br>
;          <b>The event structure</b>
;          > x = { mrdisplayEvt          ,$
;          >     type      :   0           ,$ ; (all) an integer denoting the event type, see below
;          >     top       :   0L          ,$ ; (0-4) the top widget id
;          >     display   :   obj_new()   ,$ ; (0-4) the display object itself
;          >     grWindow  :   obj_new()   ,$ ; (0-4) the IDLgrWindow object within the draw widget
;          >     lblView   :   obj_new()   ,$ ; (0-4) this objects label view window
;          >     x         :   0           ,$ ; x position of the event, device coordinates
;          >     y         :   0           ,$ ; y position of the event, device coordinates
;          >     dx        :   0           ,$ ; x difference since last event
;          >     dy        :   0           ,$ ; y difference since last event
;          >     nx        :   0           ,$ ; image number in x
;          >     ny        :   0           ,$ ; img number in y, origin is upper left (reading dir)
;          >     key       :   0L          ,$ ; key off an ASCII key press
;          >     modifiers :   0L          ,$ ; bitmask: 1 = shift, 2 = control
;          >     button    :   0           ,$ ; bitmap, left, middle right mouse button, pressed/released
;          >                                $ ; button for press/release events, current status in motion evts
;          >     buttonstatus: 0            $ ; current status of pressed/released mouse buttons, bitmap
;          > }
;
; Author : (c) Oliver Thilmann, oliver.thilmann@web.de, Oct 2002
; Licns  : This program is published under the GPL. See gpl.txt for details
;=-----------------------------------------------------------------------------


;==============================================================================
;
;                          CONSTRUCTORS AND DESTRUCTORS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::init
; Type  : Function
; Args  : long int baseid   : widget id of the base widget in which to realize
;         int      xsize    : widget size in x in pixels on screen
;         int      ysize    : widget size in y in pixels on screen
;         int      imgxsize : size in x of a single image
;         int      imgysize : size in y of a single image
;         int      nx       : number of images in x direction
;         int      ny       : number of images in y direction
; Ret   : 1 or 0, thus object pointer or null pointer
; Desc  : Instanciate an mrmultidisplay object in the base widget with the given
;         size. If imgxsize * nx (imgysize*y) should be larger than xsize,
;         scrollbars are automatically added to the widget.
;         All arguments are mandatory.
;         Rendering limits are checked for IDL versions 5.5 and greater
;=-----------------------------------------------------------------------------
function mrmultidisplay::init, baseid, xsize, ysize, imgxsize, imgysize, nx, ny
	; all arguments are mandatory!
	if( n_elements(xsize)    eq 0 or n_elements(ysize)    eq 0 or $
		n_elements(imgxsize) eq 0 or n_elements(imgysize) eq 0 or $
		n_elements(nx)       eq 0 or n_elements(ny)       eq 0 or $
		n_elements(baseid) eq 0 ) $
		then return, 0
	xlimit=1600
	ylimit=1200
	if( !VERSION.release gt '5.5' ) then begin
		xlimit=2048
		ylimit=2048
	end
	if( xsize gt xlimit or ysize gt ylimit or $
		imgxsize * nx gt xlimit or imgysize * ny gt ylimit ) then begin
		x = dialog_message(['Your version of IDL cannot render an area that large.',   $
							'Error in MRmultidisplay.'], /error)
		return, 0
	end

	xtruesize = xsize > imgxsize*nx                  ; the greater of the two
	ytruesize = ysize > imgysize*ny                  ; xtruesize is the full size

	self.wid_base = widget_base( baseid ) ;, xsize=xsize, ysize=ysize )
	self.wid_draw = widget_draw( self.wid_base, xsize=xtruesize, ysize=ytruesize, $
							x_scroll_size=xsize, y_scroll_size=ysize, /scroll, $
							graphics_level=2, renderer=1, retain=2, color_model=0 )
	self.xsize       = xtruesize
	self.ysize       = ytruesize
	self.imgXsize    = imgxsize
	self.imgYsize    = imgysize
	self.nx		     = nx
	self.ny          = ny

	widget_control, self.wid_draw, set_uvalue=self           ; for evt. handling

	; colors, colors
	device, decomposed=0

		; objects for callback functionality

	cbo                     = objarr( 7 )                  ; create 7 lists of call back objects
	for i = 0, 6 do cbo[i]  = obj_new('IDL_Container')
	self.callBackObjects    = ptr_new( cbo )

		; elements for image handling

	self.imgData            = ptr_new( ptrarr( nx, ny ), /no_copy )
	self.palette			= obj_new('IDLgrPalette')
	self.palette->loadct, 0
	self.topbytscl          = ptr_new( intarr( nx, ny ) )
	(*self.topbytscl)[*,*]  = 255
	self.center             = 0.5
	self.width              = 1.0
	self.displayRequestFlag = 0

		; vga palette is the palette for the labels

	self.vgaPalette         = obj_new('IDLgrPalette')           ; vga as defined in
	color_init, red=red, green=green, blue=blue                 ; color_init
	self.vgaPalette->setProperty, red_values=red, green_values=green, blue_values=blue

		; now for the graphics part:
		; Hierarchy (the objects in brackets are created elsewhere):
		; self.Scene -> self.imgViews -> self.imgModels -> (self.images)
        ;                                               -> (self.labels)
		;                                               -> (self.frames)
		;            -> self.lblView  -> self.lblModel

		; first: create all graphics objects

	iv = objarr( nx, ny )						; image views
	im = objarr( nx, ny )						; image models
	for i = 0, nx-1 do begin
		for j = 0, ny-1 do begin
				; don't subtract '1.' - problems with roundoff errors!
			iv[i,j] = obj_new('IDLgrView', viewplane_rect=[0,0,self.imgXsize-.99,self.imgYsize-.99], $
						location=[i*self.imgXsize, j*self.imgYsize], $
						dimensions=[self.imgXsize, self.imgYsize] )
			im[i,j] = obj_new('IDLgrModel')
		end
	end
	self.lblView  = obj_new('IDLgrView', viewplane_rect=[0.,0.,1.,1.], location=[0,0], $
							dimensions=[self.xsize, self.ysize], /transparent)
	self.lblModel = obj_new('IDLgrModel')

	self.imgViews	= ptr_new( iv, /no_copy )  ; pointer array for the image views
	self.imgModels  = ptr_new( im, /no_copy )  ; pointer array for models holding the images

		; second: create Scene and build up hierarchy as described above

	self.Scene	= obj_new('IDLgrScene')
	self.Scene->add, self.lblView                                     ; the global label view
	self.lblView->add, self.lblModel                                  ; and model
	for i = 0, nx-1 do begin
		for j = 0, ny-1 do begin
			(*(self.imgViews))[i,j]->add, (*(self.imgModels))[i,j]    ; and their models
			self.Scene->add, (*(self.imgViews))[i,j]                  ; all image views
		end
	end

		; Third: create array[nx,ny] of images, corresponding labels and frames
		; These are null pointers and set later as images and labels are added

	images      = objarr( nx, ny )
	self.images = ptr_new( images )
	labels      = objarr( nx, ny )
	self.labels = ptr_new( labels )
	frames      = objarr( nx, ny )
	self.frames = ptr_new( frames )

    return, 1
end


;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::cleanup
; Type  : Private procedure (destructor)
; Desc  : undone - thus memory leak
;=-----------------------------------------------------------------------------
pro mrmultidisplay::cleanup
	ptr_free, self.topbytscl
	ptr_free, self.event
	; some object destroys here.... later....
end


;==============================================================================
;
;                               GET AND SET METHODS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::setImg
; Syntax: obj->setImg, xpos, ypos, image_pointer, top=int, /no_contrastupdate
; Args  : <b>xpos</b>
;         x position of the image in the range [0, nx-1]
;         <b>ypos</b>
;         y position of the image in the range [0, ny-1]
;         <b>image_pointer</b>
;         pointer to the image data
;         <b>top</b>
;         top byte value to which to scale the image in the range [0,255].
;         default is 255.
;         <b>no_contrastupdate</b>
;         All images are scaled relative to the maximum value of all images,
;         such that identical grey scale values represent identical image
;         values. When setting an image with a higher maximum or lower minimum
;         value than the currently displayed images, the present images need
;         to be recalculated. This takes time of course - when setting many
;         images at once, recalculation is only necessary when setting the
;         very last image.
; Desc  : Set the image at position xpos, ypos. Coordinate system origin is at
;         the upper left and values range from 0 to nx-1 and ny-1 respectively.
;         The image will be downscaled to the size given in mrmultidisplay::init
;         Like mrdisplay, mrmultidisplay stores a local copy of the downsized
;         image data, so the user may delete the original data after setImg.
;         As of now ractangular images are resized to square format and thus
;         deformed.
;=-----------------------------------------------------------------------------
pro mrmultidisplay::setImg, xpos, ypos, imgptr, top=top, no_contrastupdate=no_contrastupdate

		; argument handling

	if( imgptr eq ptr_new() ) then return
	noc = keyword_set( no_contrastupdate )
	if( n_elements(top) ne 0 ) then begin
		top = top > 0
		top = top < 255
	end else begin
		top=255
	end

	x = xpos
	y = self.ny-1-ypos
	if( x ge self.nx or y ge self.ny or x lt 0 or y lt 0 or fix(x) ne x or fix(y) ne y ) then begin
		print, 'Error in mrmultidisplay::setImg: invalid position'
		return
	end
	(*self.topbytscl)[x,y] = top

		; Delete the old image, if it exists - this is a bit nasty due to redefinition
		; of 'y' - deleteImage also does a convertion, so we do ours before.
		; 'Call by reference' - hate it or despise it, you can't love it

	if( (*(self.imgData))[x,y] ne ptr_new() ) then begin
		self->deleteImg, xpos, ypos
	end

		; To make it easier for the user, the coordinate system has its origin
		; in the upper left, thus coordinates follow 'reading' direction.

	img = congrid( *imgptr, self.imgXsize, self.imgYsize )
	(*self.imgData)[x,y] = ptr_new( img, /no_copy )      ; no need to store more than necessary

	minval         = min( *imgptr, max=maxval )
	self.imgMinval = minval < self.imgMinval
	self.imgMaxval = maxval > self.imgMaxval
	contrast_low   = self.center - self.width > 0.
	contrast_hig   = self.center + self.width < 1.
	dispmin        = self.imgMinval + (self.imgMaxval-self.imgMinval)*contrast_low
	dispmax        = self.imgMinval + (self.imgMaxval-self.imgMinval)*contrast_hig

	(*self.images)[x,y]    = obj_new('IDLgrImage', $
								bytscl( *(*self.imgdata)[x,y], min=dispmin, max=dispmax, top=top ), $
								sub_rect = [0,0, self.imgXsize-1, self.imgYsize-1], $
								xcoord_conv = [0.,1.], $
								ycoord_conv = [0.,1.], $
								dimensions  = [self.imgXsize,self.imgYsize], $
								palette     = self.palette, $
								interpolate = 0 )
	(*self.imgModels)[x,y]->add, (*self.images)[x,y], position=0

		; this will redo the scaling

	if( not noc ) then self->setContrast

		; create ImgChange events (type 5)

	evt = ptr_new( {type:5, display:self}, /no_copy )
	objs = (*self.callbackobjects)[5]->get( /all, count=nobj )
	if( nobj gt 0 ) then begin
		for i = 0, nobj-1 do objs[i]->mrdisplayEvtHandler, evt
	end
	ptr_free, evt
	if( self.displayRequestFlag ) then self->display

end


;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::deleteImg
; Syntax: obj->deleteImg, xpos, ypos, /no_contrastupdate
;         obj->deleteImge, /all
; Args  : <b>xpos</b>
;         x position of the image in the matrix in range [0, nx-1]
;         <b>ypos</b>
;         y position of the image in the matrix in range [0, ny-1]
;         <b>all</b>
;         If set, all images are removed from the object
;         <b>no_contrastupdate</b>
;         All images are scaled relative to the maximum value of all images,
;         such that identical grey scale values represent identical image
;         values. Deleting an image might change the overall maximum and
;         minimum of course, which makes a recalculation necessary. Since
;         this takes time, it can be omitted if several images are deleted
;         at once and only be done with the last image.
; Desc  : Delete one image, i.e. the corresponding image model
;         Deleting an image does not delete possibly existing labels or
;         frames - use the corresponding functions to remove those as well.
;=-----------------------------------------------------------------------------
pro mrmultidisplay::deleteImg, xpos, ypos, all=all, no_contrastupdate=no_contrastupdate
	x   = ( n_elements(xpos) eq 0 ) ? -1 : xpos
	y   = ( n_elements(ypos) eq 0 ) ? -1 : self.ny-1-ypos
	all = keyword_set( all )
	noc = keyword_set( no_contrastupdate )
	if( all eq 0 and (x ge self.nx or x lt 0 or y ge self.ny or y lt 0) ) then return
	if( all eq 0 ) then begin							; delete just one image
		if( (*(self.imgModels))[x,y] eq obj_new() ) then return
		(*(self.imgViews))[x,y]->remove, (*(self.imgModels))[x,y]
		obj_destroy, (*(self.imgModels))[x,y]
		(*(self.images))[x,y]    = obj_new()
		(*(self.labels))[x,y]    = obj_new()
		(*(self.frames))[x,y]    = obj_new()
		(*(self.imgModels))[x,y] = obj_new('IDLgrModel')
		(*(self.imgViews))[x,y]->add, (*(self.imgModels))[x,y]
		ptr_free, (*self.imgData)[x,y]
		(*self.imgData)[x,y] = ptr_new()

			; recalculate img minimum and maximum values (brute force)

		self.imgMinval = 0.
		self.imgMaxval = 1.
		for x = 0, self.nx-1 do begin
			for y = 0, self.ny-1 do begin
				if( (*(self.imgData))[x,y] eq ptr_new() ) then continue
				minval = min( *(*(self.imgData))[x,y], max=maxval )
				self.imgMinval = minval < self.imgMinval
				self.imgMaxval = maxval > self.imgMaxval
			end
		end
		if( not noc ) then self->setContrast
	end else begin										; or loop over all
		for x = 0, self.nx-1 do begin
			for y = 0, self.ny-1 do begin
				if( (*(self.imgModels))[x,y] eq obj_new() ) then continue
				(*(self.imgViews))[x,y]->remove, (*(self.imgModels))[x,y]
				obj_destroy, (*(self.imgModels))[x,y]
				(*(self.images))[x,y]    = obj_new()
				(*(self.labels))[x,y]    = obj_new()
				(*(self.frames))[x,y]    = obj_new()
				(*(self.imgModels))[x,y] = obj_new('IDLgrModel')
				(*(self.imgViews))[x,y]->add, (*(self.imgModels))[x,y]
				ptr_free, (*self.imgData)[x,y]
				(*self.imgData)[x,y] = ptr_new()
			end
		end
		self.imgMinval = 0.
		self.imgMaxval = 1.
	end

		; create ImgChange events (type 5)

	evt = ptr_new( {type:5, display:self}, /no_copy )
	objs = (*self.callbackobjects)[5]->get( /all, count=nobj )
	if( nobj gt 0 ) then begin
		for i = 0, nobj-1 do objs[i]->mrdisplayEvtHandler, evt
	end
	ptr_free, evt
	if( self.displayRequestFlag ) then self->display

end


;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::contrast
; Syntax: ctr = obj->contrast()
; Ret   : float[2] contrast
;         Current contrast center and width (both in range [0.,1.])
;=-----------------------------------------------------------------------------
function mrmultidisplay::contrast
	return, [ self.center, self.width ]
end

;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::setContrast
; Syntax: obj->setContrast, center, width
; Args  : <b>center</c>
;         Relative center of contrast, 0. < center < 1.
;         <b>width</b>
;         Relative contrast width, 0. < width < 1.
; Desc  : Contrast settings are such that identical greyscale values (or colors)
;         correspond to identical image values (for equal 'top' settings).
;=-----------------------------------------------------------------------------
pro mrmultidisplay::setContrast, center, width
	if( n_elements(center) eq 0 ) then center = self.center
	if( n_elements(width)  eq 0 ) then width  = self.width
	self.center = 0. > center < 1.
	self.width  = 0. > width  < 1.
	contrast_low = self.center - self.width > 0.
	contrast_hig = self.center + self.width < 1.
	for x = 0, self.nx-1 do begin
		for y = 0, self.ny-1 do begin
			if( (*(self.imgData))[x,y] eq ptr_new() ) then continue
			minval  = min( *(*(self.imgData))[x,y], max=maxval )
			dispmin = self.imgMinval + (self.imgMaxval-self.imgMinval)*contrast_low
			dispmax = self.imgMinval + (self.imgMaxval-self.imgMinval)*contrast_hig
			(*(self.images))[x,y]->setProperty, $
			data = bytscl( *(*(self.imgData))[x,y], min=dispmin, max=dispmax, top=(*self.topbytscl)[x,y] )
		end
	end
end

;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::setGamma
; Syntax: obj->setGamma, gamma
; Args  : <b>gamma</b>
;         Gamma value for the image's palette 0.1 < gamma < 10.
; Desc  : Set the gamma value for the image's palette
;=-----------------------------------------------------------------------------
pro mrmultidisplay::setGamma, gamma
		self.palette->setProperty, gamma=gamma
end

;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::gamma
; Syntax: gamma = obj->gamma()
; Desc  : Return the current gamma value of the image palette
;=-----------------------------------------------------------------------------
function mrmultidisplay::gamma
	self.palette->getProperty, gamma=gamma
	return, gamma
end

;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::label
; Syntax: obj->label, x, y, label, /set, /delete, /all, color=color
; Args  : <b>x</b>
;         x position of the image
;         <b>y</b>
;         y position of the image to label
;         <b>label</b>
;         the label text
;         <b>set</b>
;         set the label, this is default, if a label text is given
;         <b>delete</b>
;         delete the label, this is default, if no label text is given
;         or the label text equals an empty string
;         <b>all</b>
;         set or delete labels for all images
;         <b>color</b>
;         3 element integer array [r,g,b] denoting the label color.
;         Default is lime.
; Desc  : Set or remove a label
;=-----------------------------------------------------------------------------
pro mrmultidisplay::label, xpos, ypos, label, set=set, delete=del, all=all, color=color

		; argument checking

	x     = ( n_elements(xpos)  eq 0 ) ? -1 : xpos
	y     = ( n_elements(ypos)  eq 0 ) ? -1 : self.ny-1-ypos
	txt   = ( n_elements(label) eq 0 ) ? '' : label
	set   = ( n_elements(set)   eq 0 ) ?  0 : set
	del   = ( n_elements(del)   eq 0 ) ?  0 : del
	all   = ( n_elements(all)   eq 0 ) ?  0 : all
	label = ( n_elements(label) eq 0 ) ? '' : label
	if( n_elements(color) ne 3 ) then color=color_getrgb('lime')

	if( label eq '' ) then del=1 else set=1-del
	if( all eq 0 and (x ge self.nx or x lt 0 or y ge self.ny or y lt 0) ) then begin
		message, 'invalid position argument'
		return
	end

	if( all eq 1 ) then begin
		x0 = 0
		x1 = self.nx-1
		y0 = 0
		y1 = self.ny-1
	end else begin
		x0 = x
		x1 = x
		y0 = y
		y1 = y
	end
	if( set eq 1 ) then begin
		for x = x0, x1 do begin
			for y = y0, y1 do begin
				; remove possibly existing labels
				if( (*(self.labels))[x,y] ne obj_new() ) then begin
					(*(self.imgModels))[x,y]->remove, (*(self.labels))[x,y]
					obj_destroy, (*(self.labels))[x,y]
				end
				; then set the new one
				charsize = self.imgXsize/10. < self.imgYsize/10.
				charsize = charsize > 10
				(*(self.labels))[x,y] = obj_new('IDLgrText', strings=label, color=color, $
								locations=[self.imgxsize/2, 5], recompute_dimensions=0, $
								xcoord_conv=[0, 1.], ycoord_conv=[0,1.], alignment=.5, $
								char_dimensions=[charsize, charsize])
				(*(self.imgModels))[x,y]->add, (*(self.labels))[x,y]
			end
		end
	end
	if( del eq 1 ) then begin
		for x = x0, x1 do begin
			for y = y0, y1 do begin
				if( (*(self.labels))[x,y] eq obj_new() ) then return
				(*(self.imgModels))[x,y]->remove, (*(self.labels))[x,y]
				obj_destroy, (*(self.labels))[x,y]
				(*(self.labels))[x,y] = obj_new()
			end
		end
	end
end



;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::frame
; Syntax: obj->frame, x, y, /set, /delete, /all, color=color
; Args  : <b>x</b>
;         x position of the image in the matrix (not pixels, but images!)
;         <b>y</b>
;         y position of the image in the matrix
;         <b>set</b>
;         set a frame, this is the default
;         <b>delete</b>
;         remove a frame
;         <b>all</b>
;         if set, all frames will be set or removed
;         <b>color</b>
;         a 3 element integer array denoting the frame color as [r,g,b]
;         default is lime.
; Desc  : Creates a thin frame around the corresponding image
;=-----------------------------------------------------------------------------
pro mrmultidisplay::frame, xpos, ypos, set=set, delete=del, all=all, color=color

		; argument checking

	x   = ( n_elements(xpos) eq 0 ) ? -1 : xpos
	y   = ( n_elements(ypos) eq 0 ) ? -1 : self.ny-1-ypos
	set = ( n_elements(set)  eq 0 ) ? 1 : set
	del = ( n_elements(del)  eq 0 ) ? 0 : del
	all = ( n_elements(all)  eq 0 ) ? 0 : all
	if( n_elements(color) ne 3 ) then color=color_getrgb('lime')

	if( del eq 1 ) then set = 0
	if( all eq 0 and ( x ge self.nx or x lt 0 or y ge self.ny or y lt 0 )) then return

		; if all is not set, the loop will be somewhat short :)
	if( all eq 1 ) then begin
		x0 = 0
		x1 = self.nx-1
		y0 = 0
		y1 = self.ny-1
	end else begin
		x0 = x
		x1 = x
		y0 = y
		y1 = y
	end
	if( set eq 1 ) then begin
		for x = x0, x1 do begin
			for y = y0, y1 do begin
				; remove possibly existing frame
				if( (*(self.frames))[x,y] ne obj_new() ) then begin
					(*(self.imgModels))[x,y]->remove, (*(self.frames))[x,y]
					obj_destroy, (*(self.frames))[x,y]
				end
				; and set a new one, ok, we could also just change the color
				(*(self.frames))[x,y] = obj_new('IDLgrPolyline', $
					[0,self.imgXsize-1,self.imgXsize-1,0,0], [0,0,self.imgYsize-1,self.imgYsize-1,0], $
					color=color)
				(*(self.imgModels))[x,y]->add, (*(self.frames))[x,y]
			end
		end
	end

	if( del eq 1 ) then begin
		for x = 0, self.nx-1 do begin
			for y = 0, self.ny-1 do begin
				if( (*(self.frames))[x,y] eq obj_new() ) then continue
				(*(self.imgModels))[x,y]->remove, (*(self.frames))[x,y]
				obj_destroy, (*(self.frames))[x,y]
				(*self.frames)[x,y] = obj_new()
			end
		end
	end

end


;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::screenXsize
; Syntax: win = obj->screenXsize()
; Ret   : Returns the visible canvas size in x in pixels ('x' argument in init)
;=-----------------------------------------------------------------------------
function mrmultidisplay::screenXsize
	return, self.xsize
end

;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::screenYsize
; Syntax: win = obj->screenYsize()
; Ret   : Returns the visible canvas size in y in pixels ('y' argument in init)
;=-----------------------------------------------------------------------------
function mrmultidisplay::screenYsize
	return, self.ysize
end

;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::grWindow
; Type : Function
; Desc : return the IDLgrWindow object associated with the drawing widget
;=-----------------------------------------------------------------------------
function mrmultidisplay::grWindow
	widget_control, self.wid_draw, get_value=grWindow
	return, grWindow
end

;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::labelView
; Type  : Function
; Ret   : The object's label view - an IDLgrView object which extends over the
;         whole display and has normalized coordinates. Meant to add your own
;         labels.
;=-----------------------------------------------------------------------------
function mrmultidisplay::labelView
	return, self.lblView
end

;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::palette
; Type : Function
; Ret  : A pointer to the IDLgrPalette object used to display the images
;=-----------------------------------------------------------------------------
function mrmultidisplay::palette
	return, self.palette
end

;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::setPalette
; Syntax: obj->setPalette, palette
; Args  : <b>palette</b>
;         An object of class IDLgrPalette
; Desc  : Copies the RGB values from the passed palette object into the
;         internal palette object.
;=-----------------------------------------------------------------------------
pro mrmultidisplay::setPalette, palette
	if( n_elements(palette) eq 0 ) then message, 'missing palette argument'
	if( not obj_isa( palette, 'IDLgrPalette' ) ) then message, 'palette is not of class IDLgrPalette'
	palette->getProperty, red=r, green=g, blue=b
	self.palette->setProperty, red=r, green=g, blue=b

			; callback PaletteChange event

	evt = ptr_new( {type:6, display:self}, /no_copy )
	objs = (*self.callbackobjects)[6]->get( /all, count=nobj )
	if( nobj gt 0 ) then begin
		for i = 0, nobj-1 do objs[i]->mrdisplayEvtHandler, evt
	end
	ptr_free, evt
	if( self.displayRequestFlag ) then self->display
end

;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::setPaletteRGB
; Syntax: obj->setPaletteRGB, red=intarr, green=intarr, blue=intarr, colorTable=int
; Args  : <b>red, green, blue</b>
;         Integer arrays with 256 elements holding the red/green/blue values of
;         the palette
;         <b>colorTable</b>
;         An integer value between 0 and 41 indexing one of IDL's predefined
;         color tables. If set to 42, a logarithmically divided colortable is
;         loaded. If colorTable is set, keywords red, green and blue are
;         ignored.
; Desc :  Sets the red/green/blue values of the image palette used to display
;         the image.
;=-----------------------------------------------------------------------------
pro mrmultidisplay::setPaletteRGB, red=red, green=gre, blue=blu, colorTable=ct
	if( n_elements(ct) ne 0 ) then begin
		if( ct lt 0 or ct gt 42 ) then message, 'Color table index out of bounds'
		if( ct lt 42 ) then begin
			self.palette->loadct, ct
		end else begin
			red = intarr( 256 )
			gre = intarr( 256 )
			blu = intarr( 256 )
			rs = [0,0,0,0,0,0,0,0,0,255,255,255,255,255,230,210]
			gs = [0,32,32,64,0,160,160,208,255,255,204,128,102,0,26,0]
			bs = [0,112,144,192,255,255,160,64,0,0,102,64,0,0,0,0]
				; maximum ratio is '8'
			scalefactor = 16./alog( 255./32.+1.)
			for i = 0, 255 do begin
				id = alog (1.*i/32.+1. )*scalefactor < 15.
				red[i] = rs[id]
				gre[i] = gs[id]
				blu[i] = bs[id]
			end
			self.palette->setProperty, red=red, green=gre, blue=blu
		end
	end else begin
		self.palette->setProperty, red=red, green=gre, blue=blu
	end

			; callback PaletteChange event

	evt = ptr_new( {type:6, display:self}, /no_copy )
	objs = (*self.callbackobjects)[6]->get( /all, count=nobj )
	if( nobj gt 0 ) then begin
		for i = 0, nobj-1 do objs[i]->mrdisplayEvtHandler, evt
	end
	ptr_free, evt
	if( self.displayRequestFlag ) then self->display

end

;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::imgXsize
; Type : Function
; Desc : Returns image size in x in pixels
;=-----------------------------------------------------------------------------
function mrmultidisplay::imgXsize
	return, self.imgXsize
end

;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::imgYsize
; Type : Function
; Desc : Returns image size in y in pixels
;=-----------------------------------------------------------------------------
function mrmultidisplay::imgYsize
	return, self.imgYsize
end

;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::nx
; Type : Function
; Desc : Returns number of images in x direction
;=-----------------------------------------------------------------------------
function mrmultidisplay::nx
	return, self.nx
end

;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::ny
; Type : Function
; Desc : Returns number of images in y direction
;=-----------------------------------------------------------------------------
function mrmultidisplay::ny
	return, self.ny
end

;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::imgMaxval
; Type : Function
; Desc : Returns the real (unscaled) maximum value of all displayed images
;=-----------------------------------------------------------------------------
function mrmultidisplay::imgMaxval
	return, self.imgMaxval
end

;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::imgMinval
; Type : Function
; Desc : Returns the real (unscaled) minimum value of all displayed images.
;=-----------------------------------------------------------------------------
function mrmultidisplay::imgMinval
	return, self.imgMinval
end

;==============================================================================
;
;                              PUBLIC OBJECT METHODS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::display
; Type  : Procedure
; Desc  : (re-)draw the current image
;=-----------------------------------------------------------------------------
pro mrmultidisplay::display
	widget_control, self.wid_draw, get_value=grWindow
	if( grWindow eq obj_new() ) then return               ; window not realized
	widget_control, hourglass=1                           ; may take some time
	grWindow->draw, self.Scene
	widget_control, hourglass=0
	self.displayRequestFlag = 0
end


;==============================================================================
;
;                                  EVENT HANDLING
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay::requestDisplay
; Syntax: obj->requestDisplay
; Desc  : Request a redisplay, but don't call mrdisplay::display immediately.
;         Instead, a redisplay of the screen is only performed after all external
;         event handlers have been called - thus avoiding possible overhead by
;         multiple calls to mrdisplay::display from each event handler.
;=-----------------------------------------------------------------------------
pro mrmultidisplay::requestDisplay
	self.displayRequestFlag = 1
end

;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::activate
; Desc : switch on event handling. IDL versions > 5.5 also provide keyboard
;        events
;=-----------------------------------------------------------------------------
pro mrmultidisplay::activate
	widget_control, self.wid_draw, event_pro='mrmultidisplay_evthandler'
	widget_control, self.wid_draw, /draw_button_events,  $
								   /draw_motion_events
	if( !VERSION.release gt '5.5' ) then $
		widget_control, self.wid_draw, /draw_keyboard_events
end

;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::deactivate
; Desc : switch off event handling
;=-----------------------------------------------------------------------------
pro mrmultidisplay::deactivate
	widget_control, self.wid_draw, draw_button_events=0,  $
								   draw_motion_events=0
	if( !VERSION.release gt '5.5' ) then $
		widget_control, self.wid_draw, draw_keyboard_events=0
end

;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::unregisterAll
; Desc : clear all entries in the callback table
;=-----------------------------------------------------------------------------
pro mrmultidisplay::unregisterAll
	for i = 0, n_elements(*self.callBackObjects)-1 do $
		(*self.callBackObjects)[i]->remove, /all
end


;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::registerEvt
; Type : Procedure
; Args : object callBackObject : the object which provides the call back function
;        string typestr        : string denoting the event type (s.b.)
; Desc : register an event
;        The object which registers with mrmultidisplay has to provide a
;        method named ::mrmultidisplay_evthandler which will then be called
;        with a pointer to mrmultidisplay's event structure as an argument.
;        (See above for the event structure definition).
;        Usually, the first thing this routine does is to check for the
;        button field, to find out, whether this event is of interest to
;        the object.
;=-----------------------------------------------------------------------------
pro mrmultidisplay::registerEvt, object, typestr
	if( n_elements( typestr ) eq 0 ) then return
	for i = 0, n_elements(typestr)-1 do begin
		type = self->resolveTypeStr(typestr[i])
		if( type eq -1 ) then continue
		(*self.callBackObjects)[type]->remove, object
		(*self.callBackObjects)[type]->add, object
	end
end

;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::unregisterEvt
; Type : Procedure
; Args : object callBackObject  : the object which is to be removed
;        string eventType       : a string denoting the event type to remove
; Desc : unregister an event from mrmultidisplay callback
;=-----------------------------------------------------------------------------
pro mrmultidisplay::unregisterEvt, cbobj, typestr
	if( n_elements(typestr) eq 0 ) then return
	for i = 0, n_elements(typestr)-1 do begin
		type = self->resolveTypeStr(typestr)
		if( type eq -1 ) then continue
		(*self.callBackObjects)[type]->remove, cbobj
	end
end

;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::unregisterObj
; Type : Procedure
; Args : object callBackObject  : the call back object to remove
; Desc : unregister a whole object from mrdisplay callback
;=-----------------------------------------------------------------------------
pro mrmultidisplay::unregisterObj, cbobj
	for i = 0, n_elements(*self.callBackObjects)-1 do begin
		((*self.callBackObjects)[i])->remove, cbobj
	end
end



;==============================================================================
;
;                             PRIVATE OBJECT METHODS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::resolveTypeStr
; Type : Private function
; Desc : Convert the type string into an index
;=-----------------------------------------------------------------------------
function mrmultidisplay::resolveTypeStr, typestr
	case strlowcase(typestr) of
		'press'				: return, 0
		'dblclick'			: return, 1
		'release'			: return, 2
		'motion'			: return, 3
		'keypress'          : return, 4
		'imgchange'         : return, 5
		'palettechange'     : return, 6
		else				: return, -1
	end
end


;+=----------------------------------------------------------------------------
; Name : mrmultidisplay_evthandler
; Type : Private procedure
; Calls: mrmultidisplay::evtHandler
; Desc : call the object's own event dispatcher
;=-----------------------------------------------------------------------------
pro mrmultidisplay_evthandler, evt
	widget_control, evt.id, get_uvalue=display
	if( not obj_valid(display) ) then begin
		print, "Invalid object pointer in event handler"
		return
	end
	display->evtHandler, evt
end


;+=----------------------------------------------------------------------------
; Name : mrmultidisplay::evthandler
; Type : Private procedure
; Desc : the event handler which does the callbacks
;=-----------------------------------------------------------------------------
pro mrmultidisplay::evthandler, evt

		; draw events evt structures contain: type, x, y, press, release, clicks
		; type=0: button press   1: button release  2: motion
		; press/release: bitmaps of the mousebuttons
		; clicks: 1 for single, 2 for double clicks
		; fill some of the fields

	widget_control, self.wid_draw, get_uvalue=grWindow

	mrevt = { type:-1, display:self, top:evt.top, grWindow:grWindow, x:evt.x, y:evt.y, $
			  dx:evt.x-self.lastx, dy:evt.y-self.lasty, nx:fix(evt.x/self.imgXsize), $
			  ny:self.ny-1-fix( evt.y/self.imgYsize), modifiers:evt.modifiers, key:0L, button:0, buttonStatus:0 }

		; finally dispatch the event

	case evt.type of
		0:	begin                                                  ; button press
				button = (255-self.buttonStatus) and evt.press     ; wrong IDL docu!
				self.buttonStatus = self.buttonStatus or evt.press
				mrevt.type = fix(evt.clicks-1)
			end
		1:	begin                                                  ; button release
				button = evt.release
				self.buttonStatus = self.buttonStatus and (255 - evt.release )
				mrevt.type = 2
			end
		2:	begin                                                  ; motion
				button = self.buttonStatus
				mrevt.type = 3
			end
		5:	begin                                                  ; ASCII press
				button = self.buttonStatus
				if( evt.press eq 1 ) then begin
					mrevt.type = 4
					mrevt.key  = long( evt.ch )
				end
			end
		6:	begin
				button = self.buttonStatus
				if( evt.press eq 1 ) then begin
					mrevt.type = 4
					mrevt.key = evt.key
				end
			end
	end

	if( mrevt.type eq -1 ) then return

	mrevt.button       = button
	mrevt.buttonStatus = self.buttonStatus

	self.lastx = evt.x
	self.lasty = evt.y

		; finally call all objects in our list

	objs = (*self.callBackObjects)[mrevt.type]->get( /all, count=nobj )
    mrevt = ptr_new( mrevt, /no_copy )
	if( nobj gt 0 ) then begin
		for i = 0, nobj-1 do objs[i]->mrdisplayEvtHandler, mrevt
	end
	ptr_free, mrevt
	if( self.displayRequestFlag ) then self->display
end

;+=----------------------------------------------------------------------------
; Name  : mrmultidisplay__define
; Type  : Private Procedure (constructor)
;=-----------------------------------------------------------------------------
pro mrmultidisplay__define
	obj = { MRMULTIDISPLAY, $
			wid_base		:	0			,$ ; the base widget's id (not used?)
			wid_draw		:	0			,$ ; the drawing widget's id
			xsize			:	0			,$ ; size in x of base and draw widget [pixels]
			ysize			:	0			,$ ; size in y of base and draw widget [pixels]
			nx				:	0			,$ ; number of images in x
			ny				:	0			,$ ; number of images in y
			imgXsize		:	0			,$ ; image size in pixels in x
			imgYsize		:	0			,$ ; image size in pixels in y
			imgMaxval       :   1.          ,$ ; maximum absolute value of all images
			imgMinval       :   0.          ,$ ; minimum absolute value of all images, usually 0.
			width			:   0.			,$ ; width of contrast setting
			center			:   0.			,$ ; center of contrast settings
			topbytscl		:	ptr_new()	,$ ; each img's top value for image bytescaling
			Scene:              obj_new()   ,$ ; the scene to draw in wid_draw
			imgViews:           ptr_new()   ,$ ; ptr to an [nx,ny] array of IDLgrView objects
			imgModels:          ptr_new()   ,$ ; dito
			images:             ptr_new()   ,$ ; dito encore
			imgData:            ptr_new()   ,$ ; *[nx,ny]: local scaled copy of all image data
			lblView:            obj_new()   ,$ ; one global label view and then
			lblModel:           obj_new()   ,$ ; its model
			lblViews:           ptr_new()   ,$ ; an array of views to hold labels for each img
			lblModels:          ptr_new()   ,$ ; the corresponding models
			labels:             ptr_new()   ,$ ;
			frames:             ptr_new()   ,$ ; the 'active' frames
			palette			:	obj_new()	,$ ; the palette used for all images
			vgaPalette		:	obj_new()	,$ ; filled with the 16 colors from color_init
         	displayRequestFlag: 0			,$ ; set to '1', if redisplay is requested
			callBackObjects	:	ptr_new()	,$ ; EVT: an array of IDL_Containers
			event			:	ptr_new()	,$ ; pointer to the event structure
			lastx			:	0			,$ ;
			lasty			:	0			,$ ;
			buttonStatus	:	0			 $ ; EVT: mouse buttons pressed or released - bitmap
	}
end
