;+=----------------------------------------------------------------------------
; Name   : mrNotify
; RCS    : $Id: mrnotify__define.pro,v 2.2 2004/11/24 10:44:31 olivert Exp olivert $
; Methods:
;         <b>Constructor and init</b>
;         mrNotify::init             : constructor, no arguments
;         <br>
;         <b>User part (public)</b>
;         mrNotify::notify           : register an external object for notification
;         mrNotify::enableNotify     : switch on notification (default!)
;         mrNotify::disableNotify    : switch off notification
;         <br>
;         <b>Class part (protected)</b>
;         mrNotify::setEvtTypes      : set available event types (string array)
;         mrNotify::sendNotification : send the notification message
;         <br>
;         <b>Private functions</b>
;         mrNotify::evtBitmap        : creates bitmap from event string array
;         mrNotify__define           : the constructor
;         mrNotify::cleanup          : the destructor
; Desc  : Implements a simple notification mechanism. Classes implementing this
;         mechanism should subclass from mrNotify.
;         Notifications send by mrNotify have either two or three arguments
;         of the form
;         pro some_other_class::notificationHandler, sender, event_type, evt
;         where sender is an object reference to the sending obejct, event_type
;         a string, e.g. 'add' or 'delete' in mrContainer, and the optional
;         argument evt a pointer to an event structure defined by the implementing
;         superclass.
; Author: (c) Oliver Thilmann, oliver.thilmann@radfys.lu.se
; Licns : This program is published under the GPL. See gpl.txt for details
;=-----------------------------------------------------------------------------

;==============================================================================
;
;                          CONSTRUCTORS AND DESTRUCTORS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrNotify::init
; Syntax: obj = obj_new('mrNotify', evtTypes)
; Args  : <b>evtTypes</b>
;         A string or array of strings holding the available event types
;         (e.g. ['add','remove','move'] for a container class.
; Desc  : s.a.
;         Notification is switched on by default.
;=-----------------------------------------------------------------------------
function mrNotify::init, evtTypes

	self.notify = 1

	ntypes = n_elements(evtTypes)
	if( ntypes eq 0 ) then return, 1         ; dunno the use, but wrong it isn't
	arr = evtTypes
	for i = 0, ntypes-1 do arr[i] = strlowcase(arr[i]) ; convert to lowercase
	self.evttypes = ptr_new( arr, /no_copy )

	return, 1
end

;+=----------------------------------------------------------------------------
; Name  : mrNotify::cleanup
; Desc  : Free allocated memory
;=-----------------------------------------------------------------------------
pro mrNotify::cleanup
	ptr_free, self.evttypes
	for i = 0, self.cbSize-1 do ptr_free, (*self.cbList)[i]
	ptr_free, self.cbList
	self.cbSize = 0           ; if some weirdo should call cleanup from outside
end

;==============================================================================
;
;                   INITIALIZATION AND SETTING OF NOTIFICATIONS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrNotify::setEvtTypes
; Syntax: <b>deprecated!</b>
; Desc  : As setting the event types should be called only once, it is now
;         included in the constructor. The method is only kept for compatibility
;         reasons, but is deprecated and will be removed sooner or later.
;=-----------------------------------------------------------------------------
pro mrNotify::setEvtTypes, evtTypes
	x = dialog_message('deprecated procedure mrNotify::setEvtTypes')
	ntypes = n_elements(evtTypes)
	if( ntypes eq 0 ) then return
	for i = 0, self.cbSize-1 do ptr_free, (*self.cbList)[i]
	ptr_free, self.cbList
	ptr_free, self.evttypes
	self.cbSize = 0
	arr = evtTypes
	for i = 0, ntypes-1 do arr[i] = strlowcase(arr[i]) ; convert to lowercase
	self.evttypes = ptr_new( arr, /no_copy )
end

;+=----------------------------------------------------------------------------
; Name  : mrNotify::notify
; Syntax: obj->notify, object, method, event, /remove, /all
; Args  : <b>object</b>
;         The object to be notified
;         <b>method</b>
;         A string holding the object's method name to call
;         <b>event</b>
;         A string or string array specifying the events upon which to send
;         a notification event. The list of possible events is set via
;         mrNotify::setEvtTypes. Subclasses to mrNotify (the usual case)
;         will specify their specific event types - see the documentation of
;         the respective classes.
;         <b>remove</b>
;         If set, remove all instances of the specified object or object/method
;         combination from the notification list.
;         <b>all</b>
;         Only meaningful in combination with remove, in which case the
;         internal notification list is cleared.
; Desc  : Objects are notified in arbitrary order, there's no guarantee that
;         objects which register first are notified first. If an object
;         registers multiple times for notification, it will be notified
;         multiple times.
;         The registered method is called with two or three arguments: an
;         object reference of the calling object (i.e. self), a string denoting
;         the event type and optionally a pointer to a class specific event
;         structure. Again see the subclass documentation.
;=-----------------------------------------------------------------------------
pro mrNotify::notify, object, method, event, remove=remove, all=all
	obj = n_elements( object ) eq 0 ? obj_new() : object
	met = n_elements( method ) eq 0 ?        '' : method
	evt = n_elements( event  ) eq 0 ?       255 : self->evtBitmap(event)

	if( evt eq -1 ) then message, 'Unknown event type'

		; remove items from list

	if( keyword_set( remove ) ) then begin
		if( self.cbSize eq 0 ) then return      ; list is empty
		if( keyword_set( all ) ) then begin
			for i = 0, self.cbSize-1 do ptr_free, (*self.cbList)[i]
			ptr_free, self.cbList
			self.cbSize = 0
			return
		end
		if( obj eq obj_new() ) then return      ; ignore silently
		for i = 0, self.cbSize-1 do begin
			if( (*self.cbList)[i] eq ptr_new() ) then continue
			if( (*(*self.cbList)[i]).o ne obj ) then continue
			if( (met ne '') and ((*(*self.cbList)[i]).m ne met) ) then continue
			ptr_free, (*self.cbList)[i]
			(*self.cbList)[i] = ptr_new()
		end
		return
	end

		; add items to list - create list, if empty

	if( met eq '' ) then message, 'Method name not specified'
	newEntry = {o:obj, m:met, e:evt }
	if( self.cbSize gt 0 ) then begin
		for i = 0, self.cbSize-1 do begin
			if( (*self.cbList)[i] eq ptr_new() ) then begin
				(*self.cbList)[i] = ptr_new( newEntry, /no_copy )
				return
			end
		end
		newcblist = [ *self.cbList, ptr_new( newEntry, /no_copy ) ]
		ptr_free, self.cblist
		self.cbList = ptr_new( newcblist, /no_copy )
		self.cbSize = self.cbSize + 1
		return
	end
	self.cbList = ptr_new( ptr_new( newEntry, /no_copy ), /no_copy )
	self.cbSize = 1

end

;+=----------------------------------------------------------------------------
; Name  : mrNotify::sendNotification
; Syntax: self->sendNotification, evtType, evt
; Args  : <b>evtType</b>
;         Case insensitive string - one of the event types as set in
;         mrNotify::setEvtTypes
;         <b>evt</b>
;         Pointer to an individual event structure which is passed as third
;         argument. If not present, it is omitted.
; Desc  : Sends a notification event to objects which registered for the
;         specified event type. A notification call has the arguments
;         sender, event_type, evt
;         Where sender is the object that sends the notification (i.e. self),
;         event_type a string denoting the event and evt, if present, a user-
;         definable pointer to an event structure.
; Bugs  : If an object is removed from the internal list this is not considered
;         within the current call-back loop.
;=-----------------------------------------------------------------------------
pro mrNotify::sendNotification, evtType, evt

		; catch errors - this will mostly happen when trying to call non-existent
		; functions
		; Actually it also catches all my other programming errors which happen
		; all too often during call-backs.

;	error = 0
;	catch, error
;	if( error ne 0 ) then begin
;		catch, /cancel
;		print, 'mrNotify: ', !ERROR_STATE.msg
;		return
;	end

	if( not self.notify ) then return
	if( self.cbSize eq 0 ) then return
	bitmap = self->evtBitmap( evtType )
	if( bitmap le 0 ) then return

		; create a list of object/method pairs to call - this ensures that
		; possible changes performed during callback will not lead to an internal
		; inconsistent state

	objects = objarr( self.cbSize )
	methods = strarr( self.cbSize )
	nCalls  = 0
	for i = 0, self.cbSize-1 do begin
		if( (*self.cbList)[i] eq ptr_new() ) then continue
		o = (*(*self.cbList)[i]).o
		e = (*(*self.cbList)[i]).e
		if( ( (e and bitmap) gt 0 ) and obj_valid(o) ) then begin ; talking about precedence... !@#$^%
			objects[nCalls] = o
			methods[nCalls] = (*(*self.cbList)[i]).m
			nCalls = nCalls+1
		end
	end

		; now loop over the list

	if( n_elements(evt) eq 0 ) then begin
		for i = 0, nCalls-1 do call_method, methods[i], objects[i], self, evtType
	end else begin
		for i = 0, nCalls-1 do call_method, methods[i], objects[i], self, evtType, evt
	end
end


;+=----------------------------------------------------------------------------
; Name  : mrNotify::enableNotify
; Syntax: obj->enableNotify
; Desc  : Switches on notification
;=-----------------------------------------------------------------------------
pro mrNotify::enableNotify
	self.notify = 1
end

;+=----------------------------------------------------------------------------
; Name  : mrNotify::disableNotify
; Syntax: obj->disableNotify
; Desc  : Switches off notification
;=-----------------------------------------------------------------------------
pro mrNotify::disableNotify
	self.notify = 0
end


;==============================================================================
;
;                                   INTERNAL
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrNotify::evtBitmap
; Syntax: obj->evtBitmap
; Desc  : Turns a string array into a bitmap for easier and faster storage/comparison
;=-----------------------------------------------------------------------------
function mrNotify::evtBitmap, events
    evt = 0
    for i = 0, n_elements(events)-1 do begin
       pos = where( strlowcase(events[i]) eq *self.evttypes )
       if( pos[0] eq -1 ) then return, -1
       evt = evt or 2^pos[0]
    end
    return, evt
end


;+=----------------------------------------------------------------------------
; Name : mrNotify__define
; Desc : The constructor.
;=-----------------------------------------------------------------------------
pro mrNotify__define
	obj = { MRNOTIFY,              	 $
			cbList:		ptr_new()	,$ ; pointer to an array of pointers to structs :)
			cbSize:				0	,$ ; size of the list
			evttypes:   ptr_new()   ,$ ; string array of events
			notify:     		0    $ ; on (1) or off (0)
	}
end