;+=----------------------------------------------------------------------------
; Name   : mrparameterset
; RCS    : $Id: mrparameterset__define.pro,v 2.0 2004/08/05 18:09:36 olivert Exp olivert $
; Methods:
;          <b>Constructor</b>
;          mrparameterset::init         : constructor, 2nd part
;          <br>
;          <b>Public variables (get)</b>
;          mrparameterset::name         : returns the name of this parameter set
;          <br>
;          <b>Public methods</b>
;          mrparameterset::getByName    : get a parameter element by its name
;          mrparameterset::check        : calls the check method of all members
;          mrparameterset::load         : load a parameter set from disk
;          mrparameterset::save         : save a parameter set to disk
;          mrparameterset::dump         : dump contents to screen - meant for debugging
;          <br>
;          <b>Inherited methods from IDL_Container</b>
;          mrparameterset::count        : returns number of elements
;          mrparameterset::get          : /all, position=index, count=var
;          mrparameterset::move         : source_index, destination_index
;          mrparameterset::remove       : obj, /all, position=index
;          <br>
;          <b>Overloaded methods</b>
;          mrparameterset::add          : overloaded to check for class
;          mrparameterset::isContained  : name, position=var
;          <br>
;          <b>XML Parser</b>
;          mrparxml__define             : constructor, part I
;          mrparxml::init               : constructor, part II
;          mrparxml::startElement       : only start element,
;          mrparxml::endElement         : end element,
;          mrparxml::characters         : and data values are parsed
;          <br>
;          <b>Private</b>
;          mrparameterset__define       : constructor, 1st part
;          mrparameterset::cleanup      : the destructor
; Desc  :
;         <b>File format:</b>
;         The file format is XML, described by a DTD directly in the file and
;         should be pretty self explaining. A parameter entry has the form
;         <tt>
;         <Parameter name="a_parameter", type="int', size="1">
;         <Low>0</Low>
;         <Up>100</Up>
;         <Val>10</Val>
;         </Paramter>
;         </tt>
;         Destroying a parameterset will automatically destroy all contained
;         parameter objects.
; Author: (c) Oliver Thilmann, oliver.thilmann@web.de
; Licns : This program is published under the LGPL. See gpl.txt for details
;=-----------------------------------------------------------------------------


;==============================================================================
;
;                          CONSTRUCTORS AND DESTRUCTORS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrparameterset::init
; Syntax: obj = obj_new('mrparameterset', name, filename=file )
; Args  : <b>name</b>
;         String, which defines a name for the parameter set
;         <b>filename</b>
;         If given, a parameterset is loaded from the given filename.
;         Should that fail, '0' is returned, i.e. the object is not created.
; Desc  : Instanciates a new object of class mrparameterset and calls the
;         constructor of the superclass IDL_Container
;=-----------------------------------------------------------------------------
function mrparameterset::init, name, filename=file
 	if (self->IDL_Container::Init() ne 1) then return, 0
 	if( n_elements(name) eq 1 ) then self.name = name
 	if( n_elements(file) ne 0 ) then begin
 		res = self->load( file )
 		return, res
	end
	return, 1
end


;+=----------------------------------------------------------------------------
; Name  : mrparameterset::cleanup
; Syntax: obj_destroy, obj
; Desc  : The destructor - destroying a container will automatically destroy
;         all contained objects.
;=-----------------------------------------------------------------------------
function mrparameterset::cleanup
	self->IDL_Container::cleanup
end


;==============================================================================
;
;                              PUBLIC CLASS METHODS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrparameterset::add
; Syntax: obj->add, objects [, position=index]
; Args  : <b>objects</b>
;         An object instance or an array of objects of class mrparameter to
;         be added to the parameter set.
;         <b>index</b>
;         Scalar or array of zero based index values. The number of elements
;         must equal the number of object references specified by the objects
;         argument, or the method will return immediately.
;         Each index specifies the position at which the corresponding object
;         is to be placed. Defauls is to add new objects at the end of the list.
; Desc  : Add an object to the parameter set.
;         Objects which are not of class mrparameter are silently ignored.
;         Objects of identical names in the set will be replaced (and destroyed!)
;         This might possibly screw up position settings.
;         If no position is given, the object is added to the end of the list.
;=-----------------------------------------------------------------------------
pro mrparameterset::add, obj, position=index
	if( n_elements(obj) eq 0 ) then return
	if( n_elements(index) ne 0 and n_elements(obj) ne n_elements(index) ) then return

	for i = 0, n_elements(obj)-1 do begin
		if( not obj_isa( obj[i], 'mrparameter' ) ) then continue
		oldobj = self->getByName( obj[i]->name() )
		if( oldobj ne obj_new() ) then begin
			self->IDL_Container::remove, oldobj
			obj_destroy, oldobj
		end
		if( n_elements(index) ne 0 ) then begin
			self->IDL_Container::add, obj[i], position=index[i]
		end else begin
			self->IDL_Container::add, obj[i]
		end
	end
	return
end

;+=----------------------------------------------------------------------------
; Name  : mrparameterset::name
; Syntax: result = obj->name
; Desc  : Returns the name of this parameter set.
;=-----------------------------------------------------------------------------
function mrparameterset::name
	return, self.name
end

;+=----------------------------------------------------------------------------
; Name  : mrparameterset::getByName
; Syntax: result = obj->getByName, name
; Args  : <b>name</b>
;         A string naming the mrparameter object to return
; Desc  : The method returns the mrparameter object of that name or obj_new(),
;         if an object of the specified name is not contained in the set.
;=-----------------------------------------------------------------------------
function mrparameterset::getByName, name
	all = self->IDL_Container::get( /all, count=count )
	if( count eq 0 ) then return, obj_new()
	for i = 0, n_elements(all)-1 do begin
		if( all[i]->name() eq name ) then return, all[i]
	end
	return, obj_new()
end

;+=----------------------------------------------------------------------------
; Name  : mrparameterset::check
; Syntax: result = obj->check()
; Desc  : Calls mrparameter::check for all contained objects and returns true (1),
;         if check was successful for all objects or false (0), if it failed for
;         at least one.
;=-----------------------------------------------------------------------------
function mrparameterset::check
	all = self->IDL_Container::get( /all )
	for i = 0, n_elements(all)-1 do begin
		if( all[i]->check() eq 0 ) then return, 0
	end
	return, 1
end

;+=----------------------------------------------------------------------------
; Name  : mrparameterset::dump
; Syntax: obj->dump
; Desc  : Dump content of the parameterset to screen by calling the dump
;         function of each of the included parameters.
;         Mainly meant for debugging
;=-----------------------------------------------------------------------------
pro mrparameterset::dump
	print, '--- mrParameterSet dump start ---'
	print, '    Name: ', self.name
	print, '    File: ', self.filename
	all = self->IDL_Container::get( /all, count=count )
	for i = 0, count-1 do begin
		all[i]->dump
	end
	print, '--- mrParameterSet dump end   ---'
end

;+=----------------------------------------------------------------------------
; Name  : mrparameterset::isContained
; Syntax: result = obj->isContained, name, position=var
; Args  : <b>name</b>
;         A string denoting the name of the parameter to check
;         <b>position</b>
;         If set, the variable upon return holds the position at which the
;         parameter is located or -1.
; Desc  : The method returns true (1), if a mrparameter of the specified
;         name exists in the set, or false (0) otherwise.
;=-----------------------------------------------------------------------------
function mrparameterset::isContained, name, position=var
    var = -1
	if( n_elements(name) eq 0 ) then return, 0
	all = self->IDL_Container::get( /all, count=count )
	if( count eq 0 ) then return, 0
	for i = 0, n_elements(all)-1 do begin
		if( all[i]->name() eq name ) then begin
			var = i
			return, 1
		end
	end
	return, 0
end

;+=----------------------------------------------------------------------------
; Name  : mrparameterset::load
; Syntax: result = obj->load( filename, doubles=var )
; Ret   : Returns 1 if successful and an error code <= 0 otherwise
;          0: unspecified error
;         -1: error opening file
; Args  : <b>filename</b>
;         The filename (complete path) from which to load the parameter set
;         <b>doubles</b>
;         A string ('error', 'replace'(D) or 'ignore') which denotes the behaviour
;         in case a mrparamter object of the same name already exists in the list.
;         By default, the object will be destroyed and replaced by the new one.
;         The 'error' option has not been tested yet. There's anyway some work to
;         be done on catching errors from the XML parser.
;         When set to 'ignore', the object in the file is ignored.
; Desc  : The load method loads a parameter set from the specified file and adds
;         the contained data as mrparameters to the existing parameterset.
;=-----------------------------------------------------------------------------
function mrparameterset::load, filename, doubles=doubles

		; Argument checking

	doubles  = ( n_elements(doubles) eq 0 )  ? 'error' : doubles
	if( doubles ne 'error' and doubles ne 'replace' and doubles ne 'ignore' ) then begin
		print, "% mrparameterset: illegal value for parameter double"
		return, 0
	end

		; check file

	if( n_elements(filename) eq 0 ) then return, -1
	if( filename eq '' ) then return, -1
	finfo = file_info( filename )
	if( finfo.exists eq 0 ) then return, -1

		; create mrparxml object

	xmlparser = obj_new('mrparxml', self, doubles)
	if( xmlparser eq obj_new() ) then return, 0

		; parse the file

	xmlparser->parseFile, filename

		; we're done

	obj_destroy, xmlparser
	return, 1
end


;+=----------------------------------------------------------------------------
; Name  : mrparameterset::save
; Syntax: obj->save, filename, /valuesonly, error=err
; Args  : <b>filename</b>
;         A string specifying the filename to write to.
;         <b>valuesonly</b>
;         If set, only the parameter's values, but no limits or valid value
;         lists are saved. Default is to save all.
;         <b>error</b>
;         The corresponding variable is set to '1' if anything goes wrong
;         while saving.
; Desc : Saves a parameterset to disk. If the parameterset was originally
;        loaded from file, filename may be omitted and the original file
;        will be overwritten.
;=-----------------------------------------------------------------------------
pro mrparameterset::save, filename, valuesonly=vo, error=err

	vo = (n_elements(vo) eq 0) ? 0 : vo       ; default is save all

		; take filename - either from argument or from object var

	file = ( n_elements(filename) eq 0 ) ? self.filename : filename
	if( file eq '' ) then begin
		print, "% mrparameterset: no filename given."
		err=1
		return
	end

		; open file

	openw, unit, file, /get_lun, error=err
	if( err ne 0 ) then begin
		print, "% mrparameterset: error opening file ", !ERROR_STATE.MSG
		return
	end

		; write XML header and DTD

	printf, unit, '<?xml version="1.0"?>'
	printf, unit, '<!-- DTD included to check for handmade changes -->'
	printf, unit, '<!DOCTYPE Parameter_set ['
	printf, unit, '<!ELEMENT Parameter_set (Parameter*)>'
	printf, unit, '<!ATTLIST Parameter_set Version CDATA #REQUIRED>'
	mrparameter__writedtd, unit
	printf, unit, ']>'
	printf, unit, ''

		; get all parameter objects and call the write method for each

	printf, unit, '<Parameter_set Version="1.0">'
	allpars = self->IDL_Container::get( /all, count=npars)
	for i = 0, npars-1 do begin
		allpars[i]->write, unit, valuesonly=vo
	end
	printf, unit, "</Parameter_set>"

		; we're done

	self.filename = file
	free_lun, unit
end

;==============================================================================
;
;                                  XML READER
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrparxml::init
; Desc  : Initialized the parser - need the parameter set and the 'doubles'
;         keyword as arguments.
;=-----------------------------------------------------------------------------
function mrparxml::init, parset, doubles
	self.parset  = parset
	self.doubles = doubles
	return, self->IDLffXMLSAX::init()
end

;+=----------------------------------------------------------------------------
; Name  : mrparxml::startElement
; Desc  : Handles element start: parses attributes and creates a new mrparameter
;         object.
;=-----------------------------------------------------------------------------
pro mrparxml::startElement, URI, local, strName, attr, value
	case local of
		'Parameter_set': self.currentelement = -1
		'Parameter': begin
			for i = 0, 2 do begin           ; XML is great - we know that the document is well-formed
				case attr[i] of
					'name': self.name = value[i]
					'type': self.type = value[i]
					'size': self.size = fix(value[i])
				end
			end
			self.par = obj_new('mrparameter', self.name, self.type, self.size)
			if( self.par eq obj_new() ) then message, 'Error creating parameter ' + self.name
			self.nvals = 0
			self.currentelement = -1
		end
		'l': begin
			self.data = ''
			self.currentelement = 0	; Low
		end
		'u': begin
			self.data = ''
			self.currentelement = 1	; Up
		end
		's': begin
			self.data = ''
			self.currentelement = 2	; Step size
		end
		'o': begin
			self.data = ''
			self.currentelement = 3	; Valid value ('ok')
		end
		'v': begin
			self.data = ''
			self.currentelement = 4	; Value
		end
		else: print, 'Unknown startElement ', local
	end
end

;+=----------------------------------------------------------------------------
; Name  : mrparxml::endElement
; Desc  : If </Parameter>: add the mrparameter object to the set, respecting
;         settings of 'doubles' keyword
;=-----------------------------------------------------------------------------
pro mrparxml::endElement, URI, local, strName
	case local of
		'Parameter_set': self.currentelement = -1
		'Parameter': begin
			if( self.parset->isContained(self.name, position=pos )) then begin
				case doubles of
				'error': begin
					obj_destroy, newpar
					message, 'Parameter ' + self.name + ' already exists in mrparameterset
				end
				'ignore': begin
					obj_destroy, self.par
				end
				'replace': begin
					oldpar = self.parset->IDL_Container::get( position=pos )
					self.parset->IDL_Container::remove, position=pos
					obj_destroy, oldpar
					self.parset->add, self.par, position=pos
				end
				end
			end else begin
				self.parset->add, self.par
			end
			self.par = obj_new()
			self.currentelement = -1
		end
		'l': begin
			self.par->setLimits,      low  = self.data
			self.currentelement = -1
		end
		'u': begin
			self.par->setLimits,      up   = self.data
			self.currentelement = -1
		end
		's': begin
			self.par->setLimits,      step = self.data
			self.currentelement = -1
		end
		'o': begin
			self.par->setValidValues, self.data
			self.currentelement = -1
		end
		'v': begin
			self.par->set, self.data, position=self.nvals
			self.nvals = self.nvals+1
			self.currentelement = -1
		end
	end
end

;+=----------------------------------------------------------------------------
; Name  : mrparxml::characters
; Desc  : Set limits, step size, value or valid value
;=-----------------------------------------------------------------------------
pro mrparxml::characters, data
	case self.currentelement of
		0: self.data = self.data+data
		1: self.data = self.data+data
		2: self.data = self.data+data
		3: self.data = self.data+data
		4: self.data = self.data+data
		else:
	end
end

;+=----------------------------------------------------------------------------
; Name  : mrparxml__define
; Desc  : create separate object - dunno if multiple inheritance works under
;         IDL, but that's cleaner anyway
;=-----------------------------------------------------------------------------
pro mrparxml__define
	obj = { MRPARXML, $
			inherits IDLffXMLSAX		,$
			parset:			 obj_new()  ,$ ; the parameter set for which we work
			doubles:				''  ,$ ; handling of doubles
			par:			 obj_new()	,$ ; current parameter
			data:                   ''  ,$ ; current data
			name:					''	,$ ; name of current parameter
			type:					''	,$ ; type of current parameter
			size:					 0	,$ ; size of current parameter
			nvals:					 0	,$ ; number of values read
			currentelement:			 0	 $ ; the current element type
		}
end




;==============================================================================
;
;                              PRIVATE CLASS METHODS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrparameterset__define
; Desc  : Private Procedure (constructor)
;=-----------------------------------------------------------------------------
pro mrparameterset__define
	obj = { MRPARAMETERSET, $
	        inherits IDL_CONTAINER		, $
	        name		:		   ''	, $ ; name of this parameter set
	        filename	:		   ''   , $ ; if loaded from file, filename is set here
			params		:	obj_new()	  $
		}
end
