;+=----------------------------------------------------------------------------
; Name   : mrRoiGroup
; RCS    : $Id: mrroigroup__define.pro,v 2.0 2004/08/05 18:13:30 olivert Exp olivert $
; Methods: <b>Constructor</b>
;          mrRoiGroup::init              : constructor
;          <br>
;          <b>Public variables (get/set)</b>
;          mrRoiGroup::id                : get the object id
;          mrRoiGroup::setId             : set the object id
;          <br>
;          <b>Copy methods</b>
;          mrRoiGroup::duplicate         : returns an exact copy of the group
;          <b>I/O methods</b>
;          mrRoiGroup::save              : saves the object to file
;          mrRoiGroup_openxml            : open xml file for writing
;          mrRoiGroup::write             : write one roigroup to xml file
;          mrRoiGroup_closexml           : close the xml file
;          mrRoiGroup_load               : parse an mrRoiGroups xml file
;          <br>
;          <b>XML Parser</b>
;          mrroixml__define             : constructor, part I
;          mrroixml::init               : constructor, part II
;          mrroixml::startElement       : only start element,
;          mrroixml::endElement         : end element,
;          mrroixml::characters         : and data values are parsed
;          <br>
;          <b>Private</b>
;          mrRoiGroup__define            : constructor 1st part
;          mrRoiGroup::cleanup           : destructor
; Desc  : Class to store mrRoi objects, subclass to IDL_Container.
;         The class provides loading and saving functionality for
;         roigroups - file format is XML. In addition, a mrRoiGroup object
;         has an 'id' which is stored with the roi group and might be used
;         as a user value.
;         There's two ways of storing/retrieving roigroups to/from file.
;         If just a single roi group is to be loaded or saved, use the
;         classes load and save methods.
;         If the file shall contain an arbitrary number of roigroups,
;         Open the file via the mrRoiGroup_openxml static method.
;         Via the write method, any number of roigroups may be written
;         to the file. Then close the file via mrRoiGroup__closexml.
;         When reading from a file which potentially contains more than
;         one roigroup, use the static mrRoiGroup_read method. The
;         method returns an IDL_Container holding all roigroups from the
;         file.
;         See also code examples.
; Code  :
;         ; rg1 and rg2 shall be two roigroups with arbitrary number of
;         ; rois.
;         ; To write just one roigroup to file, use the save method:
;         res = rg1->save('C:\roigroup1.xml')
;         ; To write both roigroups to the same file:
;         unit = mrRoiGroup_openxml, 'C:\roigroups.xml'
;         rg1->write, unit
;         rg2->write, unit
;         res  = mrRoiGroup_closexml, unit
;         ; To read this file:
;         roigroup_container = mrRoiGroup_read, 'C:\roigroups.xml'
; Author: (c) Oliver Thilmann, oliver.thilmann@web.de, Oct 2002
; Licns : This program is published under the GPL. See gpl.txt for details
;=-----------------------------------------------------------------------------


;==============================================================================
;
;                          CONSTRUCTORS AND DESTRUCTORS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrRoiGroup::init
; Syntax: x = obj_new('mrRoiGroup', id )
; Args  : <b>id</b>
;         Optional argument of type integer - sets a user chosen id.
; Desc  : The constructor.
;=-----------------------------------------------------------------------------
function mrRoiGroup::init, id
	id = n_elements(id) eq 0 ? 0 : id
	self.id = id
	return, 1
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiGroup::cleanup
; Syntax: obj_destroy, obj
; Desc  : Since mrRoiGroup is a subclass of IDL_Container, destroying a
;         roiGroup will automatically destroy all contained objects.
;=-----------------------------------------------------------------------------
pro mrRoiGroup::cleanup
	;
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiGroup::setId
; Syntax: obj->setId, id
; Args  : <b>id</b>
;         An integer id
; Desc  : Allows to set a user defined integer id
;=-----------------------------------------------------------------------------
pro mrRoiGroup::setId, id
	self.id = id
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiGroup::id
; Syntax: id = obj->id()
; Ret   : Returns the user id
;=-----------------------------------------------------------------------------
function mrRoiGroup::id
	return, self.id
end

;==============================================================================
;
;                                PUBLIC METHODS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrRoiGroup::duplicate
; Syntax: obj_copy = obj->duplicate()
; Ret   : Returns an exact copy of the roigroup
;=-----------------------------------------------------------------------------
function mrRoiGroup::duplicate
	new = obj_new('mrRoiGroup', self.id )
	if( new eq obj_new() ) then return, obj_new()
	rois = self->get( /all, count=nrois )
	for i = 0, nrois-1 do begin
		new->add, rois[i]->duplicate()
	end
	return, new
end



;+=----------------------------------------------------------------------------
; Name  : mrRoiGroup_openxml
; Syntax: res = mrRoiGroup_openxml, filename
; Args  : <b>filename</b>
;         The file to open
; Ret   : The unit number in case of success, 0 otherwise
; Desc  : Opens an mrRoiGroups XML file for writing
;=-----------------------------------------------------------------------------
function mrRoiGroup_openxml, filename
	on_error, 2
	if( n_elements(filename) eq 0 ) then message, 'Missing filename argument'
	openw, unit, filename, /get_lun, error=err
	if( err ne 0 ) then begin
		message, "% mrroigroup: error opening file " + !ERROR_STATE.MSG
		return, 0
	end
	printf, unit, '<?xml version="1.0"?>'
	printf, unit, '<MRroiGroups version="1.0">
	return, unit
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiGroup_closexml
; Syntax: mrRoiGroup_closexml, unit
; Args  : <b>unit</b>
;         The unit to close
; Desc  : Closes a file opened via mrRoiGroup_openxml
;=-----------------------------------------------------------------------------
pro mrRoiGroup_closexml, unit
	printf, unit, '</MRroiGroups>'
	free_lun, unit
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiGroup::write
; Syntax: obj->write, unit
; Args  : <b>unit</b>
;         The unit, returned from the mrRoiGroup_openxml call
; Desc  : Writes the roigroup to file
;=-----------------------------------------------------------------------------
pro mrRoiGroup::write, unit
	printf, unit, '<MRroiGroup id="' + strtrim(string(self.id),2)+ '">'
	rois = self->get( /all, count=count )
	for i = 0, count-1 do begin
		class = obj_class(rois[i])
		id    = rois[i]->id()
		printf, unit, '<Roi class="' +string(class)+ '" id="' +strtrim(string(id),2)+ '">'
		masksize = rois[i]->masksize()
		rois[i]->getProperty, data=points
		for j = 0, n_elements(points)-1, 3 do begin
			printf, unit, '<x>'+strtrim(string(points[j]),2)+'</x><y>'+strtrim(string(points[j+1]),2)+'</y>'
		end
		printf, unit, '<mx>'+strtrim( masksize[0],2 )+'</mx>'
		printf, unit, '<my>'+strtrim( masksize[1],2 )+'</my>'
		printf, unit, '</Roi>
	end
	printf, unit, '</MRroiGroup>'
end


;+=----------------------------------------------------------------------------
; Name  : mrRoiGroup_load
; Syntax: res = mrRoiGroup_load, filename
; Ret   : obj_new() in case of error, otherwise an IDL_Container object that
;         holds all roigroups
; Desc  : Instanciates an XML parser and parses the given filename. All
;         mrRoiGroup objects found in the file are stored in a container which
;         is returned to the caller.
; TODO  : catch error if parsing of file fails
;=-----------------------------------------------------------------------------
function mrRoiGroup_load, filename

	catch, error_status
	if( error_status ne 0 ) then begin
		catch, /cancel
		help, /last_message
		return, obj_new()
	end


	if( n_elements(filename) eq 0 ) then return, obj_new()
	if( filename eq '' ) then return, obj_new()
	finfo = file_info( filename )
	if( finfo.exists eq 0 ) then return, obj_new()

		; create mrparxml object

	container = obj_new('IDL_Container')
	xmlparser = obj_new('mrroixml', container )
	if( xmlparser eq obj_new() ) then return, obj_new()

		; parse the file

	xmlparser->parseFile, filename

		; if the container is empty, return obj_new() and
		; destroy the container object

	obj_destroy, xmlparser
	if( container->count() eq 0 ) then begin
		obj_destroy, container
		return, obj_new()
	end
	return, container
end


;+=----------------------------------------------------------------------------
; Name  : mrRoiGroup::save
; Syntax: res = obj->save( filename )
; Ret   : 1 in case of success, 0 for failure
; Desc  : Creates a new mrRoiGroups xml file and writes the mrRoiGroup object
;         to the file.
;=-----------------------------------------------------------------------------
function mrRoiGroup::save, filename
	unit = mrRoiGroup_openxml( filename )
	if( unit eq 0 ) then return, 0
	self->write, unit
	mrRoiGroup_closexml, unit
	return, 1
end

;==============================================================================
;
;                                  XML READER
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrroixml::init
; Desc  : Initialized the parser - need the parameter set and the 'doubles'
;         keyword as arguments.
;         XML file layout:
;         <MRroiGroups>                           main element
;         <MRroiGroup>                            one roigroup consists of...
;         <Roi class="MRROIDOT" id="1">           ... one or more Rois
;         <x>1</x><y>1</y>                        corner coordinates
;         <x>....
;         <mx>128</mx><my>128</my>                the mask size
;         </Roi>
;         <Roi ...></Roi>                         any number of rois per roigroup
;         </MRroiGroup>
;         <MRroiGroup>....</MRroiGroup>           any number of roigroups per file
;         </MRroiGroups>                          closing tag
;=-----------------------------------------------------------------------------
function mrroixml::init, container
	self.roigroups = container
	return, self->IDLffXMLSAX::init()
end

;+=----------------------------------------------------------------------------
; Name  : mrroixml::startElement
;=-----------------------------------------------------------------------------
pro mrroixml::startElement, URI, local, strName, attr, value
	case local of
		'MRroiGroups': begin
			self.ce = 0
		end
		'MRroiGroup': begin
			self.ce = 1
			id = value[0]
			self.roigroup = obj_new('mrroigroup', id )
		end
		'Roi': begin
			self.ce = 2
			self.n = 0
			for i = 0, 1 do begin
				case attr[i] of
					'class': self.roiclass = value[i]
					'id':    self.roiid    = value[i]
				end
			end
		end
		'mx': self.ce = 3
		'my': self.ce = 4
		'x' : self.ce = 5
		'y' : self.ce = 6
	end
end

;+=----------------------------------------------------------------------------
; Name  : mrroixml::endElement
;=-----------------------------------------------------------------------------
pro mrroixml::endElement, URI, local, strName
	case local of
		'MRroiGroups':
		'MRroiGroup': begin
			self.roigroups->add, self.roigroup
			self.ce = -1
		end
		'Roi': begin
			self.roi->setId, self.roiid
			self.roigroup->add, self.roi
			self.ce = -1
		end
		'mx': self.ce = -1
		'my': self.ce = -1
		'x' : self.ce = -1
		'y' : self.ce = -1
	end
end

;+=----------------------------------------------------------------------------
; Name  : mrroixml::characters
; Desc  : Collect data
;=-----------------------------------------------------------------------------
pro mrroixml::characters, data
	case self.ce of
	3: self.mx = data
	4: self.roi->setMasksize, [self.mx, data]
	5: self.x = double( data )
	6: begin
		self.y = double( data )
		if( self.n eq 0 ) then begin
			case self.roiclass of
				'MRROIDOT' : self.roi = obj_new('MRroiDot',  self.x, self.y)
				'MRROIRECT': self.roi = obj_new('MRroiRect', self.x, self.y)
				'MRROICIRC': self.roi = obj_new('MRroiCirc', self.x, self.y)
				'MRROIPOLY': self.roi = obj_new('MRroiPoly', self.x, self.y)
			end
			self.n = self.n+1
		end else begin
			self.n = self.n+1
			self.roi->appendData, self.x, self.y
		end
	end
	else:
	end
end

;+=----------------------------------------------------------------------------
; Name  : mrroixml__define
; Desc  : The XML parser
;=-----------------------------------------------------------------------------
pro mrroixml__define
	obj = { MRROIXML, $
			inherits IDLffXMLSAX		,$
			roigroups:		 obj_new()  ,$ ; the container to store the groups in
			roigroup:		 obj_new()	,$ ; current roigroup
			roi:             obj_new()  ,$ ; the current roi
			roiclass:				''	,$ ; class of the current roi
			roiid:                   0  ,$ ; id of current roi
			mx:                      0  ,$ ; temporary storage for mask size in x
			x:		        		 0. ,$ ;
			y:              		 0. ,$ ;
			n:                       0  ,$ ; number of points
			ce:                      0	 $ ; the current element (some integer id)
		}
end




;+=----------------------------------------------------------------------------
; Name  : mrRoiGroup__define
; Desc  : The class definition
;=-----------------------------------------------------------------------------
pro mrroigroup__define
	obj = { MRROIGROUP, $
			inherits IDL_Container   ,$ ; enough?
			id:                   0   $ ; an id
		}
end


