;+=----------------------------------------------------------------------------
; Name  : mrRoiUI
; RCS   : $Id: mrroiui__define.pro,v 2.4 2004/11/26 10:30:14 olivert Exp olivert $
; Methods:
;         <b>Constructor</b>
;         mrRoiUI::init                 : Constructor, the main part
;         <br>
;         <b>Public variables (get)</b>
;         mrRoiUI::roiGroup             : return the current mrroigroup
;         <br>
;         <b>Public methods</b>
;         mrRoiUI::setRoiGroup          : set mrroigroup for this object
;         mrRoiUI::addRoi               : add a roi to the roigroup by program
;         mrRoiUI::deleteRoi            : delete a roi from the roigroup
;         mrRoiUI::showRois             : display the rois in the mrdisplay object
;         mrRoiUI::hideRois             : do not display the rois
;         mrRoiUI::calcAreaImg          : recalculate the rois - necessary after mrRoiUI::setRoi
;         mrRoiUI::palette              : get the used palette, meant for read-only!
;         mrRoiUI::roiColor             : returns rgb color-vector of a certain roi
;         mrRoiUI::print                : print some info on the current rois
;         <br>
;         <b>Notification</b>
;         mrRoiUI::notify               : see mrNotify::notify
;         mrRoiUI::enableNotify         : see mrNotify::enableNotify
;         mrRoiUI::disableNotify        : see mrNotify::disableNotify
;         <br>
;         <b>Communication with mrDisplay, user part</b>
;         mrRoiUI::selfregister         : automatically register with a mrdisplay object
;         mrRoiUI::unregister           : and unregister again (see the tips page!)
;         mrRoiUI::activate             : activate the roi button window
;         mrRoiUI::deactivate           : or deactivate the buttons
;         mrRoiUI::setLoadMethod        : link a user defined load method to the load button
;         mrRoiUI::setSaveMethod        : link a user defined save method to the save button
;         <br>
;         <b>Communication with mrDisplay, internal part</b>
;         mrRoiUI::mrdisplayEvtHandler  : callback method for the display object
;         mrRoiUI::buttonMotionEvt      : handles motion events
;         mrRoiUI::buttonPressEvt       : handles button press events
;         mrRoiUI::buttonReleaseEvt     : handles button release events
;         <br>
;         <b>Event handling for the button window</b>
;         mrRoiUI::updatewidgets        : update button widget (highlighting)
;         mrRoiUI_evthandler            : the event handler for the button window
;         mrRoiUI::evthandler           : and the object version of the handler
;         <br>
;         <b>Private methods</b>
;         mrRoiUI::load                 : load a roigroup from file
;         mrRoiUI::save                 : save a roigroup
;         mrRoiUI::updateAreaImg        : update only the active roi's area
;         mrRoiUI::resetRoiColor        : i.e. make a roi non-active
;         mrRoiUI::findClosestRoi       : find the closest roi to a certain point
;         mrRoiUI::findTxtObj           : find a txt object with specified uvalue
;         mrRoiUI::findFreeRoiNumber    : find a roinumber not in use
;         mrRoiUI::definePalette        : define the palette colors
;         mrRoiUI__define               : Constructor part I
;         mrRoiUI::cleanup              : Destructor
;         <br>
; Desc  : This object provides complete ROI definition and evaluation
;         functionality. See below for some examples on how to use.
;         <b>Events</b>
;         Objects may be notified upon modification of existing ROIs or change
;         of the roi set (i.e. ROIs are added or removed). mrRoiUI knows two
;         event types: 'modify' and 'change'. 'modify' notifications are sent
;         continuously, <i>while</i> a new ROI is created or an existing ROI
;         is modified. 'change' notifications are only sent once <i>after</i>
;         creation or modification of an ROI (i.e. upon mouse button release).
;         The event structure passed during call to the objects notification
;         handler is of the following form:
;         >	x = { mrroiuievt                ,$
;         >       type    :    0            ,$ ;
;         >       top     :    0L           ,$ ; the top widget id (see below)
;         >       roi     :    obj_new()    ,$ ; the currently active roi
;         > }
;         where type equals '1' for modify and '2' for change events. The currently active
;         roi is NULL (resp. obj_new()) for change events.
;         <br>
;         To provide a value for the top widget is not straight forward as mrRoiUI self
;         receives IDL events only indirectly via callback from mrDisplay. Nevertheless
;         it is needed as the structure holding global variables might be attached
;         to the uvalue of the top-widget. Usually this should be ok, as events are
;         only created after selfregistration and activation of mrdisplay, but it might
;         nevertheless be a good idea to check for a value of 0L.
;         <b>Loading and saving of ROIs</b>
;         Loading and saving is now done via the (new) mrRoiGroup class and its
;         load/save methods which read/write xml and will store one single
;         roigroup to file, resp. read one single roigroup from file.
;         This might be inappropriate in cases where e.g. rois may be defined
;         in several slices and should be stored/retrieved together.
;         Use the load/save hook in these cases by linking user defined methods
;         via setLoadMethod, setSaveMethod, thus overriding the built-in
;         functionality. mrRoiGroup contains some example code for that case.
; Author: (c) Oliver Thilmann, oliver.thilmann@web.de
; Licns : This program is published under the GPL. See gpl.txt for details
;=-----------------------------------------------------------------------------


;==============================================================================
;
;                          CONSTRUCTORS AND DESTRUCTORS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::init
; Syntax: obj = obj_new('mrroiui', parent, row=row, column=column, /no_iobuttons,
;         spacing=spacing, path=path )
; Args  : <b>parent</b>
;         Widget id of the parent widget in which to place the mrroiui buttons. If parent
;         is not given, a separate window holding the buttons will be created.
;         <b>row</b>
;         may equal 1, 2 or 4 and defines the number of button rows to use
;         <b>column</b>
;         as row, defining the number of columns. Only one of the row and column
;         keywords may be used (the use of column overrides any row setting).
;         row=4 and column=4 yield the same layout.
;         <b>no_iobuttons</b>
;         If set, do not display the load and save button.
;         <b>spacing</b>
;         Number of pixels between the button groups. Default is 5.
;         (only for row=1 or column=1)
;         <b>path</b>
;         Initial path - the object will remember the last used path at subsequent
;         calls to load and save.
;         <b>extra</b>
;         Any additional keywords are passed on to the base widget which contains
;         the ROI buttons.
; Desc  : Instanciates an mrroiui object, creates the button window and registers with
;         xmanager.
;=-----------------------------------------------------------------------------
function mrRoiUI::init, parent, no_iobuttons=no_iobuttons, path=path, row=row, column=column,	spacing=spacing, _ref_extra = ex

	on_error, 2

	parent  = ( n_elements(parent)  eq 0 ) ?    0L : parent
	path    = ( n_elements(path)    eq 0 ) ? 'C:\' : path
	spacing = ( n_elements(spacing) eq 0 ) ?     5 : spacing
	no_io   = keyword_set( no_iobuttons )

		; initialize notification

	if( self->mrNotify::init(['modify','change']) eq 0 ) then return, 0

		; map to old keyword

	ui = ''
	if( keyword_set( row ) ) then begin
		case row of
			1: ui = 'h'
			2: ui = 'h2'
			4: ui = 'h4'
			else: message, 'Invalid row specification'
		end
	end
	if( keyword_set( column ) ) then begin
		case column of
			1: ui = 'v'
			2: ui = 'v2'
			4: ui = 'h4'
			else: message, 'Invalid column specification'
		end
	end
	if( ui eq '' ) then ui = 'h' ; default

		; if no baseid given, we create our own base window

	if( parent eq 0L ) then	begin
		parent = widget_base(title='ROI menu', /column, _extra = ex )
	end
	case ui of
		'h' : baseid = widget_base( parent, /row,  scr_ysize=30, _extra=ex )
		'h2': begin
			baseid = widget_base( parent, /column, ysize=58, _extra=ex )
			b1  = widget_base( baseid, /row, ysize=27)
			b2  = widget_base( baseid, /row, ysize=27)
		end
		'h4': begin
			baseid = widget_base( parent, /column, ysize=120, _extra=ex )
			b1     = widget_base( baseid, /row, ysize=27 )
			b2     = widget_base( baseid, /row, ysize=27 )
			b3     = widget_base( baseid, /row, ysize=27 )
			b4     = widget_base( baseid, /row, ysize=27 )
		end
		'v' : baseid = widget_base( parent, /column, xsize=30, ysize=512, _extra=ex )
		'v2': begin
			baseid = widget_base( parent, /row, xsize=68, _extra=ex)
			b1     = widget_base( baseid, /column, xsize=30 )
			b2     = widget_base( baseid, /column, xsize=30 )
		end
		else: message, 'Invalid ui layout specification'
	end

	self.Ibase  = baseid
	self.Itop   = 0L
	self.ui     = ui
	self.Iload  = -1

		; NEW UI

	if( ui eq 'h' or ui eq 'v' ) then begin
		self.Iroi    = widget_button( baseid, value='mrroi.bmp',       uvalue=self, /bitmap, sensitive=1 )
		x            = widget_label ( baseid, value='',                xsize=spacing, ysize=spacing )
		if( no_io eq 0 ) then begin
		self.Iload   = widget_button( baseid, value='open.bmp',        uvalue=self, /bitmap, sensitive=0 )
		self.Isave   = widget_button( baseid, value='save.bmp',        uvalue=self, /bitmap, sensitive=0 )
		x            = widget_label ( baseid, value='',                xsize=spacing, ysize=spacing )
		end
		self.Idot    = widget_button( baseid, value='mrroidot.bmp',    uvalue=self, /bitmap, sensitive=0 )
		self.Icirc   = widget_button( baseid, value='ellipse.bmp',     uvalue=self, /bitmap, sensitive=0 )
		self.Irect   = widget_button( baseid, value='rectangl.bmp',    uvalue=self, /bitmap, sensitive=0 )
		self.Ipoly   = widget_button( baseid, value='mrroifree.bmp',   uvalue=self, /bitmap, sensitive=0 )
		x            = widget_label ( baseid, value='',                xsize=spacing, ysize=spacing )
		self.Iundo   = widget_button( baseid, value='undo.bmp',        uvalue=self, /bitmap, sensitive=0 )
		self.Imove   = widget_button( baseid, value='pan.bmp',         uvalue=self, /bitmap, sensitive=0 )
		self.Iresize = widget_button( baseid, value='mrroizoom.bmp',   uvalue=self, /bitmap, sensitive=0 )
		self.Irotate = widget_button( baseid, value='rotate.bmp',      uvalue=self, /bitmap, sensitive=0 )
		self.Idel    = widget_button( baseid, value='mrroidel.bmp',    uvalue=self, /bitmap, sensitive=0 )
		x            = widget_label ( baseid, value='',                xsize=spacing, ysize=spacing )
		self.Ifilled = widget_button( baseid, value='mrroifilled.bmp', uvalue=self, /bitmap, sensitive=0 )
		self.Isolid  = widget_button( baseid, value='mrroitrans.bmp',  uvalue=self, /bitmap, sensitive=0 )
		x            = widget_label ( baseid, value='',                xsize=spacing, ysize=spacing )
		self.Iinfo   = widget_button( baseid, value='mrroiinfo.bmp',   uvalue=self, /bitmap, sensitive=0 )
	end
	if( ui eq 'h2' or ui eq 'v2' ) then begin
		self.Iroi    = widget_button( b1, value='mrroi.bmp',       uvalue=self, /bitmap, sensitive=1 )
		x            = widget_label ( b2, value='',                xsize=16, ysize=16 )
		x            = widget_label ( b2, value='',                xsize=4, ysize=4 )
		x            = widget_label ( b1, value='',                xsize=spacing, ysize=spacing )
		x            = widget_label ( b2, value='',                xsize=spacing, ysize=spacing )
		if( no_io eq 0 ) then begin
		self.Iload   = widget_button( b1, value='open.bmp',        uvalue=self, /bitmap, sensitive=0 )
		self.Isave   = widget_button( b2, value='save.bmp',        uvalue=self, /bitmap, sensitive=0 )
		x            = widget_label ( b1, value='',                xsize=spacing, ysize=spacing )
		x            = widget_label ( b2, value='',                xsize=spacing, ysize=spacing )
		end
		self.Idot    = widget_button( b1, value='mrroidot.bmp',    uvalue=self, /bitmap, sensitive=0 )
		self.Icirc   = widget_button( b2, value='ellipse.bmp',     uvalue=self, /bitmap, sensitive=0 )
		self.Irect   = widget_button( b1, value='rectangl.bmp',    uvalue=self, /bitmap, sensitive=0 )
		self.Ipoly   = widget_button( b2, value='mrroifree.bmp',   uvalue=self, /bitmap, sensitive=0 )
		x            = widget_label ( b1, value='',                xsize=spacing, ysize=spacing )
		x            = widget_label ( b2, value='',                xsize=spacing, ysize=spacing )
		self.Iundo   = widget_button( b1, value='undo.bmp',        uvalue=self, /bitmap, sensitive=0 )
		self.Imove   = widget_button( b2, value='pan.bmp',         uvalue=self, /bitmap, sensitive=0 )
		self.Iresize = widget_button( b1, value='mrroizoom.bmp',   uvalue=self, /bitmap, sensitive=0 )
		self.Irotate = widget_button( b2, value='rotate.bmp',      uvalue=self, /bitmap, sensitive=0 )
		self.Idel    = widget_button( b1, value='mrroidel.bmp',    uvalue=self, /bitmap, sensitive=0 )
		self.Iinfo   = widget_button( b2, value='mrroiinfo.bmp',   uvalue=self, /bitmap, sensitive=0 )
		x            = widget_label ( b1, value='',                xsize=spacing, ysize=spacing )
		x            = widget_label ( b2, value='',                xsize=spacing, ysize=spacing )
		self.Ifilled = widget_button( b1, value='mrroifilled.bmp', uvalue=self, /bitmap, sensitive=0 )
		self.Isolid  = widget_button( b2, value='mrroitrans.bmp',  uvalue=self, /bitmap, sensitive=0 )
	end
	if( ui eq 'h4' ) then begin
		self.Iroi    = widget_button( b1, value='mrroi.bmp',       uvalue=self, /bitmap, sensitive=1 )
		x            = widget_label ( b1, value='',                xsize=spacing, ysize=spacing )
		self.Idot    = widget_button( b1, value='mrroidot.bmp',    uvalue=self, /bitmap, sensitive=0 )
		self.Icirc   = widget_button( b1, value='ellipse.bmp',     uvalue=self, /bitmap, sensitive=0 )
		x            = widget_label ( b1, value='',                xsize=spacing, ysize=spacing )
		self.Ifilled = widget_button( b1, value='mrroifilled.bmp', uvalue=self, /bitmap, sensitive=0 )

		self.Iundo   = widget_button( b2, value='undo.bmp',        uvalue=self, /bitmap, sensitive=0 )
		x            = widget_label ( b2, value='',                xsize=spacing, ysize=spacing )
		self.Irect   = widget_button( b2, value='rectangl.bmp',    uvalue=self, /bitmap, sensitive=0 )
		self.Ipoly   = widget_button( b2, value='mrroifree.bmp',   uvalue=self, /bitmap, sensitive=0 )
		x            = widget_label ( b2, value='',                xsize=spacing, ysize=spacing )
		self.Isolid  = widget_button( b2, value='mrroitrans.bmp',  uvalue=self, /bitmap, sensitive=0 )

		if( no_io eq 0 ) then begin
			self.Iload   = widget_button( b3, value='open.bmp',    uvalue=self, /bitmap, sensitive=0 )
			self.Isave   = widget_button( b4, value='save.bmp',    uvalue=self, /bitmap, sensitive=0 )
		end else begin
			x            = widget_label ( b3, value='',            xsize=16, ysize=16 )
			x            = widget_label ( b3, value='',            xsize=spacing, ysize=spacing )
			x            = widget_label ( b4, value='',            xsize=16, ysize=16 )
			x            = widget_label ( b4, value='',            xsize=spacing, ysize=spacing )
		end


		x            = widget_label ( b3, value='',                xsize=spacing, ysize=spacing )
		self.Imove   = widget_button( b3, value='pan.bmp',         uvalue=self, /bitmap, sensitive=0 )
		self.Iresize = widget_button( b3, value='mrroizoom.bmp',   uvalue=self, /bitmap, sensitive=0 )

		x            = widget_label ( b4, value='',                xsize=spacing, ysize=spacing )
		self.Irotate = widget_button( b4, value='rotate.bmp',      uvalue=self, /bitmap, sensitive=0 )
		self.Idel    = widget_button( b4, value='mrroidel.bmp',    uvalue=self, /bitmap, sensitive=0 )
		x            = widget_label ( b4, value='',                xsize=spacing, ysize=spacing )
		self.Iinfo   = widget_button( b4, value='mrroiinfo.bmp',   uvalue=self, /bitmap, sensitive=0 )

	end

		; if we opened our own window, we also start our own xmanager
		; otherwise this is taken care of by the user's xmanager

	if( baseid ne -1 ) then begin
		xmanager, 'mrroiui', baseid, event_handler='mrroiui_evthandler', /just_reg
	end

		; finally initialize the roi stuff - most of the initialization is done
		; when registering with the display

	self.mode      = -self.Idot           ; negative means we're not active
	self.filled    = 1
	self.solid     = 0
	self->definePalette
	self.path      = path

		; declaration of the event structure

	x = { mrroiuievt				,$
		  type	:		0			,$ ; 1 for modify and 2 for change events
		  top   :		0L			,$ ; top widget id
		  roi   :       obj_new()    $ ; currently active roi
		}

	return, 1
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::cleanup
; Syntax: obj_destroy, obj
; Desc  : There's still (minor) memory leaks
;=-----------------------------------------------------------------------------
pro mrRoiUI::cleanup
	if( obj_valid(self.nrmodel) ) then  self.nrmodel->remove, /all
	ptr_free, self.evtstruct
	self->mrNotify::cleanup
end

;==============================================================================
;
;                         EVENT HANDLING USER INTERFACE
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrRoiUI_evthandler
; Syntax: mrRoiUI_evthandler, evt
; Calls : mrRoiUI::evthandler
; Args  : <b>evt</b>
;         A button evt structure
; Desc  : Get the mrRoiUI object from the buttons uvalue and call this
;         mrRoiUI object's event handler
;=-----------------------------------------------------------------------------
pro mrRoiUI_evthandler, evt
	widget_control, evt.id, get_uvalue=roiobj
	roiobj->evthandler, evt.id
end


;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::evthandler
; Syntax: obj->evthandler, button_id
; Args  : <b>button_id</b>
;         The widget id of the pressed button (as evaluated by mrRoiUI_evthandler)
; Desc  : The object's event handler, handling the roi button events. Mouse
;         events from the canvas are handled via mrdisplay call back.
;         There's still a little bit of functionality in here, but it's
;         almost all gone now.
;=-----------------------------------------------------------------------------
pro mrRoiUI::evthandler, buttonid

	if( self.display eq obj_new() ) then return      ; not registered

		; In principle this should not happen, but if the user managed to
		; press a button while a roi was active, either delete it (if it
		; was about to be defined) or render it inactive again.

	if( self.activeroi ne obj_new() ) then begin
		if( self.mode eq self.Idot or self.mode eq self.Icirc or self.mode eq self.Irect or self.mode eq self.Ipoly ) then begin
			self->deleteRoi, self.activeRoi
			self->calcAreaImg
		end else begin
			self->resetRoiColor, self.activeRoi
		end
		self.activeRoi = obj_new()
		self.display->display
	end

		; now for the single buttons:

	case buttonid of
		self.IRoi: begin
			if( self.mode ge 0 ) then begin
				self->deactivate
				self->hideRois
				self.display->display
			end else begin
				self->activate
				self->showRois
				self.display->display
			end
		end
		self.Iload: begin
			res = self->load(/destroy)
			return
		end
		self.Isave: begin
			res = self->save()
		end
		self.Idot: begin
			self.mode = self.Idot
		end
		self.Icirc: begin
			self.mode = self.Icirc
		end
		self.Irect: begin
			self.mode = self.Irect
		end
		self.Ipoly: begin
			self.mode = self.Ipoly
		end
		self.Iundo: begin
			if( self.lastDefined eq obj_new() ) then begin
				x = dialog_message("Don't know how to undo - sorry")
			end else begin
				self->deleteRoi, self.lastDefined
				self.lastDefined = obj_new()
			end
			widget_control, self.Iundo, sensitive=0
			self->calcAreaImg
			self.display->display
		end
		self.Imove: begin
			self.mode = self.Imove
			self.lastDefined = obj_new()
		end
		self.Iresize: begin
			self.mode = self.Iresize
			self.lastDefined = obj_new()
		end
		self.Irotate: begin
			self.mode = self.Irotate
			self.lastDefined = obj_new()
		end
		self.Idel: begin
			if( self.mode eq self.Idel ) then begin  ; twice click = delete all
				if( self.roigroup->count() eq 0 ) then begin
					x = dialog_message('No rois defined')
				end else begin
					x = dialog_message('Are you sure you want to delete all displayed rois?', /question)
					if( x eq 'Yes' ) then begin
						self->deleteRoi, /all
						self.lastDefined = obj_new()
						self->calcAreaImg
						self.mode = self.Idot
						self.display->display
					end
				end
			end else begin
				self.lastDefined = obj_new()
				self.mode = self.Idel
			end
		end
		self.Ifilled: begin
			self.filled = 1-self.filled
			self->calcAreaImg, /no_notify             ; no need for notifications here
			rois = self.roigroup->get( /all, count=nof )
			if( nof gt 0 ) then begin
				for i = 0, nof-1 do self->resetRoiColor, rois[i]
			end
			self.display->display
		end
		self.Isolid: begin
			self.solid = 1-self.solid
			if( self.solid eq 0 ) then begin
				self.areaimg->setProperty, blend_function=[1,1]
			end else begin
				self.areaimg->setProperty, blend_function=[1,4]
			end
			self.display->display
		end
		self.Iinfo: begin
			self.mode = self.Iinfo
		end
	end
	self->updatewidgets
end


;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::updatewidgets
; Syntax: obj->updatewidgets
; Desc  : Set sensitivity and bitmaps of all widgets according to the
;         internal status, described by self.mode, self.filled and self.solid.
;=-----------------------------------------------------------------------------
pro mrRoiUI::updatewidgets
	if( self.mode lt 0 ) then begin
		s = 0
		mode = -self.mode
		widget_control, self.Iroi, set_value='mrroi.bmp', /bitmap
	end else begin
		widget_control, self.Iroi, set_value='mrroi_active.bmp', /bitmap
		s = 1
		mode = self.mode
	end
	if( self.Iload gt 0 ) then begin
		widget_control, self.Iload, sensitive=s
		widget_control, self.Isave, sensitive=s
	end
	if( mode eq self.Idot ) then $
		widget_control, self.Idot, set_value='mrroidot_active.bmp', /bitmap, sensitive=s $
	else $
		widget_control, self.Idot, set_value='mrroidot.bmp', /bitmap, sensitive=s
	if( mode eq self.Icirc ) then $
		widget_control, self.Icirc, set_value='ellipse_active.bmp', /bitmap, sensitive=s $
	else $
		widget_control, self.Icirc, set_value='ellipse.bmp', /bitmap, sensitive=s
	if( mode eq self.Irect ) then $
		widget_control, self.Irect, set_value='rectangl_active.bmp', /bitmap, sensitive=s $
	else $
		widget_control, self.Irect, set_value='rectangl.bmp', /bitmap, sensitive=s
	if( mode eq self.Ipoly ) then $
		widget_control, self.Ipoly, set_value='mrroifree_active.bmp', /bitmap, sensitive=s $
	else $
		widget_control, self.Ipoly, set_value='mrroifree.bmp', /bitmap, sensitive=s
	if( self.lastDefined eq obj_new() ) then $
		widget_control, self.Iundo, sensitive=0 $
	else $
		widget_control, self.Iundo, sensitive=1*s
	if( mode eq self.Imove ) then $
		widget_control, self.Imove, set_value='pan_active.bmp', /bitmap, sensitive=s $
	else $
		widget_control, self.Imove, set_value='pan.bmp', /bitmap, sensitive=s
	if( mode eq self.Iresize ) then $
		widget_control, self.Iresize, set_value='mrroizoom_active.bmp', /bitmap, sensitive=s $
	else $
		widget_control, self.Iresize, set_value='mrroizoom.bmp', /bitmap, sensitive=s
	if( mode eq self.Irotate ) then $
		widget_control, self.Irotate, set_value='rotate_active.bmp', /bitmap, sensitive=s $
	else $
		widget_control, self.Irotate, set_value='rotate.bmp', /bitmap, sensitive=s
	if( mode eq self.Idel ) then $
		widget_control, self.Idel, set_value='mrroidel_active.bmp', /bitmap, sensitive=s $
	else $
		widget_control, self.Idel, set_value='mrroidel.bmp', /bitmap, sensitive=s
	if( mode eq self.Iinfo ) then $
		widget_control, self.Iinfo, set_value='mrroiinfo_active.bmp', /bitmap, sensitive=s $
	else $
		widget_control, self.Iinfo, set_value='mrroiinfo.bmp', /bitmap, sensitive=s
	if( self.filled eq 1 ) then $
		widget_control, self.Ifilled, set_value='mrroifilled.bmp', /bitmap, sensitive=s $
	else $
		widget_control, self.Ifilled, set_value='mrroioutline.bmp', /bitmap, sensitive=s
	if( self.solid eq 1 ) then $
		widget_control, self.Isolid, set_value='mrroisolid.bmp', /bitmap, sensitive=s $
	else $
		widget_control, self.Isolid, set_value='mrroitrans.bmp', /bitmap, sensitive=s
	widget_control, self.Iinfo, sensitive=s
end



;==============================================================================
;
;                               GET AND SET METHODS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::roiGroup
; Syntax: roigroup = obj->roiGroup()
; Ret   : returns the mrroigroup object currently displayed in mrRoiUI. The
;         obtained roigroup object should not be changed while being displayed,
;         or unexpected side-effects may occur.
;=-----------------------------------------------------------------------------
function mrRoiUI::roiGroup
	return, self.roiGroup
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::setRoiGroup
; Syntax: obj->setRoiGroup, roigroup, /destroy_current
; Calls : mrRoiUI::calcAreaImg
; Args  : <b>roigroup</b>
;         An object of class mrRoiGroup
;         <b>destroy_current</b>
;         If set, the currently used roigroup is destroyed (and with it all
;         rois contained therein). Default is '0'.
; Desc  : Replaces the current roigroup by a new one.
;         If argument roigroup is not provided, a new and empty roigroup object
;         is created.
;         The method calls mrRoiUI::calcAreaImg and thus implicitly creates
;         roichange events, if notification is switched on.
; Code  : The method call
;         <i> obj->setRoiGroup, /destroy_current </i>
;         is similar to
;         <i> obj->deleteRoi, /all </i>
;         except that in the former case, a new object of class mrRoiGroup is
;         created, while the latter would conserve the old one (thus preserving
;         possible references to the existing object).
;=-----------------------------------------------------------------------------
pro mrRoiUI::setRoiGroup, roigroup, destroy_current=destroy

	if( self.display eq obj_new() ) then return

		; parameter checking

	destroy = keyword_set( destroy )
	if( n_elements(roigroup) eq 0 ) then roigroup = obj_new()

		; clean up the old stuff

	self.nrmodel->remove, /all              ; remove all old numbers from model
	if( destroy eq 1 ) then obj_destroy, self.roigroup

	if( obj_valid(roigroup) ) then begin

			; prepare rois for display

		xsize = self.display->imgXsize()
		ysize = self.display->imgYsize()
		xs    = 1./xsize
		ys    = 1./ysize
		rois = roigroup->get( /all, count=nrois )
		for i = 0, nrois-1 do begin
			rois[i]->setMaskSize, [xsize,ysize]
			rois[i]->setProperty, xcoord_conv=[0.,xs], ycoord_conv=[0.,ys]
			x = rois[i]->computeGeometry( centroid=textloc )
			number = rois[i]->id()
			numbertxt = string( number, format='(i2)')
			txt = obj_new('IDLgrText', Palette=self.palette, $
					strings=numbertxt, locations=textloc, recompute_dimensions=2, $
					xcoord_conv=[0, xs], ycoord_conv=[0,ys], uvalue=number )
			self.nrModel->add, txt
			self->resetRoiColor, rois[i]   ; set color of roi and text
		end
		self.roigroup = roigroup
	end else begin
		self.roigroup = obj_new('mrroigroup')
	end

		; recalculate the area - this will cause events

	self.activeroi   = obj_new()
	self.lastdefined = obj_new()
	self->calcAreaImg
end



;==============================================================================
;
;                         EVENT HANDLING - EXTERNAL
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::selfregister
; Syntax: obj->selfregister, display, button=button
; Args  : <b>display</b>
;         The mrdisplay object to register with. The object must have been
;         realized on screen and an image must be loaded at the time of
;         registration.
;         <b>button</b>
;         Mouse button to use. 1 = left (default), 2 = middle, 3 = right.
; Desc  : Self-registration must only be done <b>after</b> an image has been
;         loaded in mrdisplay because some internal variables of mrRoiUI depend
;         on the image size.
;         You also have to unregister and reregister, if the image size is
;         changed in mrdisplay, that is before and after the call to
;         mrdisplay::setImg. mrRoiUI catches image set events from the display:
;         If the image size changes unexpectedly, mrRoiUI creates an error
;         window and unregisters itself.
;         <b>Important:</b> registering basically means a 'reset' of the mrRoiUI
;         object - that means among other things that a new roigroup is created
;         with each call to 'register'.
;         To avoid memory leaks, 'old' roigroups need to be manually destroyed
;         after unregistering.
;         Hah! And who would understand the setting of the blendfunction? (Evil
;         grin) It's either [1,1] or [1,4].<br>
;=-----------------------------------------------------------------------------
pro mrRoiUI::selfregister, display, button=button
	if( n_elements(button) eq 0 ) then button=1
	if( button lt 1 ) then begin
		message, 'Illegal button number in mrRoiUI::selfregister'
		return
	end
	if( display->imgXsize() eq 0 ) then message, 'Unknown image size in display. Cannot register'
	self.display   = display
	self.button    = 2^(button-1)
	self.imgview   = display->imgView()
	if( obj_valid( self.nrmodel   ) ) then begin
		self.nrmodel->remove, /all
		obj_destroy, self.nrmodel
	end
	if( obj_valid( self.areamodel ) ) then obj_destroy, self.areamodel
	if( obj_valid( self.areaimg   ) ) then obj_destroy, self.areaimg
	self.nrmodel   = obj_new('IDLgrModel')                      ; for the numbers
	self.areamodel = obj_new('IdlgrModel')                      ; for the roi area
	self.areaimg   = obj_new('IDLgrImage', $
						sub_rect = [0,0,display->imgXsize(), display->imgYsize()], $
						xcoord_conv = [0.,1.], ycoord_conv = [0.,1.], $
						dimensions  = [1.,1.], blend_function=[1,1+3*self.solid])
	self.areamodel->add, self.areaimg
	self.imgview->add, self.areamodel                ; here we display the roi regions
	self.imgview->add, self.nrmodel                  ; and here we display the roi numbers
	self.roigroup    = obj_new('mrroigroup')
	self.lastDefined = obj_new()
	self.activeRoi   = obj_new()
	data             = intarr( 4, self.display->imgXsize(), self.display->imgYsize() )
	self.areadata    = ptr_new( data, /no_copy )
	data             = intarr( 4, self.display->imgXsize(), self.display->imgYsize() )
	self.areadataa   = ptr_new( data, /no_copy )
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::unregister
; Syntax: obj->unregister
; Desc  : Unregisters with a mrdisplay object. It is a good idea to unregister
;         before destroying the mrdisplay object or setting an image with a
;         different size (s.a.)
;         Unregistering automatically deactivates the roi buttons.
;         Unregistering will <b>not</b> destroy the roigroup.
;=-----------------------------------------------------------------------------
pro mrRoiUI::unregister
	if( self.display eq obj_new() ) then return
	self->deactivate
	self.imgView->remove, self.nrmodel
	self.imgView->remove, self.areamodel
	self.imgView     = obj_new()
	self.lastDefined = obj_new()
	self.activeRoi   = obj_new()
	self.display     = obj_new()
	ptr_free, self.areadata
	ptr_free, self.areadataa
	self.areadata    = ptr_new()
	self.areadataa   = ptr_new()
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::activate
; Syntax: obj->activate
; Desc  : Activate roi widget buttons, reset roi system
;=-----------------------------------------------------------------------------
pro mrRoiUI::activate
	if( self.mode ge 0 ) then return                    ; we are already active
	if( self.display eq obj_new() ) then return
	self.mode = -self.mode                              ; deactive is denoted by neg. mode
	self->updatewidgets                                 ; set button sensitivity and pics
	self.display->registerEvt, self, ['Press','Motion','Release']
	self.currtype = 0
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::deactivate
; Syntax: obj->deactivate
; Desc  : Disable internal callback and all buttons
;=-----------------------------------------------------------------------------
pro mrRoiUI::deactivate
	if( self.mode lt 0 ) then return               ; we are already deactivated
	if( self.display eq obj_new() ) then return

		; react on a possibly active roi

	if( self.activeRoi ne obj_new() ) then begin
		if( self.mode eq self.add ) then begin
			self->deleteRoi, self.activeRoi
			self->calcAreaImg
		end else begin
			self->resetRoiColor, self.activeRoi
		end
		self.activeRoi = obj_new()
		self.display->display
	end

		; then unregister all callback and render all buttons insensitive

   	self.display->unregisterObj, self
	self.mode = -self.mode
	self->updatewidgets
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::mrdisplayEvtHandler
; Syntax: obj->mrdisplayEvtHandler, evt
; Args  : <b>evt</b>
;         An event structure (pointer) from mrdisplay
; Desc  : The event dispatcher for events from the mrdisplay object
;         We also fill our own evt.top here to be used in mrRoiUI events.
;=-----------------------------------------------------------------------------
pro mrRoiUI::mrdisplayEvtHandler, evt
	self.Itop = (*evt).top
	if( ((*evt).button and self.button) ne self.button ) then return
	case (*evt).type of
		0:	self->buttonPressEvt, evt
		2:	self->buttonReleaseEvt, evt
		3:	self->buttonMotionEvt, evt
	 else:	return
	end
end



;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::buttonPressEvt
; Syntax: obj->buttonPressEvt, evt
; Desc  : A button press event makes sense under two conditions:
;         adding:
;            dot: draw dot, but don't save it yet (happens @ button 1 release)
;           circ: draw circle with radius '0'
;           rect: draw rectangle with start point @ mouse x/y
;           poly: start drawing a polygon
;         editing:
;           - find closest object (conditions differ for the different forms)
;           - set the roi active and remember its number
;           - redraw screen with active roi marked red
;=-----------------------------------------------------------------------------
pro mrRoiUI::buttonPressEvt, evt

	; if mouseclick is outside image do nothing
	if( (*evt).imgx lt 0 or (*evt).imgy lt 0 ) then return
	x = (*evt).imgx
	y = (*evt).imgy
	xsize = self.display->imgXsize()
	ysize = self.display->imgYsize()
	xs = 1./xsize
	ys = 1./ysize

	case self.mode of
		self.Idot: begin
			self.currtype  = self.Idot
			self.activeroi = obj_new('MRroiDot', [x-1,x+1,x+1,x-1], [y-1,y-1,y+1,y+1], $
									maskSize=[xsize,ysize], $
									xcoord_conv=[0, xs], ycoord_conv=[0, ys])
			self.activeroi->setId, 999
			self.roigroup->add, self.activeroi
			self->updateAreaImg
			self.display->requestDisplay
			self.roitype = self.Idot
			return
		end
		self.Irect: begin
			self.currtype = self.Irect
			self.x0 = x								        ; store starting edge
			self.y0 = y
			self.activeroi = obj_new('MRroiRect', x, y, masksize=[xsize,ysize], $
									xcoord_conv=[0,xs], ycoord_conv=[0,ys] )
			self.activeroi->setId, 999
			self.roigroup->add, self.activeroi
			self->updateAreaImg
			self.display->requestDisplay
			self.roitype = self.Irect
			return
		end
		self.Icirc: begin
			self.currtype = self.Icirc
			self.x0 = x					                     ; store circle center
			self.y0 = y
			cx = dblarr(30)
			cy = dblarr(30)
			for i = 0, 29 do begin
				cx[i] = sin(2*!pi*i/30.)
				cy[i] = cos(2*!pi*i/30.)
			end
			self.activeroi = obj_new('mrRoiCirc', cx, cy, masksize=[xsize,ysize], $
									xcoord_conv=[x*xs,xs], ycoord_conv=[y*ys,ys] )
			self.activeroi->setId, 999
			self.roigroup->add, self.activeroi
			self->updateAreaImg
			self.display->requestDisplay
			self.roitype = self.Icirc
			return
		end
		self.Ipoly: begin          ; two starting points, as we'll replace the last one while moving
			if( self.activeroi eq obj_new() ) then begin		; new definition
				self.currtype = self.Ipoly
				self.x0 = x								        ; store starting edge
				self.y0 = y
				self.activeroi = obj_new('mrRoiPoly', [x,x], [y,y], masksize=[xsize,ysize], $
										xcoord_conv=[0,xs], ycoord_conv=[0,ys] )
				self.activeroi->setId, 999
				self.roigroup->add, self.activeroi
			end else begin
				self.activeroi->getProperty, n_verts=np
				self.activeroi->appenddata, x, y
			end
			self->updateAreaImg
			self.display->requestDisplay
			self.roitype = self.Ipoly
			return
		end
		self.Imove: begin
			self.x0 = x
			self.y0 = y
			roi = self->findClosestRoi( x, y )
			if( roi eq obj_new() ) then return               ; no close roi found -> end of story
			self.activeroi = roi
			txt   = self->findTxtObj( self.activeroi->id() )
			if( obj_class( roi ) eq 'MRROIDOT' ) then begin
				txt->setProperty, color=2                        ; always red for dots
			end else begin
				txt->setProperty, color=2*(1-self.solid)      ; otherwise black or red
			end
			self->calcAreaImg, /no_notify
			self.display->requestDisplay
		end
		self.Irotate: begin
			self.x0 = x
			self.y0 = y
			roi = self->findClosestRoi( x, y )
			if( roi eq obj_new() ) then return
			if( obj_class(roi) eq 'MRROIDOT' ) then return
			self.activeroi = roi
			txt   = self->findTxtObj( self.activeroi->id() )
			if( obj_class( roi ) eq 'MRROIDOT' ) then begin
				txt->setProperty, color=2                        ; always red for dots
			end else begin
				txt->setProperty, color=2*(1-self.solid)      ; otherwise black or red
			end
			self->calcAreaImg, /no_notify
			self.display->requestDisplay
		end
		self.Iresize: begin
			self.x0 = x
			self.y0 = y
			roi = self->findClosestRoi( x, y )
			if( roi eq obj_new() ) then return
			if( obj_class(roi) eq 'MRROIDOT' ) then return
			self.activeroi = roi
			txt   = self->findTxtObj( roi->id() )
			if( obj_class( roi ) eq 'MRROIDOT' ) then begin
				txt->setProperty, color=2                        ; always red for dots
			end else begin
				txt->setProperty, color=2*(1-self.solid)      ; otherwise black or red
			end
			self->calcAreaImg, /no_notify
			self.display->requestDisplay
		end
		self.Idel: begin
			roi = self->findClosestRoi( x, y )
			if( roi eq obj_new() ) then return
			self->deleteRoi, roi
			self->calcAreaImg
			self.display->requestDisplay
		end
		self.Iinfo: begin
			roi = self->findClosestRoi( x, y )
			if( roi eq obj_new() ) then return
			self->roiInfo, roi
		end
	end
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::buttonMotionEvt
; Syntax: obj->buttonMotionEvt, evt
; Desc  : Handles motion event, i.e. resizing, drawing, rotating, etc.
;=-----------------------------------------------------------------------------
pro mrRoiUI::buttonMotionEvt, evt
	if( self.activeroi eq obj_new() ) then return

		; !!! flaw !!!: when moving a roi, the cursor position does
		; not match the roi position! Not for points and not at all for
		; extended rois anyway

	if( (*evt).imgx lt 0 or (*evt).imgy lt 0 ) then outside=1 else outside=0

	x = (*evt).imgx
	y = (*evt).imgy
	xs = 1./self.display->imgXsize()
	ys = 1./self.display->imgYsize()

	case self.mode of
		self.Idot: begin
			if( outside eq 1 ) then return
			self.activeroi->replacedata, [x-1,x+1,x+1,x-1], [y-1,y-1,y+1,y+1]
			self->updateAreaImg
			self.display->requestDisplay
			return
		end
		self.Irect: begin
			if( outside eq 1 ) then return
			self.activeroi->replacedata, [self.x0, self.x0, x, x], [self.y0, y, y, self.y0]
			self->updateAreaImg
			self.display->requestDisplay
			return
		end
		self.Icirc: begin
			if( outside eq 1 ) then return
			if( x-self.x0 eq 0 ) then x = self.x0+1  ; to avoid illegal set property values
			if( y-self.y0 eq 0 ) then y = self.y0+1
			cx = dblarr(30)                                  ; change to intarr/dblarr
			cy = dblarr(30)                                  ; for 'edgie' or round circles
			for i = 0, 29 do begin
				cx[i] = fix(self.x0 + (self.x0 - x ) * sin(2*!pi*i/30.))
				cy[i] = fix(self.y0 + (self.y0 - y ) * cos(2*!pi*i/30.))
			end
			self.activeroi->replaceData, cx, cy
			self->updateAreaImg
			self.display->requestDisplay
			return
		end
		self.Ipoly: begin
			if( outside eq 1 ) then return
			self.activeroi->appendData, x, y
			self->updateAreaImg
			self.display->requestDisplay
			return
		end
		else:
	end
	if( outside eq 1 ) then begin
		self.activeroi = obj_new()
		self->calcAreaImg
		self.display->requestDisplay
		return
	end
	scale = self.display->scaling()
	txt   = self->findTxtObj( self.activeroi->id() )
	txt->getProperty, locations=txtloc

	case self.mode of
		self.Imove: begin
			self.activeroi->translate, (*evt).dx/scale, (*evt).dy/scale
			txt->setProperty, locations=[txtloc[0]+(*evt).dx/scale, txtloc[1]+(*evt).dy/scale]
			self->updateAreaImg
			mask = self.activeroi->mask( self.filled )
			self.display->requestDisplay
		end
		self.Iresize: begin
			dx = abs(x-txtloc[0]) - abs(self.x0-txtloc[0])
			dy = abs(y-txtloc[1]) - abs(self.y0-txtloc[1])
			self.x0=x
			self.y0=y
			scalex = 1 + dx*.05
			scaley = 1 + dy*.05
			self.activeroi->translate, [-txtloc[0], -txtloc[1]]
			self.activeroi->scale, [scalex, scaley]
			self.activeroi->translate, [txtloc[0], txtloc[1]]
			self->updateAreaImg
			self.display->requestDisplay
		end
		self.Irotate: begin
			if( y-txtloc[1] lt 0 ) then angx=(*evt).dx else angx=-(*evt).dx
			if( x-txtloc[0] gt 0 ) then angy=(*evt).dy else angy=-(*evt).dy
			a = self.activeroi->computeGeometry( centroid=center )
			self.x0=x
			self.y0=y
			dist = sqrt( (x-txtloc[0])^2 + (y-txtloc[1])^2 )
			ang = (angx+angy)*20./dist               ; experimental value :-) theory=(360/2/pi)
			self.activeroi->rotate, [0,0,1], ang, center=center
			self->updateAreaImg
			self.display->requestDisplay
		end
		else:
	end
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::buttonReleaseEvt
; Syntax: obj->buttonReleaseEvt
; Desc  : Releasing the button while adding a roi fixes this roi and makes
;         it nonactive. In edit mode, the currently active roi is set inactive.
;=-----------------------------------------------------------------------------
pro mrRoiUI::buttonReleaseEvt, evt
		; Button releases are only relevant when we have an active roi

	if( self.activeroi eq obj_new() ) then return

	if( (*evt).imgx lt 0 or (*evt).imgy lt 0 ) then outside=1 else outside=0
	x = (*evt).imgx
	y = (*evt).imgy
	xs = 1./self.display->imgXsize()
	ys = 1./self.display->imgYsize()

		; ADDED A ROI

	if( self.mode eq self.Idot or self.mode eq self.Icirc or self.mode eq self.Irect or self.mode eq self.Ipoly ) then begin

			; if button is released outside or area is null,
			; then delete this roi and do nothing

		mask = self.activeRoi->filledMask()
		t = where( *mask gt 0 )
		if( outside eq 1 or t[0] eq -1 ) then begin
			self->deleteRoi, self.activeRoi
		    self.activeRoi = obj_new()
		    self->updateAreaImg
		    self.display->requestDisplay
			return
		end

			; otherwise we have a new roi to add to our collection!

		if( self.roitype eq self.Icirc ) then begin           ; for circles we gotta recalculate
			cx = dblarr(30)                                   ; change to intarr/dblarr
			cy = dblarr(30)                                   ; for 'edgie' or round circles
			for i = 0, 29 do begin
				cx[i] = fix(self.x0 + (self.x0 - x ) * sin(2*!pi*i/30.))
				cy[i] = fix(self.y0 + (self.y0 - y ) * cos(2*!pi*i/30.))
			end
			self.activeroi->replaceData, cx, cy
			self.activeroi->setProperty, xcoord_conv=[0, xs], ycoord_conv=[0,ys]
		end

			; set string and stringposition - depends on roi shape

		case self.roitype of
			self.Idot:  textloc=[x,y]
			self.Icirc: textloc=[self.x0, self.y0 ]
			self.Irect: textloc=[(x+self.x0)/2, (y+self.y0)/2]
			self.Ipoly: a = self.activeroi->computeGeometry(centroid=textloc)
		end
		number    = self->findFreeRoiNumber()
		numbertxt = string( number, format='(i2)' )
		self.activeroi->setProperty, color=number+1
		self.activeroi->setId, number
		txt = obj_new('IDLgrText', Palette=self.palette, $
					strings=numbertxt, locations=textloc, recompute_dimensions=2, $
					xcoord_conv=[0, xs], ycoord_conv=[0,ys], uvalue=number )
		self.nrModel->add, txt
		self->resetRoiColor, self.activeroi
		self.lastDefined = self.activeroi
		widget_control, self.Iundo, sensitive=1
		self.activeroi = obj_new()                   ; now we may safely forget the reference
		self->calcAreaImg
		self.display->requestDisplay
		return
	end

		; DONE MODIFYING A ROI

	if( self.mode eq self.Iresize or self.mode eq self.Imove or self.mode eq self.Irotate ) then begin
		self->resetRoiColor, self.activeRoi
		self.activeRoi = obj_new()
		self->calcAreaImg
		self.display->requestDisplay
	end

end


;==============================================================================
;
;                              PUBLIC OBJECT METHODS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::showRois
; Syntax: obj->showRois
; Desc  : Display the ROIs in the associated mrdisplay object
;=-----------------------------------------------------------------------------
pro mrRoiUI::showRois
	if( obj_valid( self.areaimg ) ) then self.areaimg->setProperty, hide=0
	if( obj_valid( self.nrmodel ) ) then self.nrmodel->setProperty, hide=0
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::hideRois
; Syntax: obj->hideRois
; Desc  : Do not display the ROIs in the associated mrdisplay object (i.e.
;         hide them)
;=-----------------------------------------------------------------------------
pro mrRoiUI::hideRois
	if( obj_valid( self.areaimg ) ) then self.areaimg->setProperty, hide=1
	if( obj_valid( self.nrmodel ) ) then self.nrmodel->setProperty, hide=1
end


;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::calcAreaImg
; Syntax: obj->calcAreaImg, /no_notify
; Args  : <b>no_notify</b>
;         set to inhibit notifications for just this call
; Desc  : Recalculates the masks for all rois. The masks of the non active rois
;         are saved in our intermediate storage, so we don't need to recalculate
;         them when updating later.
;         Calling this routine will possibly cause a roichange event.
;=-----------------------------------------------------------------------------
pro mrRoiUI::calcAreaImg, no_notify=no_notify
	allrois = self.roigroup->get( /all, count=nrois )
	(*self.areadata)[*] = 0
	for i = 0, nrois-1 do begin
		roi = allrois[i]
		roino = roi->id()
		if( roi eq self.activeroi ) then continue
		mask = roi->mask( self.filled )
		for j = 0, 2 do begin
				(*self.areadata)[j,*,*] = (*self.areadata)[j,*,*] + *mask * self.rgb[j, roino+1]
		end
		(*self.areadata)[3,*,*] = (*self.areadata)[3,*,*] or (*mask * 255)
	end
	(*self.areadataa)[*] = 0
	if( self.activeroi ne obj_new() ) then begin
		mask = self.activeroi->mask( self.filled )
		(*self.areadataa)[0,*,*] = *mask * 128
		(*self.areadataa)[3,*,*] = (*self.areadataa)[3,*,*] or (*mask * 255)
	end
	self.areaimg->setProperty, data=(*self.areadata + *self.areadataa )

		; notifications for 'change' events

	if( keyword_set( no_notify ) ) then return
	evt = ptr_new( {type:2, top:self.Itop, roi:self.activeroi}, /no_copy )
	self->sendNotification, 'change', evt
	ptr_free, evt
end


;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::updateAreaImg
; Syntax: obj->updateAreaImg
; Desc  : As opposed to mrRoiUI::calcAreaImg this method will only recalculate
;         the currently active roi's masks. This along with the fact that
;         active rois are not accessible from the outside makes this method
;         purely 'private' (whatever that means in IDL)
;         A call to this method causes a 'roiupdate' event.
;=-----------------------------------------------------------------------------
pro mrRoiUI::updateAreaImg
	if( self.activeroi ne obj_new() ) then begin
		self.activeroi->calcFilledMask
		self.activeroi->calcFramedMask
		mask = self.activeroi->mask( self.filled )
		(*self.areadataa)[0,*,*] = *mask * 128
		(*self.areadataa)[3,*,*] = (*mask * 255)
	end else begin
		(*self.areadataa)[*] = 0
	end
	self.areaimg->setProperty, data=(*self.areadata + *self.areadataa )

		; event handling for 'modify' event

	evt = ptr_new( {type:1, top:self.Itop, roi:self.activeroi}, /no_copy )
	self->sendNotification, 'modify', evt
	ptr_free, evt
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::print
; Syntax: obj->print
; Desc  : Print a couple of roi properties - meant for debugging mostly
;=-----------------------------------------------------------------------------
pro mrRoiUI::print, imgdata
	allrois = self.roigroup->get( /all, count=nrois )
	if( nrois eq 0 ) then begin
		print, "No rois defined"
		return
	end
	for i = 0, nrois-1 do begin
		roi = allrois[i]
		roino = roi->id()
		print, 'Roi number ', roino
		print, '----------------------------------------------------------'
		x = where( *(roi->filledMask()) gt 0, npixels )
		if( npixels eq 0 ) then begin
			print, '*** no pixels in roi ***'
			print, '----------------------------------------------------------'
			continue
		end
		print, 'Number of pixels in roi:', npixels
		print, 'Mean pixel value in roi:', mean( (*imgdata)[x] )
		if( npixels gt 1 ) then print, 'Standard deviation:', stddev( (*imgdata)[x] )
		print, '----------------------------------------------------------'
	end
end


;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::palette
; Syntax: palette = obj->palette()
; Desc  : Returns an object reference to the palette used for roi display.
;         This should be regarded as read-only, as the ROI colors are determined
;         from other sources - thus changing this palette has no effect!
;=-----------------------------------------------------------------------------
function mrRoiUI::palette, red=red, blue=blue, green=green
	if( arg_present(red) ne 0 ) then self.palette->getProperty, red_values=red
	if( arg_present(blue) ne 0 ) then self.palette->getProperty, blue_values=blue
	if( arg_present(green) ne 0 ) then self.palette->getProperty, green_values=green
	return, self.palette
end


;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::roiColor
; Syntax: rgb = obj->roiColor( number )
; Args  : <b>number</b>
;         The roi number
; Desc  : Returns a 3-dim RGB-vector which holds the color of roi number
;         'number'. It plays no role, whether a roi with that number actually
;         exists on screen. Only the lower 8 bit of number are used (i.e. 0-255)
;=-----------------------------------------------------------------------------
function mrRoiUI::roiColor, roino
	roino = roino and 255
	return, [self.rgb[0,roino], self.rgb[1,roino], self.rgb[2,roino]]
end



;==============================================================================
;
;                         HANDLING ROIS FROM THE 'OUTSIDE'
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::addRoi
; Syntax: obj->addroi, type, x, y, r, /no_update
; Args  : <b>type</b>
;         One of 'dot', 'circ', 'rect' and 'poly'
;         <b>x, y, [r]</b>
;         x and y coordinates and radius for circles (see below)
;         <b>no_update</b>
;         If set to '1', roi areas are not recalculated after setting - this
;         might save time when setting many rois at once.
; Desc  : Add a MRroiDot/Rect/Circ/Poly object at the specified coordinates
;         to the roigroup. The arguments for the different roitypes need
;         to be specified as follows:
;         'dot'  : x, y define x and y position of the dot roi
;         'rect' : x, y two elements each for two opposing edges of the roi
;         'circ' : x, y define the center, r the radius
;         'poly' : x, y both have the same number of 'n' elements denoting the
;         .        n edges of the roi
;         This method does not interfere with any rois defined via the user-
;         interface and can thus be used as a complement.
;         This method <b>does not</b> call mrdisplay::display
;         If you use the /no_update flag for faster setting of several rois,
;         don't forget to either call mrRoiUI::calcAreaImg yourself after setting
;         the last roi, or add the last roi without this flag.
;         Note that /no_update also implies no notifications send
;=-----------------------------------------------------------------------------
pro mrRoiUI::addRoi, type, x, y, r, no_update=no_update

	xsize = self.display->imgXsize()
	ysize = self.display->imgYsize()
	xs=1./xsize
	ys=1./ysize

		; some error checking

	if( n_elements( type ) eq '' ) then begin
		print, "Undefined roi type in mrRoiUI::addRoi"
		return
	end
	if( n_elements(x) ne n_elements(y) ) then begin
		print, "Different number of elements in x and y in mrRoiUI::addRoi"
		return
	end

		; now for the different rois

	case strlowcase(type) of
		'dot': begin
			if( n_elements(x) ne 1 ) then begin
				print, "Illegal number of elements in 'x' in mrRoiUI::addRoi"
				return
			end
			roi = obj_new('MRroiDot', [x-1,x+1,x+1,x-1], [y-1,y-1,y+1,y+1], $
										maskSize=[xsize,ysize], $
										xcoord_conv=[0, xs], ycoord_conv=[0, ys])
			textloc=[x[0],y[0]]
			end
		'rect': begin
			if( n_elements(x) ne 2 ) then begin
				print, "Illegal number of elements in 'x' in mrRoiUI::addRoi"
				return
			end
			roi = obj_new('MRroiRect', [x[0],x[0],x[1],x[1]], [y[0],y[1],y[1],y[0]], $
			                             masksize=[xsize,ysize], $
									     xcoord_conv=[0,xs], ycoord_conv=[0,ys] )
			textloc=[fix((x[0]+x[1])/2), fix((y[0]+y[1])/2)]
			end
		'circ': begin
			if( n_elements(x) ne 1 and n_elements(r) ne 1 ) then begin
				print, "Invalid arguments in mrRoiUI::addRoi"
				return
			end
			cx = dblarr(30)
			cy = dblarr(30)
			for i = 0, 29 do begin
				cx[i] = fix( x + r * sin(2*!pi*i/30.))
				cy[i] = fix( y + r * cos(2*!pi*i/30.))
			end
			roi = obj_new('MRroiCirc', cx, cy, masksize=[xsize,ysize], $
									   xcoord_conv=[0,xs], ycoord_conv=[0,ys] )
			textloc=[x[0],y[0]]
			end
		'poly': begin
			if( n_elements(x) le 2 ) then begin
				print, "Too few elements in 'x' in mrRoiUI::addRoi"
				return
			end
			roi = obj_new('MRroiPoly', x, y, masksize=[xsize,ysize], $
									   xcoord_conv=[0,xs], ycoord_conv=[0,ys] )
			a = roi->computeGeometry(centroid=textloc)
			end
		else: begin
			message, "Illegal roi type >>", type, "<< in mrRoiUI::addRoi"
			return
		end
	end

		; add the textlabel - its position has already been defined

	number    = self->findFreeRoiNumber()
	numbertxt = string( number, format='(i2)' )        ; no more than 99 rois, ok?
	roi->setProperty, color=number+1
	roi->setId, number
	self.roigroup->add, roi
	txt = obj_new('IDLgrText', Palette=self.palette, color=number+1, $
					strings=numbertxt, locations=textloc, recompute_dimensions=2, $
					xcoord_conv=[0, xs], ycoord_conv=[0,ys], uvalue=number )
	self.nrModel->add, txt
	self->resetRoiColor, roi
	if( not keyword_set( no_update ) ) then self->calcAreaImg
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::deleteRoi
; Syntax: obj->deleteRoi, roi, /all, /no_update
; Args  : <b>roi</b>
;         The roi object to delete from the currently displayed roigroup
;         <b>all</b>
;         If set to '1', delete all rois from the current roigroup
;         <b>no_update</b>
;         If set, roi areas are not recalculated, i.e. the 'deleted' ROI
;         would still be displayed.
; Desc  : Delete the given roi and the corresponding text object
;         Note that no_update also implies that no notifications are send.
;=-----------------------------------------------------------------------------
pro mrRoiUI::deleteRoi, roi, all=all, no_update=no_update

	all = keyword_set( all )

		; delete just one roi

	if( all eq 0 ) then begin
		if( not obj_valid(roi) ) then return
		roiid = roi->id()
		self.roigroup->remove, roi
		obj_destroy, roi
		to = self->findTxtObj( roiid )
		if( to eq obj_new() ) then return
		self.nrModel->remove, to
		obj_destroy, to
		return
	end

		; delete all rois

	self.activeroi   = obj_new()
	self.lastDefined = obj_new()
	self.currtype  = 0
	to = self.nrModel->get( /all, count=nto )
	self.nrModel->remove, /all
	for i = 0, nto-1 do obj_destroy, to[i]
	obj_destroy, self.roigroup            ; this also destroys all rois therein
	self.roigroup = obj_new('mrroigroup')
	if( not keyword_set( no_update ) ) then self->calcAreaImg
end

;==============================================================================
;
;                             PRIVATE OBJECT METHODS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::resetRoiColor
; Syntax: self->resetRoiColor, roi
; Args  : <b>roi</b>
;         The roi whose color to change from active to inactive
; Desc  : Reset the color of the given roi and its text to the default index
;         (which is roinumber + 1)
;         Note that dot-like rois always get a label in the same color whereas
;         for the other rois it depends on the display style (outlined or filled)
;         For outlined rois, the label has the same color as the roi, for filled
;         rois the label is black, as it is located in the center of the roi
;         and, say, a red label on a red roi doesn't make for very good reading.
;=-----------------------------------------------------------------------------
pro mrRoiUI::resetRoiColor, roi
	if( roi eq obj_new() ) then return
	roiid = roi->id()
	to = self->findTxtObj( roiid )
	if( to eq obj_new() ) then return                     ; shouldn't happen
	if( obj_class(roi) eq 'MRROIDOT' ) then begin
		to->setProperty, color=(roiid+1)
	end else begin
		to->setProperty, color=(roiid+1)*(1-self.filled)
	end
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::findClosestRoi
; Syntax: self->findClosestroi, x, y
; Desc  : return object reference of the ROI closest to a given point
;         for now we do it the simple way and just use the center point
;=-----------------------------------------------------------------------------
function mrRoiUI::findClosestRoi, x, y
	rois  = self.roigroup->get( /all, count=nr )
	roi   = obj_new()
	mdist = 10000.
	for i = 0, nr-1 do begin
		a = rois[i]->computeGeometry( centroid=c )
		if( a eq 0 ) then continue
		dist = (c[0]-x)*(c[0]-x) + (c[1]-y)*(c[1]-y)
		if( dist lt mdist ) then begin
			mdist = dist
			roi = rois[i]
		end
	end
	return, roi
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::findFreeRoiNumber
; Syntax: self->findFreeRoiNumber
; Ret   : Lowest free roinumber
; Desc  : Return the next free roi number
;=-----------------------------------------------------------------------------
function mrRoiUI::findFreeRoiNumber
	rois = self.roigroup->get( /all, count=nrois )
	if( nrois eq 0 ) then return, 1                   ; should not be necessary
	ids = intarr(nrois)
	for i = 0, nrois-1 do ids[i] = rois[i]->id()
	ids = ids[sort(ids)]
	for i = 0, nrois-1 do begin
		if( ids[i] ne i + 1 ) then return, i+1
	end
	return, nrois+1
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::findTxtObj
; Syntax: obj = self->findTxtObj, id
; Args  : <b>id</b>
;         The uvalue of the text object (i.e. the number it displays)
; Desc  : Returns the corresponding text object or obj_new(), if not found.
;=-----------------------------------------------------------------------------
function mrRoiUI::findTxtObj, id
	to = self.nrModel->get( /all, count=nto )
	for i = 0, nto-1 do begin
		to[i]->getProperty, uvalue=toid
		if( toid eq id ) then return, to[i]
	end
	return, obj_new()
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::roiInfo
; Syntax: obj = self->roiInfo, roi
; Args  : <b>roi</b>
;         The roi to info :)
; Desc  : Displays a dialog_message with some simple roi information
;=-----------------------------------------------------------------------------
pro mrRoiUI::roiInfo, roi
	roiid = roi->id();
	x = where( *(roi->filledMask()) gt 0, npixels )

	img = self.display->img()
	message = ['Roi number      : ' + strtrim(string(roiid),2), $
	           'Pixels in roi      : ' + strtrim(string(npixels),2)]

	if( npixels gt 0 and img ne ptr_new() ) then begin
		if( n_elements(x) gt 1 ) then begin
			meanval = mean( (*img)[x] )
			stdd    = stddev( (*img)[x] )
			minval  = min( (*img)[x], max=maxval)
			message = [message, $
		           'Minimum value : ' + strtrim(string(minval),2), $
		           'Maximum value: ' + strtrim(string(maxval),2), $
		           'Mean value      : ' + strtrim(string(meanval),2), $
		           'Std deviation   : ' + strtrim(string(stdd),2)]
		end else begin
			val = (*img)[x]
			message = [message, $
		           'Pixel value       : ' + strtrim(string(val),2)]
		end
	end
	x = dialog_message( message, /info )
end


;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::definePalette
; Syntax: self->definePalette
; Desc  : Manual definition of the palette used for roi display.
;         Color index 0 must be black, index 1 is the color for active rois,
;         index 255 is white. All other colors are only defined with '128'
;         values - else they would display as white when overlapping
;=-----------------------------------------------------------------------------
pro mrRoiUI::definePalette
	obj_destroy, self.palette        ; doesn't hurt even if self.palette = NULL
	red   = bytarr(256)
	green = bytarr(256)
	blue  = bytarr(256)
	; black, red, red, blue, green, turquise, magenta
	red[0:7]      = [  0, 128, 196,   0,   0, 196, 196,   0]
	red[8:13]     = red[2:7]
	red[14:19]    = red[2:7]
	red[20:254]   = 128
	red[255]      = 255
	green[0:7]    = [  0,   0,   0,   0, 196, 196,   0, 196]
	green[8:13]   = green[2:7]
	green[14:19]  = green[2:7]
	green[20:254] = 128
	green[255]    = 255
	blue[0:7]     = [  0,   0,   0, 196,   0,   0, 196, 196]
	blue[8:13]    = blue[2:7]
	blue[14:19]   = blue[2:7]
	blue[20:254]  = 0
	blue[255]     = 255
	self.palette = obj_new('IDLgrPalette', red, green, blue)
		; our roi area image is rgb with alpha channel, so we need
		; this here
	self.rgb[0,*] = red
	self.rgb[1,*] = green
	self.rgb[2,*] = blue
	return
end

;==============================================================================
;
;                               LOADING AND SAVING
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::setLoadMethod
; Syntax: obj->setLoadMethod, object, method
; Args  : <b>object, method</b>
;         The object instance and the method name to call when the user presses
;         the 'load' button. Set object to obj_new() to remove a previously
;         set load method.
; Desc  : The given method will be called with the 'evt' structure as
;         defined above, with only elements roi and name (set to 'load') filled.
;         Note that the user is responsible for all setting of roi groups etc.
;         The 'load' and 'save' methods of mrRoiUI might serve as a prototype
;         for the functionality to implement.
;=-----------------------------------------------------------------------------
pro mrRoiUI::setLoadMethod, object, method
	if( n_elements(object) eq 0 or n_elements(method) eq 0 ) then return
	self.hookLoadObject = object
	self.hookLoadMethod = method
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::setSaveMethod
; Syntax: obj->setSaveMethod, object, method
; Args  : <b>object, method</b>
;         The object instance and the method name to call when the user presses
;         the 'save' button. Set object to obj_new() to remove a previously
;         set save hook.
; Desc  : The given method will be called with the 'evt' structure as
;         defined above, with only elements roi and name (set to 'save') filled.
;=-----------------------------------------------------------------------------
pro mrRoiUI::setSaveMethod, object, method
	if( n_elements(object) eq 0 or n_elements(method) eq 0 ) then return
	self.hookSaveObject = object
	self.hookSaveMethod = method
end


;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::load
; Syntax: res = self->load, filename, /destroy
; Args  : <b>filename</b>
;         the mrroigroup file to load. If no filename is given, the user is
;         presented a file-selector boxp
;         <b>destroy</b>
;         If set, the currently used roigroup is destroyed (and with it all
;         rois contained therein). Default is '0', i.e. do not destroy
; Ret   : 1 in case of success, 0 otherwise
;         If a user method is called, return value is always '1'.
; Desc  : The method loads the first roigroup found in the given file and
;         makes it the current roigroup. However, if the user has set a load
;         method via mrRoiUI::setLoadMethod, this method is called instead.
;=-----------------------------------------------------------------------------
function mrRoiUI::load, filename, destroy=destroy

	if( obj_valid(self.hookLoadObject) ) then begin
		evt = ptr_new( {top:0L, type:0, roiui:self}, /no_copy )
		call_method, self.hookLoadMethod, self.hookLoadObject, evt
		ptr_free, evt
		return, 1
	end

	if( self.roigroup->count() gt 0 ) then begin
		x = dialog_message(['This will destroy all currently defined ROIs','Are you sure?'], /question)
		if( x ne 'Yes' ) then return, 0
	end

		; get filename from user

	if( n_elements( filename ) eq 0 ) then begin
		filename = dialog_pickfile( path=self.path, /read, get_path=path, filter='*.roi')
		self.path=path
	end
	if( filename eq '' ) then return, 0

		; parse file

	roicont = mrroigroup_load( filename )             ; a container with all the roigroups loaded
	if( roicont eq obj_new() ) then begin
		x = dialog_message('Failed to load ROIs.', /error)
		return, 0
	end

		; This method only takes *one* roigroup - even if the file might
		; have contained more than one. So we get the first roigroup from the container,
		; remove it from the container and then destroy the container, thus destroying
		; any other roigroups

	roigroup = roicont->get( position=0 )
	roicont->remove, roigroup
	obj_destroy, roicont

		; prepare the rois in the roigroup for display

	self->setRoigroup, roigroup, destroy=destroy
	self.display->display

	x = dialog_message('ROIs loaded.', /info)
	return, 1
end

;+=----------------------------------------------------------------------------
; Name  : mrRoiUI::save
; Syntax: res = self->save( filename )
; Args  : <b>filename</b>
;         The file to store to. If no filename is given, the user is presented
;         a fileselector box
; Desc  : Saves a mrroigroup via the classes' save method.
;=-----------------------------------------------------------------------------
function mrRoiUI::save, filename

	if( obj_valid(self.hookSaveObject) ) then begin
		evt = ptr_new( {top:0L, type:0, roiui:self}, /no_copy )
		call_method, self.hookSaveMethod, self.hookSaveObject, evt
		ptr_free, evt
		return, 1
	end

	if( n_elements( filename ) eq 0 ) then begin
		filename = dialog_pickfile( path=self.path, /write, get_path=path, filter='*.roi' )
		self.path = path
	end
	if( filename eq '' ) then return, 0
	if( strpos(filename, '.') eq -1 ) then filename=filename+'.roi'
	res = self.roigroup->save( filename )
	if( res eq 0 ) then begin
		x = dialog_message('Saving failed')
	end else begin
		x = dialog_message('Rois saved', /info)
	end
	return, res
end


;+=----------------------------------------------------------------------------
; Name  : mrRoiUI__define
; Type  : Private Procedure (constructor)
;=-----------------------------------------------------------------------------
pro mrroiui__define
	obj = { MRROIUI, $
			inherits MRNOTIFY		  ,$
			display     :   obj_new() ,$ ; the associated display object
			button		:			0 ,$ ; the button to which widget_draw events are linked
			imgview     :   obj_new() ,$ ; the display object's image view
			imgdata     :   ptr_new() ,$ ; the display object's image data
			areamodel   :   obj_new() ,$ ; model to which we'll add the roi image
			areadata	:	ptr_new() ,$ ; pointer for storage of the area image bitmap
			areadataa	:	ptr_new() ,$ ; pointer for storage of active roi area bitmap
			areaimg     :   obj_new() ,$ ; the image holding the roi areas
			nrmodel     :   obj_new() ,$ ; model to hold the ROI numbers
			roigroup    :   obj_new() ,$ ; a container which holds all rois
			activeRoi   :   obj_new() ,$ ; the currently active roi
			lastactive  :   obj_new() ,$ ; the last active roi, for undo operations
			lastdefined :   obj_new() ,$ ; for undo functionality
			palette     :   obj_new() ,$ ; palette used to draw the text items
			rgb			:intarr(3,256),$ ; the single rgbvalues - used to draw the rois
			ui          :         ''  ,$ ; the UI layout
			Itop		:		   0L ,$ ; id of the top widget
			Ibase       :          0L ,$ ; the base id
			IRoi        :          0L ,$ ; id of the 'roi' main button
			Iload       :          0L ,$ ; id of the 'load' button
			Isave       :          0L ,$ ; id of the 'save' button
			Idot        :          0L ,$ ; id of the 'dot' button
			Icirc       :          0L ,$ ; id of the 'circle' button
			Irect       :          0L ,$ ; id of the 'rect' button
			Ipoly       :          0L ,$ ; id of the 'polygon' button
			Iundo       :          0L ,$ ; id of the 'undo' button
			Imove       :          0L ,$ ; id of the move button = self.dot
			Iresize     :          0L ,$ ; id of the resize button = self.circle
			Irotate     :          0L ,$ ; id of the rotate button = self.rect
			Idel        :          0L ,$ ; id of the 'delete' button = self.del
			Ifilled     :		   0L ,$ ; id of the filled/framed button
			Isolid	    :		   0L ,$ ; id of the transparent/solid button
			Iinfo       :          0L ,$ ; id of the 'help' button
			path        :          '' ,$ ; last used path
			mode        :          0  ,$ ; one of +- self.add/edit/del/delall;  < 0 means 'deactive'
			roitype     :          0  ,$ ; one of self.Idot/Icirc/Irect/Ipoly
			filled		:		   0  ,$ ; 0 = framed, 1 = filled
			solid    	:		   0  ,$ ; 0 = transparent, 1 = solid
			editmode    :          0  ,$ ; one of move, resize, del
			x0          :          0  ,$ ; starting point when creating rect and circ
			y0          :          0  ,$ ; dito
			currtype    :          0  ,$ ; the roitype we're just about to define
			evtstruct	:	ptr_new() ,$ ; the event structure
			hookLoadObject: obj_new() ,$ ; user defined load method
			hookLoadMethod:       ''  ,$ ;
			hookSaveObject: obj_new() ,$ ; user defined save method
			hookSaveMethod:       ''   $ ;
		}

end
