;+=----------------------------------------------------------------------------
; Name  : mrWidget_numfield
; RCS   : $Id: mrwidget_numfield.pro,v 2.1 2004/11/10 11:06:12 olivert Exp olivert $
; Syntax: widget_id = mrwidget_numfield( parent, /integer, /float, /editable
;         .           value=number, min=number, max=number, incr=number,
;         .           xsize=int, labeltext=string, labelsize=int, fieldsize=int, frame=int,
;         .           uname=string, uvalue=var, /all_events, format=format
; Ret   : The widget id of the created compound widget or 0L in case of errors.
; Args  : <b>parent</b>
;         widget_id of the parent widget
;         <b>integer</b>
;         value is of integer type (default>
;         <b>float</b>
;         value is of float type
;         <b>editable</b>
;         Determines wether the value field is directly editable. Default is no (=0)
;         <b>value</b>
;         the initial value, default is 0.
;         <b>min</b>
;         minmimum allowed value
;         <b>max</b>
;         maximum allowed value
;         <b>incr</b>
;         increment/decrement value for the arrow buttons. Default is 1.
;         <b>xsize</b>
;         xsize in pixels of the widget. Please note that labelsize + fieldsize
;         should be about 22 pixels less than xsize, or the up/down arrows will
;         not be displayed!
;         <b>labeltext</b>
;         text of the label
;         <b>labelsize</b>
;         size of the label in pixels
;         <b>fieldsize</b>
;         size of the input field in pixels - don't use characters because then
;         the fieldsize in pixels will depend on the font
;         <b>frame</b>
;         if set, draws a frame around the widget
;         <b>uname</b>
;         user definable name to identify this widget. See any other IDL widget
;         documentation for more info
;         <b>uvalue</b>
;         user definable value
;         <b>format</b>
;         A format string as accepted by idl, e.g. '(f4.2)' that denotes the format
;         of the output. This format string is applied whenever the user
;         a) presses one of the arrow buttons
;         b) presses return
;         c) inserts a digit which would result in a number outside possibly specified
;         boundaries. The default formats are '(i8)' for integer and '(f12.5)' for float.
;         <b>all_events</b>
;         By default, any value change in the text field will cause a corresponding
;         event, the keyword here being <i>value change</i> - i.e. removal of e.g.
;         a leading zero will <i>not</i> cause an event.
;         If all_events is set to '0', the user has to explicitly type
;         return to cause a value change event - but even then events are only caused
;         if the value has changed, i.e. two return keys cause only one event.
;         Mouse button presses on the arrows however always cause a change event.
; Desc  : A compound widget for integer or float value input with two arrow buttons
;         to change the value in predefined steps via mouseclick. Maximum and minimum
;         values are automatically controlled and obeyed, if set.
;         The returned events have the following structure:
;		  ret = { mrwidget_numfield_evt, $
;		  	id      : 0L,  $ ; widget id of the mrwidget_numfield
;		  	top     : 0L,  $ ; top widget id
;			handler : 0L,  $ ; handler widget id = mrwidget_numfield
;			type    : 0,   $ ; 0 for button event, 1 for manual change event
;			value   : 0.0  $ ; current value of the field - boundary checked
;		  }
; Author: (c) Oliver Thilmann, oliver.thilmann@web.de
;         Based on Terry Oakes' cw_field_arrows.pro, distributed with the SPAMALIZE
;         suite. See http://psyphz.psych.wisc.edu/~oakes/spam/spam_frames.htm
;         The widget is rewritten in large parts with some extra functionality,
;         esp. on boundary checking, but overall structure and the arrow bitmap
;         were taken over with kind permission from Terry's code.
; Licns : This program is published under the GPL. See gpl.txt for details
;=-----------------------------------------------------------------------------
function mrwidget_numfield, parent, integer=int, float=float, xsize=xsize, labeltext=labeltext, labelsize=labelsize, $
	fieldsize=fieldsize, column=column, row=row, frame=frame, editable=editable, uname=uname, uvalue=uvalue, $
	value=value, min=minval, max=maxval, incr=incr, all_events=all_events, format=format

		; rather use on_error here, since it's a programming error

	on_error, 2

	;error_status = 0
	;catch, error_status
	;if( error_status ne 0 ) then begin
	;	catch, /cancel
	;	; we catch the error, but do not hand it on
	;	print, '% MRWIDGET_NUMFIELD: ', + !ERROR_STATE.MSG
	;	return, 0L
	;end

		; that's my first compound widget, so please bear with me

	if( n_elements(parent) eq 0 ) then return, 0L
	type = keyword_set( int )                   ; type = 1 for integer, type = 2 for float
	type = type + 2 * keyword_set( float )
	if( type eq 3 ) then begin
		message, 'Set both /int and /float keywords'
	end
	if( type eq 0 ) then type = 1 ; default is integer

	fieldsize = ( n_elements(fieldsize ) eq 0 ) ? 100 : fieldsize  ; arbitraty default field size
	labelsize = ( n_elements(labelsize ) eq 0 ) ? 100 : labelsize
	xsize     = ( n_elements(xsize     ) eq 0 ) ?   0 : xsize
	label     = ( n_elements(label     ) eq 0 ) ?  '' : labeltext
	editable  = ( n_elements(editable  ) eq 0 ) ?   0 : editable   ; default is manual editing not allowed
	frame     = ( n_elements(frame     ) eq 0 ) ?   0 : frame
	uvalue    = ( n_elements(uvalue    ) eq 0 ) ?   0 : uvalue
	textframe = ( n_elements(textframe ) eq 0 ) ?   0 : textframe
	allevents = ( n_elements(all_events) eq 0 ) ?   1 : all_events ; default is all events
	uname     = ( n_elements(uname     ) eq 0 ) ?  '' : uname
	uvalue    = ( n_elements(uvalue    ) eq 0 ) ?   0 : uvalue
	case type of
		1: begin
			value  = ( n_elements(value ) eq 0 ) ?                  0  : fix(value)
			incr   = ( n_elements(incr  ) eq 0 ) ?                  1  : fix(incr)
			minval = ( n_elements(minval) eq 0 ) ? -!values.d_infinity : fix(minval)
			maxval = ( n_elements(maxval) eq 0 ) ?  !values.d_infinity : fix(maxval>minval)
			format = ( n_elements(format) eq 0 ) ?              '(i8)' : format
			value  = fix(  value > minval )
			value  = fix(  value < maxval )
		end
		2: begin
			value  = ( n_elements(value ) eq 0 ) ?                  0. : value
			incr   = ( n_elements(incr  ) eq 0 ) ?                  1  : incr
			minval = ( n_elements(minval) eq 0 ) ? -!values.d_infinity : minval
			maxval = ( n_elements(maxval) eq 0 ) ?  !values.d_infinity : maxval>minval
			format = ( n_elements(format) eq 0 ) ?           '(f12.5)' : format
			value  = value > minval
			value  = value < maxval
		end
	end

	; note that this image is drawn upside down! (yep, drove me crazy)
	; 0/2 = up arrow, 1/3 = down arrow
	arrows = [$
			[1,1,1,1,1,1,1,1,1,1,1,1,1,1,1], $
    		[1,1,1,1,1,1,1,1,1,1,1,1,1,1,1], $
			[1,1,1,1,1,1,1,3,1,1,1,1,1,1,1], $
			[1,1,1,1,1,1,3,3,3,1,1,1,1,1,1], $
			[1,1,1,1,1,3,3,1,3,3,1,1,1,1,1], $
			[1,1,1,1,3,3,1,1,1,3,3,1,1,1,1], $
			[1,1,1,3,3,1,1,1,1,1,3,3,1,1,1], $
			[1,1,1,3,1,1,1,1,1,1,1,3,1,1,1], $
			[1,1,1,1,1,1,1,1,1,1,1,1,1,1,1], $
			[1,1,1,1,1,1,1,1,1,1,1,1,1,1,1], $
			[4,4,4,4,4,4,4,4,4,4,4,4,4,4,4], $
			[4,4,4,4,4,4,4,4,4,4,4,4,4,4,4], $
			[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0], $
			[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0], $
			[0,0,0,2,0,0,0,0,0,0,0,2,0,0,0], $
			[0,0,0,2,2,0,0,0,0,0,2,2,0,0,0], $
			[0,0,0,0,2,2,0,0,0,2,2,0,0,0,0], $
			[0,0,0,0,0,2,2,0,2,2,0,0,0,0,0], $
			[0,0,0,0,0,0,2,2,2,0,0,0,0,0,0], $
			[0,0,0,0,0,0,0,2,0,0,0,0,0,0,0], $
			[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0], $
			[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]]
	sz = size(arrows)

	baseid = widget_base( parent, /row, uvalue=uvalue, $
				event_func     = 'mrwidget_numfield_evt', $ ; widget_event will call this func
				pro_set_value  = 'mrwidget_numfield_set', $ ; called @ widget_control, set_value
				func_get_value = 'mrwidget_numfield_get', $ ; called @ widget_control, get_value
				/base_align_center, uname=uname, frame=frame )

	if( xsize ne 0 ) then widget_control, baseid, scr_xsize=xsize
	labelid = widget_label( baseid, value=labeltext, scr_xsize=labelsize )
	fieldid = widget_text( baseid, value=strtrim(string(value, format=format),2), editable=editable, /all_events, /align_right, scr_xsize=fieldsize, uvalue=labelid )
	drawid  = widget_draw( baseid, /button_events, xsize=sz[1], ysize=sz[2], uvalue=labelid, $
				retain=2, event_func='mrwidget_numfield_arrowevt')

	val = float(value)
	state={ baseid:baseid, drawid: drawid, fieldid: fieldid, type:type, incr:incr, minval:minval, maxval:maxval, value:val, arrows:arrows, allevents:allevents, bpc:0, format:format }
	widget_control, labelid, set_uvalue=state

		; Why not set the realize function directly?
		; Because the widget might be created into an already existing base and thus be realized
		; immediately - in which case the state is not yet attached to the label widget.
		; So we do it later.
		; However (in that very case) the widget is realized by now and setting notify_realize afterwards
		; doesn't do anything. Which makes sense, but means that we have to call the realize function
		; ourselves, if the widget is visible.
		; "Oh, no, gotta learn this some day
		; Why, I, gotta learn the hard way"
		; (Nickelback, though not referring to IDL programming)
		; [afterthought: actually the realize should be attached to the base, but it works,
		; so no need to change for beauty]

	widget_control, drawid, notify_realize='mrwidget_numfield_realize'
	if( widget_info(drawid, /visible) eq 1 ) then mrwidget_numfield_realize, drawid
	return, baseid
end


;+=----------------------------------------------------------------------------
; Name  : mrwidget_numfield_realize
; Desc  : Draw the arrows upon realization.
;=-----------------------------------------------------------------------------
pro mrwidget_numfield_realize, drawid
	widget_control, drawid, get_uvalue=labelid
	widget_control, labelid, get_uvalue=state
	widget_control, state.drawid, get_value=winid
	wset, winid
	tvlct, r, g, b, /get
	tvlct, 255, 255, 255, 0
	tvlct, 255, 255, 255, 1
	if( state.value lt state.maxval ) then tvlct, 0, 0, 0, 2 else tvlct, 212, 208, 200, 2
	if( state.value gt state.minval ) then tvlct, 0, 0, 0, 3 else tvlct, 212, 208, 200, 3
	tvlct, 212, 208, 200, 4
	tv, state.arrows
	tvlct, r, g, b
end

;+=----------------------------------------------------------------------------
; Name  : mrwidget_numfield_arrowevt
; Desc  :
;=-----------------------------------------------------------------------------
function mrwidget_numfield_arrowevt, evt
	widget_control, evt.id, get_uvalue=labelid, get_value=winid
	widget_control, labelid, get_uvalue=state

	ret = { mrwidget_numfield_evt, $
			id		: state.drawid, 	$
			top		: evt.top, 			$
			handler	: state.drawid,		$
			type    : 0,				$ ; 0 for release button, 1 for set button
			value	: 0.0 $
		  }

	if( evt.release eq 1 ) then begin
		wset, winid
		tvlct, r, g, b, /get
		tvlct, 255, 255, 255, 0
		tvlct, 255, 255, 255, 1
		if( state.value lt state.maxval ) then tvlct, 0, 0, 0, 2 else tvlct, 212, 208, 200, 2
		if( state.value gt state.minval ) then tvlct, 0, 0, 0, 3 else tvlct, 212, 208, 200, 3
		tvlct, 212, 208, 200, 4
		tv, state.arrows
		tvlct, r, g, b
		ret.type = 0
		return, ret
	end
	if( evt.press eq 1 ) then begin
		tvlct, 255, 255, 255, 0
		tvlct, 255, 255, 255, 1
		if( state.value lt state.maxval ) then tvlct, 0, 0, 0, 2 else tvlct, 212, 208, 200, 2
		if( state.value gt state.minval ) then tvlct, 0, 0, 0, 3 else tvlct, 212, 208, 200, 3
		tvlct, 212, 208, 200, 4
		newval = state.value
		if( evt.y lt 10 and state.value gt state.minval ) then begin
			tvlct, 255, 255, 255, 3
			tvlct,   0,   0,   0, 1
			newval = newval - state.incr
		end
		if( evt.y gt 10 and state.value lt state.maxval ) then begin
			tvlct,  0,   0,    0, 0
			tvlct, 255, 255, 255, 2
			newval = newval + state.incr
		end
		wset, winid
		tvlct, r, g, b, /get
		tv, state.arrows
		tvlct, r, g, b
		newval = newval < state.maxval
		newval = newval > state.minval
		if( state.type eq 1 ) then newval = fix( newval )
		ret.value = newval
		ret.type  = 1 ; press
		if( newval eq state.value ) then state.bpc=0 else state.bpc=1
		state.value = newval
		widget_control, state.fieldid, set_value=strtrim(string(newval, format=state.format),2)
		widget_control, labelid, set_uvalue=state, /no_copy
		return, ret
	end
	return, 0    ; any other unknown events?
end


;+=----------------------------------------------------------------------------
; Name  : mrwidget_numfield_evt
; Desc  : The main event handling routine. We only deal with events from
;         arrow presses (issued above) and from manual changes in the
;         text field.
;         Only release events from the arrows are passed on to the user
;         As for text events it depends on the /all_events settings.
;         If this is not set, the user needs to press return for a value to
;         take effect. Otherwise every change is recognized and handed to the
;         calling routine.
;         event type is '0' for button change events, 1 for text change events
;=-----------------------------------------------------------------------------
function mrwidget_numfield_evt, evt

	error_status = 0
	catch, error_status
	if( error_status ne 0 ) then begin
		catch, /cancel
		; we catch the error, but do not hand it on
		print, '% MRWIDGET_NUMFIELD_EVT: ', + !ERROR_STATE.MSG
		return, 0
	end

	origin = widget_info( evt.id, /type )
	case origin of

			;
			; Events from a value change via the arrow buttons
			; only button release events are handed on
			;

		4: begin	                                    ; from draw event
			widget_control, evt.id, get_uvalue=labelid
			widget_control, labelid, get_uvalue=state
			ret = { mrwidget_numfield_evt,  	$
				id		: state.baseid, 		$
				top		: evt.top, 				$
				handler	: state.baseid, 		$
				type	: 0,					$
				value	: float(state.value)	$
			}
			if( state.bpc eq 0 ) then return, 0  ; did the last button press change the value?
			state.bpc = 0
			if( evt.type eq 0 ) then return, ret else return, 0
		end

			;
			; Events from the text widget
			;

		3: begin			; from the text widget
			widget_control, evt.id, get_uvalue=labelid
			widget_control, labelid, get_uvalue=state
			widget_control, state.fieldid, get_value=value
			if( evt.type gt 2 ) then return, 0                         ; take only character insert events
			ret = { mrwidget_numfield_evt,$
				id		: state.baseid, 	$
				top		: evt.top, 			$
				handler	: state.baseid, 	$
				type	: 1, 				$
				value	: state.value		$
			}
			; now check wether the change produced a legal number.
			; if not, state.value (last known good number) is written back
			; and no event is issued
			if( value eq '' or value eq '-' or value eq '.' ) then begin
				widget_control, state.fieldid, set_value='0'
				if( state.value eq 0 ) then return, 0
				state.value = 0
				allevents = state.allevents ; save intermediate
				widget_control, labelid, set_uvalue=state, /no_copy
				if( evt.type eq 2 and allevents eq 0 ) then return, 0
				if( evt.type ne 2 ) then if( evt.ch ne 10 and allevents eq 0 ) then return, 0
				ret.value = 0.
				return, ret
			end
			value = float(value)
			if( state.value eq value ) then return, 0                  ; cursor move event
			val   = value
			val = val < state.maxval
			val = val > state.minval
			if( state.type eq 1 ) then val = fix(val)
			if( val ne value ) then begin ; illegal number
				state.value = state.value > state.minval
				state.value = state.value < state.maxval
				if( state.type eq 1 ) then begin
					state.value = fix(state.value)
					widget_control, state.fieldid, set_value=strtrim(string(state.value,format=state.format),2)
					return, 0
				end else begin
					widget_control, state.fieldid, set_value=strtrim(string(state.value,format=state.format),2)
					return, 0
				end
			end
			; ok, at this point we know that value contains a valid new entry
			state.value=value
			ret = { mrwidget_numfield_evt,$
				id		: state.baseid, 	$
				top		: evt.top, 			$
				handler	: state.baseid, 	$
				type	: 1, 				$
				value	: state.value		$
			}
			widget_control, state.drawid, get_value=winid
			wset, winid
			tvlct, r, g, b, /get
			tvlct, 255, 255, 255, 0
			tvlct, 255, 255, 255, 1
			if( state.value lt state.maxval ) then tvlct, 0, 0, 0, 2 else tvlct, 212, 208, 200, 2
			if( state.value gt state.minval ) then tvlct, 0, 0, 0, 3 else tvlct, 212, 208, 200, 3
			tvlct, 212, 208, 200, 4
			tv, state.arrows
			tvlct, r, g, b
			; things (and if statements) would be easier if IDL knew about conditional logical operators...
			if( evt.type eq 2 ) then begin
				if( state.allevents eq 0 ) then return, 0
			end else begin
				if( evt.ch ne 10 and state.allevents eq 0 ) then return, 0
			end
			widget_control, labelid, set_uvalue=state, /no_copy
			return, ret
		end
		else: print, 'Unknown origin'  ; we should never get here
	end
	return, 0
end


;+=----------------------------------------------------------------------------
; Name  : mrwidget_numfield_set
; Desc  : Called from widget_control, this_id, set_value=value
;         Values that are out of the set boundaries are silently ignored.
;=-----------------------------------------------------------------------------
pro mrwidget_numfield_set, baseid, value
	labelid = widget_info( baseid, /child )
	widget_control, labelid, get_uvalue=state
	if( value gt state.maxval or value lt state.minval ) then begin
		print, 'MRWIDGET_NUMFIELD: set value out of boundaries'
		return
	end
	if( state.type eq 1 ) then val = fix(value) else val=value
	widget_control, state.fieldid, set_value=strtrim(string(val, format=state.format),2)
	state.value = val
	widget_control, state.drawid, get_value=winid
	wset, winid
	tvlct, r, g, b, /get
	tvlct, 255, 255, 255, 0
	tvlct, 255, 255, 255, 1
	if( state.value lt state.maxval ) then tvlct, 0, 0, 0, 2 else tvlct, 212, 208, 200, 2
	if( state.value gt state.minval ) then tvlct, 0, 0, 0, 3 else tvlct, 212, 208, 200, 3
	tvlct, 212, 208, 200, 4
	tv, state.arrows
	tvlct, r, g, b
	widget_control, labelid, set_uvalue=state, /no_copy
	return
end

;+=----------------------------------------------------------------------------
; Name  : mrwidget_numfield_get
; Desc  : Called via widget_control, this_id, get_value=value
;         state.value contains the last known good value, so we return that
;=-----------------------------------------------------------------------------
function mrwidget_numfield_get, base
	labelid = widget_info( base, /child )
	widget_control, labelid, get_uvalue=state
	return, state.value
end

