;+=----------------------------------------------------------------------------
; Name   : mrZoom
; RCS    : $Id: mrzoom__define.pro,v 2.1 2004/11/10 11:06:28 olivert Exp olivert $
; Methods: <b>Constructor</b>
;          mrZoom::init                : constructor, part II
;          <br>
;          <b>Event handling</b>
;          mrZoom::selfregister        : register with a display object
;          mrZoom::unregister          : unregister
;          <br>
;          <b>Private</b>
;          mrZoom::mrDisplayEvtHandler : the event handler and broker for mr*display
;          mrZoom::buttonDblClickEvt   : handling of double clicks
;          mrZoom::buttonPressEvt      : handling of button presses
;          mrZoom::buttonReleaseEvt    : handling of button release
;          mrZoom::buttonMotionEvt     : handling of motion while button pressed
;          mrZoom::update              : get scaling and update text
;          mrZoom__define              : the constructor
;          mrZoom::cleanup             : the destructor
; Desc  : A plugin object for mrdisplay to provide zoom functionality
; Author : (c) Oliver Thilmann, oliver.thilmann@web.de
; Licns  : This program is published under the GPL. See gpl.txt for details
; Code   :
;         The following code snippet creates a mrdisplay and a zoom object, then
;         registers the zoom with mrdisplay on left mouse button events.
;         <&nbsp;>
;         display = obj_new('MRdisplay', baseid, 256, 256, 256, 256 )
;         zoom    = obj_new('mrZoom')
;         zoom->selfregister( display, button=1 )  ; register on left mouse button
;         display->setImg, imgpointer              ; set the image
;         display->display                         ; display image
;         display->activate                        ; so mrdisplay creates events
;=-----------------------------------------------------------------------------


;==============================================================================
;
;                          CONSTRUCTORS AND DESTRUCTORS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrZoom::init
; Syntax: x = obj_new('mrZoom', /inscreen, sensitivity=float)
; Args  : <b>inscreen</b>
;         flag - display zoom factor in display window if set (=default)
;         <b>sensitivity</b>
;         float: sensitivity to mouse movements, default = 0. (=automatic)
; Desc  : The constructor
;         Resonable values for sensitivity are around 1./canvas size in pixels
;         If unset or set to '0.', mrZoom automatically sets the sensitivity
;         depending on the canvas size
;=-----------------------------------------------------------------------------
function mrZoom::init, inscreen=inscreen, sensitivity=sensitivity
	if( n_elements(inscreen)    eq 0 ) then inscreen    = 1
	if( n_elements(sensitivity) eq 0 ) then sensitivity = 0.
	self.inscreen    = inscreen
	self.sensitivity = sensitivity
	self.border      = 0
	self.active      = 0
	self.grZModel = obj_new('IDLgrModel', hide=0)
	self.grZText  = obj_new('IDLgrText', locations=[0.5,0.01], color=[0,255,0], $
							alignment=0.5, char_dimensions=[0.025,0.025])
	self.grZModel->add, self.grZText
	return, 1
end

;+=----------------------------------------------------------------------------
; Name  : mrZoom::cleanup
; Desc  : cleanup at object destruction
;=-----------------------------------------------------------------------------
function mrZoom::cleanup
	if( self.display ne obj_new() ) then self->unregister
	obj_destroy, self.grZModel
	obj_destroy, self.grZText
end


;==============================================================================
;
;                                  EVENT HANDLING
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mrZoom::selfregister
; Syntax: obj->selfregister, display, button=int, /border
; Args  : <b>display</b>
;         The mrDisplay object with which to register
;         <b>button</b>
;         The button number upon which to zoom; left=1, middle=2 (def), right=3
;         <b>border</b>
;         If set, the mouse has to be in the right 15% of the mrDisplay object
;         to trigger zooming. To be combined with the corresponding /noborder
;         keyword of mrPan.
; Desc  : Register with an mrDisplay object
;=-----------------------------------------------------------------------------
pro mrZoom::selfregister, display, button=button, border=border

		; no error checking on button - who knows how many are out there?

	if( n_elements(button) eq 0 ) then button = 2
	if( button lt 1 ) then begin
		print, 'Illegal button number in mrZoom::selfregister'
		return
	end
	self.border = keyword_set( border )

	display->registerEvt, self, ['Press', 'DblClick', 'Release','Motion']
	self.display = display
	self.button  = 2^(button-1)          ; use bitmap for easier comparison
	(display->labelView())->remove, self.grZModel
	(display->labelView())->add, self.grZModel

		; set text size depending on canvas size

	win = self.display->grwindow()
	win->getProperty, dimensions=res
	charsize = 13./min(res)
	self.grZText->setProperty, char_dimensions=[charsize,charsize]

		; set sensitivity if not done by the user

	if( self.sensitivity eq 0. ) then self.sensitivity = 1./min(res)

	return
end

;+=----------------------------------------------------------------------------
; Name  : mrZoom::unregister
; Syntax: obj->unregister
; Desc  : unregisters with a mrdisplay object. It is a good idea to unregister
;        before destroying the mrdisplay or the mrZoom object.
;=-----------------------------------------------------------------------------
pro mrZoom::unregister
	if( self.display eq obj_new() ) then return
	self.display->unregisterObj, self
	(self.display->labelView())->remove, self.grZModel
	self.display = obj_new()
end

;+=----------------------------------------------------------------------------
; Name  : mrZoom::mrDisplayEvtHandler
; Syntax: obj->mrDisplayEvtHandler, evt
; Args  : <b>evt</b>
;         Pointer to the mrDisplay event structure
; Desc  : Dispatch the event coming from mrDisplay
;=-----------------------------------------------------------------------------
pro mrZoom::mrdisplayEvtHandler, evt

		; button is a bitmap holding the pressed/released button for such
		; events or a bitmap of currently pressed buttons for motion events
		; As of IDL6, 'true' and 'false' are finally reasonably defined.
		; But we want to run on 5.x, too...
		; And then there's the ingenious idea to have the same operator
		; for bitmap and logical 'AND'...
		; hate it or despise it - you can't love it.

	if( ((*evt).button and self.button) ne self.button ) then return
	case (*evt).type of
		0:	self->buttonPressEvt,    evt
		1:	self->buttonDblClickEvt, evt
		2:	self->buttonReleaseEvt,  evt
		3:	self->buttonMotionEvt,   evt
		else: return
	end
end

;+=----------------------------------------------------------------------------
; Name  : mrZoom::buttonPressEvt
; Syntax: obj->buttonPressEvt, evt
; Args  : <b>evt</b>
;         Pointer to the mrDisplay event structure
; Desc  : Event handler for button press events
;         Prints the current contrast scaling vals in the lower right corner
;         Interpolation of the display object is switched off while zooming
;=-----------------------------------------------------------------------------
pro mrZoom::buttonPressEvt, evt
	if( self.border and (*evt).x lt 0.85 * self.display->screenXsize() ) then return
	self->update
	self.active = 1
	self.grZModel->setProperty, hide=0
	self.interpolate = self.display->interpolate()
	self.display->setInterpolate, 0
	self.display->requestDisplay
end

;+=----------------------------------------------------------------------------
; Name  : mrZoom::buttonDblClickEvt
; Syntax: obj->buttonDblClickEvt, evt
; Args  : <b>evt</b>
;         Pointer to the mrDisplay event structure
; Desc  : Fit image to display screen size on double click
;=-----------------------------------------------------------------------------
pro mrZoom::buttonDblClickEvt, evt
	if( self.border and (*evt).x lt 0.85 * self.display->screenXsize() ) then return
	self.display->fitToScreen
	self->update
	self.display->requestDisplay
end

;+=----------------------------------------------------------------------------
; Name  : mrZoom::buttonReleaseEvt
; Syntax: obj->buttonReleaseEvt, evt
; Args  : <b>evt</b>
;         Pointer to the mrDisplay event structure
; Desc  : Event handler for button release events
;=-----------------------------------------------------------------------------
pro mrZoom::buttonReleaseEvt, evt
	self.active=0
	self.grZModel->setProperty, hide=1
	self.display->setInterpolate, self.interpolate
	self.display->requestDisplay
end

;+=----------------------------------------------------------------------------
; Name  : mrZoom::buttonMotionEvt
; Syntax: obj->buttonMotionEvt, evt
; Args  : <b>evt</b>
;         Pointer to the mrDisplay event structure
; Desc  : Event handler for motion events while a button is pressed
;=-----------------------------------------------------------------------------
pro mrZoom::buttonMotionEvt, evt
	if( self.active eq 0 ) then return
	sc = self.display->scaling()
		; reducing in size is a bit more sensitive to mouse-movement
	if( sc lt 1. ) then             $
		sc = sc + .7 * self.sensitivity * (*evt).dy    $
	else                            $
		sc = sc + 2. * self.sensitivity * (*evt).dy
	if( sc gt 10. ) then sc = 10.
	if( sc lt .01 ) then sc = .01
	self->update
	self.display->setScaling, sc
	self.display->requestDisplay
end


;==============================================================================
;
;                             PRIVATE OBJECT METHODS
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mrZoom::update
; Syntax: obj->update
; Desc  : get scaling from display object and set text
;=-----------------------------------------------------------------------------
pro mrZoom::update
	if( not obj_valid( self.display ) ) then return
	sc = self.display->scaling()
	s = 'Zoom: ' + string(sc, format='(f5.2)')
	self.grZText->setProperty, strings=s
end

;+=----------------------------------------------------------------------------
; Name  : mrZoom__define
; Type  : Private Procedure (constructor)
;=-----------------------------------------------------------------------------
pro mrZoom__define
	obj = { MRZOOM, $
			display		:	obj_new()	,$ ; the display object we're associated with
			button		:	0			,$ ; button bitmap on which we react (only one bit set)
			border		:	0			,$ ; sensitive only in right 15% of screen
			active      :   0           ,$ ; needed for border operation
			sensitivity :   0.0         ,$ ; sensitivity on mouse movements
			inscreen	:	0			,$ ; flag: display zoom factor in mrdisplay object?
			interpolate	:	0 			,$ ; intermediate storage between button press/release
			grZModel	:	obj_new() 	,$
			grZText		:	obj_new()	 $
		}
end
