;+=----------------------------------------------------------------------------
; Name  : vcolorbar
; RCS   : $Id: vcolorbar__define.pro,v 2.0 2004/08/05 18:12:02 olivert Exp olivert $
; Desc  : The purpose of this program is to create a vertical
;         colorbar object to be used in conjunction with other
;         IDL 5 graphics objects.
; Author: (c) FANNING SOFTWARE CONSULTING
;         David Fanning, Ph.D.
;         1645 Sheely Drive
;         Fort Collins, CO 80526 USA
;         Phone: 970-221-0438
;         E-mail: davidf@dfanning.com
;         Coyote's Guide to IDL Programming: http://www.dfanning.com
;         Included with kind permission by the author.
;
;         Oliver, 14.10.02 - adapted to our doc style
; Licns : This program is published under the GPL. See gpl.txt for details
;=-----------------------------------------------------------------------------

;==============================================================================
;
;                          CONSTRUCTORS AND DESTRUCTORS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name : vcolorbar::init
; Type : Function (constructor)
; Args : COLOR  : A three-element array representing the RGB values of a color
;                 for the colorbar axes and annotation. The default value is
;                 white: [255,255,255].
;        NAME   : The name associated with this object.
;        NCOLORS: The number of colors associated with the colorbar. The
;                 default is 256.
;        MAJOR  : The number of major tick divisions on the colorbar axes.
;                 The default is 5.
;        MINOR  : The number of minor tick marks on the colorbar axes.
;                 The default is 4.
;        PALETTE: A palette object for the colorbar. The default palette
;                 is a gray-scale palette object.
;        POSITION: A four-element array specifying the position of the
;                 colorbar in the arbitary coordinate system of the viewplane
;                 rectangle. The default position is [0.90, 0.10, 0.95, 0.90].
;        RANGE  : The range associated with the colorbar axis. The default
;                 is [0, NCOLORS].
;        TITLE  : A string containing a title for the colorbar axis
;                 annotation. The default is a null string.
; Ret  : 1 or 0 for true/false. For '0', IDL will return the null-object
; Desc : A VCOLORBAR object is created. The colorbar INHERITS IDLgrMODEL.
;        Thus, all IDLgrMODEL methods and keywords can also be used. It is
;        the model that is selected in a selection event, since the SELECT_TARGET
;        keyword is set for the model.
;=-----------------------------------------------------------------------------
function VColorBar::INIT, Position=position, $
    NColors=ncolors, Title=title, Palette=palette, $
    Major=major, Minor=minor, Range=range, Color=color, $
    _Extra=extra, Name=name

     ; Catch possible errors.

   Catch, theError
   IF theError NE 0 THEN BEGIN
      Catch, /Cancel
      ok = Dialog_Message(!Error_State.Msg)
      Message, !Error_State.Msg, /Informational
      RETURN, 0
   ENDIF
   forward_function vcolorbar_normalize

        ; Initialize model superclass.

   IF (self->IDLgrModel::Init(_EXTRA=extra) NE 1) THEN RETURN, 0

      ; Define default values for keywords, if necessary.

   IF N_Elements(name) EQ 0 THEN name=''
   IF N_Elements(color) EQ 0 THEN self.color = [255,255,255] $
      ELSE self.color = color
   thisFont = Obj_New('IDLgrFont', 'Helvetica', Size=8.0)
   self.thisFont = thisFont
   IF N_Elements(title) EQ 0 THEN title=''

   thisTitle = Obj_New('IDLgrText', title, Color=self.color, $
       Font=thisFont, Recompute_Dimensions=2, /Enable_Formatting, _Extra=extra)

   IF N_Elements(ncolors) EQ 0 THEN self.ncolors = 256 $
   ELSE self.ncolors = ncolors
   IF N_Elements(palette) EQ 0 THEN BEGIN
       red = (green = (blue = BIndGen(self.ncolors)))
       self.palette = Obj_New('IDLgrPalette', red, green, blue)
   ENDIF $
   ELSE self.palette = palette
   IF N_Elements(range) EQ 0 THEN self.range = [0, self.ncolors] $
   ELSE self.range = range

   IF N_Elements(major) EQ 0 THEN self.major = 5 $
   ELSE self.major = major

   IF N_Elements(minor) EQ 0 THEN self.minor = 4 $
   ELSE self.minor = minor

   IF N_Elements(position) EQ 0 THEN self.position = [0.90, 0.10, 0.95, 0.90] $
   ELSE self.position = position

     ; Create the colorbar image. Get its size.

   bar = REPLICATE(1B,10) # BINDGEN(self.ncolors)
   s = SIZE(bar, /Dimensions)
   xsize = s[0]
   ysize = s[1]

     ; Create the colorbar image object. Add palette to it.

   thisImage = Obj_New('IDLgrImage', bar, Palette=self.palette)
   xs = vcolorbar_normalize([0,xsize], Position=[0,1.])
   ys = vcolorbar_normalize([0,ysize], Position=[0,1.])
   thisImage->SetProperty, XCoord_Conv=xs, YCoord_Conv=ys

     ; Create a polygon object. Add the image as a texture map. We do
        ; this so the image can rotate in 3D space.

   thisPolygon = Obj_New('IDLgrPolygon', [0, 1, 1, 0], [0, 0, 1, 1], [0,0,0,0], $
               Texture_Map=thisImage, $
               Texture_Coord = [[0,0], [1,0], [1,1], [0,1]], $
               color=[255,255,255])
   self.thisPolygon = thisPolygon

     ; Scale the polygon into the correct position.

   xs = vcolorbar_normalize([0,1], Position=[self.position(0), self.position(2)])
   ys = vcolorbar_normalize([0,1], Position=[self.position(1), self.position(3)])
   thispolygon->SetProperty, XCoord_Conv=xs, YCoord_Conv=ys

     ; Create scale factors to position the axes.

   longScale  = vcolorbar_normalize(self.range, Position=[self.position(1), self.position(3)])
   shortScale = vcolorbar_normalize([0,1], Position=[self.position(0), self.position(2)])

     ; Create the colorbar axes. 1000 indicates this location ignored.

   shortAxis1 = Obj_New("IDLgrAxis", 0, Color=self.color, Ticklen=0.025, $
               Major=1, Range=[0,1], /NoText, /Exact, XCoord_Conv=shortScale,  $
              Location=[1000, self.position(1), 0.001])
   shortAxis2 = Obj_New("IDLgrAxis", 0, Color=self.color, Ticklen=0.025, $0.001
             Major=1, Range=[0,1], /NoText, /Exact, XCoord_Conv=shortScale,  $
             Location=[1000, self.position(3), 0.001], TickDir=1)

   textAxis = Obj_New("IDLgrAxis", 1, Color=self.color, Ticklen=0.025, $
             Major=self.major, Minor=self.minor, Title=thisTitle, $
             Range=self.range, /Exact, $
    YCoord_Conv=longScale, Location=[self.position(0), 1000, 0.001], _Extra=extra)
   textAxis->GetProperty, TickText=thisText
   thisText->SetProperty, Font=self.thisFont, Recompute_Dimensions=2

   longAxis2 = Obj_New("IDLgrAxis", 1, Color=self.color, /NoText, Ticklen=0.025, $
             Major=self.major, Minor=self.minor, Range=self.range, TickDir=1, $
            YCoord_Conv=longScale, Location=[self.position(2), 1000, 0.001], /Exact)

     ; Add the parts to the colorbar model.

   self->Add, shortAxis1
   self->Add, shortAxis2
   self->Add, textAxis
   self->Add, longAxis2
   self->Add, thisPolygon

     ; Assign the name.


   self->IDLgrModel::SetProperty, Name=name, Select_Target=1

     ; Create a container object and put the objects into it.

   thisContainer = Obj_New('IDL_Container')
   thisContainer->Add, thisFont
   thisContainer->Add, thisImage
   thisContainer->Add, thisText
   thisContainer->Add, thisTitle
   thisContainer->Add, self.palette
   thisContainer->Add, textAxis
   thisContainer->Add, shortAxis1
   thisContainer->Add, shortAxis2
   thisContainer->Add, longAxis2

     ; Update the SELF structure.

   self.thisImage     = thisImage
   self.thisFont      = thisFont
   self.thisText      = thisText
   self.textAxis      = textAxis
   self.shortAxis1    = shortAxis1
   self.shortAxis2    = shortAxis2
   self.longAxis2     = longAxis2
   self.thisContainer = thisContainer
   self.thisTitle     = thisTitle

   RETURN, 1
end

;+=----------------------------------------------------------------------------
; Name : vcolorbar::cleanup
; Type : Procedure (destructor)
; Args :
; Desc :
;=-----------------------------------------------------------------------------
PRO VColorBar::Cleanup
   Obj_Destroy, self.thisContainer
   self->IDLgrMODEL::Cleanup
END


;==============================================================================
;
;                               GET AND SET METHODS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name : vcolorbar::getProperty
; Type : Procedure
; Desc : Returns colorbar properties in keyword
;        parameters as defined for the INIT method. Keywords allowed are:
;          COLOR
;          MAJOR
;          MINOR
;          NAME
;          PALETTE
;          POSITION
;          RANGE
;          TITLE
;          TRANSFORM
;=-----------------------------------------------------------------------------
pro VColorBar::GetProperty, Position=position, Text=text, $
    Title=title, Palette=palette, Major=major, Minor=minor, $
    Range=range, Color=color, Name=name, $
    Transform=transform, _Ref_Extra=extra

     ; Get the properties of the colorbar.

   IF Arg_Present(position) THEN position = self.position
   IF Arg_Present(text) THEN text = self.thisText
   IF Arg_Present(title) THEN self.thisTitle->GetProperty, Strings=title
   IF Arg_Present(palette) THEN palette = self.palette
   IF Arg_Present(major) THEN major = self.major
   IF Arg_Present(minor) THEN minor = self.minor
   IF Arg_Present(range) THEN range = self.range
   IF Arg_Present(color) THEN color = self.color
   IF Arg_Present(name) THEN self->IDLgrMODEL::GetProperty, Name=name
   IF Arg_Present(transform) THEN self->IDLgrMODEL::GetProperty, Transform=transform
   IF Arg_Present(extra) THEN self->IDLgrMODEL::GetProperty, _Extra=extra
end

;+=----------------------------------------------------------------------------
; Name : vcolorbar::setProperty
; Type : Procedure
; Desc : Sets colorbar properties in keyword
;        parameters as defined for the INIT method. Keywords allowed are:
;          COLOR
;          NAME
;          MAJOR
;          MINOR
;          PALETTE
;          POSITION
;          RANGE
;          TITLE
;          TRANSFORM
;=-----------------------------------------------------------------------------
PRO VColorBar::SetProperty, Position=position, $
    Title=title, Palette=palette, Major=major, Minor=minor, $
    Range=range, Color=color, Name=name, Transform=transform, _Extra=extra

   forward_function vcolorbar_normalize
   IF N_Elements(position) NE 0 THEN BEGIN
      self.position = position

        ; Move the image polygon into its new positon.

      xs = vcolorbar_normalize([0,1], Position=[position(0), position(2)])
      ys = vcolorbar_normalize([0,1], Position=[position(1), position(3)])
      self.thisPolygon->SetProperty, XCoord_Conv=xs, YCoord_Conv=ys

      ; Create new scale factors to position the axes.

      longScale  = vcolorbar_normalize(self.range, Position=[self.position(1), self.position(3)])
      shortScale = vcolorbar_normalize([0,1], Position=[self.position(0), self.position(2)])

      ; Position the axes. 1000 indicates this position ignored.

      self.textaxis->SetProperty, YCoord_Conv=longScale, $
               Location=[self.position(0), 1000, 0]
      self.longaxis2->SetProperty, YCoord_Conv=longScale, $
               Location=[self.position(2), 1000, 0]
     self.shortAxis1->SetProperty, XCoord_Conv=shortScale, $
                Location=[1000, self.position(1), 0]
      self.shortAxis2->SetProperty, XCoord_Conv=shortScale, $
                Location=[1000, self.position(3), 0]

   end
   IF N_Elements(title) NE 0 THEN self.thisTitle->SetProperty, Strings=title
   IF N_Elements(transform) NE 0 THEN self->IDLgrMODEL::SetProperty, Transform=transform
   IF N_Elements(palette) NE 0 THEN BEGIN
       self.palette = palette
      self.thisImage->SetProperty, Palette=palette
   end
   IF N_Elements(major) NE 0 THEN BEGIN
       self.major = major
       self.textAxis->SetProperty, Major=major
       self.longAxis2->SetProperty, Major=major
   END
   IF N_Elements(minor) NE 0 THEN BEGIN
       self.minor = minor
       self.textAxis->SetProperty, Minor=minor
       self.longAxis2->SetProperty, Minor=minor
   END
   IF N_Elements(range) NE 0 THEN BEGIN
       self.range = range
       longScale = vcolorbar_normalize(range, $
          Position=[self.position(1), self.position(3)])
       self.textAxis->SetProperty, Range=range, YCoord_Conv=longScale
       self.longAxis2->SetProperty, Range=range, YCoord_Conv=longScale
   ENDIF
   IF N_Elements(color) NE 0 THEN BEGIN
       self.color = color
       self.textAxis->SetProperty, Color=color
       self.longAxis2->SetProperty, Color=color
       self.shortAxis1->SetProperty, Color=color
       self.shortAxis2->SetProperty, Color=color
       self.thisText->SetProperty, Color=color
   ENDIF
   IF N_Elements(name) NE 0 THEN self->IDLgrMODEL::SetProperty, Name=name
   IF N_Elements(extra) NE 0 THEN BEGIN
      self->IDLgrMODEL::SetProperty, _Extra=extra
      self.textAxis->SetProperty, _Extra=extra
      self.thisTitle->SetProperty, _Extra=extra
   ENDIF
END

;==============================================================================
;
;                             PUBLIC OBJECT FUNCTIONS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name : vcolorbar::clamp
; Type : Procedure
; Args : datarange : two element array
; Desc : This method clamps the data to a particular data range:
;        Given a two-element array in the data range of
;        the colorbar, the colorbar image is clamped to this range. In
;        other words, the range of colors is clamped to the specified
;        range. Values above or below the range in the colorbar are set to
;        the minimum and maximum range values, respectively.
;=-----------------------------------------------------------------------------
pro VColorBar::Clamp, datarange
   self->GetProperty, Range=currentRange
   thisclamp = Bytscl(datarange, Max=currentRange[1], Min=currentRange[0])
   if( thisclamp[0] eq thisclamp[1] ) then thisclamp[1]=thisclamp[0]+1
   bar = BytScl(Replicate(1B,10) # Bindgen(self.ncolors), Min=thisclamp[0], Max=thisclamp[1])
   self.thisImage->SetProperty, Data=bar
end

;==============================================================================
;
;                          PRIVATE OBJECT FUNCTIONS
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name : vcolorbar_normalize
; Type : Function
; Ret  : the scaling factor
; Desc : This is a utility function to calculate the scale factor
;        required to position a vector of specified range at a
;        specific position given in normalized coordinates.
;=-----------------------------------------------------------------------------
function vcolorbar_normalize, range, Position=position
   IF (N_Elements(position) EQ 0) THEN position = [0.0, 1.0]
   scale = [((position[0]*range[1])-(position[1]*range[0])) / $
       (range[1]-range[0]), (position[1]-position[0])/(range[1]-range[0])]
   return, scale
end

;+=----------------------------------------------------------------------------
; Name : vcolorbar__define
; Type : Private procedure
; Desc : the constructor, inherits the IDLgrModel object
;=-----------------------------------------------------------------------------
pro vcolorbar__define

colorbar = { VCOLORBAR, $
             INHERITS IDLgrMODEL, $      ; Inherits the Model Object.
             Position:FltArr(4), $       ; The position of the colorbar.
             Palette:Obj_New(), $        ; The colorbar palette.
             thisImage:Obj_New(), $      ; The colorbar image.
             imageModel:Obj_New(), $     ; The colorbar image model.
             thisPolygon:Obj_New(), $    ; The image polygon.
             thisContainer:Obj_New(), $  ; Container for cleaning up.
             thisFont:Obj_New(), $       ; The annotation font object.
             thisText:Obj_New(), $       ; The bar annotation text object.
             thisTitle: Obj_New(), $     ; The title of the colorbar.
             textAxis:Obj_New(), $       ; The axis containing annotation.
             shortAxis1:Obj_New(), $     ; A short axis.
             shortAxis2:Obj_New(), $     ; A second short axis.
             longAxis2:Obj_New(), $      ; The other long axis.
             NColors:0, $                ; The number of colors in the bar.
             Major:0, $                  ; Number of major axis intervals.
             Minor:0, $                  ; Number of minor axis intervals.
             Color:BytArr(3), $          ; Color of axes and annotation.
             Range:FltArr(2) }           ; The range of the colorbar axis.
end
