;+=----------------------------------------------------------------------------
; Name    : mod_calc
; RCS     : $Id: mod_calc__define.pro,v 1.1 2004/11/11 11:24:01 olivert Exp olivert $
; Desc    : This module contains some very basic calculation routines for
;           baseline averaging and concentration image calculation
;           Make sure you set the parameters correctly before calling the
;           calculation methods.
; Author  : (c) 2004 Oliver Thilmann
;=------------------------------------------------------------------------------
function mod_calc::revision
	rev = '$Revision: 1.1 $'
	return, rev
end

;+=----------------------------------------------------------------------------
; Name  : mod_calc::init
; Syntax: obj = obj_new('lupemod', type, exclusive )
; Args  : <b>type</b>
;         One of 'noUI', 'top', 'embedded' (case-insensitive)
; Desc  :
;=-----------------------------------------------------------------------------
function mod_calc::init, g
	return, self->lupemod::init( 'mod_calc', 'noui', 0, g )
end

;+=----------------------------------------------------------------------------
; Name  : mod_calc::readyToExecute
; Desc  : No matter which calculation method is called - a 'raw/signal'
;         image set has to be present.
;=-----------------------------------------------------------------------------
function mod_calc::readyToExecute, option
	i18n = (*self.g).i18n
	if( n_elements(option) eq 0 ) then return, 0

	x = mrGetImageSet( (*self.g).imgSets, class='raw', name='signal' )
	if( x eq obj_new() ) then begin
		x = dialog_message( i18n->get('err_noPerfusionSetLoaded'), title=i18n->get('Info') )
		return, 0
	end
	return, 1
end

;+=----------------------------------------------------------------------------
; Name  : mod_calc::execute
; Desc  :
;=-----------------------------------------------------------------------------
pro mod_calc::execute, option
	if( n_elements(option) eq 0 ) then return
	case option of
		'baseline': self->calcBaselineImgSet, /progressbar, /verbose
		'ctr_t2*' :	self->calcCtrImgsetFromT2star, /progressbar
		'ctr_t1'  :	x = dialog_message('Not yet implemented', /error )
		'ctr_ct'  : self->calcCtrImgsetFromCT, /progressbar
		else      : x = dialog_message('mod_calc::execute: unknown option >' + option +'<', /error )
	end
	return
end

;+=----------------------------------------------------------------------------
; Name  : mod_calc__define
; Desc  : Private Procedure (constructor)
;=-----------------------------------------------------------------------------
pro mod_calc__define
	obj = { MOD_CALC		   ,$
			inherits LUPEMOD    $
		  }
end



;+=----------------------------------------------------------------------------
; Name   : mod_calc::calcBaselineImgSet
; Desc   : Needs a valid raw-signal mrImageSet in the imgSets container
;          If no condensed raw_signal_xyt mrImageSet exists, this method creates one
;          If no raw-baseline mrImageSet exists, this method calculates one
;          This method also checks, whether the t0/t1 values of a possibly
;          existing raw-baseline set match the current preference values.
;          If not, the existing set is destroyed and a new one calculated.
;          If yes, nothing happens and the method returns immediately.
;          Input:  raw, signal
;          Output: raw, signal_xyt
;                  raw, baseline
;=-----------------------------------------------------------------------------
pro mod_calc::calcBaselineImgset, progressbar=pb, verbose=verbose

	verbose = keyword_set( verbose ) ; only needed once

	pb = ( n_elements(pb) eq 0 ) ? 0 : pb

	i18n = (*self.g).i18n

		; check for available datasets

	imgsets = (*self.g).imgSets
	imgset  = mrGetImageSet( imgsets, class='raw', name='signal' )
	timeset = mrGetImageSet( imgsets, class='raw', name='signal_xyt' )
	baseset = mrGetImageSet( imgsets, class='raw', name='baseline' )

	if( imgset eq obj_new() ) then return  ; no image set available???? Then how did we get here???

		; baseline values are averaged from t0 to t1 as defined in the preferences

	t0 = (*(*self.g).preferences).baseline_t0
	t1 = (*(*self.g).preferences).baseline_t1

		; check, whether the image set contains enough time points to calculate the baseline

	if( imgset->nshots() lt t1 ) then begin
		x = mrDialog_select([ i18n->get('calc_baselineErr1'), $
		                      i18n->get('calc_baselineErr2'), $
		                      i18n->get('calc_baselineErr3')],$
 		                      'Ok', title=i18n->get('Error'), /skull, group_leader=(*self.g).topid )
		return
	end

		; if no time condensed set is available, we need to create one and add to the imgSets container
		; dim 0 in image set is the slice number, dim 1 is the time
		; testing for existing baseline values is only done if the timeset did exist. Actually these two
		; conditions should be mutually exclusive anyway, as all perfusion data is destroyed @ loading,
		; but you never know.
		; Anyway, if a baseline set with correct t0/t1 parameters exists, return now

	if( timeset eq obj_new() ) then begin
		timeset = imgset->condense( 1 )
		timeset->setName,   'signal_xyt'
		timeset->setClass,  'raw'
		(*self.g).imgSets->add, timeset
	end else begin
		if( baseset ne obj_new() ) then begin
			times = *(baseset->uvalue())
			if( times[0] eq t0 and times[1] eq t1 ) then begin
				if( verbose ) then begin
					x = mrDialog_select(i18n->get('calc_baselineAlreadyDone'), /info, $
					                      'Ok', title=i18n->get('Information'), group_leader=(*self.g).topid )
				end
				return
			end
		end
	end

		; now we know that we'll calculate a new set

	if( pb ) then pbar = obj_new('mrprogressbar', title=i18n->get('calc_baselineImgs'), /floating, group_leader=(*self.g).topid )

		; ok, since we'll calculate a new one, we may delete any existing baseline set

	if( baseset ne obj_new() ) then begin
		(*self.g).imgSets->remove, baseset
		obj_destroy, baseset
	end

		; create new set

	gs = indgen(256)
	baseset = obj_new('mrImageSet', imgset->nSlices(), name='baseline', class='raw', red=gs, green=gs, blue=gs )

	timeimgs = timeset->imgData()

		; loop over all slices, create an image matrix, fill it and save it in baseset

	nslices = timeset->d0()
	for i = 0, nslices-1 do begin
		bimg = fltarr( imgset->xsize(), imgset->ysize() )
		for x = 0, imgset->xsize()-1 do begin
			for y = 0, imgset->ysize()-1 do begin
				bimg[x,y] = mean( (*(*timeimgs)[i])[x,y,t0:t1] )
			end
		end
		baseset->set, i, ptr_new( bimg, /no_copy ), /no_copy
		if( pb ) then pbar->update, 1.*(i+1)/nslices
	end
	if( pb ) then obj_destroy, pbar

		; set baseset uvalue = [t0,t1], so we can check next time

	baseset->setUvalue, ptr_new( [t0,t1], /no_copy )
	(*self.g).imgSets->add, baseset

		; and finally set the parent for the new created set

	baseset->setParent, imgSet
end

;+=----------------------------------------------------------------------------
; Name   : mod_calc::calcCtrImgsetFromT2star
; Desc   : Calculates the concentration images, following the simple
;          linear approach
;          Input:  raw, signal
;          Output: raw, signal_xyt (via calcBaselineImgset)
;                  raw, baseline  (via calcBaselineImgSet)
;                  raw, ctr_xyt
;=-----------------------------------------------------------------------------
pro mod_calc::calcCtrImgsetFromT2star, progressbar=pb

	pb = ( n_elements(pb) eq 0 ) ? 1 : pb

	i18n = (*self.g).i18n

		; make sure that baseline images exist and correspond to current
		; preference settings

	self->calcBaselineImgset, progressbar=pb

		; check whether concentration images exist, if so: destroy the object

	imgSets   = (*self.g).imgSets
	ctr       = mrGetImageSet( imgSets, class='raw', name='ctr_xyt'    )
	sigxyt    = mrGetImageSet( imgSets, class='raw', name='signal_xyt' )
	baseline  = mrGetImageSet( imgSets, class='raw', name='baseline'   )
	signal    = mrGetImageSet( imgSets, class='raw', name='signal'     )

	if( baseline eq obj_new() ) then return ; baseline calculation failed

	if( ctr ne obj_new() ) then begin
		imgsets->remove, ctr
		obj_destroy, ctr
	end

		; check whether the data contains a valid TE. Warn the user (if preferences)
		; if we have to use the default value.

	dicom = (*(signal->dicomData()))[0,0]
	te    = dicom->getTagValue('acq_echo_time')
	if( te eq '' ) then begin
		if( (*(*self.g).preferences).default_warn eq 1 ) then begin
			x = dialog_message( [i18n->get('calc_TEDefaultWarning1'),i18n->get('calc_TEDefaultWarning2')], title=i18n->get('Info'), /info )
		end
		te = (*(*self.g).preferences).default_te
	end

		; ok, now we may do the calculation.

	sigthres = (*(*self.g).preferences).ssvd_sigthres
	ctrthres = (*(*self.g).preferences).ssvd_ctrthres
	nslices  = signal->d0()
	nshots   = signal->d1()
	xsize    = signal->xsize()
	ysize    = signal->ysize()
	ctr      = obj_new('mrImageSet', nslices, class='raw', name='ctr_xyt', xsize=xsize, ysize=ysize )
	ctrset   = ctr->imgData()

	if( pb ) then pbar = obj_new('mrprogressbar', title=i18n->get('calc_concentrationImgs'))

	for i = 0, nslices-1 do begin
		base = (*(baseline->imgData()))[i]
		sig  = (*(sigxyt->imgData()))[i]
		res  = fltarr( xsize, ysize, nshots )
		for j = 0, nshots-1 do begin
			res[*,*,j] = (1000./te * alog( (((*base)[*,*] / (*sig)[*,*,j]) > 1.)  ) * ( (*base)[*,*] gt sigthres )) > 0.
		end
		res = res < ctrthres
		(*ctrset)[i] = ptr_new( res, /no_copy )
		if( pb ) then pbar->update, 1.*(i+1)/nslices
	end
	if( pb ) then obj_destroy, pbar

	imgSets->add, ctr

	return
end


;+=----------------------------------------------------------------------------
; Name   : mod_calc::calcCtrImgsetFromCT
; Desc   : Calculates concentration images from a CT image set
;=-----------------------------------------------------------------------------
pro mod_calc::calcCtrImgsetFromCT, progressbar=pb

	pb = (n_elements(pb) eq 1 ) ? pb : 0

	i18n = (*self.g).i18n

		; as above

	self->calcBaselineImgset, progressbar=pb

	imgSets   = (*self.g).imgSets
	ctr       = mrGetImageSet( imgSets, class='raw', name='ctr_xyt'    )
	sigxyt    = mrGetImageSet( imgSets, class='raw', name='signal_xyt' )
	baseline  = mrGetImageSet( imgSets, class='raw', name='baseline'   )
	signal    = mrGetImageSet( imgSets, class='raw', name='signal'     )

	if( baseline eq obj_new() ) then return ; baseline calculation failed

	if( ctr ne obj_new() ) then begin
		imgsets->remove, ctr
		obj_destroy, ctr
	end

		; perform the calculation

	if( pb ) then pbar = obj_new('mrprogressbar', title=i18n->get('calc_concentrationImgs'))

	nslices = signal->d0()
	nshots  = signal->d1()
	xsize    = signal->xsize()
	ysize    = signal->ysize()
	ctr      = obj_new('mrImageSet', nslices, class='raw', name='ctr_xyt', xsize=xsize, ysize=ysize )
	ctrset   = ctr->imgData()

	for i = 0, nslices-1 do begin
		if( i eq 5 ) then begin
			print, 'stop'
		end
		base = (*(baseline->imgData()))[i]
		sig  = (*(sigxyt->imgData()))[i]
		res  = fltarr( xsize, ysize, nshots )
		for j = 0, nshots-1 do begin
			res[*,*,j] = ((*sig)[*,*,j] - (*base)[*,*] ) > 0.
		end
		(*ctrset)[i] = ptr_new( res, /no_copy )
		if( pb ) then pbar->update, 1.*(i+1)/nslices
	end
	if( pb ) then obj_destroy, pbar
	imgSets->add, ctr
end

