;+=----------------------------------------------------------------------------
; Name    : mod_colorcode
; RCS     : $Id: mod_colorcode__define.pro,v 1.3 2004/11/25 20:09:58 olivert Exp olivert $
; Desc    : Module for color coding of result images.
;           There's two ways:
;           - replace the greyscale color map by any of the predefined IDL color
;             tables
;           - application of a logarithmic color scale. The user needs to
;             define the normalization value by ROI placement. Besides he
;             may set a multiplication factor which defines how the scale
;             'stretches'
;           Color coding always applies to the full set - a preview of the
;           color encoded set is displayed in a mrMultidisplay object at the
;           lower third of the screen.
;           The color encoded image sets are stored under the same name,
;           but with class specifier 'color'
; Author  : (c) 2004 Oliver Thilmann
;=------------------------------------------------------------------------------
function mod_colorcode::revision
	rev = '$Revision: 1.3 $'
	return, rev
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::init
; Syntax: obj = obj_new('lupemod', type, exclusive )
; Args  :
; Desc  :
;=-----------------------------------------------------------------------------
function mod_colorcode::init, g

    res = self->lupemod::init( 'mod_colorcode', 'embedded', 0, g )
    if( res eq 0 ) then return, 0

	self.currSource  = ''
	self.currScheme  = 0   ; start logarithmically :)
	self.currSlice   = -1
	self.interpolate = 0
	self.dispDicom   = 0
	self.contUpdate  = 0
	self.logfactor   = 1.
	self.roigroups   = ptr_new()

	return, 1
end


;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::readyToExecute
; Desc  : Need image sets with class 'result' for execution
;=-----------------------------------------------------------------------------
function mod_colorcode::readyToExecute, option
	i18n = (*self.g).i18n
	sets = mrGetImageSet( (*self.g).imgSets, class='result', count=nsets )
	if( nsets eq 0 ) then begin
		x = dialog_message( i18n->get('err_noResultImageSet'), title=i18n->get('Info'), /info )
		return, 0
	end
	return, 1
end


;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::map
; Desc  : Embedded module.
;=-----------------------------------------------------------------------------
pro mod_colorcode::map, option

	i18n = (*self.g).i18n

		; create the base widget

	res = self->createBaseWidget( /column )
	if( res eq 0 ) then return

		; calculate layout of the multidisplay - IDL does not allow for more than
		; 2048x2048 pixels canvas size, so when there's more than 16 images, we use
		; a layout of 10 times x, otherwise it's 2 rows
		; All result images should have the same number of slices, but we check'em
		; all anyway and use the largest
		; We won't (because we can't) update that later, though. So if the user
		; creates a result set with more slices, while mod_colorcode is displayed
		; we're lost. But what's the odds?

	nslices = 0
	sets = mrGetImageSet( (*self.g).imgSets, class='result', count=nsets )
	for i = 0, nsets-1 do nslices = nslices > sets[i]->nslices()
	self.nx = fix( nslices/2. + .5 )
	self.ny = 2
	if( self.nx gt 16 ) then begin
		self.nx = 10
		self.ny = fix( nslices/10. + .99 )
	end
		; upper block: t1, lower block (only multidisplay): t2

	t1 = widget_base( self.baseid, /row )
	t2 = widget_base( self.baseid, /column, /align_center, /frame)
	self.multidisplay = obj_new('mrmultidisplay', t2, 996, 256, 128, 128, self.nx, self.ny )

		; upper block left: t2 (display, slider, roi)

	t2 = widget_base( t1, /frame, /row )
	self.display       = obj_new('mrdisplay', t2, 384, 384, interpolate=self.interpolate )
	self.idSliceslider = widget_slider( t2, xsize=30, ysize=256, /vertical, uvalue={o:self, m:'sliceSelectEvt'} )
	self.roiui         = obj_new('mrroiui', t2, /column, ysize=256, spacing=2 )

		; upper block right: t3 with frame

	t3 = widget_base( t1, /column, /frame )
    x  = widget_label( t3, value=i18n->get('colc_title'), font='fixedsys', /align_center, /sunken_frame, frame=1, xsize=531)


		; the four frames on the right side: t4/t5/t6

	x  = widget_base ( t3, xsize=10, ysize=3 ) ; vertical spacer

	t4 = widget_base ( t3, /row )
	x  = widget_base ( t4, xsize=10, ysize = 10 ) ; horizontal spacer
	t5 = widget_base ( t4, /column )
	x  = widget_label( t5, value=i18n->get('ImageSets')+':', /align_left )
	t6 = widget_base ( t5, /column, /frame, xsize=230, ysize=125 )
	t7 = widget_base ( t6, /row )
	x  = widget_label( t7, scr_xsize=90, value=i18n->get('SourceClass')+':')
	x  = widget_text ( t7, scr_xsize=120, value='result' )
	t7 = widget_base ( t6, /row )
	x  = widget_label( t7, scr_xsize=90, value=i18n->get('SourceName')+':')
	self.idSourcelist = widget_droplist( t7, scr_xsize=120, xsize=120, uvalue={o:self, m:'sourceNameSelectEvt'} )
	t7 = widget_base ( t6, /row )
	x  = widget_label( t7, scr_xsize=90, value=i18n->get('TargetClass')+':' )
	x  = widget_text ( t7, scr_xsize=120, value='color' )
	t7 = widget_base ( t6, /row )
	x  = widget_label( t7, scr_xsize=90, value=i18n->get('TargetName')+':' )
	self.idTargetname = widget_text ( t7, scr_xsize=120, /editable, uvalue={o:self, m:'targetNameSelectEvt'} )

	x  = widget_base ( t4, xsize=20, ysize = 10 ) ; horizontal spacer

	t5 = widget_base ( t4, /column )
	x  = widget_label( t5, value=i18n->get('Display')+':', /align_left )
	t6 = widget_base ( t5, /column, /frame, xsize=230, ysize=125 )
	t7 = widget_base  ( t6 , /row, ysize=30 )
	x  = widget_label ( t7, value=i18n->get('Interpolate')+':', xsize=190 )
	t8 = widget_base  ( t7, /row, /nonexclusive )
	x  = widget_button( t8, value='', uvalue={o:self, m:'interpolateEvt'} )
	if( self.interpolate ) then widget_control, x, /set_button
	t7 = widget_base  ( t6 , /row, ysize=30 )
	x  = widget_label ( t7, value=i18n->get('colc_continuousUpdate'), xsize=190 )
	t8 = widget_base  ( t7, /row, /nonexclusive )
	self.idContUpdate = widget_button( t8, value='', uvalue={o:self, m:'contUpdateButtonEvt'} )
	if( self.contUpdate ) then widget_control, x, /set_button
	t7 = widget_base  ( t6 , /row, ysize=30 )
	x  = widget_label ( t7, value=i18n->get('colc_displayDicomInfo'), xsize=190 )
	t8 = widget_base  ( t7, /row, /nonexclusive )
	self.idDicom      = widget_button( t8, value='', uvalue={o:self, m:'dicomViewButtonEvt'} )
	if( self.dispDicom ) then widget_control, x, /set_button
	t7 = widget_base  ( t6 , /row, ysize=30 )
	x  = widget_label ( t7, value=i18n->get('Button3')+':', xsize=95 )
	t8 = widget_base  ( t7, /exclusive, /row )
	x  = widget_button( t8, value=i18n->get('C/W'), uvalue={o:self, m:'contrastModeEvt'} )
	y  = widget_button( t8, value=i18n->get('Gamma'), uvalue={o:self, m:'gammaModeEvt'} )
 	widget_control, x, /set_button

	x  = widget_base ( t3, xsize=10, ysize=3 ) ; vertical spacer

	t4 = widget_base ( t3, /row )
	x  = widget_base ( t4, xsize=10, ysize=10 ) ; horizontal spacer

	t5 = widget_base ( t4, /column )
	x  = widget_label( t5, value=i18n->get('colc_encoding'), /align_left )
	t6 = widget_base ( t5, /column, /frame, xsize=230, ysize=70 )
	t7 = widget_base ( t6, /row )
	x  = widget_label( t7, scr_xsize=90, value=i18n->get('colc_colorscheme') )
	self.idColorlist = widget_droplist( t7, scr_xsize=120, xsize=120, uvalue={o:self, m:'colorSchemeSelectEvt'} )
	t7 = widget_base ( t6, /row )
	self.idLogfactor = mrwidget_numfield( t7, /float, min=.5, max=10, incr=.5, /editable, labeltext=i18n->get('colc_logStretchfactor'), $
	                   value=self.logfactor, xsize=210, labelsize=158, fieldsize=30, uvalue={o:self, m:'logScaleSelectEvt'}, format='(f4.1)' )

	x  = widget_base ( t4, xsize=20, ysize=10 ) ; horizontal spacer

	t5 = widget_base ( t4, /column )
	x  = widget_label( t5, value='Rois:', /align_left )
	t6 = widget_base ( t5, /column, /frame, xsize=230, ysize=70 )
	t7 = widget_base ( t6, /row )
	x  = widget_label( t7, scr_xsize=90, value=i18n->get('RoiMeanValue')+':' )
	self.idMeanRoi = widget_text ( t7, scr_xsize=120, value='0.' )
	t7 = widget_base ( t6, /row )
	self.idStoreRoi   = widget_button(t7, value=i18n->get('Store'),     xsize=70, uvalue={o:self, m:'storeRoisEvt'} )
	self.idRecallRoi  = widget_button(t7, value=i18n->get('Recall'),    xsize=70, uvalue={o:self, m:'recallRoisEvt'} )
	self.idDeleteRois = widget_button(t7, value=i18n->get('DeleteAll'), xsize=70, uvalue={o:self, m:'deleteAllRoisEvt'} )

		; the contrast bar sits in an extra frame

	x  = widget_base ( t3, xsize=10, ysize=3 ) ; vertical spacer

	t4 = widget_base ( t3, /row )
	x  = widget_base ( t4, xsize=10, ysize=10 ) ; horizontal spacer
	t5 = widget_base ( t4, /column )
	x  = widget_label( t5, value=i18n->get('colc_selectedEncoding'), /align_left )
	t6 = widget_base ( t5, /row, /frame, xsize=494, ysize=60 )
	self.contrast2 = obj_new('mrcontrast', t6, xsize=488, ysize=55 )

		; realize the widget and register with the xmanager

	self->placeAndRegister

		; final registration, settings and then display an image

	loadct, get_names=colorschemes
	widget_control, self.idColorlist, set_value= [ i18n->get('Logarithmic'), colorschemes]

	self.contrast1 = obj_new('mrcontrast')
	self.contrast1->selfregister, self.display, button=3
	self.display->activate
	self.contrast2->selfregister, self.multidisplay, button=0 ; no user interaction here!
	self.multidisplay->activate
	self.currSlice  = 0
	self.currScheme = 0
	self.roiMean    = 0.
	self->update
	(*self.g).imgSets->notify, self, 'update', ['add','remove']
	self.roiui->setLoadMethod, self, 'load'
	self.roiui->setSaveMethod, self, 'save'
	self.roiui->notify, self, 'roiChangeEvt', ['change', 'modify']

end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::unmap
; Syntax:
; Args  : <b>killed</b>
;         If the user killed the window via Ctrl-F4 or the close handle (the
;         cross at the upper right), killed is set to '1'. In this case it
;         might be (I have not tested this) that the system's window manager
;         has already removed the window from screen.
; Desc  :
;=-----------------------------------------------------------------------------
pro mod_colorcode::unmap, _ref_extra=ex

	if( self.baseid eq 0L ) then return  ; uups - we don't exist :( Shouldn't happen

		; 'reset' module, remove from notification list

	self.currSource = ''
	(*self.g).imgSets->notify, self, /remove

		; destroy the module - taken care off by the superclass

	self->lupemod::unmap, _strict_extra=ex
	obj_destroy, self.contrast1
	obj_destroy, self.contrast2
	obj_destroy, self.roiui
	obj_destroy, self.display
	obj_destroy, self.multidisplay

		; destroy the saved roigroups

	; !!! TODO !!!
end



;==============================================================================
;
;                           E V E N T   H A N D L I N G
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::sourceNameSelectEvt
; Desc  :
;=-----------------------------------------------------------------------------
pro mod_colorcode::sourceNameSelectEvt, evt
	widget_control, self.idSourcelist, get_value=list
	self.currSource = list[evt.index]
	self->imgSetChange
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::targetNameSelectEvt
; Desc  :
;=-----------------------------------------------------------------------------
pro mod_colorcode::targetNameSelectEvt, evt
	print, 'targetNameSelectEvt called'
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::sliceSelectEvt
; Desc  :
;=-----------------------------------------------------------------------------
pro mod_colorcode::sliceSelectEvt, evt

		; if no ROIs are displayed, this is simple

	if( self.currScheme ne 0 ) then begin
		self.currSlice = evt.value
		self->display, /update
		return
	end

		; Otherwise we got to do some ROI handling
		; save current roi group, if it is not empty

	currRoiGroup = self.roiui->roiGroup()
	if( currRoiGroup->count() ne 0 ) then begin
		(*self.roiGroups)[self.currSlice] = currRoiGroup
	end else begin
		obj_destroy, (*self.roiGroups)[self.currSlice]
		(*self.roiGroups)[self.currSlice] = obj_new()
	end

		; change slice

	self.currSlice = evt.value

		; set new roigroup, but do not notify!

	self.roiui->disableNotify
	self.roiui->setRoiGroup, (*self.roiGroups)[self.currSlice]
	self.roiui->enableNotify

		; and display

	self->display, /update
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::colorSchemeSelectEvt
; Desc  : Activates or deactivates the roiUI and then applies the selected
;         colorscheme via call to self->applyColorScheme
;=-----------------------------------------------------------------------------
pro mod_colorcode::colorSchemeSelectEvt, evt
	if( evt.index eq 0 ) then begin
		widget_control, self.idLogfactor, /sensitive
		self.roiui->selfregister, self.display, button=1
		self.roiui->activate
		self.display->display
	end else begin
		widget_control, self.idLogfactor, sensitive=0
		self.roiui->deactivate
		self.roiui->unregister
		self.display->display
	end
	self.currScheme = evt.index
	self->applyColorScheme
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::logScaleSelectEvt
; Desc  :
;=-----------------------------------------------------------------------------
pro mod_colorcode::logScaleSelectEvt, evt
	widget_control, self.idLogFactor, get_value=val
	self.logfactor=val
	self->applyColorScheme
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::interpolateEvt
; Desc  :
;=-----------------------------------------------------------------------------
pro mod_colorcode::interpolateEvt, evt
	self.display->setInterpolate, evt.select
	self.interpolate = evt.select
	self.display->display
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::storeRoisEvt
; Desc  :
;=-----------------------------------------------------------------------------
pro mod_colorcode::storeRoisEvt, evt

		; store reference of currently displayed roi

	currRoiGroup = self.roiui->roiGroup()
	if( currRoiGroup->count() ne 0 ) then begin
		(*self.roiGroups)[self.currSlice] = self.roiui->roiGroup()
	end else begin
		obj_destroy, (*self.roiGroups)[self.currSlice]
		(*self.roiGroups)[self.currSlice] = obj_new()
	end

		; delete formerly stored rois

	if( ptr_valid( self.storedRoiGroups ) ) then begin
		for i = 0, n_elements( *self.storedRoiGroups )-1 do begin
			obj_destroy, (*self.storedRoiGroups)[i]
		end
		ptr_free, self.storedRoiGroups
	end

		; make a copy of the current roigroups

	minmax = widget_info( self.idSliceSlider, /slider_min_max )
	newRoiGroups = objarr( minmax[1]+1 )
	for i = 0, minmax[1] do begin
		if( obj_valid( (*self.roiGroups)[i] ) ) then begin
			newRoiGroups[i] = (*self.roiGroups)[i]->duplicate()
		end
	end
	self.storedRoiGroups = ptr_new( newRoiGroups, /no_copy )
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::recallRoisEvt
; Desc  : Copy the roigroups from self.storedRoiGroups to self.roiGroups
;         and to the roiUI. The currently valid self.roiGroups are deleted.
;=-----------------------------------------------------------------------------
pro mod_colorcode::recallRoisEvt, evt
	i18n = (*self.g).i18n
	if( self.storedRoiGroups eq ptr_new() ) then return
	if( n_elements( *self.storedRoiGroups ) ne n_elements( *self.roiGroups ) ) then begin
		x = dialog_message(i18n->get('err_numberOfSlicesNotMatch'), title=i18n->get('Error'), /error)  ; should rarely happen
		return
	end
	for i = 0, n_elements( *self.storedRoiGroups )-1 do begin
		obj_destroy, (*self.roiGroups)[i]
		if( obj_valid( (*self.storedRoiGroups)[i] ) ) then begin
			(*self.roiGroups)[i] = (*self.storedRoiGroups)[i]->duplicate()
		end else begin
			(*self.roiGroups)[i] = obj_new()
		end
	end
	self.roiui->setRoiGroup, (*self.roiGroups)[self.currSlice]
	self.display->display
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::deleteAllRoisEvt
; Desc  : Delete all roi groups and the currently displayed one
;=-----------------------------------------------------------------------------
pro mod_colorcode::deleteAllRoisEvt, evt

		; destroy current roiGroup in widget, then destroy all current rois

	self.roiUI->disableNotify
	self.roiUI->setRoiGroup, obj_new(), /destroy
	for i = 0, n_elements( *self.roiGroups )-1 do begin
		if( obj_valid( (*self.roiGroups)[i] ) ) then obj_destroy, (*self.roiGroups)[i]
		(*self.roiGroups)[i] = obj_new()
	end
	self.display->display
	self->roiChangeEvt, obj_new(), 'change', 0
	self.roiUI->enableNotify
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::contrastModeEvt
; Desc  :
;=-----------------------------------------------------------------------------
pro mod_colorcode::contrastModeEvt, evt
	self.contrast1->setMode, 1-evt.select
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::gammaModeEvt
; Desc  : Dummy. Handled by contrastModeEvt
;=-----------------------------------------------------------------------------
pro mod_colorcode::gammaModeEvt, evt
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::dicomViewButtonEvt
; Desc  :
;=-----------------------------------------------------------------------------
pro mod_colorcode::dicomViewButtonEvt, evt
	self.dispDicom = evt.select
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::contUpdateButtonEvt
; Desc  :
;=-----------------------------------------------------------------------------
pro mod_colorcode::contUpdateButtonEvt, evt
	self.contUpdate = evt.select
	self->displayColorSet
end



;==============================================================================
;
;                                 I N T E R N A L
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::update
; Desc  : The callback function for the image set container.
;         - any image sets left? No -> unmap
;         - Currently displayed set still existing?
;         - no -> do nothing and return
;         - update name list of the source
;         - yes-> any image sets left?
;         -       no -> inform the user, then unmap
;         -      yes -> inform the user, select new image set, call imgSetChange
;         This method can also be used when mapping the UI.
;=-----------------------------------------------------------------------------
pro mod_colorcode::update, dummy1, dummy2  ; dummy args for notification from mrContainer
	i18n = (*self.g).i18n

	sets = mrGetImageSet( (*self.g).imgSets, class='result', count=nsets )

		; if no image sets are left: unmap and return

	if( nsets eq 0 ) then begin
		x = dialog_message([i18n->get('colc_errNoImagesLeft'),i18n->get('colc_closingModule')], title=i18n->get('Info'), /info )
		self.currSource = ''
		self->unmap
		return
	end

		; get names of image sets

	setnames = strarr( nsets )
	for i = 0, nsets-1 do setnames[i] = sets[i]->name()

		; currently displayed image set (if any) still present?
		; yes -> ok, nothing to do; no->select new list

	widget_control, self.idSourcelist, set_value=setnames
	curr = where( setnames eq self.currSource )

	if( curr eq -1 ) then begin
		if( self.currSource ne '' ) then $
			x = dialog_message([i18n->get('colc_title')+':',i18n->get('colc_currentSetDeleted')], title=i18n->get('Info'), /info)
		self.currSource = setnames[0]
		widget_control, self.idSourcelist, set_droplist_select=0
		self->imgSetChange
	end else begin
		widget_control, self.idSourcelist, set_droplist_select=curr
	end
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::imgSetChange
; Desc  : The user has selected a new image set
;         -> update all UI elements such that they reflect the current status
;         - update source and target name, keep current name, if possible
;         - activate or inactivate 'display dicom info'
;         - activate or inactivate 'log scale factor'
;         - activate or inactivate mrRoi widget
;         display the new image set
;=-----------------------------------------------------------------------------
pro mod_colorcode::imgSetChange
	i18n = (*self.g).i18n

		; changing the image set will create a new roiGroup array, since
		; we can't be 100% sure that all roigroups will have the same number
		; of slices (although it is very, very likely)

	self.roiui->setRoiGroup, obj_new(), /destroy_current
	if( ptr_valid( self.roigroups ) ) then begin
		for i = 0, n_elements( *self.roiGroups )-1 do begin
			if( obj_valid( (*self.roiGroups)[i] ) ) then obj_destroy, (*self.roiGroups)[i]
		end
		ptr_free, self.roiGroups
	end

		; unregister the ROI

	self.roiui->unregister

		; get the new set

	set = mrGetImageSet( (*self.g).imgSets, class='result', name=self.currSource )
	if( set eq obj_new() ) then begin                     ; should not happen
		x = dialog_message( i18n->get('colc_terribleMistake'), title=i18n->get('Error'), /error)
		self->unmap
		return
	end

	if( obj_isa( set, 'mrImageSetDicom' ) ) then begin
		widget_control, self.idDicom, sensitive=1
	end else begin
		widget_control, self.idDicom, sensitive=0
	end

		; update default target name

	targetname = self.currSource + ' color'
	widget_control, self.idTargetname, set_value=targetname

		; update slider values

	nslices = set->nSlices()
	widget_control, self.idSliceSlider, set_slider_max=nslices-1
	self.currSlice = self.currSlice < (nslices-1)
	widget_control, self.idSliceSlider, set_value=self.currSlice

		; display the new image

	self->display
	self->displayColorSet

		; create a new object array for the roi groups

	self.roiGroups = ptr_new( objarr( set->nSlices() ), /no_copy )

		; re-register the roi, if scheme is logarithmic

	if( self.currScheme eq 0 ) then begin
		self.roiui->selfregister, self.display, button=1
		self.roiui->activate
	end
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::display
; Desc  : display the new slice, keeping all contrast/gamma/zooming settings
;=-----------------------------------------------------------------------------
pro mod_colorcode::display, update=update

	set = mrGetImageSet( (*self.g).imgSets, class='result', name=self.currSource )
	img = (*(set->imgData()))[self.currSlice]

	if( keyword_set( update ) ) then begin
		self.display->updateImg, img
	end else begin
		self.display->setImg, img
		self.display->fitToScreen
	end
	self.display->display
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::displayColorSet
; Desc  : Check whether a color image set with the currently selected name
;         exists and display it, if so
;=-----------------------------------------------------------------------------
pro mod_colorcode::displayColorSet, cont

	widget_control, self.idTargetName, get_value=target

	set = mrGetImageSet( (*self.g).imgSets, class='color', name=target )

		; no color set available? -> clear all

	if( set eq obj_new() ) then begin
		for x = 0, self.nx-1 do begin
			for y = 0, self.ny-1 do begin
				self.multidisplay->setImg, x, y, ptr_new(intarr(2,2),/no_copy), /no_contrastupdate
			end
		end
		self.multidisplay->setContrast, .5, 1.
		self.multidisplay->display
		return
	end

		; otherwise display the set

	imgs = set->imgData()
	self.multidisplay->deactivate
	for i = 0, set->nslices()-1 do begin
		x = i mod self.nx
		y = i / self.nx
		self.multidisplay->setImg, x, y, (*imgs)[i], /no_contrastupdate
	end
	self.multidisplay->setContrast, .5, 1.
	self.multidisplay->activate
	self.multidisplay->setPalette, set->Palette()  ; updates contrast bar via callback
	self.multidisplay->display
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::applyColorScheme
; Desc  : Apply the selected colorscheme to the currently displayed imageset
;         Save the created colored image set in a new mrImageSet object
;=-----------------------------------------------------------------------------
pro mod_colorcode::applyColorScheme

	scheme = self.currScheme
	widget_control, self.idTargetName, get_value=targetname
	source = mrGetImageSet( (*self.g).imgSets, class='result', name=self.currSource )
	if( source eq obj_new() ) then message, 'Prog error in applyColorScheme'

		; nothing to do, if no rois defined for logs

	if( scheme eq 0 and self.roiMean eq 0. ) then return

		; The target image set equals the source image set with image values scaled to [0,255]
		; For logarithmic scaling and very large maximum values, this may lead to some of the lower
		; boundary colors not being displayed at all, but the effect is small and we won't tell the doctors :)
		;
		; If the target set already exists, we use the old one (enabling the image browser to smoothly
		; updated changed palettes).


		; find maximum in whole set for scaling

	sourceimgs = source->imgData()
	maxval = 0.
	for i = 0, source->nslices()-1 do begin
		maxval = max( *(*sourceimgs)[i] ) > maxval
	end

		; create new set if necessary

	target    = mrGetImageSet( (*self.g).imgSets, class='color', name=targetname )
	if( target ne obj_new() ) then existed=1 else existed=0
	if( target eq obj_new() ) then begin
		target = obj_new('mrImageSet', source->nslices(), class='color', name=targetname, xsize=source->xsize(), ysize=source->ysize() )
		if( target eq obj_new() ) then message, 'Error creating image set'
	end
	target->setParent, source
	targetimgs = target->imgData()

		; by working directly with the arrays, we avoid the sending of notifications
		; !!!
		; This is not entirely clean, since possibly existing pointers are lost in
		; the process. But as of now, there's no module that stores the image pointers
		; locally, so that's ok at least within this program.

	for i = 0, source->nslices()-1 do begin
		if( ptr_valid( (*targetimgs)[i] ) ) then ptr_free, (*targetimgs)[i]
		newimg = ( fix( 255. * *(*sourceimgs)[i]/maxval ) < 255 )
		(*targetimgs)[i] = ptr_new( newimg )
	end
	if( not existed ) then (*self.g).imgSets->add, target  ; add only if it wasn't there before
	target->sendNotification, 'set'                        ; send notifications now manually

		;
		; Logarithmic scheme
		;

	if( scheme eq 0 ) then begin

			; red/green/blue of the 16 colors for color encoding

		rs    = [0,0,0,0,0,0,0,0,0,255,255,255,255,255,230,255]           ; 210
		gs    = [0,32,32,64,0,160,160,208,255,255,204,128,102,0,26,255]
		bs    = [0,112,144,192,255,255,160,64,0,0,102,64,0,0,0,255]
		red   = intarr( 256 )
		gre   = intarr( 256 )
		blu   = intarr( 256 )

			; percentage scale. For cs[i] <= imgval/normalization_val < cs[i+1], use color 'i'
			; the somewhat weird 2nd formula has historic reasons
			; anyway - in the default (ColLogRatio=4.0), 0.9 to 1.15 are colored yellow
			; Image values with value/mean > colLogRatio are drawn with color id 15

		widget_control, self.idLogfactor, get_value=fact
		cs = [0,0.05,0.1,0.2,0.3,0.4,0.5,0.6,0.75,0.9,1.15,1.4,1.7,2.1,2.8,4.0]
		cs = cs*fact

			; find the maximum value in this image set

		for i = 14, 0, -1 do begin
			low = fix( 255.*cs[i]*self.roimean/maxval ) < 255
			hig = fix( 255.*cs[i+1]*self.roimean/maxval ) <255
			red[low:hig] = rs[i]
			gre[low:hig] = gs[i]
			blu[low:hig] = bs[i]
		end
		hig = fix( 255.*cs[15]*self.roimean/maxval ) <255
		red[hig:255] = rs[15]
		gre[hig:255] = gs[15]
		blu[hig:255] = bs[15]

		if( self.roimean eq 0. ) then 								$
			target->setPaletteRGB, colorTable=0 					$
		else 														$
			target->setPaletteRGB, red=red, green=gre, blue=blu

	end else begin

		; Standard color schemes. This is very easy as we just need
		; to update the palette

		target->setPaletteRGB, colorTable=scheme-1     ; this causes an update in the contrast
	end
	self->displayColorSet
end

;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::calcRoiMean
; Desc  : Calculate the roi mean value and store it in self.roiMean
;         In most cases, only rois in the present slice need to be recalcu-
;         lated, but we skip this optimization option for now.
;=-----------------------------------------------------------------------------
pro mod_colorcode::calcRoiMean
	self.roiMean = 0.
	sum   = 0.
	ndots = 0
	set  = mrGetImageSet( (*self.g).imgSets, class='result', name=self.currSource )
	imgs = set->imgData()
	for sliceno = 0, n_elements(*self.roiGroups)-1 do begin
		roiGroup = (*self.roiGroups)[sliceno]
		if( roiGroup eq obj_new() ) then continue
		rois = roiGroup->get( /all, count=nrois )
		for n = 0, nrois-1 do begin
			mask = where( *(rois[n]->filledMask()) gt 0, npixels )
			if( npixels eq 0 ) then continue
			ndots = ndots + npixels
			sum = sum + mean( (*(*imgs)[sliceno])[mask] )*npixels
		end
	end
	if( ndots ne 0 ) then self.roiMean = sum/ndots
end


;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::roiChangeEvt, evt
; Desc  :
;=-----------------------------------------------------------------------------
pro mod_colorcode::roiChangeEvt, sender, evtType, evt

		; store currently defined rois in our internal list, then calculate the mean

	currRoiGroup = self.roiui->roiGroup()
	;if( currRoiGroup->count() ne 0 ) then begin                     ; might even store empty roigroups!
		(*self.roiGroups)[self.currSlice] = currRoiGroup
	;end else begin
	;	obj_destroy, (*self.roiGroups)[self.currSlice]
	;	(*self.roiGroups)[self.currSlice] = obj_new()
	;end
	self->calcRoiMean
	widget_control, self.idMeanRoi, set_value=strcompress( string(self.roiMean),/remove_all )

	if( evtType eq 'modify' and self.contUpdate eq 0 ) then return
	self->applyColorScheme
end


;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::load
; Desc  : Hooked into mrroi's load button - as opposed to mrroi's build-in
;         load function, this one is able to load a roigroup for each slice.
;=-----------------------------------------------------------------------------
pro mod_colorcode::load, evt
	i18n = (*self.g).i18n

	nrois = 0
	for i = 0, n_elements( *self.roigroups )-1 do begin
		if( not obj_valid( (*self.roigroups)[i] )) then continue
		nrois = nrois + (*self.roigroups)[i]->count()
	end

	if( nrois gt 0 ) then begin
		x = mrdialog_select([i18n->get('colc_loadWarning'),i18n->get('Continue')+'?'], $
		                    [i18n->get('Yes'),i18n->get('No')], default=1, group_leader=self.baseid, $
		                    title=i18n->get('Question'), /warning )
		if( x eq 1 ) then return
	end

	file = dialog_pickfile( /read, title=i18n->get('colc_selectRoiFile'), path=(*self.g).workingdir, get_path=path, /must_exist )
	if( file eq '' ) then return

	(*self.g).workingdir = path
	roigroups = mrRoiGroup_load( file )
	if( roigroups eq obj_new() ) then begin
		x = dialog_message( i18n->get('err_readingFile'), title=i18n->get('Error'), /error )
		return
	end

		; delete ROIs

	self.roiUI->disableNotify
	self.roiUI->setRoiGroup, obj_new(), /destroy
	for i = 0, n_elements( *self.roiGroups )-1 do begin
		if( obj_valid( (*self.roiGroups)[i] ) ) then obj_destroy, (*self.roiGroups)[i]
		(*self.roiGroups)[i] = obj_new()
	end

		; loop over loaded ROIs and set them

	nslices = n_elements( *self.roigroups )
	nrois = roigroups->count()
	for i = 0, nrois-1 do begin
		roigroup = roigroups->get( position=0 )
		id = roigroup->id()
		if( id ge nslices ) then begin
			x = dialog_message( i18n->get('colc_errRoiSliceTooLarge') + strtrim(id, 2), title=i18n->get('Error'), /error )
			roigroups->remove, roigroup
			obj_destroy, roigroup
		end	else begin
			(*self.roigroups)[id] = roigroup
			roigroups->remove, roigroup
		end
	end

	obj_destroy, roigroups

	self.roiui->setRoigroup, (*self.roigroups)[self.currSlice]

	self.display->display
	self->roiChangeEvt, obj_new(), 'change', 0
	self.roiUI->enableNotify
end


;+=----------------------------------------------------------------------------
; Name  : mod_colorcode::save
; Desc  : Hooked into mrroi's save button. As opposed to mrroi's built-in save
;         function this one here saves all roigroups from all slices
;=-----------------------------------------------------------------------------
pro mod_colorcode::save, evt
	i18n = (*self.g).i18n

	nrois = 0
	for i = 0, n_elements( *self.roiGroups )-1 do begin
		if( not obj_valid( (*self.roiGroups)[i] )) then continue
		nrois = nrois + (*self.roiGroups)[i]->count()
	end

	if( nrois eq 0 ) then begin
		x = dialog_message( i18n->get('colc_noRoisToSave'), title=i18n->get('Info'), /info)
		return
	end

	file = dialog_pickfile( /write, title=i18n->get('colc_selectRoiFileForWrite'), path=(*self.g).workingdir, get_path=path )
	if( file eq '' ) then return

	(*self.g).workingdir = path
	unit = mrRoiGroup_openxml( file )
	if( unit eq 0 ) then begin
		x = dialog_message( i18n->get('err_openFile'), title=i18n->get('Error'), /error )
		return
	end
	for i = 0, n_elements( *self.roigroups )-1 do begin
		roigroup = (*self.roiGroups)[i]
		if( roigroup eq obj_new() ) then continue
		if( roigroup->count() eq 0 ) then continue
		roigroup->setId, i                         ; the slice number
		roigroup->write, unit
	end
	mrroigroup_closexml, unit
	x = dialog_message( i18n->get('colc_roisSaved'), title=i18n->get('Info'), /info)
end


;+=----------------------------------------------------------------------------
; Name  : mod_colorcode__define
; Desc  : Private Procedure (constructor)
;=-----------------------------------------------------------------------------
pro mod_colorcode__define
	obj = { MOD_COLORCODE					,$
			inherits LUPEMOD				,$
			display:			obj_new()	,$
			multidisplay:		obj_new()	,$
			nx:							0	,$ ; number of images in x in multidisplay
			ny:							0	,$ ; number of images in y in multidisplay
			contrast1:			obj_new()	,$ ; contrast for mrDisplay
			contrast2:			obj_new()	,$ ; contrast for mrMultidisplay
			roiUI:				obj_new()	,$ ; the roi user interface object
			roiGroups:			ptr_new()	,$ ; array of mrRoiGroups objects, one per slice
			roiMean:					0.	,$ ; mean value of the defined ROIs
			storedRoiGroups:	ptr_new()   ,$ ; stored RoiGroups
			idSliceSlider:				0L	,$ ; widget id of the slice selector
			idSourcelist:				0L	,$ ; widget id of the source droplist
			idColorlist:				0L	,$ ; widget id of color scheme list
			idTargetname:				0L	,$ ; widget id of the target namefield
			idDicom:					0L	,$ ; widget id of the dicom checkmark box
			idContUpdate:				0L	,$ ; widget id of 'auto update' checkmark box
			idLogfactor:				0L	,$ ; widget id of the logarithmic factor select
			idMeanRoi:					0L	,$ ; mean Roi value text widget id
			idStoreRoi:					0L	,$ ; the 'store Rois' button
			idRecallRoi:				0L	,$ ; the 'recall Rois' button
			idDeleteRois:				0L	,$ ; the 'delete all Rois' button
			interpolate:				0   ,$ ; save 'interpolate' status
			contUpdate:					0   ,$ ; save 'continuous update' status
			dispDicom:					0   ,$ ; save 'display dicom' status
			logFactor:					0.  ,$ ; save 'logarithmic factor'
			currSource:					''	,$ ; current name of source image set
			currScheme:					 0  ,$ ; index of the currently selected coloring scheme
			currSlice:					 0	 $ ; current slice
		  }
end
