;+=----------------------------------------------------------------------------
; Name    : mod_ssvd
; RCS     : $Id: mod_ssvd__define.pro,v 1.2 2004/11/25 09:08:27 olivert Exp olivert $
; Desc    : Implementation of simple singular value decomposition
; Author  : (c) 2004 Oliver Thilmann
;=------------------------------------------------------------------------------
function mod_ssvd::revision
	rev = '$Revision: 1.2 $'
	return, rev
end

;+=----------------------------------------------------------------------------
; Name  : mod_ssvd::init
; Syntax: obj = obj_new('lupemod', type, exclusive )
; Args  : g is a pointer to the global structure
; Desc  : A module in Lupe is a class derived from class lupemod. A module is
;         instanciated only once at program startup.
;=-----------------------------------------------------------------------------
function mod_ssvd::init, g

		; call the constructor of the superclass
		; Arguments are module_name, module_type, exclusive_flag, global
		; 'mod_ssvd' is the name or identifier of this module (e.g. used in
		; the file menu.xml.
		; This is an 'embedded' module, i.e. it is realized in the program's
		; main window. Other options here are 'top' for modules which create
		; their own window and 'noui' for modules that don't need a user interface
		; If the exclusive flag is set to '1', all other modules will be unmapped
		; upon mapping of this module. This was meant for mutually exclusive
		; modules, but is not used anywhere in the program right now.
		; g is just the pointer to the global structure which is stored as
		; an object variable, thus accessible via self.g

    res = self->lupemod::init( 'mod_ssvd', 'embedded', 0, g )
    if( res eq 0 ) then return, 0

	return, 1
end


;+=----------------------------------------------------------------------------
; Name  : mod_ssvd::readyToExecute
; Desc  : When the user selects a menu entry, the framework will call the
;         'readyToExecute' function of the corresponding module. The
;         function should return '1', if all pre-conditions to map/execute
;         this module are fulfilled, otherwise it should return '0'.
;         In case '0' is returned, it is the module's responsibility to
;         inform the user *why* it is not executed.
;
;         In this case, we check whether the necessary input data (raw,
;         baseline, aif) is available. If not, we inform the user.
;=-----------------------------------------------------------------------------
function mod_ssvd::readyToExecute, option
	i18n = (*self.g).i18n

	sigset = mrGetImageSet( (*self.g).imgSets, class='raw', name='signal' )
	ctrset = mrGetImageSet( (*self.g).imgSets, class='raw', name='ctr_xyt')
	baselineset = mrGetImageSet( (*self.g).imgSets, class='raw', name='baseline')
	if( sigset eq obj_new() or ctrset eq obj_new() or baselineset eq obj_new() ) then begin
		x = dialog_message(i18n->get('svd_missingRawImgSet'), title=i18n->get('Information'))
		return, 0
	end

	aifset = mrGetImageSet( (*self.g).imgsets, class='aif', name='global' )
	if( aifset eq obj_new() ) then begin
 		x = dialog_message(i18n->get('svd_noAIFDefined'), title=i18n->get('Information'))
		return, 0
	end
	return, 1
end

;+=----------------------------------------------------------------------------
; Name  : mod_ssvd::map
; Desc  : The map method is called for 'top' and 'embedded' modules if readyToExecute
;         returned '1', immediately after 'readyToExecute' (i.e. you may rely on
;         the conditions checked in the latter.
;         Event handling:
;         All selectable widgets need to have a uvalue set which specifies which
;         object/method pair to call upon any widget event, for example
;         x = widget_button( 'buttontext', uvalue={o:self, m:'buttonEvt'} )
;         'o' denotes the object to call (that's us), 'm' the method.
;         Event handling is done via lupemod which (in this case upon mouse click
;         on the button) will then call self->buttonEvt.
;         Widget IDs:
;         Widget IDs are stored in object variables idSomething for later use.
;=-----------------------------------------------------------------------------
pro mod_ssvd::map, option

		; i18n is an object of class mrHash and holds all the translations

	i18n = (*self.g).i18n

		; create the base widget. The superclass 'lupemod' provides a
		; 'createBaseWidget' method which does exactly that.
		; Embedded modules may *not* provide any window size information
		; as this is fixed for embedded modules. See mod_imagebrowser for
		; an example of a module with a top-level window.

	res = self->createBaseWidget( /row )
	if( res eq 0 ) then return

		; the rest is pretty much straight forward UI creation.

	l = widget_base( self.baseid, /column, /frame )
	r = widget_base( self.baseid, /column, /frame, /base_align_center )

		; left side: multidisplay & three result displays

	imgset = mrGetImageSet( (*self.g).imgSets, class='raw', name='signal' )
    nrows = 6 > ( imgset->nshots() + 9)/10


	l1 = widget_base( l, /row )
	self.mdisp = obj_new('mrmultidisplay', l1, 700, 420, 70, 70, 10, nrows )
	l1 = widget_base( l, /row )
	self.disp1 = obj_new('mrdisplay', l1, 232, 232, interpolate=0 )
	self.disp2 = obj_new('mrdisplay', l1, 232, 232, interpolate=0 )
	self.disp3 = obj_new('mrdisplay', l1, 232, 232, interpolate=0 )

		; right side: all the buttons I could think of (calculate)

    x  = widget_label( r, value=i18n->get('svd_ssvdTitle'), font='fixedsys', /align_center, /sunken_frame, frame=1, xsize=250)
	r1 = widget_base ( r, xsize=230, /column )

		; image set names

	x  = widget_base ( r1, ysize=20 )
	r2 = widget_base ( r1, /row )
	x  = widget_label( r2, xsize=200, value=i18n->get('svd_resultImageSetNames') )
	r2 = widget_base ( r1, xsize=220, /column, /frame )
	r3 = widget_base ( r2, /row )
	x  = widget_label( r3, xsize=100, value=i18n->get('svd_CBVname') )
	self.idcbvname  = widget_text ( r3, xsize=130, value='rCBV sSVD', /editable, uvalue={o:self, m:'dummyEvt'})
	r3 = widget_base ( r2, /row )
	x  = widget_label( r3, xsize=100, value=i18n->get('svd_CBFname') )
	self.idcbfname  = widget_text ( r3, xsize=130, value='rCBF sSVD', /editable, uvalue={o:self, m:'dummyEvt'} )
	r3 = widget_base ( r2, /row )
	x  = widget_label( r3, xsize=100, value=i18n->get('svd_MTTname') )
	self.idmttname  = widget_text ( r3, xsize=130, value='rMTT sSVD', /editable, uvalue={o:self, m:'dummyEvt'} )

		; Display parameters
		; Here's another one: if there's no value provided for the idsigthres, etc.
		; widget_labels, IDL might, I stress, *might*, collapse the ysize of these widgets
		; to '0' - thus rendering any subsequently set value invisible. The actual behaviour
		; depends on the placement of the widget_control statement that sets a value.
		; At the point I found out about this, it was before or after creating the 'close
		; window' button).

	x  = widget_base ( r1, ysize=20 )
	r2 = widget_base ( r1, /row )
	x  = widget_label( r2, xsize=200, value=i18n->get('svd_currentParameters') )

	r2 = widget_base ( r1, xsize=220, /column, /frame )
	r3 = widget_base ( r2, /row )
	x  = widget_label( r3, xsize=150, value=i18n->get('svd_signalValueThres') )
	self.idsigthres  = widget_label ( r3, xsize=50, value='-', /align_right )
	r3 = widget_base ( r2, /row )
	x  = widget_label( r3, xsize=150, value=i18n->get('svd_ctrValueThres') )
	self.idctrthres  = widget_label ( r3, xsize=50, value='-', /align_right )
	r3 = widget_base ( r2, /row )
	x  = widget_label( r3, xsize=150, value=i18n->get('svd_wMatrixCutoff') )
	self.idwcutoff   = widget_label ( r3, xsize=50, value='-', /align_right )
	r3 = widget_base ( r2, /row )
	x  = widget_label( r3, xsize=150, value=i18n->get('svd_CBVcutoff') )
	self.idcbvcutoff   = widget_label ( r3, xsize=50, value='-', /align_right )
	r3 = widget_base ( r2, /row )
	x  = widget_label( r3, xsize=150, value=i18n->get('svd_CBFcutoff') )
	self.idcbfcutoff   = widget_label ( r3, xsize=50, value='-', /align_right )
	r3 = widget_base ( r2, /row )
	x  = widget_label( r3, xsize=150, value=i18n->get('svd_MTTcutoff') )
	self.idmttcutoff   = widget_label ( r3, xsize=50, value='-', /align_right )

		; progressbar and calculate button

	x  = widget_base  ( r1, ysize=86 )
	x  = widget_button( r1, xsize=230, value=i18n->get('svd_calculate'), uvalue={o:self, m:'calcButtonEvt'})
	self.progressbar = obj_new( 'mrprogressbar', r1, xsize=222, /frame, title=i18n->get('CalculationProgress'), /remaining )

		; close window button

	x  = widget_base  ( r1, ysize=136 )
	x  = widget_button( r1, xsize=220, value=i18n->get('CloseModule'), uvalue={o:self, m:'closeButtonEvt'} )

       ; set labels on the three result images

    mdl = obj_new('IDLgrModel')
    txt = obj_new('IDLgrText', 'rCBV', locations=[.98,.02], color=[0,255,0], alignment=1.)
    mdl->add, txt
    (self.disp1->labelView())->add, mdl

    mdl = obj_new('IDLgrModel')
    txt = obj_new('IDLgrText', 'rCBF', locations=[.98,.02], color=[0,255,0], alignment=1.)
    mdl->add, txt
    (self.disp2->labelView())->add, mdl

    mdl = obj_new('IDLgrModel')
    txt = obj_new('IDLgrText', 'MTT', locations=[.98,.02], color=[0,255,0], alignment=1.)
    mdl->add, txt
    (self.disp3->labelView())->add, mdl

		; After all widgets have been defined, realization of the UI and registration
		; with the event handler is done via a method again provided by the superclass lupemod:

	self->placeAndRegister

		; updateParameters is defined further down in this file and fills the
		; calculation parameter fields with the values as specified in preferernces

	self->updateParameters

end

;+=----------------------------------------------------------------------------
; Name  : mod_ssvd::unmap
; Desc  : Removes the module from screen. At this point, temporary objects
;         like displays and plugins have to be destroyed as well as they're
;         created anew with each 'map'.
;         Final unmapping is done via the unmap method of the superclass.
;=-----------------------------------------------------------------------------
pro mod_ssvd::unmap, _ref_extra=ex
	if( self.baseid eq 0L ) then return  ; uups - we don't exist :( Shouldn't happen

	obj_destroy, self.mdisp
	obj_destroy, self.disp1
	obj_destroy, self.disp2
	obj_destroy, self.disp3

	self->lupemod::unmap, _strict_extra=ex
end



;==============================================================================
;
;                           E V E N T   H A N D L I N G
;
;==============================================================================

;+=----------------------------------------------------------------------------
; Name  : mod_ssvd::calcButtonEvt
; Desc  : There's not too many events in this module.
;         The 'evt' argument is whatever IDL provided to the event-handler.
;         The calculate function is not called directly as UI event handling
;         and functionality are split to allow for possible batch processing.
;=-----------------------------------------------------------------------------
pro mod_ssvd::calcButtonEvt, evt
	self->calculate
end

;+=----------------------------------------------------------------------------
; Name  : mod_ssvd::closeButtonEvt
; Desc  :
;=-----------------------------------------------------------------------------
pro mod_ssvd::closeButtonEvt, evt
	self->unmap
end

;+=----------------------------------------------------------------------------
; Name  : mod_ssvd::dummyEvt
; Desc  : Does nothing. The content of the image set name text fields is
;         not stored, but read directly from the widget when needed.
;         Nevertheless the text widgets create events upon key-press and we
;         have to catch these.
;=-----------------------------------------------------------------------------
pro mod_ssvd::dummyEvt, evt
end



;==============================================================================
;
;                                 I N T E R N A L
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : mod_ssvd::updateParameters
; Desc  : Get the parameters from preferences and display them
;=-----------------------------------------------------------------------------
pro mod_ssvd::updateParameters
	prefs = (*self.g).preferences

	widget_control, self.idsigthres,  set_value=string( (*prefs).ssvd_sigthres, format='(f6.1)' )
	widget_control, self.idctrthres,  set_value=string( (*prefs).ssvd_ctrthres, format='(f6.1)' )
	widget_control, self.idwcutoff,   set_value=string( fix( 100.*(*prefs).ssvd_matrixcutoff) )
	widget_control, self.idcbvcutoff, set_value=string( (*prefs).ssvd_cbvcutoff, format='(f5.0)' )
	widget_control, self.idcbfcutoff, set_value=string( (*prefs).ssvd_cbfcutoff, format='(f5.0)' )
	widget_control, self.idmttcutoff, set_value=string( (*prefs).ssvd_mttcutoff, format='(f5.0)' )
end

;+=----------------------------------------------------------------------------
; Name  : mod_ssvd::calculate
; Desc  : Perform the SVD calculation, slice by slice, and display the results
;         after each slice
;=-----------------------------------------------------------------------------
pro mod_ssvd::calculate

	i18n = (*self.g).i18n

		; update the displayed parameter values and reset the progress bar to 0%

	self->updateParameters
	self.progressbar->reset

		; All image sets are stored in a central container 'imgSets' and is identified
		; via a 'class' string (this has got nothing to do with OO programming but is
		; just a way to group image sets) and a 'name' string.
		; All input data is of class 'raw'.
		; Given class and name, mrGetImageSet returns the mrImageSet or mrImageSetDicom
		; object with that identifiers - or obj_new(), if a set with these specifiers
		; does not exist.
		; Please see the documentation on the two library classes mrImageSet and mrImageSetDicom!

	sigset      = mrGetImageSet( (*self.g).imgSets, class='raw', name='signal' )
	ctrset      = mrGetImageSet( (*self.g).imgSets, class='raw', name='ctr_xyt' )
	baselineset = mrGetImageSet( (*self.g).imgSets, class='raw', name='baseline')

		; Even though checked in readyToExecute, an evil user might have deleted
		; necessary image sets from memory. In that case, we inform the user and unmap
		; the module.

	if( sigset eq obj_new() or ctrset eq obj_new() or baselineset eq obj_new() ) then begin
		x = dialog_message([i18n->get('svd_missingSets1'),i18n->get('svd_missingSets2')], /error, title=i18n->get('Error') )
		self->unmap
		return
	end

		; get number of slices and x/y size (in pixels) of the perfusion image set

	nslices = sigset->nslices()
	xsize   = sigset->xsize()
	ysize   = sigset->ysize()

		; the calculated result image sets are given a class-name 'result' and
		; a name as specified by the user in the UI.
		; First we check whether the user has specified names for all result sets.
		; Next it is checked that the specified names differ
		; Last we check if image sets with these names already exist in memory.
		; If yes, the user is asked whether he wants to replace these sets or
		; rather use a different name.

	resultsets = mrGetImageSet( (*self.g).imgSets, class='result', count=nsets )
	widget_control, self.idcbvname, get_value=cbvname
	cbvname = strtrim(cbvname, 2)
	widget_control, self.idcbfname, get_value=cbfname
	cbfname = strtrim(cbfname, 2)
	widget_control, self.idmttname, get_value=mttname
	mttname = strtrim(mttname, 2)
	if( cbvname eq '' or cbfname eq '' or mttname eq '' ) then begin
		x = dialog_message(i18n->get('svd_missingSetName'), title=i18n->get('Error'), /error)
		return
	end
	names = [cbvname, cbfname, mttname]
	if( n_elements( uniq(names[sort(names)]) ) ne 3 ) then begin
		x = dialog_message(i18n->get('svd_imageSetNamesMustDiffer'), title=i18n->get('Error'), /error)
		return
	end
	if( nsets gt 0 and (*(*self.g).preferences).warnResultReplace eq 1 ) then begin
		setnames = resultSets[0]->name()
		for i = 1, nsets-1 do setnames = [setnames, resultSets[i]->name()]
		doubles = intersect( names, setnames )
		if( doubles[0] ne -1 ) then begin          ; yields a type conversion error, but so what?
			x = mrdialog_select( [i18n->get('svd_resultSetExistsWarn1'),doubles,i18n->get('svd_resultSetExistsWarn2'),$
								  i18n->get('svd_resultSetExistsWarn3')], ['Yes', 'No'], $
								  default=1, group_leader=self.baseid, bsize=80, title=i18n->get('Question'), /question )
			if( x eq 1 ) then return
		end
	end

		; Now create the result sets
		; First try to get them from the central image set container and destroy
		; possibly existing image sets with the same name

	cbvset = mrGetImageSet( (*self.g).imgSets, class='result', name=cbvname )
	cbfset = mrGetImageSet( (*self.g).imgSets, class='result', name=cbfname )
	mttset = mrGetImageSet( (*self.g).imgSets, class='result', name=mttname )

	if( obj_valid(cbvset) ) then begin
		(*self.g).imgSets->remove, cbvset   ; remove from container before destruction!
		obj_destroy, cbvset
	end
	if( obj_valid(cbfset) ) then begin
		(*self.g).imgSets->remove, cbfset
		obj_destroy, cbfset
	end
	if( obj_valid(mttset) ) then begin
		(*self.g).imgSets->remove, mttset
		obj_destroy, mttset
	end

		; Create the new result image sets

	cbvset = obj_new('mrimageset', nslices, xsize=xsize, ysize=ysize, class='result', name=cbvname )
	cbfset = obj_new('mrimageset', nslices, xsize=xsize, ysize=ysize, class='result', name=cbfname )
	mttset = obj_new('mrimageset', nslices, xsize=xsize, ysize=ysize, class='result', name=mttname )

		; The AIF is an image set as well, albeit a peculiar one: it contains only one image
		; with only one dimension (and number_of_timepoints elements)
		; Again it is checked whether the AIF still exists or whether the user has deleted it
		; in the meantime.

	aifset = mrGetImageSet( (*self.g).imgsets, class='aif', name='global' )
	if( aifset eq obj_new() ) then begin       ; just for safety, already checked in 'readyToExecute'
		x = dialog_message(i18n->get('svd_noAIFDefined'), title=i18n->get('Error'), /error)
		self->unmap
		return
	end

		; The 'imgData' method returns a pointer to a matrix which contains pointers to all images
		; in this image set, i.e.
		; aifset->imgData()         is a pointer to the matrix
		; *(aifset->imgData())      is the matrix of pointers itself
		; (*(aifset->imgData()))[0] is the pointer to the first image - since this imageSet object
		;                           contains only one image, in this case it is also the only one.

	aif = (*(aifset->imgData()))[0] ; so aif is a pointer to an array of size 'number_of_timepoints'

		; The 'dicomData()' method is analogous to the 'imgData' method above, except that it returns
		; a pointer to a matrix of pointers to mrDicom objects.
		; To obtain the repetition time, we use the very first image in the data series.
		; If the DICOM data contains no TR (I experienced that once), we use the default value
		; without warning. The user has already been warned about that when calculating concentration
		; images

	dicom = (*(sigset->dicomData()))[0,0]
	tr    = dicom->getTagValue('acq_repetition_time')             ; dicom is an object of class mrDicom
	if( tr eq '' ) then tr = (*(*self.g).preferences).default_tr  ; take from defaults, if not found
	tr    = tr/1000.                                              ; convert time into seconds

		; Finally perform the calculation

	imgdata = sigset->imgdata()      ; imagedata is a pointer to a [number_of_slices, number_of_timepoints] matrix of pointers
	for i = 0, nslices-1 do begin    ; loop over all slices

			; display the time sequence of the slice which is about to be calculated
			; In the case of SVD where calculation is fast, this takes more time than the
			; actual calculation. But the user might want to 'see' something happening besides
			; the progress bar and otoh it is not *that* much time wasted.
			; self->mdisp is the mrMultidisplay object - see the library documentation for more info.

		self.mdisp->label, /delete, /all
		for k = 0, sigset->nshots()-1 do begin
			row = k/10
			col = k mod 10
			self.mdisp->setImg, col, row, (*imgdata)[i, k], /no_contrastupdate ; no_contrastupdate to speed up things, anyway it's slow
			self.mdisp->label, col, row, string( k, format='(i2)' )
		end
		self.mdisp->setContrast
		self.mdisp->display

			; the actual calculation

		self->calcPerfusionImages, i, tr, aif, baselineset, sigset, ctrset, cbvset, cbfset, mttset

			; update the progress abr

		self.progressbar->update, (i+1.)/nslices

			; display the calculated result images in the three mrDisplay objects at the lower border

		self.disp1->setImg, (*(cbvset->imgdata()))[i]
		self.disp1->fitToScreen
		self.disp1->display
		self.disp2->setImg, (*(cbfset->imgdata()))[i]
		self.disp2->fitToScreen
		self.disp2->display
		self.disp3->setImg, (*(mttset->imgdata()))[i]
		self.disp3->fitToScreen
		self.disp3->display
	end

		; Remember the ancestry
		; This is important when saving in Dicom format: every calculated image set should have a
		; parent from which it was derived. When writing Dicom, the writing routine will search for
		; a parent set of DICOM images and use this set as source when copying all the DICOM tags.
		; An image set (mrImageSet) without a DICOM image set (mrImageSetDicom) somewhere in its
		; ancestry can not be written in DICOM format.

	cbvset->setParent, sigset
	cbfset->setParent, sigset
	mttset->setParent, sigset

		; Finally add the image sets to the central container

	(*self.g).imgSets->add, cbfset
	(*self.g).imgSets->add, cbvset
	(*self.g).imgSets->add, mttset

end

;+=----------------------------------------------------------------------------
; Name   : mod_ssvd::calcPerfusionImages
; Desc   : Calculates the perfusion images via simple SVD
;=-----------------------------------------------------------------------------
pro mod_ssvd::calcPerfusionImages, slice, tr, ai, baselineset, sigset, ctrset, cbvset, cbfset, mttset

	aif = *ai

	nshots       = sigset->nshots()
	xsize        = sigset->xsize()
	ysize        = sigset->ysize()
	matrixcutoff = (*(*self.g).preferences).ssvd_matrixcutoff
	sigthres     = (*(*self.g).preferences).ssvd_sigthres
	bl           = baselineset->uvalue()  ; region over which average was calculated

       ; preparation of the arterial input function - make local copy and
       ; have the aif always be positive.
       ; As indices up to ns0[1] are used for baseline averaging, the
       ; corresponding values are not used - thus all field and matrix sizes
       ; within this routine are of length nshots - ns0[1] - 1

    aif  = aif[(*bl)[1]+1:*]
    aif  = aif > 0
    aifc = aif              ; aif with cut off recirculation effects

       ; cut end of aif function to get rid of recirculation effects
       ; achieved by setting all values to zero after bump at a level < 40% maxvalue
       ; according to Ronnie, this doesn't make much of a difference, though

    maxval = max( aifc, maxid )
    for i = maxid+1, n_elements(aifc)-1 do begin
       if( aifc[i] le aifc[i-1] or aifc[i] gt .4*maxval ) then continue
       aifc[i:n_elements(aifc)-1] = 0.
       break
    end

       ; create the diagonal matrix for unfolding - see the Oestergaard paper
       ; the dimension of the aifmatrix does not need to include the leading (*bl)[1]
       ; images which are used for baseline determination as there's no need to
       ; include them in the calculation
       ; the matrix is calculated from the aif with cut off second 'bumps'

    aifmatrix = fltarr( nshots-(*bl)[1]-1, nshots-(*bl)[1]-1 )
    for i = 0, nshots-(*bl)[1]-2 do begin
       for j = 0, nshots-(*bl)[1]-i-2 do begin
         aifmatrix[i+j, i] = aifc[j]
       end
    end

       ; define the result arrays for cbf, cbv, mtt

    cbf = fltarr( xsize, ysize )
    cbv = fltarr( xsize, ysize )
    mtt = fltarr( xsize, ysize )

       ; one residue

    R=fltarr( nshots-(*bl)[1]-1 )

       ; now do the single value decomposition of the aif_matrix such that
       ; aifmatrix = U x W x V where U and V are orthonormal and W diagonal

    SVDC, aifmatrix, W, U, V, /double, /column

       ; for the diagonal matrix, remove the singular elements, i.e. those that have
       ; a value of less then xx% (as defined in preferences) of the maximum value

    wcut = matrixcutoff * max(abs(W))
    W    = (W ge wcut ) * W

		; do the calculation for each image pixel of this slice
		; The loop itself consumes about 6 seconds of time on my machine (as compared
		; to 10 seconds for the svsol) - 3 seconds of which are needed to fill 'signal'.
		; It actually makes no difference whether signal is defined (signal[*] = fltarr(something))
		; or newly created in each loop - the latter is even a little bit faster.
		; another 1.5 seconds are spend on the s0[x,y] gt sigthres comparison

    s0     = *(*(baselineset->imgdata()))[slice]
    ctr    = (*(ctrset->imgdata()))[slice]

	signal = fltarr( nshots-(*bl)[1]-1 )
	totaif = total(aif)
	for x = 0, xsize-1 do begin
		for y = 0, ysize-1 do begin
			signal[*] = (*ctr)[x,y,(*bl)[1]+1:nshots-1]
			totsig = total(signal)
			if( s0[x,y] gt sigthres and totsig gt 0. ) then begin
				R[*] = svsol( u, w, v, signal, /double, /column )
				cbf[x,y] = 6000 * 0.705 / tr * max(R[*]) * totsig / ( totaif * total(R[*]) )
				cbv[x,y] = 0.705 * 100. * totsig / totaif
				mtt[x,y] = total(R[*]) * tr / max(R[*])
			end
		end
	end

       ; cut off extreme values as defined in preferences

    cbv = cbv < (*(*self.g).preferences).ssvd_cbvcutoff
    cbv = cbv > 0.
    cbf = cbf < (*(*self.g).preferences).ssvd_cbfcutoff
    cbf = cbf > 0.
    mtt = mtt < (*(*self.g).preferences).ssvd_mttcutoff
    mtt = mtt > 0.

		; store the result in the result imagesets

	ptr_free, (*(cbvset->imgdata()))[slice]
	ptr_free, (*(cbfset->imgdata()))[slice]
	ptr_free, (*(mttset->imgdata()))[slice]

	(*(cbvset->imgdata()))[slice] = ptr_new( cbv, /no_copy )
	(*(cbfset->imgdata()))[slice] = ptr_new( cbf, /no_copy )
	(*(mttset->imgdata()))[slice] = ptr_new( mtt, /no_copy )

end


;+=----------------------------------------------------------------------------
; Name  : mod_ssvd__define
; Desc  : Private Procedure (constructor)
;=-----------------------------------------------------------------------------
pro mod_ssvd__define
	obj = { MOD_SSVD		   			,$
			inherits LUPEMOD 	  		,$
			mdisp:			 obj_new()	,$
			disp1:			 obj_new()	,$
			disp2:			 obj_new()	,$
			disp3:			 obj_new()	,$
			progressbar:	 obj_new()	,$
			idcbvname:				0L	,$
			idcbfname:				0L	,$
			idmttname:				0L	,$
			idsigthres:				0L	,$
			idctrthres:				0L	,$
			idwcutoff:				0L	,$
			idcbvcutoff:			0L	,$
			idcbfcutoff:			0L	,$
			idmttcutoff:			0L	 $
		  }
end

