;+=----------------------------------------------------------------------------
; Name    : widget_selectImgSet
; RCS     : $Id: widget_selectImgSet__define.pro,v 1.1 2004/11/11 11:37:48 olivert Exp olivert $
; Desc    : A convenient little widget object that displays a 'class' and 'name'
;           droplist for selection of an image set from the central container.
;           The object automatically keeps in sync with the contents of the
;           container and notifications upon user action (i.e. selection of
;           an image set) are send out upon request.
; Author  : (c) 2004 Oliver Thilmann
;=------------------------------------------------------------------------------


;+=----------------------------------------------------------------------------
; Name  : widget_selectImgSet::init
; Syntax: obj = obj_new('widget_selectImgSet', baseid, g, classTitle=string,
;         nameTitle=string, titleSize=int, listSize=int, exclude=stringArr
; Desc  :
;=-----------------------------------------------------------------------------
function widget_selectImgSet::init, baseid, g, classTitle=ct, nameTitle=cn, titleSize=ts, $
							listSize=ls, exclude=ex

	i18n = (*g).i18n
	if( self->mrNotify::init('change') eq 0 ) then return, 0

	b1           = widget_base    ( baseid, /row )
	x            = widget_label   ( b1, value=ct, xsize=ts )
	self.idclass = widget_droplist( b1, xsize=ls, uvalue=self, event_pro='widget_selectImgSet_classSelectEvt' )
	b1           = widget_base    ( baseid, /row )
	x            = widget_label   ( b1, value=cn, xsize=ts )
	self.idname  = widget_droplist( b1, xsize=ls, uvalue=self, event_pro='widget_selectImgSet_nameSelectEvt' )


	self.g      = g
	self.baseid = baseid
	if( n_elements(ex) gt 0 ) then self.exclude = ptr_new(ex) else self.exclude = ptr_new()

		; we want to be informed about changes in the central image container

	(*g).imgSets->notify, self, 'imgSetChangeEvt'

		; initial fill of the lists

	self->update

	return, 1
end


;+=----------------------------------------------------------------------------
; Name  : widget_selectImgSet::cleanup
; Desc  :
;=-----------------------------------------------------------------------------
pro widget_selectImgSet::cleanup
	(*self.g).imgSets->notify, self, /remove
	ptr_free, self.exclude
end




;==============================================================================
;
;        E V E N T   H A N D L I N G   A N D   N O T I F I C A T I O N
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : widget_selectImgSet::imgSetChangeEvt
; Desc  : Callback from image set container
;=-----------------------------------------------------------------------------
pro widget_selectImgSet::imgSetChangeEvt, origin, type, evt
	class = self.class
	name  = self.name
	self->update
	if( class ne self.class or name ne self.name ) then self->sendNotification, 'change'
end

;+=----------------------------------------------------------------------------
; Name  : widget_selectImgSet_classSelectEvt
; Desc  : Wrapper procedure
;=-----------------------------------------------------------------------------
pro widget_selectImgSet_classSelectEvt, evt
	widget_control, evt.id, get_uvalue=obj
	obj->classSelectEvt, evt
end

;+=----------------------------------------------------------------------------
; Name  : widget_selectImgSet_nameSelectEvt
; Desc  : Wrapper procedure
;=-----------------------------------------------------------------------------
pro widget_selectImgSet_nameSelectEvt, evt
	widget_control, evt.id, get_uvalue=obj
	obj->nameSelectEvt, evt
end

;+=----------------------------------------------------------------------------
; Name  : widget_selectImgSet::classSelectEvt
; Desc  :
;=-----------------------------------------------------------------------------
pro widget_selectImgSet::classSelectEvt, evt
	widget_control, self.idclass, get_value=classes
	class = self.class
	name  = self.name
	self.class = classes[evt.index]
	self->update
	if( class ne self.class or name ne self.name ) then self->sendNotification, 'change'
end

;+=----------------------------------------------------------------------------
; Name  : widget_selectImgSet::nameSelectEvt
; Desc  :
;=-----------------------------------------------------------------------------
pro widget_selectImgSet::nameSelectEvt, evt
	widget_control, self.idname, get_value=names
	class     = self.class
	name      = self.name
	self.name = names[evt.index]
	self->update
	if( class ne self.class or name ne self.name ) then self->sendNotification, 'change'
end

;==============================================================================
;
;                               I N T E R N A L
;
;==============================================================================


;+=----------------------------------------------------------------------------
; Name  : widget_selectImgSet::update
; Desc  : Update class and name lists - try to keep currently displayed
;         set, if possible. Otherwise chose next set and send notification
;=-----------------------------------------------------------------------------
pro widget_selectImgSet::update

		; During an update, we check whether we still exist
		; if we don't, this is a programming error!

	widget_control, self.baseid, bad_id=error   ; check whether widget is mapped
	if( error ne 0L ) then begin
		(*self.g).imgSets->notify, self, /remove
		message, 'Update call to destroyed widget_selectImgSet'
		self.baseid = 0L
		return
	end

		; get all current image sets, excluding the 'exclude' list

	imgSets = (*self.g).imgSets
	all     = imgSets->get( /all, count=asets )

	if( asets gt 0 ) then begin
		allsets = objarr(asets)
		nsets   = 0
		for i = 0, asets-1 do begin
			if( self.exclude ne ptr_new() ) then begin
				if( (where(*self.exclude eq all[i]->class()))[0] ne -1 ) then continue
			end
			allsets[nsets] = all[i]
			nsets++
		end
		allsets = allsets[0:nsets-1]
	end else begin
		nsets = 0
	end

		; if no image sets available, set currName and currClass to '' and
		; send a notification

	if( nsets eq 0 ) then begin
		widget_control, self.idclass, set_value='', sensitive=0
		widget_control, self.idname, set_value='', sensitive=0
		self.class = ''
		self.name  = ''
		self->sendNotification, 'change'
		return
	end
	widget_control, self.idclass, /sensitive
	widget_control, self.idname, /sensitive

		; get current set

	currset = mrGetImageSet( imgSets, class=self.class, name=self.name )

		; find all available classes

	classes = strarr(nsets)
	lastset = 0
	for i = 0, nsets-1 do begin                ; a bit complicated to keep the ordering
		classname = allsets[i]->class()
		x = where( classes eq classname )
		if( x ne -1 ) then continue
		classes[lastset] = classname
		lastset++
	end
	classes = classes[0:lastset-1]

	cid = where( classes eq self.class )
	if( cid[0] eq -1 ) then cid[0] = 0
	widget_control, self.idclass, set_value=classes
	widget_control, self.idclass, set_droplist_select=cid[0]
	classname = classes[cid[0]]

		; ok, we got the class, now go for the name

	allsets = mrGetImageSet( imgSets, class=classname, count=nnames )
	names   = strarr( nnames )
	for i = 0, nnames-1 do names[i] = allsets[i]->name()
	nid = where( names eq self.name )
	if( nid[0] eq -1 ) then nid[0]=0

	widget_control, self.idname, set_value=names
	widget_control, self.idname, set_droplist_select=nid[0]

	self.class = classname
	self.name  = names[nid[0]]

end


;+=----------------------------------------------------------------------------
; Name  : widget_selectImgSet::class
; Desc  : returns the currently selected class name
;=-----------------------------------------------------------------------------
function widget_selectImgSet::class
	return, self.class
end

;+=----------------------------------------------------------------------------
; Name  : widget_selectImgSet::name
; Desc  : returns currently selecte name
;=-----------------------------------------------------------------------------
function widget_selectImgSet::name
	return, self.name
end


;+=----------------------------------------------------------------------------
; Name  : widget_selectImgSet__define
; Desc  :
;=-----------------------------------------------------------------------------
pro widget_selectImgSet__define
	obj = { WIDGET_SELECTIMGSET			,$
			inherits MRNOTIFY           ,$
			g:				ptr_new()	,$ ; pointer to globals
			baseid:					0L	,$ ; widget id of the base
			exclude:		ptr_new()	,$ ; pointer to a string array
			idclass:				0L	,$ ; widget id of the class listbox widget
			idname:					0L	,$ ; widget id of the name listbox widget
			class:				 	''	,$ ; the current class
			name:					''	 $ ; the current name
	}
end
